/*!
 * @file fas_log.h
 * @brief Fastec Error and Debug Logging support.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FAS_LOG_H_
#define FAS_LOG_H_
#pragma once

#if !defined(_WIN32)
#include <syslog.h>
#define FI_FUNCTION __PRETTY_FUNCTION__
#else
// These would normally reside in the syslog.h file.
#define LOG_EMERG   0   /* system is unusable */
#define LOG_ALERT   1   /* action must be taken immediately */
#define LOG_CRIT    2   /* critical conditions */
#define LOG_ERR     3   /* error conditions */
#define LOG_WARNING 4   /* warning conditions */
#define LOG_NOTICE  5   /* normal but significant condition */
#define LOG_INFO    6   /* informational */
#define LOG_DEBUG   7   /* debug-level messages */

#define FI_FUNCTION __FUNCTION__
#endif  /* _WIN32 */

#include <memory>
#include <mutex>
#include <string>


namespace FI_LOG
{
class FiLogger
{
private:
    static std::unique_ptr<FiLogger> instance;
    static std::mutex instance_mutex;

private:
    FiLogger();

public:
    ~FiLogger();

    // Singletons should not be cloneable.
    FiLogger(FiLogger& other) = delete;

    // Singletons should not be assignable.
    void operator=(const FiLogger&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static FiLogger *factory();

    void fi_log(unsigned severity, const char *fmt, ...);

    void loggerUseSyslog(bool enable);
    void loggerUseStdout(bool enable);
    int loggerFile(const char *filename);
    void loggerReopen();

private:
    // Flag to indicate messages from debug_log() are be displayed.
    bool debug_to_log;

    // Flag to indicate if the syslog facility should be used.
    bool use_syslog;

    // Flag to indicate if log messages are to be sent to STDOUT.
    bool use_stdout;

    // Private log file name.
    std::string logFileName;

    // Flag to indicate if log messages are to be saved to a file.
    int logFileFd;
};
} // End namespace FI_LOG

#endif /* !FAS_LOG_H_ */
