/*!
 * @file fi_rdi_packed.h
 * @brief Definitions and declarations for Fastec Imaging RDI packed images.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FIVID_FI_RDI_PACKED_
#define FIVID_FI_RDI_PACKED_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <memory>
#include <vector>

namespace FI_RDI_PACKED
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! @brief RDIP: RDI Packed Image Format.
//! 8,10, or 12 bit packed image data for RDI packed raw images.
const uint32_t FOURCC_RDI   = 0x52444950;

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiRDIPacked
{
public:
    FiRDIPacked();
    ~FiRDIPacked() {}

    //! @name Public Property Accessor Functions.
    //!
    //! These are the public functions used to access the properties of this object.
    //! @{
    static uint32_t getHeaderSize();

    constexpr int getBitsPerPixel()         { return bitsPerPixel; }
    constexpr uint32_t getImageWidth()      { return width; }
    constexpr uint32_t getImageHeight()     { return height; }
    constexpr uint32_t getPackedDataSize()  { return packedDataSize; }

    const uint8_t *getHeader();
    const uint8_t *getPackedData();

    bool isHeaderInitialized() const  { return !header.empty(); }
    //! @}

    void setHeader(uint32_t frameNumber, time_t timeSeconds, uint32_t timeMicroseconds);
    void setTiffImageData(const uint8_t *imagePtr, uint32_t imageWidth, uint32_t imageHeight, int imageBpp);
    void setPackedData(const uint8_t *dataPtr, uint32_t dataLength, uint32_t imageWidth, uint32_t imageHeight, int imageBpp);

private:
    std::vector<uint8_t> header;
    std::unique_ptr<uint8_t[]> packedData;

    int bitsPerPixel;                       //!< Number of bits per pixel in the original image data (8, 10, 12).
    uint32_t packedDataSize;                //!< Number of bytes of packed image data. NOT the original image size!
    uint32_t height;                        //!< The width (in pixels) of the original image.
    uint32_t width;                         //!< The height of the original image.
};

} // End of FI_RDI_PACKED namespace.

#endif  /* FIVID_FI_RDI_PACKED_ */

