/*!
 * @file fas_stream_rdi_packed.h
 * @brief The RDI Packed stream object.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_RDI_PACKED_STREAM_H_
#define FASTEC_RDI_PACKED_STREAM_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>

#include <deque>
#include <fstream>
#include <memory>
#include <mutex>

#include "fas_internal.h"

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
namespace FI_RDI_PACKED
{
class FiRDIPacked;
}

namespace FAS_RDI
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiRDIPackedStream
{
public:
    FiRDIPackedStream();
    ~FiRDIPackedStream();

    // This object, for now, is non-copyable and non-copy assignable.
    FiRDIPackedStream(const FiRDIPackedStream& i) = delete;
    FiRDIPackedStream& operator=(const FiRDIPackedStream& i) = delete;

    FAS_INTERNAL::FIP_STATUS addFrame(std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> imagePtr);
    FAS_INTERNAL::FIP_STATUS fiAddPlaceholderFrame();
    FAS_INTERNAL::FIP_STATUS closeStream();
    FAS_INTERNAL::FIP_STATUS editVideoFrame(uint32_t frameNumber, std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> imagePtr);
    FAS_INTERNAL::FIP_STATUS isStreamWriterComplete();
    FAS_INTERNAL::FIP_STATUS prepareForWriting(const std::string& streamName, const FAS_INTERNAL::FIP_STREAM_CONFIGURATION& configInfo);

    uint32_t getFrameCount();
    const std::string getVideoName()            { return streamNameMinusExt; }
    uint32_t getLastFrameAdded()                { return frameSeqNum; }
    uint32_t getFramesEncodedCount()            { return frameSeqNum; }

    int getBitsPerPixel();
    uint32_t getWidth();
    uint32_t getHeight();

private:
    void calcDivFileInfo(uint32_t imageFrameNbr, uint32_t& divFileIndex, uint32_t& divFileOffset);
    void calcFrameTime(uint32_t imageFrameNbr, uint64_t& timeSec, uint32_t& timeMicrosec);
    void guidGen(uint8_t *uuid_ptr);
    bool initializeHeader(std::ofstream *streamPtr);
    void queueFrame(std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> imagePtr);
    void setConfiguration(const FAS_INTERNAL::FIP_STREAM_CONFIGURATION& configInfo);

private:
    // A collection of static string used by the class object.
    static const std::string RDI_PACKED_HEADER_EXTENSION;
    static const std::string RDI_PACKED_DIV_EXTENSION;
    static const std::string RDI_PACKED_FILE_v3_0;
    static const std::string RDI_PACKED_CONVERTED;
    static const std::string FASTEC_GENERATED;

    //! The internal status of the object.
    FAS_INTERNAL::FIP_STATUS streamStatus;

    //! The current ".div" file being written to.
    std::ofstream divFile;

    //! The current ".div" file index.
    uint32_t divIndex;

    //! The configuration information object.
    std::unique_ptr<FAS_INTERNAL::FIP_STREAM_CONFIGURATION> configInfo;

    //! How big is each RDI Packed image frame? This does not include the image header just the frame.
    uint32_t imagePackedFrameSize;

    //! How many RDI packed images fit into a single DIV file?
    uint32_t imagesPerDiv;

    //! Packed stream name components.
    std::string streamName;
    std::string streamNameMinusExt;
    std::string streamDivPath;

    //!< Sequence number of next frame to be queued to be written.
    uint32_t frameSeqNum;
};

} // End of FAS_RDI Namespace

#endif /* !FASTEC_RDI_PACKED_STREAM_H_ */
