/*!
 * @file fas_stream_rdi_packed.cpp
 * @brief The RDI Packed stream object.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <inttypes.h>

#include <iostream>
#include <fstream>
#include <filesystem>
#include <vector>

#include <fmt/printf.h>

#include "fi_stream_rdi_packed.h"
#include "fi_rdi_packed.h"
#include "fi_log.h"

namespace FAS_RDI
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The RDI maximum size for a DIV file.
const uint32_t MAX_DIV_FILESIZE                 = 0x7FFFFFFF;

// We do not have a mechanism to integrate RDI software headers and for a temporary solution we
// will populate a list of constants that will be utilized for the missing enumerations.
const uint32_t RDI_DISTANCE_UNIT            = 0;    //!< Default to feet(0).
const uint32_t RDI_LIGHTING_TYPE            = 1;    //!< Default to OTHER.
const uint32_t RDI_ACQUISITION_MODE         = 0;    //!< 0 is MotionAmplification.
const uint32_t RDI_ACQUISITION_TYPE         = 1;    //!< 1 is Time.
const uint32_t RDI_FRAME_ORIENTATION        = 0;    //!< 0 is standard frame orientation.
const uint32_t RDI_TIME_UNIT                = 0;    //!< Default to SECONDS;
const uint32_t RDI_ANALOG_GAIN              = 0;    //!< Default to ONE.
const uint32_t RDI_LINE_FREQUENCY           = 60;   //!< From the example, tif_rdi_convert.
const uint32_t RDI_2X_BINNING               = 1;    //!< 1 if 2x binning is enabled.
const uint32_t RDI_4X_BINNING               = 2;    //!< 2 if 4x binning is enabled.

//! For RDI Packed Stream writers they must know how they are to pack
//! the input image frame data for RDI Packed output.
typedef enum
{
    RDI_08_PACK,
    RDI_12_PACK,
    RDI_16_PACK
} RDI_IMAGE_FORMAT;

//! Define what a GUID is.
typedef uint8_t RDI_GUID[16];

//! The contents of an RDI Metadata/Header file with extension ".RDI".
//! Note that this structure is packed in the RDI C# source.
#pragma pack(1)
typedef struct RDI_PACKED_STREAM_HDR
{
    char                    fileVersion[5];                 // +5 = 5
    char                    userDesc[151];                  // +151 = 156
    double                  focalLength;                    // +8 = 164
    double                  distance;                       // +8 = 172
    uint32_t                distanceUnits;                  // +4 = 176 - use RDI_DISTANCE_UNIT
    uint16_t                imageWidth;                     // +2 = 178
    uint16_t                imageHeight;                    // +2 = 180
    uint32_t                imageBPP;                       // +4 = 184
    double                  frameRate;                      // +8 = 192
    double                  gain;                           // +8 = 200
    double                  shutter;                        // +8 = 208
    char                    driverName[70];                 // +70 = 278
    char                    firmwareBuildTime[30];          // +30 = 308
    char                    firmwareVersion[12];            // +12 = 320
    uint32_t                iidcVersion;                    // +4 = 324
    char                    modelName[60];                  // +60 = 384
    char                    sensorInfo[60];                 // +60 = 444
    char                    sensorResolution[16];           // +16 = 460
    uint32_t                serialNumber;                   // +4 = 464
    uint32_t                numImagesEachFile;              // +4 = 468
    uint32_t                totalNumImages;                 // +4 = 472
    uint8_t                 demoFile;                       // +1 = 473
    uint32_t                lightingType;                   // +4 = 477 - use RDI_LIGHTING_TYPE
    uint32_t                aquisitionMode;                 // +4 = 481 - use RDI_ACQUISITION_MODE
    uint16_t                brightness;                     // +2 = 483
    double                  lineFrequency;                  // +8 = 491 - use RDI_LINE_FREQUENCY
    char                    collectedFileVersions[5];       // +5 = 496
    uint8_t                 stabilized;                     // +1 = 497
    uint8_t                 filterCount;                    // +1 = 498
    uint8_t                 appliedFilters[1200];           // +1200 = 1698
    RDI_GUID                recordingId;                    // +16 = 1714
    RDI_GUID                originalRecordingId;            // +16 = 1730
    RDI_GUID                parentRecordingId;              // +16 = 1746
    uint32_t                frameOrientationRequested;      // +4 = 1750 - use RDI_FRAME_ORIENTATION
    uint32_t                frameOrientationApplied;        // +4 = 1754 - use RDI_FRAME_ORIENTATION
    uint16_t                imageLeft;                      // +2 = 1756
    uint16_t                imageTop;                       // +2 = 1758
    double                  fMax;                           // +8 = 1766
    uint32_t                acquisitionType;                // +4 = 1770 - use RDI_ACQUISITION_TYPE
    double                  actualFramerate;                // +8 = 1778
    char                    serialNumberString[24];         // +24 = 1802
    uint32_t                analogGain;                     // +4 = 1806 - use RDI_ANALOG_GAIN
    uint32_t                binning;                        // +4 = 1810 - use RDI_BINNING
    double                  assetSpeed;                     // +8 = 1818
    double                  strobeFreq;                     // +8 = 1826
    double                  strobeCycles;                   // +8 = 1834
    double                  degPerFrame;                    // +8 = 1842
    double                  dBrightness;                    // +8 = 1850
    uint32_t                acquisitionDevice;              // +4 = 1854
    uint32_t                preTriggerImages;               // +4 = 1858
    double                  prevFocalLength;                // +8 = 1866
    double                  prevDistance;                   // +8 = 1874
    double                  secPerFrame;                    // +8 = 1882
    uint32_t                frameTimeUnit;                  // +4 = 1886 - use RDI_TIME_UNIT
    uint32_t                ltmDuration;                    // +4 = 1890
    uint8_t                 varFramerate;                   // +1 = 1891
    uint8_t                 stMasterIdx;                    // +1 = 1892
    double                  efsShutterMin;                  // +8 = 1900
    char                    stSerialNumberString[24];       // +24 = 1924
    double                  stBaseline;                     // +8 = 1932
    uint32_t                stOlapRonL[4];                  // +16 = 1948
    float                   spare_floats[4];                // +16 = 1964
    double                  timerDelay;                     // +8 = 1972
    double                  spareDoubles[8];                // +64 = 2036
    float                   spareFloats2[8];                // +32 = 2058
    double                  spareDoubles2[8];               // +64 = 2132
    double                  prevSpeed;                      // +8 = 2140
    uint8_t                 stLeftIdx;                      // +1 = 2141
    uint8_t                 changedVals;                    // +1 = 2142
    uint8_t                 distanceCalculated;             // +1 = 2143
    uint8_t                 spareBytes[855];                // +855 = 2998

} RDI_PACKED_STREAM_HDR;
#ifdef _WIN32
#pragma pack()
#else
#pragma pack(0)
#endif

//**************************************************
// RDI Packed Streamer Specific
//**************************************************

typedef std::vector<uint8_t> PLACE_HOLDER_FRAME;

} // End of FAS_RDI

using namespace FAS_RDI;

//! The RDI Packed Header extension value.
const std::string FiRDIPackedStream::RDI_PACKED_HEADER_EXTENSION("rdi");

//! The RDI Packed collection of images comprising 2GB extension value.
const std::string FiRDIPackedStream::RDI_PACKED_DIV_EXTENSION("div");

//! Version of the RDI packed file format we implement.
const std::string FiRDIPackedStream::RDI_PACKED_FILE_v3_0("v3.0");

//! Default string for converted data.
const std::string FiRDIPackedStream::RDI_PACKED_CONVERTED("Converted");

//! A generic string to put into the metadata just to identify the creator.
//! This is temporary.
const std::string FiRDIPackedStream::FASTEC_GENERATED("Fastec Generated");

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*!
 * @brief RDI Packed stream construction.
 *
 * To write an RDI packed stream, the user must take the following steps:
 * * Create an instance of the FiRDIPackedStream.
 * * Get a destination path for where to write the stream data.
 * * Initialize the FAS_INTERNAL::FIP_STREAM_CONFIGURATION structure.
 * * Prepare the stream by calling FiRDIPackedStream::prepareForWriting.
 * * Download frames as either raw TIFF or RDI Packed and create FI_RDI_PACKED::FiRDIPacked frame objects.
 * * Add the FI_RDI_PACKED::FiRDIPacked objects to the stream using the FiRDIPackedStream::addFrame method.
 * * If a frame has been dropped, add a place holder frame with the FiRDIPackedStream::fiAddPlaceholderFrame method.
 *   * Use the FiRDIPackedStream::editVideoFrame method to replace a place holder frame with an actual image.
 * * When all frames have been added, check to see if the stream has finished writing by
 *   calling the FiRDIPackedStream::isStreamWriterComplete method.
 * * Once the stream finishes writing, call the FiRDIPackedStream::closeStream method.
 */
FiRDIPackedStream::FiRDIPackedStream()
{
    FI_LOG::FiLogger::factory()->loggerUseStdout(true);
    FI_LOG::FiLogger::factory()->loggerUseSyslog(true);

    imagePackedFrameSize = 0;
    imagesPerDiv = 0;
    frameSeqNum = 0;
    divIndex = 0;

    streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_BUILT;
}

/*!
 * @brief RDI Packed stream destruction.
 */
FiRDIPackedStream::~FiRDIPackedStream()
{
    // If the stream status isn't set to closed - close it.
    if (streamStatus != FAS_INTERNAL::FIP_STATUS::STREAM_CLOSED)
        FiRDIPackedStream::closeStream();
}

/*!
 * @brief Set up the RDI Packed stream object for writing.
 *
 * @param streamName    Full path for the created RDI packed stream including the ".rdi" extension.
 * @param configInfo    The configuration information for the stream.
 *
 * @returns The stream status setting of the object.
 */
FAS_INTERNAL::FIP_STATUS FiRDIPackedStream::prepareForWriting(const std::string& streamName, const FAS_INTERNAL::FIP_STREAM_CONFIGURATION& configInfo)
{
    char pathDel = '/';
    std::string::size_type extPos = 0;
    std::string::size_type dirPos = 0;
    std::error_code err;
    std::ofstream metadataFile;

    // The stream should be in a status of construction in order to prepare for writing.
    if (streamStatus != FAS_INTERNAL::FIP_STATUS::STREAM_BUILT)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_WARNING, "%s: Was not properly constructed, unable to create %s", FI_FUNCTION, streamName);

        // This method is expecting a newly constructed stream object.
        // Deny the request and leave the stream status set to its state.
        return FAS_INTERNAL::FIP_STATUS::REQUEST_DENIED;
    }

    // Initialize the configuration information.
    setConfiguration(configInfo);

    // Sanity check.
    if (this->configInfo.get() == nullptr)
    {
        streamStatus = FAS_INTERNAL::FIP_STATUS::ALLOCATION_FAILED;
        FI_LOG::FiLogger::factory()->fi_log(LOG_WARNING, "%s: no configuration information found!", FI_FUNCTION);
        goto exitPrepareForRDIPackedWriting;
    }

    // There should be at least one frame to stream.
    if (this->configInfo->frameCount == 0)
    {
        streamStatus = FAS_INTERNAL::FIP_STATUS::BAD_RANGE;
        FI_LOG::FiLogger::factory()->fi_log(LOG_WARNING, "%s: frameCount set to zero.", FI_FUNCTION);
        goto exitPrepareForRDIPackedWriting;
    }

    // How many of our images will fit into a single DIV file?
    // RDI Packed data where there are 2 pixels per 24 bits for all packing requests.
    if (configInfo.bitsPerPixel == 8)
        imagePackedFrameSize = configInfo.width * configInfo.height;
    else
        imagePackedFrameSize = (configInfo.width * configInfo.height * 3) / 2;
    imagesPerDiv = MAX_DIV_FILESIZE / (imagePackedFrameSize + FI_RDI_PACKED::FiRDIPacked::getHeaderSize());

    // Perform some checks and pull out some information from the streamName.
    // Find and remove the file extension for storage in our object.
    extPos = streamName.find_last_of('.');
    if (extPos == std::string::npos)
    {
        streamStatus = FAS_INTERNAL::FIP_STATUS::CODEC_SETUP_ERROR;
        FI_LOG::FiLogger::factory()->fi_log(LOG_WARNING, "%s: streamName is missing a file extension!", FI_FUNCTION);
        goto exitPrepareForRDIPackedWriting;
    }

    // Extract the sub-folder name we will use for the DIV files.
    streamDivPath = streamName.substr(0, extPos);
//    FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamDivPath='%s'", FI_FUNCTION, streamDivPath.c_str());

    // Pull out the filename for the stream without the full path or the file extension.
    dirPos = streamDivPath.find_last_of(pathDel);
    if (dirPos == std::string::npos)
    {
        // Try the windows delimiter.
        pathDel = '\\';
        dirPos = streamDivPath.find_last_of(pathDel);
    }

    // If found use it to pull the name only.
    if (dirPos != std::string::npos)
    {
        // Pull out the filename portion only.
        streamNameMinusExt = streamDivPath.substr(dirPos + 1);
    }
    else
    {
        // Just use the whole thing since there were no path separators.
        streamNameMinusExt = streamDivPath;
    }

    // Make sure the path is terminated with a path delimiter.
    if (streamDivPath.back() != pathDel)
        streamDivPath.push_back(pathDel);

    // Save the streamName.
    this->streamName.assign(streamName);

    // Create an output file stream for the RDI metadata file (.RDI) file using the standard output file streaming object.
    metadataFile.open(streamName.c_str(), std::ofstream::out | std::ofstream::binary | std::ofstream::trunc);

    // Check the status of the output file stream object.
    if (!metadataFile.is_open())
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_WARNING, "%s: Unable to open %s for binary output!", FI_FUNCTION, streamName);
        streamStatus = FAS_INTERNAL::FIP_STATUS::NO_FILE_FOUND;
        goto exitPrepareForRDIPackedWriting;
    }

    // Initialize the RDI Packed header information preparing for RDI Packed stream writer.
    if (!initializeHeader(&metadataFile))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: Unable to initialize RDI Packed header.", FI_FUNCTION);
        streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_SETUP_ERROR;
        goto exitPrepareForRDIPackedWriting;
    }

    // Make the subfolder for the DIV files.
    if (!std::filesystem::create_directories(streamDivPath, err))
    {
        if (!std::filesystem::exists(streamDivPath))
        {
            streamStatus = FAS_INTERNAL::FIP_STATUS::CODEC_SETUP_ERROR;
            FI_LOG::FiLogger::factory()->fi_log(LOG_WARNING, "%s: Unable to create subfolder %s for DIV files!", FI_FUNCTION, streamDivPath);
            goto exitPrepareForRDIPackedWriting;
        }
    }

    // Everything is ready for streaming.
    streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_READY;

exitPrepareForRDIPackedWriting:
    // Close the metadataFile object, if it is still open.
    if (metadataFile.is_open())
        metadataFile.close();

    // Return the status of this stream object as being prepared.
    return streamStatus;
}

int FiRDIPackedStream::getBitsPerPixel()
{
    if (configInfo.get() == nullptr)
        return 0;
    return configInfo->bitsPerPixel;
}

uint32_t FiRDIPackedStream::getHeight()
{
    if (configInfo.get() == nullptr)
        return 0;
    return configInfo->height;
}

uint32_t FiRDIPackedStream::getWidth()
{
    if (configInfo.get() == nullptr)
        return 0;
    return configInfo->width;
}

uint32_t FiRDIPackedStream::getFrameCount()
{
    if (configInfo.get() == nullptr)
        return 0;
    return configInfo->frameCount;
}

/*!
 * @brief Pass the configuration in to the FiRDIPackedStream object.
 *
 * The configuration information is copied to a private internal structure for
 * use during the streaming operation.
 *
 * @param configInfo    The structure containing the configuration information of the stream.
 */
void FiRDIPackedStream::setConfiguration(const FAS_INTERNAL::FIP_STREAM_CONFIGURATION& configInfo)
{
    // Allocate our configuration information object.
    this->configInfo = std::unique_ptr<FAS_INTERNAL::FIP_STREAM_CONFIGURATION>(new FAS_INTERNAL::FIP_STREAM_CONFIGURATION());
    if (this->configInfo.get() == nullptr)
    {
        // ALM_TODO: SQUAWK!
        return;
    }

    // Copy everything over.
    this->configInfo->manufacturer.assign(configInfo.manufacturer);
    this->configInfo->cameraModel.assign(configInfo.cameraModel);
    this->configInfo->cameraVersion.assign(configInfo.cameraVersion);
    this->configInfo->macAddress.assign(configInfo.macAddress);
    this->configInfo->userComment.assign(configInfo.userComment);

    this->configInfo->roiX = configInfo.roiX;
    this->configInfo->roiY = configInfo.roiY;
    this->configInfo->width = configInfo.width;
    this->configInfo->height = configInfo.height;
    this->configInfo->frameRate = configInfo.frameRate;
    this->configInfo->exposure = configInfo.exposure;
    this->configInfo->bitsPerPixel = configInfo.bitsPerPixel;
    this->configInfo->frameCount = configInfo.frameCount;

    this->configInfo->sensorType = configInfo.sensorType;
    this->configInfo->sensorBinningOpt = configInfo.sensorBinningOpt;
    this->configInfo->sensorSubSampleOpt = configInfo.sensorSubSampleOpt;

    this->configInfo->lensManufacturer.assign(configInfo.lensManufacturer);
    this->configInfo->lensModel.assign(configInfo.lensModel);
    this->configInfo->focalLength = configInfo.focalLength;
    this->configInfo->aperture = configInfo.aperture;

    this->configInfo->timeStartOfStreamSeconds = configInfo.timeStartOfStreamSeconds;
    this->configInfo->timeStartOfStreamMicrosec = configInfo.timeStartOfStreamMicrosec;
}

/*!
 * @brief Create a binary GUID.
 *
 * A GUID is <32-bit val>-<16-bit>-<16-bit>-<16-bit>-<48-bit>. In bytes
 * we have: 4bytes-2bytes-2bytes-2bytes-6bytes, for a total of 16 bytes
 * or 4 4byte values.
 *
 * @param [out] uuid_ptr    Pointer to where we are to put the GUID.
 */
void FiRDIPackedStream::guidGen(uint8_t *uuid_ptr)
{
    uint32_t rand_val[4];

    // Seed the random number generator.
    srand(time(NULL));

    // Build up our GUID.
    rand_val[0] = rand();
    rand_val[1] = rand();
    rand_val[2] = rand();
    rand_val[3] = rand();

    // Write out our GUID.
    memcpy(uuid_ptr, rand_val, sizeof(uint32_t) * 4);
}

/*!
 * @brief Adds an image or placeholder (fake) image to the stream.
 *
 * This should only be called by addFrame and fiAddPlaceholderFrame methods and
 * it is assumed that any validation has taken place prior to this call for an
 * attached frame or a placeholder frame.
 */
void FiRDIPackedStream::queueFrame(std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> imagePtr)
{
    uint32_t divSeqNbr;
    uint32_t divOffset;

    // Sanity check.
    if (imagePtr.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: imageptr is nullptr!?", FI_FUNCTION);
        return;
    }
    if (frameSeqNum >= getFrameCount())
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: seqNum=%d but frame_count=%d", FI_FUNCTION, frameSeqNum, getFrameCount());
        return;
    }

    // Calculate information needed to write the frame data.
    calcDivFileInfo(frameSeqNum, divSeqNbr, divOffset);

    // Are we writing to a new ".div" file?
    if (divIndex != divSeqNbr)
    {
        // Do we have a file open?
        if (divFile.is_open())
        {
            // Close the existing file.
            divFile.close();
        }

        // Generate the name of the new ".div" file.
        std::string divFileName = fmt::sprintf("%s%s_%04u.div", streamDivPath, streamNameMinusExt, divSeqNbr);

        // Open the new file.
        divFile.open(divFileName.c_str(), std::ofstream::out | std::ofstream::binary | std::ofstream::trunc);

        // Remember the new file's index.
        divIndex = divSeqNbr;
    }

    // Make sure our file is open.
    if (!divFile.is_open())
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: the divFile is not open.", FI_FUNCTION);
        return;
    }

    // Write out the header and image data to the file.
    if (imagePtr->getHeader() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: There is no header!", FI_FUNCTION);
        return;
    }
    if (imagePtr->getPackedData() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: There is no packed data!", FI_FUNCTION);
        return;
    }
    divFile.write((const char *)imagePtr->getHeader(), imagePtr->getHeaderSize());
    divFile.write((const char *)imagePtr->getPackedData(), imagePtr->getPackedDataSize());

//    FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: frame (%u) was added to stream.", FI_FUNCTION, frameSeqNum);

    // Increment base class progress counter for frame processing.
    frameSeqNum++;

    // We are ready for the next frame.
    streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_READY;

    // Are we done?
    if (frameSeqNum >= getFrameCount())
    {
        // Do we have a file open?
        if (divFile.is_open())
        {
            // Close the existing file.
            divFile.close();
        }
    }
}

/*!
 * @brief Initialize RDI Packed image header by writing it on disk along with the
 * subfolder that will contain all of the DIV files associated with this stream.
 *
 * @return Boolean true if successful or false if error occurred in which case check streamStatus.
 */
bool FiRDIPackedStream::initializeHeader(std::ofstream *streamPtr)
{
    RDI_PACKED_STREAM_HDR metadataRDI;
    RDI_IMAGE_FORMAT imageFormat;
    bool initializedHeader = false;

    // Need to write out the header file.
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamPtr is null!", FI_FUNCTION);
        goto exitInitializeHeader;
    }

    // Figure out what the image format should be.
    if (configInfo->bitsPerPixel == 8)
        imageFormat = RDI_08_PACK;
    else
        imageFormat = RDI_12_PACK;

    // Clear it with zeros.
    memset((void *) &metadataRDI, 0, sizeof(RDI_PACKED_STREAM_HDR));

    // Just fill in a couple of items.
    strncpy(&metadataRDI.fileVersion[0], RDI_PACKED_FILE_v3_0.c_str(), RDI_PACKED_FILE_v3_0.length());
    strncpy(&metadataRDI.userDesc[0], configInfo->userComment.c_str(), sizeof(metadataRDI.userDesc));

    metadataRDI.focalLength                 = (double)configInfo->focalLength;
    metadataRDI.distance                    = 1.0;
    metadataRDI.distanceUnits               = RDI_DISTANCE_UNIT;
    metadataRDI.imageWidth                  = configInfo->width;
    metadataRDI.imageHeight                 = configInfo->height;
    metadataRDI.imageBPP                    = imageFormat;
    metadataRDI.gain                        = 0.0;
    metadataRDI.frameRate                   = configInfo->frameRate;        // ALM_TODO: Verify units.
    metadataRDI.shutter                     = configInfo->exposure;         // ALM_TODO: Verify units.

    strncpy(&metadataRDI.driverName[0], RDI_PACKED_CONVERTED.c_str(), RDI_PACKED_CONVERTED.length());
    strncpy(&metadataRDI.firmwareBuildTime[0], RDI_PACKED_CONVERTED.c_str(), RDI_PACKED_CONVERTED.length());
    strncpy(&metadataRDI.firmwareVersion[0], RDI_PACKED_CONVERTED.c_str(), RDI_PACKED_CONVERTED.length());
    metadataRDI.iidcVersion                 = 0;

    strncpy(&metadataRDI.modelName[0], configInfo->cameraModel.c_str(), configInfo->cameraModel.length());
    strncpy(&metadataRDI.sensorInfo[0], RDI_PACKED_CONVERTED.c_str(), RDI_PACKED_CONVERTED.length());
    strncpy(&metadataRDI.sensorResolution[0], RDI_PACKED_CONVERTED.c_str(), RDI_PACKED_CONVERTED.length());

    metadataRDI.serialNumber                = 0;
    metadataRDI.numImagesEachFile           = imagesPerDiv;
    metadataRDI.totalNumImages              = configInfo->frameCount;

    metadataRDI.demoFile                    = 0;
    metadataRDI.lightingType                = RDI_LIGHTING_TYPE;
    metadataRDI.aquisitionMode              = RDI_ACQUISITION_MODE;
    metadataRDI.brightness                  = 0;
    metadataRDI.lineFrequency               = RDI_LINE_FREQUENCY;
    strncpy(&metadataRDI.collectedFileVersions[0], RDI_PACKED_FILE_v3_0.c_str(), RDI_PACKED_FILE_v3_0.length());
    metadataRDI.stabilized                  = 0;
    metadataRDI.filterCount                 = 0;

    guidGen(metadataRDI.recordingId);
    memcpy(metadataRDI.originalRecordingId, metadataRDI.recordingId, sizeof(metadataRDI.recordingId));

    metadataRDI.frameOrientationRequested   = RDI_FRAME_ORIENTATION;
    metadataRDI.frameOrientationApplied     = RDI_FRAME_ORIENTATION;
    metadataRDI.imageLeft                   = 0;
    metadataRDI.imageTop                    = 0;
    metadataRDI.fMax                        = 0;
    metadataRDI.acquisitionType             = RDI_ACQUISITION_TYPE;
    metadataRDI.actualFramerate             = metadataRDI.frameRate;

    metadataRDI.analogGain                  = RDI_ANALOG_GAIN;
    metadataRDI.frameTimeUnit               = RDI_TIME_UNIT;

    // Are we doing 2x binning?
    if (configInfo->sensorBinningOpt == FAS_INTERNAL::FIP_SENSOR_BINNING::BINNING_2X)
        metadataRDI.binning                 = RDI_2X_BINNING;
    else if (configInfo->sensorBinningOpt == FAS_INTERNAL::FIP_SENSOR_BINNING::BINNING_4X)
        metadataRDI.binning                 = RDI_4X_BINNING;

    // Seek to the beginning.
    streamPtr->seekp(0, std::ios_base::beg);

    // Write the data to the output stream.
    streamPtr->write((const char *)&metadataRDI, sizeof(RDI_PACKED_STREAM_HDR));

    // Is the output stream IO object in good state?
    if (!streamPtr->good())
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_ERR, "%s: io stream write failed.", FI_FUNCTION);
        streamStatus = FAS_INTERNAL::FIP_STATUS::NO_IO_STREAM;
        goto exitInitializeHeader;
    }

    // If made it here all is initialized and ready to proceed.
    initializedHeader = true;

exitInitializeHeader:
    return initializedHeader;
}

/*!
 * @brief Adds an image to the video stream.
 *
 * @param imagePtr      The pointer containing the image data frame.
 *
 * @returns The stream status setting of the object.
 */
FAS_INTERNAL::FIP_STATUS FiRDIPackedStream::addFrame(std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> imagePtr)
{
    // Update the image's header information.
    if (!imagePtr->isHeaderInitialized())
    {
        uint64_t timeSec;
        uint32_t timeMicrosec;

        // Get the frame time.
        calcFrameTime(frameSeqNum, timeSec, timeMicrosec);

        // Setup the header information.
        imagePtr->setHeader(frameSeqNum, timeSec, timeMicrosec);
    }

    // Call the actual logic to insert this image data frame into RDI Packed stream.
    // Note that queueFrame will increment the frameSeqNum.
    queueFrame(std::move(imagePtr));

    // Is this stream complete?
    if (configInfo->frameCount == frameSeqNum)
    {
        // Are we done writing this stream?
        if (streamStatus == FAS_INTERNAL::FIP_STATUS::STREAM_READY)
        {
            streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED;
        }
    }

    // Return our status.
    return streamStatus;
}

/*!
 * @brief Inserts an empty frame into the proper DIV file for that frame as a placeholder.
 *
 * The time information will be zero, as will be the image data. The sequence number should
 * be accurate.
 *
 * @returns The video stream status setting of the object.
 */
FAS_INTERNAL::FIP_STATUS FiRDIPackedStream::fiAddPlaceholderFrame()
{
    std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> imagePtr;

    imagePtr = std::unique_ptr<FI_RDI_PACKED::FiRDIPacked>(new FI_RDI_PACKED::FiRDIPacked());
    if (imagePtr.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_ERR, "%s: Unable to allocate the FiRDIPacked() placeholder image!", FI_FUNCTION);
        streamStatus = FAS_INTERNAL::FIP_STATUS::ALLOCATION_FAILED;
    }
    else
    {
        // Initialize our placeholder frame. The size of the packed image data is assumed
        // to include the image header too.
        imagePtr->setHeader(frameSeqNum, 0, 0);
        imagePtr->setPackedData(nullptr, imagePackedFrameSize + FI_RDI_PACKED::FiRDIPacked::getHeaderSize(), configInfo->width, configInfo->height, configInfo->bitsPerPixel);

        // Queue up our placeholder.
        queueFrame(std::move(imagePtr));

        // All seems well.
        streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_READY;
    }

    // Return our status.
    return streamStatus;
}

/*!
 * @brief Insert the real frame of an image into the created stream at a
 *        specified frameNumber representing a placeholder frame location.
 *
 * @param frameNumber   Zero-based frame number for the placeholder frame to edit.
 * @param imagePtr      The image frame data to edit in place.
 *
 * @returns The video stream status setting of the object.
 */
FAS_INTERNAL::FIP_STATUS FiRDIPackedStream::editVideoFrame(uint32_t frameNumber, std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> imagePtr)
{
    FAS_INTERNAL::FIP_STATUS retStatus = FAS_INTERNAL::FIP_STATUS::REQUEST_DENIED;   // Not implemented yet.

    (void)imagePtr;
    FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: Not implemented yet. Ignoring %d", FI_FUNCTION, frameNumber);

    return retStatus;
}

/*!
 * @brief Queries complete state of this stream.
 *
 * @returns The status of the stream writer.  If it is complete
 * without error the status should be FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED.
 */
FAS_INTERNAL::FIP_STATUS FiRDIPackedStream::isStreamWriterComplete()
{
    // For the RDI Packed stream once the encoded frames count and frame count are equal
    // test for placeholder frames in case they exist.
    if (configInfo->frameCount == frameSeqNum)
    {
        // We are complete.
        streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED;
    }

    // At this point return the actual stream status.
    return streamStatus;
}

/*!
 * @brief Closes the video stream and releases any allocated resources.
 *
 * @returns The video stream status setting of the object.
 */
FAS_INTERNAL::FIP_STATUS FiRDIPackedStream::closeStream()
{
    // Did this stream complete?
    bool looksLikeWeAreCompleted = false;

    // Determine if we actually completed writing the data.
//    FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: frameCount=%d, frameSeqNum=%d, status=%d", FI_FUNCTION, configInfo->frameCount, frameSeqNum, streamStatus);
    if (configInfo->frameCount == frameSeqNum)
    {
        // Should not happen but if it does catch it.
        if ((streamStatus == FAS_INTERNAL::FIP_STATUS::STREAM_READY) ||
            (streamStatus == FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED))
        {
            streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED;
            looksLikeWeAreCompleted = true;
        }
    }

    // Handle marking this video as incomplete if necessary.
    if (!looksLikeWeAreCompleted)
    {
        // Make sure this stream is marked incomplete if still ready.
        if (streamStatus == FAS_INTERNAL::FIP_STATUS::STREAM_READY)
        {
            streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_INCOMPLETE;
        }

        std::string videoNameWithErrorTag = streamName + std::string(".incomplete.txt");
        std::string statusLog;
        FI_LOG::FiLogger::factory()->fi_log(LOG_ERR, "%s: RDI Packed streamStatus is not marked complete.  Writing %s to document incomplete status.", FI_FUNCTION, videoNameWithErrorTag.c_str());

        // using ofstream constructors.
        std::ofstream outfile(videoNameWithErrorTag.c_str());
        outfile << "Packed stream is incomplete and was not fully processed." << std::endl;
        outfile << "Please review log data for possible error conditions." << std::endl;
        outfile << "Status:" << statusLog << std::endl;
        outfile.close();
    }

    // Make sure the ".div" file is closed.
    if (divFile.is_open())
        divFile.close();

    // Set the stream status of this object as being closed.
    streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_CLOSED;

    // The stream status of the object should be FAS_INTERNAL::FIP_STATUS::STREAM_CLOSED
    return streamStatus;
}

/*!
 * @brief Determines the ".div" file for the specified image (imageFrameNbr) and the
 *        offset within that file.
 *
 * @param [in]imageFrameNbr     Zero-based image frame number for the image frame to replace.
 * @param [out]divFileIndex     Number (1 based) of the ".div" file associated with the imageFrameNbr.
 * @param [out]divFileOffset    Offset within the ".div" file for the image data.
 */
void FiRDIPackedStream::calcDivFileInfo(uint32_t imageFrameNbr, uint32_t& divFileIndex, uint32_t& divFileOffset)
{
    // Calculate the DIV sequence number based on the current image frame sequence number.
    // Image frame sequence numbers are 0-based, but sequence numbers used for DIV
    // filenames are one-based.
    divFileIndex = (imageFrameNbr / imagesPerDiv) + 1;

    // Calculate the offset into that DIV file for the image location.
    divFileOffset = imageFrameNbr % imagesPerDiv;
}

/*!
 * @brief Calculate the time for the frame specified by imageFrameNbr.
 *
 * The calculations are based on the start time of the stream (frame 0)
 * using the configInfo->timeStartOfStreamSeconds, configInfo->timeStartOfStreamMicrosec,
 * and configInfo->frameRate.
 *
 * @param [in] imageFrameNbr    The frame number we are interested in.
 * @param [out] timeSec         The seconds portion of the time for frame 'frameSeqNum'.
 * @param [out] timeMicrosec    The microseconds portion of the time for frame 'frameSeqNum'.
 */
void FiRDIPackedStream::calcFrameTime(uint32_t imageFrameNbr, uint64_t& timeSec, uint32_t& timeMicrosec)
{
    double timeCalc;
    double timeToFrame;

    // Get the start of the stream.
    timeCalc = (double)configInfo->timeStartOfStreamSeconds + (((double)configInfo->timeStartOfStreamMicrosec) / 1000000.0);

    // Figure out the time for the current frame based on the frame rate.
    timeToFrame = imageFrameNbr * (1.0 / ((double)configInfo->frameRate));

    // Now figure out the stream time.
    timeCalc += timeToFrame;

    // Convert back to seconds.
    timeSec = (uint64_t)trunc(timeCalc);

    // Remove the seconds part.
    timeCalc -= (double)timeSec;

    // Bring up the microseconds and convert to an integer, rounding down.
    timeCalc *= 1000000.0;
    timeMicrosec = (uint32_t)trunc(timeCalc);
}
