/*!
 * @file fi_message.cpp
 * @brief The main implementation file for the Message class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "fas_thread.h"
#include "fi_exception.h"
#include "fi_message.h"

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} /* namespace FAS_THREAD */

using namespace FAS_THREAD;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*!
 * @brief Base Message class for inter-thread communication.
 *
 * Messages may be passed among Threads in whatever means is convenient. Threads
 * may explicitly pass Messages to one another by using the receiving Thread's
 * messageAdd() method.
 *
 * Alternatively, a Message's originator can fill in the list of Threads through
 * which a Message should pass, in turn. The Message's own forward() method will
 * pass the Message to the next Thread's input queue. When the route has been
 * completed (no Threads remain in the route), forward() will return the Message
 * to its owner.
 */
Message::Message() :
    req(NOP),
    resp(NONE),
    owner(nullptr)
{
    route = std::unique_ptr<std::stack<FAS_THREAD::MsgThread *>>(new std::stack<FAS_THREAD::MsgThread *>());
    stream = std::unique_ptr<std::ostream>();
}

Message::~Message()
{
}

/*!
 * @brief Forward this Message to next Thread.
 *
 * Forward this Message to the next Thread given in the 'route' list,
 * or (if the route list is empty), to the Message owner.
 */
void Message::forward()
{
    // Are we at the end of the routing list, or is there
    // a fault?
    if (route->empty() || (resp == FAS_THREAD::FAULT))
    {
        // Make sure we have an owner.
        if (owner == nullptr)
        {
            FI_LOGIC_ERROR("forward() with empty route and nullptr owner");
        }
        else
        {
            // Send to owner
            owner->messageAdd(this);
        }
    }
    else
    {
        // Get the first element from the queu.
        MsgThread *t = route->top();
        route->pop();

        // Add this message to the thread's queue.
        t->messageAdd(this);
    }
}

