/*!
 * @file fi_msg_thread.cpp
 * @brief The implementation of the MsgThread class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#ifndef _WIN32
// For the abi::__forced_unwind exception.
#include <cxxabi.h>
#endif

#include <chrono>

#include "fas_thread.h"
#include "log_if.h"

namespace FAS_THREAD
{

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} /* namespace FAS_THREAD */

using namespace FAS_THREAD;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*! @brief Base class for Fastec MsgThreads.
 *
 * A Fastec thread incorporates a POSIX thread (pThread) and a message
 * queue. Any time the message queue is empty, the thread will not execute.
 * When one or more messages reside in the queue, the MsgThread may execute
 * to process them, subject to scheduling constraints and policies of the
 * host system.
 *
 * Classes derived from FAS_THREAD::MsgThread must define a MsgThread::messageHandle()
 * method, which processes one message taken from the queue.
 */
MsgThread::MsgThread() :
    Thread(),
    startup_ok(true),
    messagesIn(0),
    messagesOut(0),
    halted(false)
{
    // Create all our dynamically allocated objects.
    msg_mutex = std::unique_ptr<std::mutex>(new std::mutex());
    startup_mutex = std::unique_ptr<std::mutex>(new std::mutex());
    startup_condition = std::unique_ptr<std::condition_variable>(new std::condition_variable());
    jobs = std::unique_ptr<std::condition_variable>(new std::condition_variable());
    queue = std::unique_ptr<std::queue<FAS_THREAD::Message *>>(new std::queue<FAS_THREAD::Message *>());
}

/*!
 * @brief Standard destructor.
 */
MsgThread::~MsgThread()
{
}

/*!
 * @brief The thread is not actually started until this method is
 *    called. This method does all the setup and creation of
 *    the thread.
 */
void MsgThread::start(FAS_THREAD::THREAD_STARTUP threadFn)
{
    // If a starting function was supplied, use that. Otherwise, we'll use the
    // default.
    if (threadFn == nullptr)
        Thread::start(static_cast<THREAD_STARTUP>(MsgThread::threadEntry));
    else
        Thread::start(threadFn);
}

/*!
 * @brief Hold at thread startup until released.
 *
 * This is intended for use by a thread's startup function. MsgThread
 * execution will block here until released by a call to the
 * @ref release "release()" method.
 */
void MsgThread::hold()
{
    // Wait here for the constructor to finish initializing stuff.
    std::unique_lock<std::mutex> locker(*startup_mutex);
    while (!startup_ok)
        startup_condition->wait(locker);
}


/*!
 * @brief Release thread for execution.
 *
 * This is intended for use by sub-class constructors that have
 * created a thread using the _hold_ = true parameter. This will
 * allow the new thread, blocked in @ref hold "hold()", to continue
 * execution. Sub-class constructors can use this to prevent
 * threads' _run()_ methods from accessing derived class members
 * before they are initialized.
 */
void MsgThread::release()
{
    std::unique_lock<std::mutex> locker(*startup_mutex);
    startup_ok = true;
    startup_condition->notify_one();
}

/*!
 * @brief Wait for a message to arrive.
 * @param delay_ms  The amount of time to wait, in milliseconds.
 *          A value of 0 means to wait forever.
 * @returns Pointer to the message if a message was received, nullptr otherwise.
 */
FAS_THREAD::Message *MsgThread::wait_for_message(int delay_ms)
{
    FAS_THREAD::Message *message = nullptr;

    try
    {
        // Block on Message Queue until a Message arrives.
        // Call handler (which derived classes must provide)
        // for each Message.
        {
            std::unique_lock<std::mutex> locker(*msg_mutex);
            std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
            using ms = std::chrono::milliseconds;

            // If there's nothing in the queue, and we're stopping,
            // then just leave.
            if ((queue->size() == 0) && halted)
                goto bail;

            // Keep looping while there is nothing in the queue.
            while (queue->size() == 0)
            {
                // Do we want to wait for a certain amount of time?
                if (delay_ms > 0)
                {
                    if (jobs->wait_until(locker, now + ms(delay_ms)) == std::cv_status::timeout)
                    {
                        // Got a timeout, bail out of the loop.
                        break;
                    }
                }
                else
                {
                    jobs->wait(locker);
                }
            }

            // Is there a message on the queue?
            if (queue->size() > 0)
            {
                // Grab the first message off the queue.
                message = queue->front();
                queue->pop();
            }
            else
            {
                // No message.
                message = nullptr;
            }
        } // End of block releases locker

#if !defined(__APPLE__) && !defined(_WIN32)
    }
    catch (abi::__forced_unwind&)
    {
        // thread termination support (?)
        throw;
#endif // __APPLE__
    }
    catch (Exception& e)
    {
        // This is the end of the line for exceptions in this
        // MsgThread. Send the exception message to the system
        // Logger, and soldier on.
        e << "\n  Unhandled Fastec exception reached root of MsgThread";
        fi_log(LOG_ERR, "\n  %s", e.what().c_str());
    }
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, e.what());
        fi_log(LOG_ERR, "  Unhandled C++ exception reached root of MsgThread");
    }
    catch (...)
    {
        fi_log(LOG_ERR, "Unhandled exception of unknown type reached root of MsgThread");
    }

bail:
    return message;
}

/*!
 * @brief MsgThread entry point
 *
 * The base implementation blocks on the MsgThread's Queue, and passes
 * Messages from the Queue to the messageHandle() method as they arrive.
 * Derived classes may override this method if they need different
 * behavior.
 */
void MsgThread::run()
{
    for (;;)
    {
        FAS_THREAD::Message *message;

        // Wait (forever) for a message.
        message = wait_for_message();

        // Sanity check the return value.
        if (message == nullptr)
        {
            fi_log(LOG_DEBUG, "Unexpected nullptr from wait_for_message");
        }

        // Are we the owner of the message?
        else if (message->owner == this)
        {
            // Handle the disposal of this message.
            ownerMessage(message);
        }
        else
        {
            // Go process the message.
            messageHandle(message);
        }

        // Are we done?
        if ((queue->size() == 0) && halted)
            break;
    }
}

/*!
 * @brief Send a message to this MsgThread.
 *
 * Append a message to the MsgThread's queue.
 * This method executes in the calling MsgThread's context.
 *
 * @param message A copy of the message is added to the MsgThread's queue.
 */
void MsgThread::messageAdd(FAS_THREAD::Message *message)
{
    std::unique_lock<std::mutex> locker(*msg_mutex);
    if (halted)
    {
        // See if we can dispose of this Message
        if ((message->owner != this) && (message->owner != nullptr))
            message->forward();
        // FI_RUNTIME_ERROR("Message received after HALT");
        // Exit scope on exception releases locker
        fi_log(LOG_DEBUG, "Message received after HALT");
        return;
    }

    if (message->req == HALT)
        halted = true;

    if (queue->size() == 0)
    {
        // Add message and signal thread.
        queue->push(message);
        messagesIn++;
        jobs->notify_one();
    }
    else
    {
        // Just add the message
        messagesIn++;
        queue->push(message);
    }
}

/*!
 * @brief Default processing if we are the owner of the message
 *    is to delete the message.
 */
void MsgThread::ownerMessage(FAS_THREAD::Message *message)
{
    // See if we are the owner of the message.
    if (message->owner == this)
    {
        // Delete the message. We're done with it.
        delete message;
    }
}

std::string FAS_THREAD::MsgThread::reqToString(int req_code)
{
    std::string req_string;

    // Figure out the conversion.
    // Parse the message.
    switch (req_code)
    {
    case FAS_THREAD::DUMP:
        req_string.append("FAS_THREAD::DUMP");
        break;
    case FAS_THREAD::HALT:
        req_string.append("FAS_THREAD::HALT");
        break;
    case FAS_THREAD::NOP:
        req_string.append("FAS_THREAD::NOP");
        break;
    default:
        req_string.append("unknown");
        break;
    }

    // Return our string.
    return req_string;
}

