/*!
 * @file fi_ticker.cpp
 * @brief The implementation file for the Ticker class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#ifndef _WIN32
// For the abi::__forced_unwind exception.
#include <cxxabi.h>
#endif

#include <errno.h>
#include <signal.h>

#include <chrono>

#include "fi_ticker.h"
#include "fas_thread.h"
#include "fas_log.h"

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

} /* namespace FAS_THREAD */

using namespace FAS_THREAD;

/**
 * @brief Ticker class
 *
 * The ticker arranges to call its tick() method at regular (periodic) intervals.
 * The interval is given as a parameter to the constructor. The Ticker establishes
 * its own independent thread of execution, and executes the tick()
 * method on that thread. Ticker is a virtual class. A derived class must implement
 * the virtual tick() method.
 *
 * A ticker is initially stopped, and must be started by a call to the start()
 * before tick() executions will begin.
 */

Ticker::Ticker(uint32_t period) :
    Thread(),
    interval(period),
    paused(false),
    single_shot(false),
    stopThread(false)
{
}

Ticker::~Ticker()
{
    // If the thread is still running, make sure we stop it.
    if (!stopThread)
        stop();
}

void Ticker::setTime(uint32_t period)
{
    std::unique_lock<std::mutex> locker(timerMutex);

    // Change our interval.
    interval = period;
    timerCond.notify_one();
}

void Ticker::stop()
{
    // Pass it up.
    Thread::stop();

    std::unique_lock<std::mutex> locker(timerMutex);
    stopThread = true;
    timerCond.notify_one();
}

void Ticker::run()
{
    try
    {
        bool done(false);           // Flag to indicate when we are done.

        // Let folks know we've started.
        notify_ready();

        // Loop until told to stop.
        while (!done)
        {
            uint32_t old_interval(interval);    // Remember the interval before we go to sleep.

            {
                std::unique_lock<std::mutex> locker(timerMutex);

                // Calculate when the wait will be done.
                using microsec = std::chrono::microseconds;
                std::chrono::system_clock::time_point tick_time = std::chrono::system_clock::now() + microsec(interval);

                // Wait for our timer to expire, or the request comes through to stop the thread.
                done = timerCond.wait_until(locker, tick_time, [this]() { return stopThread; });
            }

            // Make sure we're not supposed to quit.
            if (!done && !paused && (old_interval == interval))
            {
                // Time to TICK.
                tick();
            }

            // Are we single-shot?
            if (single_shot)
                done = true;
        }
#if !defined(__APPLE__) && !defined(_WIN32)
    }
    catch (abi::__forced_unwind&)
    {
        // thread termination support (?)
        throw;
#endif // __APPLE__
    }
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, e.what());
        fi_log(LOG_ERR, "  Unhandled C++ exception reached root of Ticker");
    }
    catch (...)
    {
        fi_log(LOG_ERR, "Unhandled exception of unknown type reached root of Ticker");
    }
}
