
/*!
 * @file ffmpeg_container.cpp
 * @brief Manage FFmpeg video container object.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <iostream>
#include <string>
#include <memory>
#include <mutex>

#include <fmt/format.h>

#ifdef  __cplusplus
extern "C" {
#endif

#include "libavformat/avformat.h"
#include "libavformat/avio.h"
#include "libavutil/pixfmt.h"
#include "libavutil/opt.h"
#include "libavcodec/avcodec.h"
#ifdef  __cplusplus
}
#endif

#include "fas_video.h"
#include "fi_stream.h"
//#include "ffmpeg_io.h"
#include "ffmpeg_encoder.h"
#include "ffmpeg_status.h"

#include "ffmpeg_container.h"

#include "log_if.h"

namespace FI_FFMPEG
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! If you want to turn on verbose logging for this object set this to 1
//static const bool VERBOSE_SAVE_LOG = true;
const bool VERBOSE_SAVE_LOG = false;

//! A threshold value to determine when too many frames have been lost in a row
//! if this value is reached the entire stream will be considered too degraded
//! to be viable and the error condition will occur terminating the stream.
//const uint32_t MAX_LOST_FRAME_GAP  = 10;   // 10 lost frames in a row are fatal right now.

//! This label will precede any FFmpeg av_log entries to identify it as
//! coming from the FFmpeg library.
const char FFMPEG_LOG_LABEL[] = "[FFMPEG AVLOG]";

//! This setting determines the FFmpeg log level used for its log messages.
#ifdef USE_VERBOSE_INFO_DEBUG
const int REPORT_FFMPEG_AV_LOG_LEVEL = AV_LOG_DEBUG;
#else
const int REPORT_FFMPEG_AV_LOG_LEVEL = AV_LOG_ERROR;
#endif
// For your information - Changing it to Error it was Verbose.
//#define AV_LOG_ERROR    16
//#define AV_LOG_WARNING  24
//#define AV_LOG_INFO     32
//#define AV_LOG_VERBOSE  40
//#define AV_LOG_DEBUG    48
//#define AV_LOG_TRACE    56

//! This constant can be set if you want to force a software encoder only.
const bool forceLIBX264 = false;

} // End of FI_FFMPEG

// Make things easier for usage.
using namespace std;
using namespace FAS_VIDEO;
using namespace FI_FFMPEG;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------
extern std::mutex theFFmpegLogMutex;
std::mutex theFFmpegLogMutex;

//! Must be thread-safe!
static void log_callback_report(void *ptr, int level, const char *fmt, va_list vl)
{
    std::lock_guard<std::mutex> guard(theFFmpegLogMutex);

    va_list vl2;
    char line[2048];
    static int print_prefix = 1;

    va_copy(vl2, vl);
    av_log_default_callback(ptr, level, fmt, vl);
    av_log_format_line(ptr, level, fmt, vl2, line, sizeof(line), &print_prefix);
    va_end(vl2);
    if (REPORT_FFMPEG_AV_LOG_LEVEL >= level)
    {
        // Get rid of extra newline character at the end of these lines.
        int len = (int)strlen(line);
        int x;
        bool replacedLastOne = false;
        for (x = len - 1; !replacedLastOne && (x >= 0); x--)
        {
            if (line[x] == 0x0a)
            {
                line[x] = ' ';
                replacedLastOne = true;
            }
        }

        fi_log(LOG_INFO, "%s: %s", FFMPEG_LOG_LABEL, line);
    }
}

static void setup_ffmpeg_logfile()
{
    av_log_set_level(REPORT_FFMPEG_AV_LOG_LEVEL);
    av_log_set_callback(log_callback_report);
}

FFmpegContainer::FFmpegContainer() :
    videoNameForContainer(""),
    containerType(CONTAINER_TYPE::NO_CONTAINER),
    containerOp(CONTAINER_OPS::NO_OPERATION),
    containerStatus(),
    videoFrameToProcess(nullptr),
    inFIVideoFrameFormat(FAS_VIDEO::IMAGE_FORMAT::FI_INVALID_IMAGE),
    hwFFmpegDeviceFmt(AV_PIX_FMT_NONE),
    inFFMpegVideoFrameFmt(AV_PIX_FMT_NONE),
    outFFMpegVideoFrameFmt(AV_PIX_FMT_NONE),
    frameCount(0),
    frameRate(0),
    fullFrameSize(0),
    framesProcessed(0),
    totalFramesLost(0),
    crfValue(22),
    presetSetting("medium"),
    keyframeInterval(FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL),
    containerIsClosed(true)
{
    fi_log(LOG_DEBUG, "%s: Constructing.", __PRETTY_FUNCTION__);
}

FFmpegContainer::~FFmpegContainer()
{
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_DEBUG, "%s: Destructing.", __PRETTY_FUNCTION__);
    }

    // This shouldn't be the case but check.
    if (!containerIsClosed)
    {
        fi_log(LOG_DEBUG, "%s: Container wasn't Closed Properly!", __PRETTY_FUNCTION__);
        CloseDown();
    }
}

/*!
 * @brief Factory to create an FFmpeg Container for MP4 video.
 *
 * @param firstVideoFrame   An FiImage containing the first video frame for video.
 * @param frameRateToUse    How many frames per second for a frame rate is requested.
 * @param frameCountToUse   How many frames are expected to be received for video.
 * @param crfValue          What is the H.264 Constant Rate Factor for frame quality.
 * @param presetSettingStr  What is the H.264 Preset value.
 *
 * @returns A std::unique_ptr<FI_FFMPEG::FFmpegContainer> created.
 */
std::shared_ptr<FI_FFMPEG::FFmpegContainer> FFmpegContainer::CreateMP4Container(const std::string& videoNameToCreate,
                                                                                std::shared_ptr<FAS_VIDEO::FiImage> firstVideoFrame,
                                                                                uint32_t frameRateToUse,
                                                                                uint32_t frameCountToUse,
                                                                                uint32_t crfValueToSet,
                                                                                const char *presetValueToSet,
                                                                                uint32_t keyframeIntervalToUse,
                                                                                bool swOverrideFlag,
                                                                                uint32_t lostFrameCountInFront)
{
    std::shared_ptr<FI_FFMPEG::FFmpegContainer> ffmpegContainerPtr;
    std::string fiFrameImageTypeStr ;
    std::string myName = "[CreateMP4Container]";

    try
    {

        // Validate firstVideoFrame is allocated.
        if (!firstVideoFrame)
        {
            fi_log(LOG_ERR, "%s: firstVideoFrame empty!", __PRETTY_FUNCTION__);
            goto exitCreateMP4Container;
        }

        fiFrameImageTypeStr = FAS_VIDEO::FiVideoStream::GetStringForFiImageID(firstVideoFrame->GetImageFormat());

#ifdef USE_VERBOSE_INFO_DEBUG
        {
            const char *avutilVersionVerbose = "";
            const char *avutilConfiguration = "";
            const char *avutilLicense = "";
            unsigned int avutilVersion = 0;

            avutilVersion = avutil_version();
            avutilVersionVerbose = av_version_info();
            avutilConfiguration = avutil_configuration();
            avutilLicense = avutil_license();

            fi_log(LOG_DEBUG, "%s: Name[%s] Framerate[%d] FrameCount[%d] Keys[%d] FramesIn[%s]",
                   myName.c_str(), videoNameToCreate.c_str(),
                   frameRateToUse, frameCountToUse,
                   keyframeIntervalToUse, fiFrameImageTypeStr.c_str());

            fi_log(LOG_DEBUG, "\n**%s:\n**[FFmpeg Version: %d %s]\n**[FFmpeg Configuration: %s]\n**[FFmpeg License: %s]\n",
                   myName.c_str(), avutilVersion, avutilVersionVerbose,
                   avutilConfiguration, avutilLicense);
        }
#endif

        // Create the ffmpegContainer.
        ffmpegContainerPtr = std::shared_ptr<FI_FFMPEG::FFmpegContainer>(new FI_FFMPEG::FFmpegContainer());

        // If unable to create the ffmpegContainer warn and exit.
        if (!ffmpegContainerPtr)
        {
            fi_log(LOG_ERR, "%s: Unable to create ffmpegContainer object.", __PRETTY_FUNCTION__);
            goto exitCreateMP4Container;
        }

        // Validate firstVideoFrame is allocated.
        if (!firstVideoFrame)
        {
            fi_log(LOG_ERR, "%s: firstVideoFrame empty!", __PRETTY_FUNCTION__);
            goto exitCreateMP4Container;
        }

        // Set the type of container.
        ffmpegContainerPtr->containerType = FI_FFMPEG::CONTAINER_TYPE::MP4_CONTAINER;

        // Verify the image input format is compatible with the container type.
        switch (firstVideoFrame->GetImageFormat())
        {
        case FAS_VIDEO::FI_TIF_IMAGE:
            ffmpegContainerPtr->containerOp = CONTAINER_OPS::ENCODER;
            break;
        case FAS_VIDEO::FI_BMP_IMAGE:
        case FAS_VIDEO::FI_BMP_PLANAR_IMAGE:
            ffmpegContainerPtr->containerOp = CONTAINER_OPS::ENCODER;

            // NOTE:  Important.  There is an order implied here of preference
            // with regard to the HW encoder of choice when more than one is
            // available on a system.  We prefer VAAPI over QSV.  Do not move
            // this order for this reason.

            // Is a VAAPI HW Encoder available?  If so set up to use it.
            if (FFmpegEncoder::IsVAAPIAvailable())
            {
                ffmpegContainerPtr->hwFFmpegDeviceFmt = AV_PIX_FMT_VAAPI;
                ffmpegContainerPtr->outFFMpegVideoFrameFmt = AV_PIX_FMT_NV12;
            }
            // Is an Intel QSV HW Encoder available?  If so set up to use it.
            else if (FFmpegEncoder::IsQSVAvailable())
            {
                ffmpegContainerPtr->hwFFmpegDeviceFmt = AV_PIX_FMT_QSV;
                ffmpegContainerPtr->outFFMpegVideoFrameFmt = AV_PIX_FMT_NV12;
            }
            // Is an NVIDIA HW Encoder available?  If so set up to use it.
            else if (FFmpegEncoder::IsNVIDIAAvailable())
            {
                ffmpegContainerPtr->hwFFmpegDeviceFmt = AV_PIX_FMT_CUDA;
                ffmpegContainerPtr->outFFMpegVideoFrameFmt = AV_PIX_FMT_NV12;
            }
            // Is an AMD AMF HW Encoder available?  If so set up to use it.
            else if (FFmpegEncoder::IsAMFAvailable())
            {
                // NOTE: The constant H264_AMD_AV_PIX_FMT can be different see its declaration.
                ffmpegContainerPtr->hwFFmpegDeviceFmt = H264_AMD_AV_PIX_FMT;
                // Both of the device types support source input of NV12.
                ffmpegContainerPtr->outFFMpegVideoFrameFmt = AV_PIX_FMT_NV12;
            }
            // Use the libx264 software encoder.
            else
            {
                ffmpegContainerPtr->hwFFmpegDeviceFmt = AV_PIX_FMT_NONE;
                ffmpegContainerPtr->outFFMpegVideoFrameFmt = AV_PIX_FMT_YUV420P;
            }

            // If forcing a software encoding pathway override any HW setting from above.
            if (forceLIBX264 || swOverrideFlag)
            {
                ffmpegContainerPtr->hwFFmpegDeviceFmt = AV_PIX_FMT_NONE;
                ffmpegContainerPtr->outFFMpegVideoFrameFmt = AV_PIX_FMT_YUV420P;
            }
            break;
        default:
            break;
        }

        // Verify that the container operation is valid.
        if (ffmpegContainerPtr->containerOp == CONTAINER_OPS::NO_OPERATION)
        {
            fi_log(LOG_ERR, "%s: Missing container operation.", __PRETTY_FUNCTION__);
            goto exitCreateMP4Container;
        }

        // Save information about the video frames coming to this video stream.
        ffmpegContainerPtr->videoNameForContainer = videoNameToCreate;
        ffmpegContainerPtr->inFIVideoFrameFormat = firstVideoFrame->GetImageFormat();
        ffmpegContainerPtr->inFFMpegVideoFrameFmt = FFmpegContainer::GetFFmpegPixelFormat(firstVideoFrame->GetFourCC());
        ffmpegContainerPtr->frameRate = frameRateToUse;
        ffmpegContainerPtr->frameCount = frameCountToUse;
        ffmpegContainerPtr->crfValue = crfValueToSet;
        ffmpegContainerPtr->presetSetting = presetValueToSet;
        ffmpegContainerPtr->keyframeInterval = keyframeIntervalToUse;
        ffmpegContainerPtr->videoFrameToProcess = firstVideoFrame;

        {
            std::string hwPixelFmtStr = FFmpegContainer::GetStringForAVPixelFormat(ffmpegContainerPtr->hwFFmpegDeviceFmt);
            std::string inPixelFmtStr = FFmpegContainer::GetStringForAVPixelFormat(ffmpegContainerPtr->inFFMpegVideoFrameFmt);
            std::string outPixelFmtStr = FFmpegContainer::GetStringForAVPixelFormat(ffmpegContainerPtr->outFFMpegVideoFrameFmt);
            // Mark in the log the information after analysis of frame type and hw availability.
            fi_log(LOG_INFO, "%s: [%s] frames HW[%s] IN[%s] OUT[%s].",
                   myName.c_str(), fiFrameImageTypeStr.c_str(),
                   hwPixelFmtStr.c_str(),
                   inPixelFmtStr.c_str(),
                   outPixelFmtStr.c_str());
        }

        // Let the container prepare its internal objects.
        if (!ffmpegContainerPtr->PrepareInternalObjects(lostFrameCountInFront))
        {
            fi_log(LOG_DEBUG, "%s: Unable to prepare the FFmpeg container.", myName.c_str());
            ffmpegContainerPtr.reset();
            goto exitCreateMP4Container;
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught creating MP4 Streamer.", __PRETTY_FUNCTION__, e.what());
        goto exitCreateMP4Container;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: Exception (...) caught creating MP4 Streamer.", __PRETTY_FUNCTION__);
        goto exitCreateMP4Container;
    }

exitCreateMP4Container:

    if (!ffmpegContainerPtr)
        fi_log(LOG_ERR, "%s: Leaving and returning an empty ffmpegContainerPtr.", myName.c_str());
    else
        fi_log(LOG_DEBUG, "%s: Leaving and returning an valid ffmpegContainerPtr.", myName.c_str());

    return ffmpegContainerPtr;
}

/*!
 * @brief Factory to create an FFmpeg Container object for ProRes Video.
 *
 * @param imageinType   What type of image data is expected for input.
 * @param imageOutType  What type of image data is expected for output.
 *
 * @returns A std::shared_ptr<FI_FFMPEG::FFmpegContainer> containing the
 * newly created FFmpegContainer object.
 */
std::shared_ptr<FI_FFMPEG::FFmpegContainer>
FFmpegContainer::CreateProResContainer(std::shared_ptr<FAS_VIDEO::FiImage> firstVideoFrame)
{
    std::shared_ptr<FI_FFMPEG::FFmpegContainer> ffmpegContainerPtr = nullptr;

    // Validate firstVideoFrame is allocated at least.
    if (!firstVideoFrame)
    {
        fi_log(LOG_ERR, "%s: firstVideoFrame empty!", __PRETTY_FUNCTION__);
        goto exitCreateProResContainer;
    }

    fi_log(LOG_DEBUG, "%s: Creating FFmpeg Container for receiving images format[%u].",
           __PRETTY_FUNCTION__, (uint32_t)firstVideoFrame->GetImageFormat());

exitCreateProResContainer:
    return ffmpegContainerPtr;
}

/*!
 * @brief Prepares the internal objects used by this container.
 *
 * @returns A boolean true value if prepared or false if error condition.
 */
bool FFmpegContainer::PrepareInternalObjects(uint32_t framesLostInFront)
{
    bool preparedObjects = false;
    std::string myNameStr = "[FFmpegContainer::PrepareInternalObjects]";

    try
    {
        // Make sure the ffmpeg logging output gets funnelled to our logger.
        setup_ffmpeg_logfile();

        // Prepare the encoder object using the factory.
        std::shared_ptr<FFmpegEncoder> ffmpegEncoder;               //!< The FFmpeg Encoder object for this container.
        ffmpegEncoder = FFmpegEncoder::CreateEncoder(containerType,
                                                     videoNameForContainer,
                                                     hwFFmpegDeviceFmt,
                                                     inFFMpegVideoFrameFmt,
                                                     outFFMpegVideoFrameFmt,
                                                     frameRate,
                                                     frameCount,
                                                     videoFrameToProcess->GetImageWidth(),
                                                     videoFrameToProcess->GetImageHeight(),
                                                     crfValue,
                                                     presetSetting,
                                                     keyframeInterval);
        if (!ffmpegEncoder)
        {
            fi_log(LOG_ERR, "%s: Unable to create encoder for containerType[%d].",
                   myNameStr.c_str(), (uint32_t)containerType);
            containerStatus.set_no_ctx_encoder();
            goto exitPrepareInternalObjects;
        }

        // Make sure that the encoder is in a prepared state.
        if (! ffmpegEncoder->IsPrepared())
        {
            fi_log(LOG_ERR, "%s: Unable to prepare encoder for containerType[%d].",
                   myNameStr.c_str(), (uint32_t)containerType);
            containerStatus.set_no_ctx_encoder();
            goto exitPrepareInternalObjects;
        }

        // The container is prepared mark it as such.
        fi_log(LOG_DEBUG, "%s: Setting the ffmpeg container as prepared.", myNameStr.c_str());
        containerStatus.set_prepared();

        // This is the first frame for the video stream write it out to the video stream.
#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "%s: Encoding the videoFrameToProcess.", myNameStr.c_str());
#endif
        if (!EncodeFrame(videoFrameToProcess, framesLostInFront))
        {
            fi_log(LOG_ERR, "%s: Encoder error while processing videoFrameToProcess.", myNameStr.c_str());
            goto exitPrepareInternalObjects;
        }

        // If any error report the problem and reset the container.
#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "%s: Checking at end to make sure still in prepared state.", myNameStr.c_str());
#endif
        if (!containerStatus.isPrepared())
        {
            fi_log(LOG_ERR, "%s: Unable to encode videoFrameToProcess into FFMpeg stream.", myNameStr.c_str());
            goto exitPrepareInternalObjects;
        }

        // NOTE_JLC - this is how you set up a function to call if you want to
        // have a callback but it's too difficult to utilize - I have left this
        // comment in place to note it:  av_log_set_callback(log_callback_report);
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing FFmpeg containerType[%d].", myNameStr.c_str(), e.what(), (uint32_t)containerType);
        goto exitPrepareInternalObjects;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: Exception (...) caught while preparing FFmpeg containerType[%d].", myNameStr.c_str(), (uint32_t)containerType);
        goto exitPrepareInternalObjects;
    }

    preparedObjects = true;
    containerIsClosed = false;  // It's open now and needs to be cleaned upon exit.

exitPrepareInternalObjects:
    if (!preparedObjects)
    {
        fi_log(LOG_ERR, "%s: Unable to prepare FFmpeg containerType[%d]!", myNameStr.c_str(), (uint32_t)containerType);
    }
    return preparedObjects;
}

/*!
 * @brief Increments the frames that have been processed and sent to
 * the encoder.  It is here that if we hit the number of frames expected
 * that we spawn off the flushframe to the frame writer thread.
 *
 * @returns True if the frames processed was incremented, false if an error condition.
 */
bool FFmpegContainer::IncrementFramesProcessed()
{
    bool allOK = true;

    framesProcessed++;
    //fi_log(LOG_INFO, "%s: framesProcessed[%d] frameCount[%d].", __PRETTY_FUNCTION__, framesProcessed, frameCount);

    if (framesProcessed == frameCount)
    {
        // This means all of the frames for the FFMpeg video stream have been processed through
        // the system.  Go ahead and send a flushing frame to the encoder.
        std::shared_ptr<FFmpegEncoder> ffmpegEncoder;
        ffmpegEncoder = FFmpegEncoder::GetEncoderInstance();
        if (ffmpegEncoder != nullptr)
            allOK = ffmpegEncoder->SendFlushEncoderFrame();
    }
    return allOK;
}

/*!
 * @brief Translates an Fastec Imaging FOURCC code into FFmpeg AVPixelFormat.
 *
 * @param FIFourCC  The Fastec Imaging FOURCC code.
 *
 * @returns A valid AVPixelFormat enumeration or AV_PIX_FMT_NONE if unable to translate.
 */
AVPixelFormat FFmpegContainer::GetFFmpegPixelFormat(uint32_t FIFourCC)
{
    AVPixelFormat theAVPixelFormat = AV_PIX_FMT_NONE;

    // Looking at the known FAS_VIDEO fourcc values make a translation to the known
    // AVPixelFormat equivalent.

    // Notes:  From the test code ffmpeg_image logic
    // TIFF RAW - return (int)AV_PIX_FMT_GRAY16LE;
#if 0
    //! ARGB: packed 8-bit RGBA.
    //! Four bytes per pixel: alpha, red, green, blue.
    const uint32_t FOURCC_ARGB  = 0x42475241

                                  //! @brief BYR2: Raw Bayer data with 12-bits per sample stored
                                  //!    in 16-bit words.
    const uint32_t FOURCC_BYR2  = 0x32525942;
#endif

    switch (FIFourCC)
    {
    case FAS_VIDEO::FOURCC_Y8:
    case FAS_VIDEO::FOURCC_BYR1:
        theAVPixelFormat = AV_PIX_FMT_GRAY8;
        // AV_PIX_FMT_GRAY8,     ///<        Y        ,  8bpp
        break;
    case FAS_VIDEO::FOURCC_RGB:
        //NOTE_JLC - The bitmap frames coming from our cameras are store in BGRBGR format not RGBRGB.
        //theAVPixelFormat = AV_PIX_FMT_RGB24;
        theAVPixelFormat = AV_PIX_FMT_BGR24;
        //AV_PIX_FMT_RGB24,     ///< packed RGB 8:8:8, 24bpp, RGBRGB...
        //AV_PIX_FMT_BGR24,     ///< packed RGB 8:8:8, 24bpp, BGRBGR...
        break;
    case FAS_VIDEO::FOURCC_NV12:
        theAVPixelFormat = AV_PIX_FMT_NV12;
        // NOTE_JLC - double check this value with code in camera server.
        // AV_PIX_FMT_NV12,      ///< planar YUV 4:2:0, 12bpp, 1 plane for Y and 1 plane for the UV components,
        //                       ///< which are interleaved (first byte U and the following byte V)
        // AV_PIX_FMT_NV21,      ///< as above, but U and V bytes are swapped
        break;
    case FAS_VIDEO::FOURCC_Y16:
    case FAS_VIDEO::FOURCC_BYR2:
        theAVPixelFormat = AV_PIX_FMT_GRAY16LE;
        // NOTE_JLC - double check this value.
        // From fas_video.h:
        // Y16: 16-bit greyscale.
        // Only a single sample, 16 bit Y plane for monochrome images.
        // From pixfmt.h:
        //AV_PIX_FMT_GRAY16BE,  ///<        Y        , 16bpp, big-endian
        //AV_PIX_FMT_GRAY16LE,  ///<        Y        , 16bpp, little-endian
        //AV_PIX_FMT_BAYER_BGGR16 AV_PIX_FMT_NE(BAYER_BGGR16BE,    BAYER_BGGR16LE)
        //AV_PIX_FMT_BAYER_RGGB16 AV_PIX_FMT_NE(BAYER_RGGB16BE,    BAYER_RGGB16LE)
        //AV_PIX_FMT_BAYER_GBRG16 AV_PIX_FMT_NE(BAYER_GBRG16BE,    BAYER_GBRG16LE)
        //AV_PIX_FMT_BAYER_GRBG16 AV_PIX_FMT_NE(BAYER_GRBG16BE,    BAYER_GRBG16LE)
        break;
    default:
        fi_log(LOG_ERR, "%s: Unable to translate FI fourcc (0x%08x)", __PRETTY_FUNCTION__, FIFourCC);
        break;
    }
    return theAVPixelFormat;
}

/*!
 * @brief Translates an FFmpeg AVPixelFormat into readable string.
 *
 * @param avPixelFormatId   The FFmpeg AVPixelFormat code.
 *
 * @returns The result of the translation.
 */
const char *FFmpegContainer::GetStringForAVPixelFormat(enum AVPixelFormat avPixelFormatId)
{
    static std::string avPixelFormatStr = fmt::format(FMT_STRING("AvPixelFormat[{:d}] Untranslated."), (uint32_t)avPixelFormatId);

    // Only translate a few of these as list is huge.
    switch (avPixelFormatId)
    {
    case AV_PIX_FMT_YUV420P:
        avPixelFormatStr = "AV_PIX_FMT_YUV420P";
        break;
    case AV_PIX_FMT_YUYV422:
        avPixelFormatStr = "AV_PIX_FMT_YUYV422";
        break;
    case AV_PIX_FMT_RGB24:
        avPixelFormatStr = "AV_PIX_FMT_RGB24";
        break;
    case AV_PIX_FMT_BGR24:
        avPixelFormatStr = "AV_PIX_FMT_BGR24";
        break;
    case AV_PIX_FMT_YUV422P:
        avPixelFormatStr = "AV_PIX_FMT_YUV422P";
        break;
    case AV_PIX_FMT_YUV444P:
        avPixelFormatStr = "AV_PIX_FMT_YUV444P";
        break;
    case AV_PIX_FMT_GRAY8:
        avPixelFormatStr = "AV_PIX_FMT_GRAY8";
        break;
    case AV_PIX_FMT_NV12:
        avPixelFormatStr = "AV_PIX_FMT_NV12";
        break;
    case AV_PIX_FMT_VAAPI:
        avPixelFormatStr = "AV_PIX_FMT_VAAPI";
        break;
    case AV_PIX_FMT_QSV:
        avPixelFormatStr = "AV_PIX_FMT_QSV";
        break;
    case AV_PIX_FMT_CUDA:
        avPixelFormatStr = "AV_PIX_FMT_CUDA";
        break;
    case AV_PIX_FMT_D3D11VA_VLD:
        avPixelFormatStr = "AV_PIX_FMT_D3D11VA_VLD";
        break;
    case AV_PIX_FMT_D3D11:
        avPixelFormatStr = "AV_PIX_FMT_D3D11";
        break;
    case AV_PIX_FMT_DXVA2_VLD:
        avPixelFormatStr = "AV_PIX_FMT_DXVA2_VLD";
        break;
    case AV_PIX_FMT_NONE:
        avPixelFormatStr = "AV_PIX_FMT_NONE";
        break;
    default:
        break;
    }
    return avPixelFormatStr.c_str();
}

/*!
 * @brief Translates an FFmpeg AVCodecID into readable string.
 *
 * @param avCodecID   The FFmpeg AVCodecID code.
 *
 * @returns The result of the translation.
 */
const char *FFmpegContainer::GetStringForAVCodecID(AVCodecID avCodecID)
{
    static std::string avCodecIDStr = fmt::format(FMT_STRING("AVCodecID[{:d}] Untranslated."), (uint32_t)avCodecID);

    // These are the "known" encoder_ids that are supported in this container.
    switch (avCodecID)
    {
    case AV_CODEC_ID_H264:
        avCodecIDStr = "AV_CODEC_ID_H264";
        break;
    default:
        break;
    }
    return avCodecIDStr.c_str();
}

/*!
 * @brief Log the data inside AVFrame object.
 *
 * @returns It returns nothing.
 */
void FFmpegContainer::LogAVFrameInfo(AVFrame *avFrameToLog,  const char *frameNameStr, bool onlyShowPictTypeInfo)
{
    const char *setFrameName = "No AVFrame Caller Information Provided";
    const char *divider = "------------------------------------------------";
    if (frameNameStr != nullptr)
        setFrameName = frameNameStr;

    std::string pixel_fmt_str;
    std::string pict_type_str;
    std::string color_space_str;
    std::string color_primary_str;

    fi_log(LOG_DEBUG, "%s", divider);
    fi_log(LOG_DEBUG, "LOG AVFrame[%s]", setFrameName);

    // Translate some of the settings into readable strings.
    if (avFrameToLog != nullptr)
    {
        switch (avFrameToLog->pict_type)
        {
        case AV_PICTURE_TYPE_NONE:
            pict_type_str = "None";
            break;
        case AV_PICTURE_TYPE_I:
            pict_type_str = "Intra";
            break;
        case AV_PICTURE_TYPE_P:
            pict_type_str = "Predicted";
            break;
        case AV_PICTURE_TYPE_B:
            pict_type_str = "Bi-dir predicted";
            break;
        case AV_PICTURE_TYPE_S:
            pict_type_str = "S(GMC)-VOP MPEG-4";
            break;
        case AV_PICTURE_TYPE_SI:
            pict_type_str = "Switching Intra";
            break;
        case AV_PICTURE_TYPE_SP:
            pict_type_str = "Switching Predicted";
            break;
        case AV_PICTURE_TYPE_BI:
            pict_type_str = "BI type";
            break;
        default:
            pict_type_str = "Unknown";
            break;
        }

        // Translate known avPixelFormat values.
        enum AVPixelFormat theFmt = AV_PIX_FMT_NONE;
        theFmt = (enum AVPixelFormat) avFrameToLog->format;

        pixel_fmt_str = FFmpegContainer::GetStringForAVPixelFormat(theFmt);

        // Only translate a few of these as list is huge.
        switch (avFrameToLog->colorspace)
        {
        case AVCOL_SPC_RGB:
            color_space_str = "AVCOL_SPC_RGB";
            break;
        case AVCOL_SPC_BT709:
            color_space_str = "AVCOL_SPC_BT709";
            break;
        case AVCOL_SPC_UNSPECIFIED:
            color_space_str = "AVCOL_SPC_UNSPECIFIED";
            break;
        case AVCOL_SPC_RESERVED:
            color_space_str = "AVCOL_SPC_RESERVED";
            break;
        case AVCOL_SPC_FCC:
            color_space_str = "AVCOL_SPC_FCC";
            break;
        case AVCOL_SPC_BT470BG:
            color_space_str = "AVCOL_SPC_BT470BG";
            break;
        case AVCOL_SPC_SMPTE170M:
            color_space_str = "AVCOL_SPC_SMPTE170M";
            break;
        case AVCOL_SPC_SMPTE240M:
            color_space_str = "AVCOL_SPC_SMPTE240M";
            break;
        case AVCOL_SPC_YCGCO:
            color_space_str = "AVCOL_SPC_YCGCO";
            break;
        case AVCOL_SPC_BT2020_NCL:
            color_space_str = "AVCOL_SPC_BT2020_NCL";
            break;
        case AVCOL_SPC_BT2020_CL:
            color_space_str = "AVCOL_SPC_BT2020_CL";
            break;
        case AVCOL_SPC_SMPTE2085:
            color_space_str = "AVCOL_SPC_SMPTE2085";
            break;
        case AVCOL_SPC_CHROMA_DERIVED_NCL:
            color_space_str = "AVCOL_SPC_CHROMA_DERIVED_NCL";
            break;
        case AVCOL_SPC_CHROMA_DERIVED_CL:
            color_space_str = "AVCOL_SPC_CHROMA_DERIVED_CL";
            break;
        case AVCOL_SPC_ICTCP:
            color_space_str = "AVCOL_SPC_ICTCP";
            break;
        case AVCOL_SPC_NB:
            color_space_str = "AVCOL_SPC_NB";
            break;
        default:
            color_space_str = "Unknown";
            break;
        }

        // Only translate a few of these as list is huge.
        switch (avFrameToLog->color_primaries)
        {
        case AVCOL_PRI_RESERVED0:
            color_primary_str = "AVCOL_PRI_RESERVED0";
            break;
        case AVCOL_PRI_BT709:
            color_primary_str = "AVCOL_PRI_BT709";
            break;
        case AVCOL_PRI_UNSPECIFIED:
            color_primary_str = "AVCOL_PRI_UNSPECIFIED";
            break;
        case AVCOL_PRI_RESERVED:
            color_primary_str = "AVCOL_PRI_RESERVED";
            break;
        case AVCOL_PRI_BT470M:
            color_primary_str = "AVCOL_PRI_BT470M";
            break;
        case AVCOL_PRI_BT470BG:
            color_primary_str = "AVCOL_PRI_BT470BG";
            break;
        case AVCOL_PRI_SMPTE170M:
            color_primary_str = "AVCOL_PRI_SMPTE170M";
            break;
        case AVCOL_PRI_SMPTE240M:
            color_primary_str = "AVCOL_PRI_SMPTE240M";
            break;
        case AVCOL_PRI_BT2020:
            color_primary_str = "AVCOL_PRI_BT2020";
            break;
        case AVCOL_PRI_SMPTE428:
            color_primary_str = "AVCOL_PRI_SMPTE428";
            break;
        case AVCOL_PRI_SMPTE431:
            color_primary_str = "AVCOL_PRI_SMPTE431";
            break;
        case AVCOL_PRI_SMPTE432:
            color_primary_str = "AVCOL_PRI_SMPTE432";
            break;
        case AVCOL_PRI_JEDEC_P22:
            color_primary_str = "AVCOL_PRI_JEDEC_P22";
            break;
        case AVCOL_PRI_NB:
            color_primary_str = "AVCOL_PRI_NB";
            break;
        default:
            color_primary_str = "Unknown";
            break;
        }

        // Always output the format, keyframe, picture type and frame number.
        fi_log(LOG_DEBUG, "  format[%d,%s] keyframe[%d,%s] picture type[%d,%s] pkt_dts(frame#)[%" PRId64 " 0x%" PRIx64 "]",
               avFrameToLog->format, pixel_fmt_str.c_str(),
               ((avFrameToLog->flags & AV_FRAME_FLAG_KEY) == AV_FRAME_FLAG_KEY),
               ((avFrameToLog->flags & AV_FRAME_FLAG_KEY) == AV_FRAME_FLAG_KEY) ? "keyframe" : "not keyframe",
               (uint32_t)avFrameToLog->pict_type, pict_type_str.c_str(),
               avFrameToLog->pkt_dts, avFrameToLog->pkt_dts);

        // If user only wanted to see the pict_type line don't log anymore information.
        if (!onlyShowPictTypeInfo)
        {
            // Dump the data pointer for the planes only look at 3.
            fi_log(LOG_DEBUG, "  plane0[0x%" PRIx64 "] plane1[0x%" PRIx64 "] plane2[0x%" PRIx64 "]", ((uint64_t)avFrameToLog->data[0]), ((uint64_t)avFrameToLog->data[1]), ((uint64_t)avFrameToLog->data[2]));
            fi_log(LOG_DEBUG, "  lineSize0[%d] lineSize1[%d] lineSize2[%d] width[%d] height[%d] aspect ratio[%d/%d]",
                   avFrameToLog->linesize[0], avFrameToLog->linesize[1], avFrameToLog->linesize[2],
                   avFrameToLog->width, avFrameToLog->height,
                   avFrameToLog->sample_aspect_ratio.num, avFrameToLog->sample_aspect_ratio.den);
#ifdef USE_VERBOSE_INFO_DEBUG
            fi_log(LOG_DEBUG, "  pts[%lld 0x%llx] quality(1=good to FF_LAMBDA_MAX[%d]=bad)[%d] coded pict#(bitstream order)[%d] display order pict#[%d]",
                   avFrameToLog->pts, avFrameToLog->pts,
                   FF_LAMBDA_MAX,
                   avFrameToLog->quality,
                   avFrameToLog->coded_picture_number, avFrameToLog->display_picture_number);
            fi_log(LOG_DEBUG, "  color range[%d] color primaries[%d, %s] "
                   "color space[%d,%s] chroma location[%d]",
                   avFrameToLog->color_range, avFrameToLog->color_primaries, color_primary_str.c_str(),
                   avFrameToLog->colorspace, color_space_str.c_str(), avFrameToLog->chroma_location);
#endif
        }
    }
    else
    {
        fi_log(LOG_ERR, "%s: AVFrame pointer was a nullptr.", __PRETTY_FUNCTION__);
    }

    fi_log(LOG_DEBUG, "%s", divider);
}

void FFmpegContainer::log_callback_report(void *ptr, int level, const char *fmt, va_list vl)
{
    (void)ptr;
    (void)level;
    (void)fmt;
    (void)vl;
#if 0
#define ff_mutex_init    pthread_mutex_init
#define ff_mutex_lock    pthread_mutex_lock
#define ff_mutex_unlock  pthread_mutex_unlock
#define ff_mutex_destroy pthread_mutex_destroy
#define AVMutex char
#define AV_MUTEX_INITIALIZER 0
    static AVMutex mutex = AV_MUTEX_INITIALIZER;

    // From the log.c file in
    // C:\Users\jcurtis\Documents\sandbox\ffmpeg - old downloads from beginning\ffmpeg\libavutil
    // In QT if we assign this can we mimic console output?
    void av_log_default_callback(void *ptr, int level, const char *fmt, va_list vl)
    {
        static int print_prefix = 1;
        static int count;
        static char prev[LINE_SZ];
        AVBPrint part[4];
        char line[LINE_SZ];
        static int is_atty;
        int type[2];
        unsigned tint = 0;

        if (level >= 0)
        {
            tint = level & 0xff00;
            level &= 0xff;
        }

        if (level > av_log_level)
            return;
        ff_mutex_lock(&mutex);

        format_line(ptr, level, fmt, vl, part, &print_prefix, type);
        snprintf(line, sizeof(line), "%s%s%s%s", part[0].str, part[1].str, part[2].str, part[3].str);

#if HAVE_ISATTY
        if (!is_atty)
            is_atty = isatty(2) ? 1 : -1;
#endif

        if (print_prefix && (flags & AV_LOG_SKIP_REPEATED) && !strcmp(line, prev) &&
            *line && line[strlen(line) - 1] != '\r')
        {
            count++;
            if (is_atty == 1)
                fprintf(stderr, "    Last message repeated %d times\r", count);
            goto end;
        }
        if (count > 0)
        {
            fprintf(stderr, "    Last message repeated %d times\n", count);
            count = 0;
        }
        strcpy(prev, line);
        sanitize(part[0].str);
        colored_fputs(type[0], 0, part[0].str);
        sanitize(part[1].str);
        colored_fputs(type[1], 0, part[1].str);
        sanitize(part[2].str);
        colored_fputs(av_clip(level >> 3, 0, NB_LEVELS - 1), tint >> 8, part[2].str);
        sanitize(part[3].str);
        colored_fputs(av_clip(level >> 3, 0, NB_LEVELS - 1), tint >> 8, part[3].str);

#if CONFIG_VALGRIND_BACKTRACE
        if (level <= BACKTRACE_LOGLEVEL)
            VALGRIND_PRINTF_BACKTRACE("%s", "");
#endif
end:
        av_bprint_finalize(part + 3, NULL);
        ff_mutex_unlock(&mutex);
    }
#endif
}

/*!
 * @brief Validate the FFmpeg stream container.
 *
 * @returns A boolean true if all is ok or false if error.  It will also
 * log any errors to the system log.
 */
bool FFmpegContainer::isFFmpegStreamOK()
{
    bool FFmpegStreamIsOK = false;
    std::shared_ptr<FFmpegEncoder> ffmpegEncoder;

    try
    {
        // Validate the container type.
        // NOTE_JLC add the container type for PRORES when implemented.
        if (containerType != CONTAINER_TYPE::MP4_CONTAINER)
        {
            fi_log(LOG_ERR, "%s: FFmpeg stream has invalid container type[%u].", __PRETTY_FUNCTION__, (uint32_t)containerType);
            containerStatus.set_no_ctx_type();
            goto exitIsFFmpegStreamOK;
        }

        // Validate the container operation.
        if (containerOp == CONTAINER_OPS::NO_OPERATION)
        {
            fi_log(LOG_ERR, "%s: FFmpeg stream has no set container operation.", __PRETTY_FUNCTION__);
            containerStatus.set_no_ctx_ops();
            goto exitIsFFmpegStreamOK;
        }

        // What is the status?
        if (!containerStatus.isPrepared())
        {
            // The container is not valid for use.
            fi_log(LOG_ERR, "%s: FFmpeg stream is not in prepared state.", __PRETTY_FUNCTION__);
            containerStatus.logStatus();
            goto exitIsFFmpegStreamOK;
        }

        // Validate type of image frame data expected.
        if (inFIVideoFrameFormat == FAS_VIDEO::IMAGE_FORMAT::FI_INVALID_IMAGE)
        {
            fi_log(LOG_ERR, "%s: FFmpeg stream has invalid inFIVideoFrameFormat.", __PRETTY_FUNCTION__);
            containerStatus.set_no_img_info();
            goto exitIsFFmpegStreamOK;
        }

        // Is there an FFmpeg Encoder object?
        ffmpegEncoder = FFmpegEncoder::GetEncoderInstance();
        if (!ffmpegEncoder)
        {
            // The container is not valid for use.
            fi_log(LOG_ERR, "%s: FFmpeg stream encoder is missing.", __PRETTY_FUNCTION__);
            containerStatus.set_no_ctx_encoder();
            goto exitIsFFmpegStreamOK;
        }

        // Is the FFmpeg Encoder in prepared state?
        if (!ffmpegEncoder->IsPrepared())
        {
            fi_log(LOG_ERR, "%s: FFmpeg stream encoder is not prepared properly.", __PRETTY_FUNCTION__);
            containerStatus.set_no_ctx_encoder();
            goto exitIsFFmpegStreamOK;

        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught.", __PRETTY_FUNCTION__, e.what());
        goto exitIsFFmpegStreamOK;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught.", __PRETTY_FUNCTION__);
        goto exitIsFFmpegStreamOK;
    }

    // If make it here the FFmpeg stream is ok for operation.
    FFmpegStreamIsOK = true;

exitIsFFmpegStreamOK:
    return FFmpegStreamIsOK;
}

/*!
 * @brief Encodes a FAS_VIDEO::FiImage frame to this stream.
 *
 * @param videoFrameToEncode is a video frame from the system to encode into the stream.
 *
 * @returns A boolean true if all is ok or a false if not.
 */
bool FFmpegContainer::CloseDown()
{
    bool closedDownOk = false;
    try
    {
        std::shared_ptr<FFmpegEncoder> ffmpegEncoder;
        ffmpegEncoder = FFmpegEncoder::GetEncoderInstance();
        if (ffmpegEncoder != nullptr)
        {
            if (VERBOSE_SAVE_LOG)
            {
                fi_log(LOG_DEBUG, "%s: Cleanup FFMpeg Encoder.", __PRETTY_FUNCTION__);
            }
            FFmpegEncoder::CloseEncoderInstance();
            containerIsClosed = true;
        }
        else
        {
            fi_log(LOG_ERR, "%s: FFMpeg Encoder was NULLPTR!", __PRETTY_FUNCTION__);
        }
        closedDownOk = true;
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s.", __PRETTY_FUNCTION__, e.what());
        goto exitCloseDown;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception.", __PRETTY_FUNCTION__);
        goto exitCloseDown;
    }

exitCloseDown:
    return closedDownOk;
}

/*!
 * @brief Encodes a FAS_VIDEO::FiImage frame to this stream.
 *
 * @param videoFrameToEncode is a video frame from the system to encode into the stream.
 * @param duplicateFrameCount is greater than zero if we are filling in for lost frames.
 *
 * @returns A boolean true if all is ok or a false if not.
 */
bool FFmpegContainer::EncodeFrame(std::shared_ptr<FAS_VIDEO::FiImage> videoFrameToEncode, uint32_t duplicateFrameCount)
{
    int ffmpeg_result = 0;
    int encodedFrameCount = 0;
    //FiDurations timings;
    std::string frameInfoStr;
    bool encodedFrames = false;

    int frameDupStart   = -1;
    int frameDupEnd     = -1;
    int framesToDup     = (int)duplicateFrameCount;
    int theValidFrameId = (int)framesProcessed + framesToDup;

    uint32_t framesToAdd = 1;   // There is at least one frame to add.

    // Is there a duplication that needs to happen?
    if (duplicateFrameCount > 0)
    {
        framesToAdd += duplicateFrameCount;  // Want to insert our good frame additionally framesLost times.

        // We are missing previous frames.  Setup the missing frame numbers for the log.
        frameDupStart   = (int)(framesProcessed);
        frameDupEnd     = frameDupStart + (int)duplicateFrameCount - 1;

        if (VERBOSE_SAVE_LOG)
        {
            fi_log(LOG_DEBUG, "%s: duplicateFrameCount[%d] framesProcessed[%d] frameDupStart[%d] frameDupEnd[%d].",
                   __PRETTY_FUNCTION__, duplicateFrameCount, framesProcessed, frameDupStart, frameDupEnd);
        }
    }

    AVFrame *ffmpegAVFrame = nullptr;
    FRAME_PLANAR_DATA framePlanarDataRec = {0, {0}, {nullptr}};
    const uint8_t *frameDataPtr = nullptr;
    uint32_t frameDataSize = 0;
    uint32_t rowIndex = 0;

    try
    {
        frameInfoStr = fmt::format(FMT_STRING("Post EncodeFrame Frame[{:d}]"), framesProcessed);

        // Is there a valid videoFrameToEncode?
        if (!videoFrameToEncode)
        {
            fi_log(LOG_ERR, "%s: Video frame to encode was nullptr.", __PRETTY_FUNCTION__);
            goto exitEncodeFrame;
        }

        // If the frame range has been processed and the caller is still
        // trying to encode frames they didn't set up this stream writer
        // correctly.
        if (framesProcessed == frameCount)
        {
            fi_log(LOG_ERR, "%s: Video frames processed[%d] == frameCount[%d].",
                   __PRETTY_FUNCTION__, framesProcessed, frameCount);
            goto exitEncodeFrame;
        }

        // Need to add this frame framesToAdd times.
        while (framesToAdd > 0)
        {
            // Create the FFmpeg AVFrame object necessary for FFmpeg encoder.
            ffmpegAVFrame = av_frame_alloc();

            // Was the system able to create the FFmpeg AVFrame object?
            if (ffmpegAVFrame == nullptr)
            {
                fi_log(LOG_ERR, "%s: av_frame_alloc() FAILED.", __PRETTY_FUNCTION__);
                ffmpeg_result = AVERROR(ENOMEM);
                containerStatus.set_ffmpeg_err(ffmpeg_result, true, __PRETTY_FUNCTION__);
                goto exitEncodeFrame;
            }

            // Set the AVPixelFormat for the input data.
            ffmpegAVFrame->format = inFFMpegVideoFrameFmt;
            ffmpegAVFrame->width = (int) videoFrameToEncode->GetImageWidth();
            ffmpegAVFrame->height = (int) videoFrameToEncode->GetImageHeight();

            // The frame data needed for MP4 encoding is top-down.
            frameDataPtr = videoFrameToEncode->GetAttachedFrameData(frameDataSize,
                                                                    framePlanarDataRec,
                                                                    FAS_VIDEO::FRAME_ORIENTATION::TOP_DOWN);

            // If there are no image planes this frame is not eligible for encoding.
            if (!frameDataPtr || framePlanarDataRec.planeCount == 0)
            {
                fi_log(LOG_ERR, "%s: Missing planar data from image frame.", __PRETTY_FUNCTION__);
                av_frame_free(&ffmpegAVFrame);
                goto exitEncodeFrame;
            }

            // Using the planar data settings fill in the AVFrame to hold our localized image data.
            // This should preclude any image copy operations by re-using what we already have in hand.
            for (rowIndex = 0; rowIndex < framePlanarDataRec.planeCount; rowIndex++)
            {
                // SANITY CHECK!  Make sure plane 0 is identical to the start of the frame.
                if (rowIndex == 0)
                {
                    const uint8_t *plane0Ptr = framePlanarDataRec.planarData[0];
                    if (plane0Ptr != frameDataPtr)
                    {
                        fi_log(LOG_ERR, "%s: Attached Frame Error!! Plane 0[0x%" PRIx64 "] != frameDataPtr[0x%" PRIx64 "] rowsize[%u] planeCount[%u].",
                               __PRETTY_FUNCTION__, ((uint64_t)plane0Ptr), ((uint64_t)frameDataPtr), framePlanarDataRec.rowsizePerPlane[rowIndex], framePlanarDataRec.planeCount);

                        // Frame planar data is not properly setup.
                        ffmpeg_result = AVERROR(EINVAL);
                        goto exitEncodeFrame;
                    }
                }

                //#COMMENT_JLC - comment this out for now.
                //fi_log(LOG_INFO, "EncodeFRAME: Plane[%d] rowsize[%d] frameDataPtr[%p] rowPtr[%p].",
                //       rowIndex, framePlanarDataRec.rowsizePerPlane[rowIndex],
                //       frameDataPtr, (const uint8_t *) framePlanarDataRec.planarData[rowIndex]);

                // This will drop the const-ness but no choice as we don't want to copy the frame
                // data and we want the FFmpeg library to have access to the bytes.
                ffmpegAVFrame->data[rowIndex] = (uint8_t *)framePlanarDataRec.planarData[rowIndex];
                ffmpegAVFrame->linesize[rowIndex] = (int)framePlanarDataRec.rowsizePerPlane[rowIndex];
            }

            // pts is the Presentation Time Stamp - time when frame should be shown to user.
            // THIS MUST be the frame number in the sequence.  It is how the time is interleaved
            // when the encoded frames start coming out of the ffmpeg encoder.  In this object the
            // inforamtion is maintained in the framesProcessed.
            ffmpegAVFrame->pts = (long long)framesProcessed;
            ffmpegAVFrame->pkt_dts = (long long)framesProcessed;
            ffmpegAVFrame->sample_aspect_ratio.num = 1;
            ffmpegAVFrame->sample_aspect_ratio.den = 1;

            // Filter, encode and write the frame.
            std::shared_ptr<FFmpegEncoder> ffmpegEncoder;
            ffmpegEncoder = FFmpegEncoder::GetEncoderInstance();

            // Could have been duplicates pushed if lost frames waiting to go.
            ffmpeg_result = ffmpegEncoder->EncodeAVFrame(&ffmpegAVFrame);

            // If an error occurred terminate this process.
            if (ffmpeg_result != 0)
            {
                // Is this a timeout condition?
                if (ffmpeg_result == AVERROR(ETIME))
                {
                    // Mark the status as an encoder timeout.
                    ffmpeg_result = 0;
                    containerStatus.set_encoder_timeout(true);
                }
                goto exitEncodeFrame;
            }

            //TEST_JLC  - FORCE ERROR CONDITION
#if 0
            if (framesProcessed == 13)
            {
                // Forcing an error condition.
                ffmpeg_result = AVERROR(ENOMEM);
                goto exitEncodeFrame;
            }
#endif

            // Increment the count of frames sent to the encoder
            // And the framesProcessed.
            encodedFrameCount++;
            IncrementFramesProcessed();

            if (duplicateFrameCount > 0)
            {
                totalFramesLost++;      // Add to the totalFramesLost.
                duplicateFrameCount--;  // Decrement the duplicate count.
            }

            framesToAdd--;  // Decrement the overall number of frames to add.

            // Reset the ffmpegAVFrame data and linesize planes prior to freeing the ffmpegAVFrame.
            for (rowIndex = 0; rowIndex < FAS_VIDEO::MAX_PLANES; rowIndex++)
            {
                ffmpegAVFrame->data[rowIndex] = nullptr;
                ffmpegAVFrame->linesize[rowIndex] = 0;
            }

            // Call the FFmpeg free frame if allocated to prepare for next loop.
            av_frame_free(&ffmpegAVFrame);
            ffmpegAVFrame = nullptr;
        }   // End of the frame processing loop.

        // Write a log message if frames were lost.
        if (frameDupStart != -1)
        {
            std::string lostFramesDataStr = fmt::format(FMT_STRING("ADDED LOST FRAMES [{0:d}] to [{1:d}] by duplicating frame[{2:d}]. Number of frames duplicated for this frame[{3:d}] total frames duplicated for video[{4:d}]"),
                                                        frameDupStart, frameDupEnd, theValidFrameId, framesToDup, totalFramesLost);
            fi_log(LOG_ERR, "[FI_FFMPEG::FFmpegContainer::EncodeFrame] %s.", lostFramesDataStr.c_str());
        }

        // If make it here there were no errors and encodedFrameCount > 0 then success!
        if (encodedFrameCount > 0)
            encodedFrames = true;
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught.", __PRETTY_FUNCTION__, e.what());
        goto exitEncodeFrame;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught.", __PRETTY_FUNCTION__);
        goto exitEncodeFrame;
    }

exitEncodeFrame:

    // If the FFmpegAVFrame is allocated free it.
    if (ffmpegAVFrame != nullptr)
    {
        for (rowIndex = 0; rowIndex < FAS_VIDEO::MAX_PLANES; rowIndex++)
        {
            ffmpegAVFrame->data[rowIndex] = nullptr;
            ffmpegAVFrame->linesize[rowIndex] = 0;
        }
        // Call the FFmpeg free frame
        av_frame_free(&ffmpegAVFrame);
    }

    if (!encodedFrames)
    {
        // Log any ffmpeg error codes.
        if (ffmpeg_result != 0)
            log_ffmpeg_error_str(ffmpeg_result);
        fi_log(LOG_ERR, "%s: framesProcessed[%d] encodedFrame[ERROR].",
               __PRETTY_FUNCTION__, framesProcessed);
    }
    //timings.LogDurationInfo(frameInfoStr.c_str());
    return encodedFrames;
} // end of EncodeFrame()

/*!
 * @brief Returns the frame number (one-based) of the last frame added to the encoder.
 *
 * @returns 0 if no frames have been added, or > 0 to represent one-based frame number.
 */
uint32_t FFmpegContainer::GetFramesAdded()
{
    uint32_t lastFrameAdded = 0;

    std::shared_ptr<FFmpegEncoder> ffmpegEncoder;
    ffmpegEncoder = FFmpegEncoder::GetEncoderInstance();
    if (!ffmpegEncoder)
    {
        fi_log(LOG_ERR, "%s: ffmpegEncoder was nullptr.",
               __PRETTY_FUNCTION__);
        goto exitGetFramesAdded;
    }
    lastFrameAdded = framesProcessed;

exitGetFramesAdded:
    return (uint32_t) lastFrameAdded;
}

/*!
 * @brief Returns the number of frames encoded by the encoder.
 *
 * @param isPrepared reference to state of preparedness for the FFmpeg container.
 * @param isComplete reference to the state of completeness of the video being created.
 * @param hasError reference to the state of ffmpeg error condition for the encoder.
 *
 * @returns 0 if no frames have been encoded or the actual count.
 */
uint32_t FFmpegContainer::GetEncodedFrameCount(bool& isPrepared, bool& isComplete, bool& hasError)
{
    uint32_t encodedFrameCount = 0;
    isPrepared = false;
    isComplete = false;
    hasError = false;

    std::shared_ptr<FFmpegEncoder> ffmpegEncoder;
    ffmpegEncoder = FFmpegEncoder::GetEncoderInstance();
    if (!ffmpegEncoder)
    {
        fi_log(LOG_ERR, "%s: ffmpegEncoder was nullptr.",
               __PRETTY_FUNCTION__);
        goto exitGetEncodedFrameCount;
    }

    // The isComplete will represent whether or not the AV Trailer has been written which
    // is the best way to ensure the ffmpeg encoder processed the mp4 video to completion.
    encodedFrameCount = ffmpegEncoder->GetEncodedFrameCount(isPrepared, isComplete, hasError);

exitGetEncodedFrameCount:
    return encodedFrameCount;
}

/*!
 * @brief Returns user readable status data about this container.
 *
 * @param statusInfoStr place to store the status information.
 */
void FFmpegContainer::GetStatusInfo(std::string& statusInfoStr)
{
    containerStatus.get_status_msg(statusInfoStr);
}

/*!
 * @brief Determines if the Frame Writer Thread has completed the video.
 *
 * @param inError(output) - Set for caller to indicate whether
 * or not the ffmpeg encoder is registering an error condition.
 *
 * @returns True if the encoder has finished processing all
 * frames presented.
 */
bool FFmpegContainer::IsFrameWriterComplete(bool& inError)
{
    inError = false;
    bool isFrameWriterComplete = false;
    try
    {
        bool isPrepared = false;
        // All we really care about in this case is whether or not the
        // frame writer is completed.  In this call for the MP4 stream
        // the GetEncodedFrameCount will be checking to see if the AV Trailer
        // was written to the file which indicates the frame writer was
        // successfully concluded.  If the ffmpeg encoder registered an
        // error condition the inError setting will reflect that.
        GetEncodedFrameCount(isPrepared, isFrameWriterComplete, inError);
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught.", __PRETTY_FUNCTION__, e.what());
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught.", __PRETTY_FUNCTION__);
    }

    return isFrameWriterComplete;
}

/*!
 * @brief Determines if the encoder has completed the video.
 *
 * @param nFramesLeft - a place to store the number of frames still left
 * to encode.
 *
 * @returns True if the encoder is completed, false if not.
 */
bool FFmpegContainer::IsVideoComplete(uint32_t& nFramesToEncodeStill, bool& inError)
{
    bool isComplete = false;
    try
    {
        bool isPrepared = false;
        uint32_t framesEncoded = GetEncodedFrameCount(isPrepared, isComplete, inError);

        std::string hwPixelFmtStr = FFmpegContainer::GetStringForAVPixelFormat(hwFFmpegDeviceFmt);
        std::string inPixelFmtStr = FFmpegContainer::GetStringForAVPixelFormat(inFFMpegVideoFrameFmt);
        std::string outPixelFmtStr = FFmpegContainer::GetStringForAVPixelFormat(outFFMpegVideoFrameFmt);

        // If the frames encoded is equal to the frame count check to make sure it is complete.
        if (framesEncoded >= frameCount)
        {
            // Document this in the log as a warning.
            if (!isComplete && !inError)
            {
                fi_log(LOG_WARNING, "[FFmpegContainer::IsVideoComplete] [ENCODED COUNT GOOD BUT NO EOF YET] isPrepared[%s] isComplete[%s] inError[%s] frameCount[%d] framesEncoded[%d].",
                       isPrepared ? "T" : "F", isComplete ? "T" : "F",
                       inError ? "T" : "F", frameCount, framesEncoded);

                // If isComplete is false it may be that the frame writer thread hadn't written
                // the trailer record and in that case let's try again to see if we can get a change
                // in the complete status.  Hold for a couple of seconds.  This is really a fault of
                // the library caller making a call to close the stream before it was complete.
                std::this_thread::sleep_for(std::chrono::milliseconds(2000));

                // Ask the encoder again if it is done.
                framesEncoded = GetEncodedFrameCount(isPrepared, isComplete, inError);

                // Check again to see if we became complete and log results.
                fi_log(LOG_WARNING, "[FFmpegContainer::IsVideoComplete] [RE-CHECK AFTER 2 SECOND SLEEP] isPrepared[%s] isComplete[%s] inError[%s] frameCount[%d] framesEncoded[%d].",
                       isPrepared ? "T" : "F", isComplete ? "T" : "F",
                       inError ? "T" : "F", frameCount, framesEncoded);

            }

            // Keep the framesEncoded in range (this theoretically should never happen).
            if (framesEncoded > frameCount)
            {
                framesEncoded = frameCount;
            }
        }

        nFramesToEncodeStill = frameCount - framesEncoded;

        // This is an information log message.  It stamps this inforamation in the log file
        // and provides a good snapshot of a stream.
        fi_log(LOG_INFO, "[FFmpegContainer::IsVideoComplete] isPrepared[%s] isComplete[%s] inError[%s] frameCount[%d] framesEncoded[%d] HW[%s] IN[%s] OUT[%s].",
               isPrepared ? "T" : "F", isComplete ? "T" : "F",
               inError ? "T" : "F", frameCount, framesEncoded,
               hwPixelFmtStr.c_str(),
               inPixelFmtStr.c_str(),
               outPixelFmtStr.c_str());
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught.", __PRETTY_FUNCTION__, e.what());
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught.", __PRETTY_FUNCTION__);
    }

    return isComplete;
}

/*!
 * @brief When a stream is unable to duplicate frames due to missing
 * frames at the end we need a mechanism to pad out the frames processed
 * to allow for the flush frame to be sent.  This will in effect complete
 * the video with missing end frames.  The count and placement of these
 * missing frames is noted in the log data.
 *
 * @returns True if all is ok or false if an error condition occurred.
 */
bool FFmpegContainer::PadWithEmptyFramesAtEnd()
{
    bool allOK = IncrementFramesProcessed();
    fi_log(LOG_DEBUG, "%s: framesProcessed[%d] allOK[%s].", __PRETTY_FUNCTION__, framesProcessed, (allOK ? "Y" : "N"));
    return allOK;
}
