/*!
 * @file ffmpeg_status.cpp
 * @brief Provides status support object.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#ifdef  __cplusplus
extern "C" {
#endif

#include "libavutil/opt.h"

#ifdef  __cplusplus
}
#endif

#include "ffmpeg_status.h"
#include "log_if.h"

using namespace FI_FFMPEG;

//! @brief Translates FFmpeg error codes into user readable error string and logs it.
void log_ffmpeg_error_str(int av_code)
{
    char err_buff[320] = { 0 };
    av_make_error_string(err_buff, sizeof(err_buff), av_code);
    fi_log(LOG_ERR, "[FFmpeg Error: %s]", err_buff);
}

//--------------------------------------------------------------------
// Class Definitions
//--------------------------------------------------------------------

FFmpegStatus::FFmpegStatus() :
    status(CTX_STATUS::CONSTRUCTED),
    ffmpeg_result(0)
{
}
FFmpegStatus::~FFmpegStatus()
{
}

/*!
 * @brief Queries state of being constructed.
 *
 * @returns A boolean true if in constructed state or false if not.
 */
bool FFmpegStatus::isConstructed()
{
    if (status == CTX_STATUS::CONSTRUCTED)
        return true;
    return false;
}

/*!
 * @brief Queries state of being prepared.
 *
 * @returns A boolean true if in prepared state or false if not.
 */
bool FFmpegStatus::isPrepared()
{
    if (status == CTX_STATUS::PREPARED)
        return true;
    return false;
}

/*!
 * @brief Set the status state as prepared.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_prepared(bool log_it, const char *more_info)
{
    status = CTX_STATUS::PREPARED;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for allocation failure.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_bad_alloc(bool log_it, const char *more_info)
{
    status = CTX_STATUS::BAD_ALLOC;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state as for exception failure.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_bad_except(bool log_it, const char *more_info)
{
    status = CTX_STATUS::BAD_EXCEPT;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for bad file condition.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_bad_file(bool log_it, const char *more_info)
{
    status = CTX_STATUS::BAD_FILE;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing file condition.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_file_info(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_FILE_CTRL;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing AVFormatContext condition.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_fmt_ctx(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_FMT_CTX;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing AVStream condition.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_stream(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_STREAM;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing AVCodec condition.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_codec(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_CODEC;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing AVFrame condition.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_img_info(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_IMG_INFO;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing container context operation.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_ctx_type(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_CTX_TYPE;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing container context operation.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_ctx_ops(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_CTX_OPS;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing container context decoder for transcoding.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_ctx_decoder(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_CTX_DECODER;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing container context filters for encoding.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_ctx_filters(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_CTX_FILTERS;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing container context encoder for encoding.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_ctx_encoder(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_CTX_ENCODER;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for missing container context FFmpeg I/O object.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_no_ctx_ffmpeg_io(bool log_it, const char *more_info)
{
    status = CTX_STATUS::NO_CTX_FFMPEG_IO;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state as an Encoder Timeout error.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_encoder_timeout(bool log_it, const char *more_info)
{
    status = CTX_STATUS::ENCODER_TIMEOUT;
    if (log_it)
        logStatus(more_info);
}

/*!
 * @brief Set the status state for general FFmpeg error condition.
 *
 * @param log_it    Determines whether or not to log the status.
 * @param more_info Provides more information for the log if logging status.
 *
 * @returns This is a void function.
 */
void FFmpegStatus::set_ffmpeg_err(int ffmpeg_error, bool log_it, const char *more_info)
{
    (void)ffmpeg_error;
    status = CTX_STATUS::FFMPEG_ERR;
    if (log_it)
        logStatus(more_info);
}


/*!
 * @brief Retrieve a readable status message.
 *
 * @param statusMsg - will hold the result of this call.
 *
 * @returns The user readable string describing the FFmpeg Status.
 */
void FFmpegStatus::get_status_msg(std::string& statusMsg)
{
    // Build in the actual error that occurred.
    switch (status)
    {
    case CTX_STATUS::CONSTRUCTED:
        statusMsg = "[CONSTRUCTED] ";
        break;
    case CTX_STATUS::PREPARED:
        statusMsg = "[PREPARED] ";
        break;
    case CTX_STATUS::BAD_ALLOC:
        statusMsg = "[BAD_ALLOC] ";
        break;
    case CTX_STATUS::BAD_EXCEPT:
        statusMsg = "[BAD_EXCEPT] ";
        break;
    case CTX_STATUS::BAD_FILE:
        statusMsg = "[BAD_FILE] ";
        break;
    case CTX_STATUS::NO_FILE_CTRL:
        statusMsg = "[NO_FILE_CTRL] ";
        break;
    case CTX_STATUS::NO_FMT_CTX:
        statusMsg = "[NO_FMT_CTX] ";
        break;
    case CTX_STATUS::NO_STREAM:
        statusMsg = "[NO_STREAM] ";
        break;
    case CTX_STATUS::NO_CODEC:
        statusMsg = "[NO_CODEC] ";
        break;
    case CTX_STATUS::NO_IMG_INFO:
        statusMsg = "[NO_IMG_INFO] ";
        break;
    case CTX_STATUS::NO_CTX_TYPE:
        statusMsg = "[NO_CTX_TYPE] ";
        break;
    case CTX_STATUS::NO_CTX_OPS:
        statusMsg = "[NO_CTX_OPS] ";
        break;
    case CTX_STATUS::NO_CTX_DECODER:
        statusMsg = "[NO_CTX_DECODER] ";
        break;
    case CTX_STATUS::NO_CTX_FILTERS:
        statusMsg = "[NO_CTX_FILTERS] ";
        break;
    case CTX_STATUS::NO_CTX_ENCODER:
        statusMsg = "[NO_CTX_ENCODER] ";
        break;
    case CTX_STATUS::NO_CTX_FFMPEG_IO:
        statusMsg = "[NO_CTX_FFMPEG_IO] ";
        break;
    case CTX_STATUS::ENCODER_TIMEOUT:
        statusMsg = "[ENCODER_TIMEOUT WAITING FOR FRAMEWRITER QUEUE THRESHOLD] ";
        break;
    case CTX_STATUS::FFMPEG_ERR:
        statusMsg = "[FFMPEG_ERR] ";
        break;
    default:
        statusMsg = "[UNKNOWN] ";
        break;
    }
}

/*!
 * @brief Log the status to logging mechanism.
 *
 * @param addToLogMessage       Additional information to put in the log message.
 * @param logToFFmpegConsole    If true this will log to the std console.
 *
 * @returns This function is a void method.
 */
void FFmpegStatus::logStatus(const char *addToLogMessage)
{
    std::string msg;
    std::string statusMsgType;

    uint32_t severity = LOG_ERR;
    std::string additionalInfo = "";

    // Get the status message type.
    get_status_msg(statusMsgType);

    msg = "[FFmpegStatus]";
    msg = msg + statusMsgType;

    if (addToLogMessage != nullptr)
    {
        additionalInfo += addToLogMessage;
        msg += " [";
        msg += additionalInfo.c_str();
        msg += "] ";
    }

    // Build in the actual error that occurred.
    switch (status)
    {
    case CTX_STATUS::CONSTRUCTED:
        severity = LOG_INFO;
        break;
    case CTX_STATUS::PREPARED:
        severity = LOG_INFO;
        break;
    case CTX_STATUS::FFMPEG_ERR:
        msg += "[FFMPEG_ERR] ";
        log_ffmpeg_error_str(ffmpeg_result);
        break;
    default:
        msg += "[UNKNOWN] ";
        break;
    }

    // Always log to the our own mechanism.
    fi_log(severity, "%s", msg.c_str());
}

// End of ffmpeg_utils.cpp


