/*!
 * @file fi_bmp.h
 * @brief Contains the implementation of the FiBMP class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

#if defined(_LINUX)
#include <linux/limits.h>
#endif

#include "fi_bmp.h"
#include "fas_video.h"
#include "log_if.h"

namespace FI_BMP
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! Bit depth values for greyscale BMP headers.
const uint16_t BMP_GREY_BIT_DEPTH   = 8;

//! Bit depth values for color BMP headers.
const uint16_t BMP_COLOR_BIT_DEPTH  = 24;

//! Color planes must be set to 1.
const uint16_t BMP_COLOR_PLANES     = 1;

//! We need 2 alignment bytes after the header (or after header+colortable) to
//! align the image on a 4 byte boundary for DMA read purposes.
const int IMAGE_ALIGNMENT_BYTES     = 2;

//! The BMP magic 2 byte code.  This is kept out of the header structure to
//! avoid alignment problems on processors that have trouble with unaligned
//! structure members.
const char *BMP_MAGIC               = "BM";

} // End of FI_BMP

// Make things easier for ourselves.
using namespace FI_BMP;
using namespace FAS_VIDEO;

FiBMP::FiBMP() : FiImage(),
    createdHeader(false),
    frameData(nullptr)
{
    hdr_size = 0;   // We have no valid header until this >= sizeof BMP_HEADER.

    // No frame plane data exists yet.
    framePlaneData = {0, {0}, {nullptr}};
}

/*!
 * @brief Initialize the BMP's color lookup table.
 *
 * For greyscale 8-bit images, we use a color lookup table
 * to remap the 8-bit value into an RGBA value.
 */
void FiBMP::init_color_table()
{
    uint32_t i;

    // Do we need to initialize our color table?
    if (color_table.size() > 0)
        return;

    // Make sure our array is large enough.
    color_table.reserve(DIB_COLOR_TABLE_SIZE);
    for (i = 0; i < DIB_COLOR_TABLE_SIZE; i += sizeof(uint32_t))
    {
        char table_entry = static_cast<char>(i / sizeof(uint32_t));

        // Initialize the RGBA color table.
        // Set R, G, and B to be equal. A is set to 0.
        color_table.push_back(table_entry);
        color_table.push_back(table_entry);
        color_table.push_back(table_entry);
        color_table.push_back(0);
    }
}

/*!
 * @brief Does BMP Header exist.
 *
 * @returns A true value if header exists, false if not.
 */
bool FiBMP::BMPHeaderExists()
{
    bool bmpHeaderExists = false;
    int bmpHeaderSize = (int) sizeof(BMP_HEADER);

    // NOTE_JLC - Do we need to test this too? (headerFourCC == FAS_VIDEO::FOURCC_NONE)

    // If the hdr_size is zero then there is no header information.
    // It can be exactly equal to sizeof(BMP_HEADER) if there is no color table.
    // It will be the sizeof(BMP_HEADER) + sizeof(color_table) if it has a color table.
    // Right now we are only testing for the size to be >= sizeof(BMP_HEADER).
    if (hdr_size >= bmpHeaderSize)
        bmpHeaderExists = true;

    return bmpHeaderExists;
}

/*!
 * @brief Copy the BMP header information from the front of an image data array.
 *
 * @returns A true value if BMP header was copied, false if unable to copy.
 */
bool FiBMP::CopyBMPHeaderData(const uint8_t *memDataPtr, uint32_t imageDataSize)
{
    bool imageDataCopied = false;                                       // Has the data been copied to internal structures?
    uint32_t bmpMagicNbrSize = (uint32_t) strlen(FI_BMP::BMP_MAGIC);    // How big is BMP Magic string.
    uint32_t bmpHeaderSize = sizeof(BMP_HEADER);                        // How big is BMP Header structure.

    const uint8_t *curImageDataPtr = memDataPtr;
    uint32_t curImageDataSize = imageDataSize;

    // The memDataPtr must not be a nullptr.
    if (memDataPtr == nullptr)
        goto exitCopyBMPHeaderData;

    // The image data must have a BMP_HEADER + BMP_MAGIC size.
    if (imageDataSize < (bmpMagicNbrSize + bmpHeaderSize))
        goto exitCopyBMPHeaderData;

    // The image data must have a BMP_MAGIC indicator.
    if (!isBmpHdr((const char *)memDataPtr))
        goto exitCopyBMPHeaderData;

    // Adjust our interim data by the size of the BMP_MAGIC number.
    curImageDataPtr += bmpMagicNbrSize;
    curImageDataSize -= bmpMagicNbrSize;

    // There is enough data to get the bitmap header so copy the data.
    memcpy(&header, curImageDataPtr, bmpHeaderSize);

    // The header is from image data.
    createdHeader = false;

    // Adjust our interim data by the size of the BMP_HEADER.
    curImageDataPtr += bmpHeaderSize;
    curImageDataSize -= bmpHeaderSize;

    // Set the hdr_size to the size of the bmp header data.
    hdr_size = (int) bmpHeaderSize;

    // Should there be a color table here?
    if (header.dib.depth == BMP_GREY_BIT_DEPTH)
    {
        // Initialize the color table if it isn't initialized.
        init_color_table();

        // Make sure there is enough data to copy the color table.
        if (curImageDataSize < color_table.size())
        {
            fi_log(LOG_ERR, "%s: curImageDataSize[%u] color_table.size()[%d].",
                   __PRETTY_FUNCTION__, curImageDataSize, color_table.size());
            goto exitCopyBMPHeaderData;
        }

        // It should be ok to copy the "color table" into our color table.
        memcpy(((void *)color_table.data()), curImageDataPtr, color_table.size());

        //curImageDataSize -= color_table.size();   // Not being used by anything.

        // Increase the hdr_size by the color table size.
        hdr_size += color_table.size();
    }

    // We can assume the data was copied correctly.
    imageDataCopied = true;

exitCopyBMPHeaderData:
    return imageDataCopied;
}

/*!
 * @brief Create a BMP header based on the specified parameters.
 *
 * @param fourcc    One of the supported FAS_VIDEO @ref fv_fourcc "FourCC" codes.
 * @param meta      Pointer to the image's metadata.
 * @returns The number of bytes the header uses, or a negative number
 *      indicating an error.
 */
int FiBMP::CreateHeader(uint32_t fourcc, const pm_metadata *meta)
{
    uint32_t bpp = 3;       // BYTES-per-pixel, assume color.

    // Sanity check.
    if (meta == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: meta pointer is null!", __PRETTY_FUNCTION__);
        return -1;
    }

    // Save the data in base class properties so it persists.
    headerFourCC = fourcc;
    memcpy(&headerMetaData, meta, sizeof(pm_metadata));

    // Initialize our length and zero the header.
    hdr_size = sizeof(header);
    memset(&header, 0, sizeof(header));

    // Initialize the constant settings.
    header.dib.depth = BMP_COLOR_BIT_DEPTH;
    header.dib.data_header_size = sizeof(header.dib);
    header.dib.color_planes = BMP_COLOR_PLANES;

    // If we're creating a monochrome image, then we need
    // to account for the color table.
    switch (fourcc)
    {
    case FOURCC_Y8:
    case FOURCC_BYR1:
        // Make sure our color lookup table is initialized.
        init_color_table();

        // We only deal with 8-bit greyscale, and so
        // we only have 1 byte per pixel.
        header.dib.depth = BMP_GREY_BIT_DEPTH;
        bpp = 1;

        // Adjust the header size by the lookup table.
        hdr_size += color_table.size();
        break;
    case FOURCC_RGB:
        // Everything we need initialized is already initialized.
        break;

    default:
        // Oops. This isn't right.
        hdr_size = -1;
        fi_log(LOG_ERR, "%s: Invalid fourcc (0x%08x)", __PRETTY_FUNCTION__, fourcc);
        break;
    }
    if (hdr_size < 0)
        goto head_bail;

    // Add these alignment bytes since the BMP header size is
    // not a multiple of words.
    hdr_size += IMAGE_ALIGNMENT_BYTES;

    // Calculate the image size. Start with the image width (in BYTES).
    header.dib.data_size = meta->roi_width * bpp;

    // Align the row width.
    header.dib.data_size = (header.dib.data_size + (ROW_ALIGNMENT - 1)) & (~(ROW_ALIGNMENT - 1));

    // Now multiply by the height to get total size.
    header.dib.data_size *= meta->roi_height;

    // The image data will be after the header.
    header.data_offset = static_cast<uint32_t>(hdr_size);

    // The length is the total length of the file.
    header.length = header.dib.data_size + header.data_offset;

    // Save the width.
    header.dib.width = meta->roi_width;

    // Negative height indicates image stored top row at head of file
    header.dib.height = -1 * ((int32_t)meta->roi_height);

    // We have created a header!
    createdHeader = true;

head_bail:
    return hdr_size;
}

/*!
 * @brief Get a unique pointer to an array of bytes containing
 *    the BMP header.
 *
 * The returned data is ready to be prepended onto an image to
 * create a BMP file.
 *
 * @returns A std::unique_ptr<char> containing the BMP header.
 */
std::unique_ptr<FAS_VIDEO::HEADER_DATA> FiBMP::GetHeader()
{
    uint8_t *current_position;      // Write position in buffer.
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> hdr_ptr = std::unique_ptr<FAS_VIDEO::HEADER_DATA>(new FAS_VIDEO::HEADER_DATA(static_cast<uint64_t>(hdr_size), 0));

    // Get our pointer.
    current_position = hdr_ptr->data();

    // Set the magic number
    memcpy(current_position, BMP_MAGIC, strlen(BMP_MAGIC));
    current_position += strlen(BMP_MAGIC);

    // Copy the header into the buffer
    memcpy(current_position, ((const char *)&header), sizeof(header));
    current_position += sizeof(header);

    // Copy the color table if we need one.
    if (color_table.size() > 0)
    {
        // Append the table.
        memcpy(current_position, ((const char *)color_table.data()), color_table.size());
    }

#ifdef USE_VERBOSE_INFO_DEBUG
    uint8_t *magicPtr = hdr_ptr->data();      // Write position in buffer.
    fi_log(LOG_DEBUG, "%s: BMP(DIB), MAGIC[0]=%d MAGIC[1]=%d BMP_MAGIC[0]=%d BMP_MAGIC[1]=%d.",
           __PRETTY_FUNCTION__, magicPtr[0], magicPtr[1], BMP_MAGIC[0], BMP_MAGIC[1]);
#endif

    // Return the pointer.
    return hdr_ptr;
}

/*!
 * @brief Query the image format supported by this object.
 *
 * @returns The IMAGE_FORMAT value.
 */
FAS_VIDEO::IMAGE_FORMAT FiBMP::GetImageFormat()
{
    // We are BMPs.
    return FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE;
}

/*!
 * @brief Query the image width of created header or attached image.
 *
 * @returns The image width or 0 if there is no known value.
 */
uint32_t FiBMP::GetImageWidth()
{
    uint32_t imageWidth = 0;

    // If there is a header created for this image use its data.
    if (BMPHeaderExists())
    {
        // This negative test doesn't make sense since it's only the height that is
        // negative in a DIB image coming from camera server.  But I am going to leave
        // this in place because we will never have a negative width but if we did we
        // would return the proper width.
        imageWidth = (header.dib.width < 0 ? (uint16_t)(-header.dib.width) : (uint16_t)header.dib.width);
    }
    else if (HasKnownImageInfo())
    {
        // At least one image was attached saving this information it is valid to use it.
        imageWidth = FiImage::GetImageWidth();
    }
    return imageWidth;
}

/*!
 * @brief Query the image height of created header or attached image.
 *
 * @returns The image width or 0 if there is no known value.
 */
uint32_t FiBMP::GetImageHeight()
{
    uint32_t imageHeight = 0;

    // If there is a header created for this image use its data.
    if (BMPHeaderExists())
    {
        // Extract the width, handle the - height case used to flip image
        imageHeight = (header.dib.height < 0 ? (uint16_t)(-header.dib.height) : (uint16_t)header.dib.height);
    }
    else if (HasKnownImageInfo())
    {
        // At least one image was attached saving this information it is valid to use it.
        imageHeight = FiImage::GetImageHeight();
    }
    return imageHeight;
}

/*!
 * @brief Query the image height of created header or attached image.
 * The bit depth in the BMP header will be either 8 (greyscale) or
 * 24 (RGB).
 *
 * @returns The bit depth from header or 0 if there is no known value.
 */
uint32_t FiBMP::GetBitsPerPixel()
{
    uint32_t bitDepth = 0;

    // If there is a header created for this image use its data.
    if (BMPHeaderExists())
    {
        // Extract the BMP header's bit depth value.
        bitDepth = header.dib.depth;
    }
    else if (HasKnownImageInfo())
    {
        // At least one image was attached saving this information it is valid to use it.
        bitDepth = FiImage::GetBitsPerPixel();
    }
    return bitDepth;
}

/*!
 * @brief Returns the row alignment for BMP image data.
 *
 * @returns The row alignment for bitmap frames.
 */
uint32_t FiBMP::GetRowAlign()
{
    uint32_t rowAlign = FI_BMP::ROW_ALIGNMENT;
    return rowAlign;
}

/*!
 * @brief Returns total number of uint8_t bytes in a frame of data.
 *
 * @returns 0 in this base class.
 */
uint32_t FiBMP::GetFullFramesize()
{
    uint32_t frameSize = 0;

    // If there is a header created for this image use its data.
    if (BMPHeaderExists())
    {
        uint32_t rowWidth = 0;

        // If 8 bits per pixel - it's greyscale data.  (old pixfmt=FI_PIX_FMT_GRAY)
        if (header.dib.depth == BMP_GREY_BIT_DEPTH)
            rowWidth = GetImageWidth();
        else
            rowWidth = GetImageWidth() * 3;

        // Calculate the uncompressed image size for DIB.
        // Add proper padding to align the beginnings of rows.
        frameSize = (uint32_t)(rowWidth + (FI_BMP::ROW_ALIGNMENT - 1)) & (uint32_t)(~(FI_BMP::ROW_ALIGNMENT - 1));
        frameSize *= GetImageHeight();
    }
    else if (HasKnownImageInfo())
    {
        // At least one image was attached saving this information it is valid to use it.
        frameSize = FiImage::GetFullFramesize();
    }
    return frameSize;
}

uint32_t FiBMP::GetHeaderSize()
{
    return (uint32_t) hdr_size;
}

bool FiBMP::isBmpHdr(const char *imageDataPtr)
{
    bool rc = false;

    // If we have "created" an external header, then there won't be an "embedded"
    // header in the data stream. If we did not create an "external" header, then
    // we need to look.
    if (!createdHeader)
    {
        // Does the image data have the Y8 signature?
        if (strncmp(imageDataPtr, FI_BMP::BMP_MAGIC, strlen(FI_BMP::BMP_MAGIC)) == 0)
        {
            // Looks like we have an embedded BMP header.
            rc = true;
        }
    }

    return rc;
}

/*!
 * @brief Test the data in a BMP to verify it and pull out dimensions.
 *
 * @param imageDataToVerify     The Bitmap (DIB) data to verify BMP format.
 * @param imageDataSize         How many uint8_t bytes are in this BMP image.
 *
 * @returns true if BMP data was verified and attach dimension data interpreted.
 */
bool FiBMP::VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize)
{
    bool verifiedImage = false;     // Keeps track of verification success.
    bool hasHeader = false;         // Does the imageDataToVerify have a header?

    // BMP specific information to verify in the image data.
    uint32_t minHdrSize = sizeof(header) + (uint32_t) strlen(FI_BMP::BMP_MAGIC);

    // Reset any normalized frame data.
    frameData.reset();
    framePlaneData = {0, {0}, {nullptr}};

    // Make sure the imageDataToVerify is valid.
    if ((imageDataToVerify == nullptr) || (imageDataSize < minHdrSize))
    {
        fi_log(LOG_ERR, "%s: BMP imageDataToVerify either empty or imageDataSize[%u] less than minHdrSize[%u].", __PRETTY_FUNCTION__, imageDataSize, minHdrSize);
        goto exitVerifyImageData;
    }

    // Look for the BMP_MAGIC number and set hasHeader condition if it exists.
    if (isBmpHdr((const char *)imageDataToVerify))
    {
        // The imageDataToVerify has header information save this condition.
        hasHeader = true;
    }

    // PROCESS BMPPlanar Raw Frame.  No BMPPlanar header data is prepended to imageDataToVerify.
    if (!hasHeader)
    {
        // Does this image object have a BMPPlanar header?
        if (!BMPHeaderExists())
        {
            // This object needs header information to process raw frames so reject the imageDataToVerify.
            fi_log(LOG_ERR, "%s: Must create Bitmap Header prior to writing raw image frames hdr_size[%u] expected [%u]. ",
                   __PRETTY_FUNCTION__, hdr_size, sizeof(header));

            fi_log(LOG_ERR, "%s: header.width[%ld] header.height[%ld] header.compression[%d] header.depth[%d].",
                   __PRETTY_FUNCTION__, header.dib.width, header.dib.height, header.dib.compression, header.dib.depth);

            goto exitVerifyImageData;
        }

#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "%s: BMP HEADER EXISTED:  header.length[%u] hdr_size[%u] header.width[%ld] header.height[%ld] header.compression[%d] header.depth[%d] header.data_offset[%d].",
               __PRETTY_FUNCTION__, header.length, hdr_size, header.dib.width, header.dib.height, header.dib.compression, header.dib.depth, header.data_offset);
#endif

        // We have verified the image.
        verifiedImage = true;
        goto exitVerifyImageData;
    }

    // PROCESS BMP IMAGE.  Includes a BMP header data in imageDataToVerify.

    // If this object has no BMP header information copy it from imageDataToVerify.
    if (!BMPHeaderExists())
    {
        if (!CopyBMPHeaderData(imageDataToVerify, imageDataSize))
        {
            // This object needs header information to process raw frames so reject the imageDataToVerify.
            fi_log(LOG_ERR, "%s: Unable to copy Bitmap Header from imageDataToVerify. ", __PRETTY_FUNCTION__);
            goto exitVerifyImageData;
        }

#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "%s: BMP HEADER COPIED FROM IMAGE:  header.length[%u] hdr_size[%u] header.width[%ld] header.height[%ld] header.compression[%d] header.depth[%d] header.data_offset[%d].",
               __PRETTY_FUNCTION__, header.length, hdr_size, header.dib.width, header.dib.height, header.dib.compression, header.dib.depth, header.data_offset);
#endif
    }
    else   // Already has existing BMP data in the object.
    {
        // Verify the existing BMP data against the one inside of imageDataToVerify.
        const BMP_HEADER *bmpHdrPtr = (const BMP_HEADER *)(imageDataToVerify + strlen(FI_BMP::BMP_MAGIC));

        // Check the header width.
        if (bmpHdrPtr->dib.width != header.dib.width)
        {
            // There should be no mis-match between image header and imageDataToVerify's header.
            fi_log(LOG_ERR, "%s: The image's width[%lu] does not match stored header's width[%lu].",
                   __PRETTY_FUNCTION__, bmpHdrPtr->dib.width, header.dib.width);
            goto exitVerifyImageData;
        }

        // Check the header height.
        if (bmpHdrPtr->dib.height != header.dib.height)
        {
            fi_log(LOG_ERR, "%s: The image's height[%lu] does not match stored header's height[%lu].",
                   __PRETTY_FUNCTION__, bmpHdrPtr->dib.height, header.dib.height);
            goto exitVerifyImageData;
        }

        // Check the header depth.
        if (bmpHdrPtr->dib.depth != header.dib.depth)
        {
            fi_log(LOG_ERR, "%s: The image's depth[%lu] does not match stored header's depth[%lu].",
                   __PRETTY_FUNCTION__, bmpHdrPtr->dib.depth, header.dib.depth);
            goto exitVerifyImageData;
        }

        // Check the header compression.
        if (bmpHdrPtr->dib.compression != header.dib.compression)
        {
            fi_log(LOG_ERR, "%s: The image's compression[%lu] does not match stored header's compression[%lu].",
                   __PRETTY_FUNCTION__, bmpHdrPtr->dib.compression, header.dib.compression);
            goto exitVerifyImageData;
        }
    }

    // Make sure there is a fourCC code for this image data too if it is not set.
    if (headerFourCC == FOURCC_NONE)
    {
        if (header.dib.depth == BMP_COLOR_BIT_DEPTH)
            headerFourCC =  FAS_VIDEO::FOURCC_RGB;      // Assuming RGB color image.
        else if (header.dib.depth == BMP_GREY_BIT_DEPTH)
            headerFourCC = FAS_VIDEO::FOURCC_Y8;        // Assuming greyscale image.
    }

    // Passed all verification checks.  Mark it as verified.
    verifiedImage = true;

exitVerifyImageData:
    // Returns the status of the verification process.
    return verifiedImage;
}

/*!
 * @brief Retrieves a read-only frame data pointer to attached image data.
 * data minus any header if one exists.
 *
 * @param attachedFrameSize     (Reference)Set to length of frame data or zero if none exist.
 * @param planarDataAttached    (Reference)Set to the planar data of attached frame.
 * @param frameOrientation      How is the frame oriented.
 *
 * @returns read-only pointer to attached frame data or nullptr if none exist in addition
 * the references (attachedFrameSize,planarDataAttached) are output settings as well.
 */
const uint8_t *FiBMP::GetAttachedFrameData(uint32_t& attachedFrameSize,
                                           FAS_VIDEO::FRAME_PLANAR_DATA& planarDataAttached,
                                           FAS_VIDEO::FRAME_ORIENTATION frameOrientation)
{
    const uint8_t *frameDataPtr = nullptr;              // Set the read-only pointer to null.

    bool frameCreatedOK         = false;                // Returned from the frame creator to catch error condition.
    bool frameReadyToUse        = false;                // Must get through all tests to be usable.

    uint32_t frameDataOffset    = 0;                    // If image data has BMP header this will be offset to frame data.
    uint32_t frameDataLenMinusHdr = 0;                  // How many bytes are in the actual frame data.
    const uint8_t *frameDataPtrMinusHdr = nullptr;      // Will point to the actual frame data skipping the header.
    uint32_t frameHeight = 0;                           // Will hold the positive frame height value.
    bool needsFlip = false;                             // If height read from bmp header is negative we need to flip.

    uint32_t fourCCImage   = FAS_VIDEO::FOURCC_NONE;    // Type of frame data.
    uint32_t rowAlign      = 4;                         // Aligning on 32-bit boundary (old do_flip() was forcing this)

    // Set the caller's size to zero.
    attachedFrameSize = 0;

    // Initialize the passed in planar record to empty state.
    planarDataAttached = {0, {0}, {nullptr}};

    // Can't process if there is no bitmap image attached.
    if (status != FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        fi_log(LOG_ERR, "%s: No bitmap registered as attached.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Because there is an attached bitmap image it means that the header
    // contains the data read and verified from it and it can be relied upon.
    if (header.dib.depth == BMP_GREY_BIT_DEPTH)
    {
        fourCCImage = FAS_VIDEO::FOURCC_Y8;    // Assuming Greyscale image.
    }
    else if (header.dib.depth == BMP_COLOR_BIT_DEPTH)
    {
        fourCCImage = FAS_VIDEO::FOURCC_RGB;   // Assuming RGB color image.
    }
    else
    {
        fi_log(LOG_ERR, "%s: The header.dib.depth was set to unsupported value[%d].", __PRETTY_FUNCTION__, header.dib.depth);
        goto exitGetAttachedFrameData;
    }

    // Make sure the header fourcc has a valid value in it.
    headerFourCC = fourCCImage;

    // Is height negative?  If so frame data is in top-down orientation
    // The frame orientation set in the object determines whether to flip
    // the frame data presented or not.
    if (header.dib.height < 0)
    {
        frameHeight = (uint32_t)(header.dib.height * -1);

        // Negative height means frame is oriented top-down.

        // If the orientation is set bottom-up then frame needs to be flipped.
        if (frameOrientation == FAS_VIDEO::FRAME_ORIENTATION::BOTTOM_UP)
            needsFlip = true;
    }
    else
    {
        frameHeight = (uint32_t) header.dib.height;

        // Positive height means frame is oriented bottom-up.

        // If the orientation is set top-down then frame needs to be flipped.
        if (frameOrientation == FAS_VIDEO::FRAME_ORIENTATION::TOP_DOWN)
            needsFlip = true;
    }

    // Look at the attached image data package to find the bitmap frame data.
    frameDataPtr = FiImage::GetAttachedImageData();
    attachedFrameSize = FiImage::GetAttachedImageSize();

    // Check to see we were able to get attached frame information.
    if (frameDataPtr && (attachedFrameSize > 0))
    {
        // Look for the BMP_MAGIC number and if so frame has prepended BMP header.
        if (isBmpHdr((const char *)frameDataPtr))
        {
            // The raw BMP frame data starts after the prepended BMP header.
            frameDataOffset = header.data_offset;
        }
        else
        {
            // There is no BMP header only a raw BMP frame data.
            frameDataOffset = 0;
        }

        // Prepare access to frame data only.
        frameDataLenMinusHdr = attachedFrameSize - frameDataOffset;
        frameDataPtrMinusHdr = frameDataPtr + frameDataOffset;
    }
    else
    {
        fi_log(LOG_ERR, "%s: The attached BMP image is null or empty.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    if (needsFlip)
    {
        // Using the FiFrame factory create an internal frame to hold the frame data
        // in the orientation requested (this requires flip of data and new memory allocation).
        frameData = FiFrame::CreateFrame(frameCreatedOK, fourCCImage, (uint32_t)header.dib.width, frameHeight, rowAlign, frameOrientation);

        // If unable to create the normalized frame then cannot continue.
        if (!frameCreatedOK)
        {
            const std::string fourCCReadable = FiVideo::GetFourCCReadableString(fourCCImage);
            fi_log(LOG_ERR, "%s: FiFrame::CreateFrame() failed.  Could external frame provided be corrupted?  Info: fourCC[%s] orient[%d] frameDataOffset[%d] attachedFrameSize[%d] w[%d] h[%d] frameDataLenMinusHdr[%d]",
                   __PRETTY_FUNCTION__, fourCCReadable.c_str(), (int)frameOrientation, frameDataOffset, attachedFrameSize, header.dib.width, frameHeight, frameDataLenMinusHdr);
            goto exitGetAttachedFrameData;
        }

        // If we don't have any frame data don't attempt set it up.
        if (!frameData)
        {
            const std::string fourCCReadable = FiVideo::GetFourCCReadableString(fourCCImage);
            fi_log(LOG_ERR, "%s: FiFrame::CreateFrame() returned empty frameData.  Info: fourCC[%s] orient[%d] frameDataOffset[%d] attachedFrameSize[%d] w[%d] h[%d] frameDataLenMinusHdr[%d]",
                   __PRETTY_FUNCTION__, fourCCReadable.c_str(), (int)frameOrientation, frameDataOffset, attachedFrameSize, header.dib.width, frameHeight, frameDataLenMinusHdr);
            goto exitGetAttachedFrameData;
        }

        // If make it to this point in execution an FiFrame object was created.
        // Can it receive frame data?
        if (!frameData->FrameReadyToReceiveData())
        {
            fi_log(LOG_ERR, "%s: The FiFrame cannot receive frame data.", __PRETTY_FUNCTION__);
            frameData.reset();
            goto exitGetAttachedFrameData;
        }

        // Can we simply return the frameDataMinusHdr information?
        if (!frameData->CopyFrameData(frameDataPtrMinusHdr, frameDataLenMinusHdr, needsFlip))
        {
            fi_log(LOG_ERR, "%s: Unable to copy frame data for FiFrame.", __PRETTY_FUNCTION__);
            goto exitGetAttachedFrameData;
        }
        // We will return the normalized bitmap data frame pointer and size.
        frameDataPtr = frameData->GetFrameDataPtr();
        attachedFrameSize = frameData->GetFrameDataSize();
    }
    else
    {
        // Using FiFrame factory create normalized frame using existing data that is already
        // set in the frame orientation that is requested.
        frameData = FiFrame::CreateFrame(frameCreatedOK, fourCCImage,
                                         (uint32_t)header.dib.width, frameHeight,
                                         rowAlign,
                                         frameOrientation,
                                         frameDataPtrMinusHdr, frameDataLenMinusHdr);

        // If unable to create the normalized frame then cannot continue.
        if (!frameCreatedOK)
        {
            const std::string fourCCReadable = FiVideo::GetFourCCReadableString(fourCCImage);
            fi_log(LOG_ERR, "%s: FiFrame::CreateFrame() failed.  Could external frame provided be corrupted?  Info: fourCC[%s] orient[%d] frameDataOffset[%d] attachedFrameSize[%d] w[%d] h[%d] frameDataLenMinusHdr[%d]",
                   __PRETTY_FUNCTION__, fourCCReadable.c_str(), (int)frameOrientation, frameDataOffset, attachedFrameSize, header.dib.width, frameHeight, frameDataLenMinusHdr);
            goto exitGetAttachedFrameData;
        }

        // If we don't have any frame data don't attempt set it up.
        if (!frameData)
        {
            const std::string fourCCReadable = FiVideo::GetFourCCReadableString(fourCCImage);
            fi_log(LOG_ERR, "%s: FiFrame::CreateFrame() returned empty frameData.  Info: fourCC[%s] orient[%d] frameDataOffset[%d] attachedFrameSize[%d] w[%d] h[%d] frameDataLenMinusHdr[%d]",
                   __PRETTY_FUNCTION__, fourCCReadable.c_str(), (int)frameOrientation, frameDataOffset, attachedFrameSize, header.dib.width, frameHeight, frameDataLenMinusHdr);
            goto exitGetAttachedFrameData;
        }

        // Simply use the attached image data's raw bitmap frame as it did not need to be flipped.
        frameDataPtr = frameDataPtrMinusHdr;
        attachedFrameSize = frameDataLenMinusHdr;
    }

    // Using the normalized frame also initialize normalized plane data in case caller wants it later.
    if (!frameData->GetPlanarData(framePlaneData))
    {
        fi_log(LOG_ERR, "%s: Unable to locate Planar data for BMP image frame.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Fill in the planar data for the caller based on the attached frame data ptr.
    if (!GetAttachedFramePlanarData(planarDataAttached))
    {
        fi_log(LOG_ERR, "%s: Unable to verify Planar data for BMP image frame.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Make it here and all checks and preparations for the frame were successful.
    frameReadyToUse = true;

exitGetAttachedFrameData:

    // If we did not reach ready-to-use status we will return a null pointer.
    if (!frameReadyToUse)
    {
        frameDataPtr = nullptr;
    }

    // The return read-only pointer to the bitmap data and the size was set in reference.
    return frameDataPtr;
}

/*!
 * @brief Retrieves planar information for frame data of attached image.
 *
 * The data returned provides read-only access to plane data if it exists.
 *
 * @param planarDataToSet  The planar information requested.
 *
 * @returns true if planar data is populated with at least one plane or false if not.
 */
bool FiBMP::GetAttachedFramePlanarData(FAS_VIDEO::FRAME_PLANAR_DATA& planarDataToSet)
{
    bool onePlaneExists = (framePlaneData.planeCount > 0) ? true : false;

    // Initialize the passed in record to empty state.
    planarDataToSet = {0, {0}, {nullptr}};

    uint32_t planeIndex;
    for (planeIndex = 0; planeIndex < FAS_VIDEO::MAX_PLANES; planeIndex++)
    {
        if (planeIndex < framePlaneData.planeCount)
        {
            planarDataToSet.rowsizePerPlane[planeIndex] = framePlaneData.rowsizePerPlane[planeIndex];
            planarDataToSet.planarData[planeIndex] = framePlaneData.planarData[planeIndex];
        }
    }
    planarDataToSet.planeCount = framePlaneData.planeCount;

    // Return indication of at least one plane existing in the frame data.
    return onePlaneExists;
}



