/*!
 * @file fi_bmp_planar.h
 * @brief Definitions and declarations for Fastec BMP planar images.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <unistd.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

#if defined(_LINUX)
#include <linux/limits.h>
#endif

#include "fi_bmp.h"
#include "fi_bmp_planar.h"
#include "fas_video.h"
#include "log_if.h"

namespace FI_BMP
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------
//! Bit depth is 8 bits per pixel.
const uint16_t BMP_PLANAR_BIT_DEPTH     = 8;

//! Color planes must be set to 1.
const uint16_t BMP_PLANAR_COLOR_PLANES  = 1;

//! We need 2 alignment bytes after the header (or after header+colortable) to
//! align the image on a 4 byte boundary for DMA read purposes.
const int IMAGE_ALIGNMENT_BYTES     = 2;

// The BMP_MAGIC_* constants are kept out of the header structure to avoid alignment
// problems on processors that have trouble with unaligned structure members.

//! The number of bytes in the BMP_MAGIC_* string must be 2.
const uint32_t BMP_MAGIC_SIZE        = 2;

//! The BMP magic 2 byte code for our Y8 planar format.
const char *BMP_MAGIC_Y8            = "Y8";

//! The BMP magic 2 byte code for our NV12 planar format.
const char *BMP_MAGIC_NV           = "NV";

} // End of FI_BMP

// Make things easier for ourselves.
using namespace FI_BMP;
using namespace FAS_VIDEO;

FiBMPPlanar::FiBMPPlanar() : FI_BMP::FiBMP()
{
}

/*!
 * @brief Create a BMP header based on the specified parameters.
 *
 * @param fourcc    One of the supported FAS_VIDEO @ref fv_fourcc "FourCC" codes.
 * @param meta      Pointer to the image's metadata.
 * @returns The number of bytes the header uses, or a negative number
 *      indicating an error.
 */
int FiBMPPlanar::CreateHeader(uint32_t fourcc, const pm_metadata *meta)
{
    uint32_t bytesPerPixel = 1;       // BYTES-per-pixel is always 1 (8 bits per pixel).

    // Sanity check.
    if (meta == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: meta pointer is null!", __PRETTY_FUNCTION__);
        return -1;
    }

    // Save the data in base class properties so it persists.
    headerFourCC = fourcc;
    memcpy(&headerMetaData, meta, sizeof(pm_metadata));

    // Initialize our length and zero the header.
    hdr_size = sizeof(header);
    memset(&header, 0, sizeof(header));

    // Initialize the constant settings.
    header.dib.depth = BMP_PLANAR_BIT_DEPTH;
    header.dib.data_header_size = sizeof(header.dib);
    header.dib.color_planes = BMP_PLANAR_COLOR_PLANES;

    // If we're creating a monochrome image, then we need
    // to account for the color table.
    switch (fourcc)
    {
    case FOURCC_Y8:
    case FOURCC_NV12:
        break;

    default:
        // Reset the headerFourCC to a FOURCC_NONE as the one passed in is invalid.
        headerFourCC = FOURCC_NONE;

        // Oops. This isn't right.
        hdr_size = -1;
        IsFourCCValid(fourcc, __PRETTY_FUNCTION__);
        break;
    }
    if (hdr_size < 0)
        goto head_bail;

    // Add these alignment bytes since the BMP header size is
    // not a multiple of words.
    hdr_size += IMAGE_ALIGNMENT_BYTES;

    // Calculate the image size. Start with the image width (in BYTES).
    header.dib.data_size = meta->roi_width * bytesPerPixel;

    // Align the row width.
    header.dib.data_size = (header.dib.data_size + (ROW_ALIGNMENT - 1)) & (~(ROW_ALIGNMENT - 1));

    // Now multiply by the height to get total size.
    header.dib.data_size *= meta->roi_height;

    // The image data will be after the header.
    header.data_offset = static_cast<uint32_t>(hdr_size);

    // The length is the total length of the file.
    header.length = header.dib.data_size + header.data_offset;

    // Save the width.
    header.dib.width = meta->roi_width;

    // Save the height.
    header.dib.height = meta->roi_height;

    // We have created a header!
    createdHeader = true;

head_bail:
    return hdr_size;
}

/*!
 * @brief Get a unique pointer to an array of bytes containing
 *    the BMP header.
 *
 * The returned data is ready to be prepended onto an image to
 * create a BMP file.
 *
 * @returns A std::unique_ptr<char> containing the BMP header.
 */
std::unique_ptr<FAS_VIDEO::HEADER_DATA> FiBMPPlanar::GetHeader()
{
    uint8_t *current_position;      // Write position in buffer.
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> hdr_ptr = std::unique_ptr<FAS_VIDEO::HEADER_DATA>(new FAS_VIDEO::HEADER_DATA(static_cast<uint64_t>(hdr_size), 0));

    // Get our pointer.
    current_position = hdr_ptr->data();

    // Set the magic number
    if (headerFourCC == FOURCC_Y8)
    {
        memcpy(current_position, BMP_MAGIC_Y8, strlen(BMP_MAGIC_Y8));
        current_position += strlen(BMP_MAGIC_Y8);

    }
    else if (headerFourCC == FOURCC_NV12)
    {
        memcpy(current_position, BMP_MAGIC_NV, strlen(BMP_MAGIC_NV));
        current_position += strlen(BMP_MAGIC_NV);
    }
    else
    {
        // Cannot verify that a valid header exists.
        std::unique_ptr<FAS_VIDEO::HEADER_DATA> emptyHdrPtr;

        // This object has not had a header created yet.
        fi_log(LOG_ERR, "%s: The image header must be created before you can ask for it.", __PRETTY_FUNCTION__);
        return emptyHdrPtr;
    }

    // Copy the header into the buffer
    memcpy(current_position, ((const char *)&header), sizeof(header));
    //current_position += sizeof(header);

#ifdef USE_VERBOSE_INFO_DEBUG
    uint8_t *magicPtr = hdr_ptr->data();      // Write position in buffer.
    fi_log(LOG_DEBUG, "%s: BMP_PLANER, MAGIC[0]=%d MAGIC[1]=%d BMP_MAGIC_NV[0]=%d BMP_MAGIC_NV[1]=%d.",
           __PRETTY_FUNCTION__, magicPtr[0], magicPtr[1], BMP_MAGIC_NV[0], BMP_MAGIC_NV[1]);
#endif

    // Return the pointer.
    return hdr_ptr;
}

/*!
 * @brief Query the image format supported by this object.
 *
 * @returns The IMAGE_FORMAT value.
 */
FAS_VIDEO::IMAGE_FORMAT FiBMPPlanar::GetImageFormat()
{
    // This is a BMP Planar image.
    return FAS_VIDEO::IMAGE_FORMAT::FI_BMP_PLANAR_IMAGE;
}

/*!
 * @brief Query the image width of created header or attached image.
 *
 * @returns The image width or 0 if there is no known value.
 */
uint32_t FiBMPPlanar::GetImageWidth()
{
    uint32_t imageWidth = 0;

    // If there is a header created for this image use its data.
    if (BMPHeaderExists())
    {
        // Pull the image width from the header information.
        imageWidth = (uint32_t) header.dib.width;
    }
    else if (HasKnownImageInfo())
    {
        // At least one image was attached saving this information it is valid to use it.
        imageWidth = FiImage::GetImageWidth();
    }
    return imageWidth;
}

/*!
 * @brief Query the image height of created header or attached image.
 *
 * @returns The image width or 0 if there is no known value.
 */
uint32_t FiBMPPlanar::GetImageHeight()
{
    uint32_t imageHeight = 0;

    // If there is a header created for this image use its data.
    if (BMPHeaderExists())
    {
        // Extract the width, handle the - height case used to flip image
        imageHeight = (uint32_t) header.dib.height;
    }
    else if (HasKnownImageInfo())
    {
        // At least one image was attached saving this information it is valid to use it.
        imageHeight = FiImage::GetImageHeight();
    }
    return imageHeight;
}

/*!
 * @brief The bits-per-pixel bit depth for bmp planar image should always be 8.
 *
 * @returns The bit depth from header or BMP_PLANAR_BIT_DEPTH.
 */
uint32_t FiBMPPlanar::GetBitsPerPixel()
{
    uint32_t bitDepth = BMP_PLANAR_BIT_DEPTH;

    // It's always 1.
    return bitDepth;
}

/*!
 * @brief Returns the row alignment for BMP planar image data.
 *
 * @returns The row alignment for bitmap frames.
 */
uint32_t FiBMPPlanar::GetRowAlign()
{
    uint32_t rowAlign = 1;

    // Always 1 (for now).
    return rowAlign;
}

/*!
 * @brief Returns total number of uint8_t bytes in a frame of data.
 *
 * @returns 0 in this base class.
 */
uint32_t FiBMPPlanar::GetFullFramesize()
{
    uint32_t frameSize  = 0;
    uint32_t rowWidth;
    uint32_t rowAlign   = GetRowAlign();

    // If Y8 there is only one plane and it is the width * height.
    if (headerFourCC == FOURCC_Y8)
    {
        rowWidth = GetImageWidth();

        // Calculate the uncompressed image size for Y Plane.
        // Add proper padding to align the beginnings of rows.
        frameSize = (uint32_t)(rowWidth + (rowAlign - 1)) & (uint32_t)(~(rowAlign - 1));
        frameSize *= GetImageHeight();
    }
    else if (headerFourCC == FOURCC_NV12)
    {
        rowWidth = GetImageWidth();
        uint32_t rowHeight = GetImageHeight();

        // The Y plane is a full width and height.
        uint32_t yPlaneSize = (uint32_t)(rowWidth + (rowAlign - 1)) & (uint32_t)(~(rowAlign - 1));
        yPlaneSize *= rowHeight;

        // The U and V planes are 2X2 subsampled areas or 1/4 the size of the Y plane.
        // The U/V values are interleaved in a single plane so there are only two planes.
        //
        // The UV-plane is on a HEXWORD (64-byte) aligned offset.
        frameSize = ((yPlaneSize + 63) >> 6) << 6;
        frameSize += (yPlaneSize >> 1);
    }
    else
    {
        // This object is not set up properly.
        IsFourCCValid(headerFourCC, __PRETTY_FUNCTION__);
    }
    return frameSize;
}

uint32_t FiBMPPlanar::GetHeaderSize()
{
    return (uint32_t) hdr_size;
}

/*!
 * @brief Test the data in a BMP Planar image to verify it and pull out dimensions.
 *
 * @param imageDataToVerify     The BMP Planar data to validate.
 * @param imageDataSize         The number of bytes in the BMP Planar data.
 *
 * @returns true if BMP Planar data was verified and attach dimension data interpreted.
 */
bool FiBMPPlanar::VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize)
{
    bool verifiedImage = false;     // Keeps track of verification success.
    bool hasHeader = false;         // Does the imageDataToVerify have a header?

    // BMP Planar specific information to verify in the image data.
    uint32_t minHdrSize = sizeof(header) + BMP_MAGIC_SIZE;
    uint32_t fourCCFromData = FOURCC_NONE;

    //const BMP_HEADER *bmpHdrPtr = nullptr;

    // Reset any normalized frame data.
    frameData.reset();
    framePlaneData = {0, {0}, {nullptr}};

    // Make sure the imageDataToVerify is valid.
    if ((imageDataToVerify == nullptr) || (imageDataSize < minHdrSize))
    {
        fi_log(LOG_ERR, "%s: imageDataToVerify either empty or imageDataSize[%u] less than minHdrSize[%u].", __PRETTY_FUNCTION__, imageDataSize, minHdrSize);
        goto exitVerifyBMPPlanarImage;
    }

    // The imageDataToVerify may or may not include embedded BMP header information.  This must
    // be verified by checking for BMP header codes that identify it as header information.
    // If there is no header data provided inside of imageDataToVerify then it is expected
    // that a header was created for this image object before any imageDataToVerify was attached
    // and that header information exists in our "this.header" data member.
    // NOTE:  It is important to understand that in the "wild" we have come across frame data that
    // has the magic BMP header bytes in the front of it which will yield a false-positive and that
    // the code makes an attempt to identify these false-positives and only report a header when
    // the imageDataToVerify truly has one in the front of its data.
    if (isBmpY8((const char *)imageDataToVerify))
    {
        // Make sure that the headerFourCC matches Y8.
        fourCCFromData = FOURCC_Y8;

        // The imageDataToVerify has header information save this condition.
        hasHeader = true;

//        fi_log(LOG_ERR, "%s: FOURCC_Y8 BMP PLANER WITH HEADER.", __PRETTY_FUNCTION__);
    }
    else if (isBmpNV((const char *)imageDataToVerify))
    {
        // Make sure that the headerFourCC matches Y8.
        fourCCFromData = FOURCC_NV12;

        // The imageDataToVerify has header information save this condition.
        hasHeader = true;

//        fi_log(LOG_ERR, "%s: FOURCC_NV12 BMP PLANER WITH HEADER.", __PRETTY_FUNCTION__);
    }

    // PROCESS BMP Planar Frame.  No BMP header data is prepended to imageDataToVerify.
    if (!hasHeader)
    {
        // Does this image object have a BMP header?
        if (!headerExists())
        {
            // This object needs header information to process raw frames so reject the imageDataToVerify.
            fi_log(LOG_ERR, "%s: Missing Bitmap Planar Header on image frame: hdr_size[%u] expected [%u]. ",
                   __PRETTY_FUNCTION__, hdr_size, sizeof(header));

            goto exitVerifyBMPPlanarImage;
        }

        // We have verified the image.
        verifiedImage = true;
        goto exitVerifyBMPPlanarImage;
    }

    // Copy the header information into our bitmap header without the magic number.
    // If this object has no BMP header information copy it from imageDataToVerify.
    if (!CopyBMPHeaderData(imageDataToVerify, imageDataSize))
    {
        // This object needs header information to process raw frames so reject the imageDataToVerify.
        fi_log(LOG_ERR, "%s: Unable to copy Bitmap Planar Header from imageDataToVerify. ", __PRETTY_FUNCTION__);
        goto exitVerifyBMPPlanarImage;
    }

    // Make sure that the headerFourCC has the valid setting.
    headerFourCC = fourCCFromData;

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: BMP_PLANER  header.length[%u] hdr_size[%u] header.width[%ld] header.height[%ld] header.compression[%d] header.depth[%d] header.data_offset[%d].",
           __PRETTY_FUNCTION__, header.length, hdr_size, header.dib.width, header.dib.height, header.dib.compression, header.dib.depth, header.data_offset);
    fi_log(LOG_DEBUG, "%s: The image's width[%lu] height[%lu] depth[%lu] fourcc (0x%08x).",
           __PRETTY_FUNCTION__, header.dib.width, header.dib.height,
           header.dib.depth, fourCCFromData);
#endif

#if 0
    {
        const BMP_HEADER *bmpHdrPtr;

        // Verify the existing BMP Planar header data against the one inside of imageDataToVerify.
        bmpHdrPtr = (const BMP_HEADER *)(imageDataToVerify + BMP_MAGIC_SIZE);

        // Check the header width.
        if (bmpHdrPtr->dib.width != header.dib.width)
        {
            // There should be no mis-match between image header and imageDataToVerify's header.
            fi_log(LOG_ERR, "%s: The image's width[%lu] does not match stored header's width[%lu].",
                   __PRETTY_FUNCTION__, bmpHdrPtr->dib.width, header.dib.width);
            goto exitVerifyBMPPlanarImage;
        }

        // Check the header height.
        if (bmpHdrPtr->dib.height != header.dib.height)
        {
            fi_log(LOG_ERR, "%s: The image's height[%lu] does not match stored header's height[%lu].",
                   __PRETTY_FUNCTION__, bmpHdrPtr->dib.height, header.dib.height);
            goto exitVerifyBMPPlanarImage;
        }

        // Check the header depth.
        if (bmpHdrPtr->dib.depth != header.dib.depth)
        {
            fi_log(LOG_ERR, "%s: The image's depth[%lu] does not match stored header's depth[%lu].",
                   __PRETTY_FUNCTION__, bmpHdrPtr->dib.depth, header.dib.depth);
            goto exitVerifyBMPPlanarImage;
        }

        // Make sure there is a fourCC code for this image data too if it is not set.
        if (headerFourCC != fourCCFromData)
        {
            fi_log(LOG_ERR, "%s: The image's fourcc (0x%08x) does not match stored header's fourcc (0x%08x).",
                   __PRETTY_FUNCTION__, fourCCFromData, headerFourCC);
            goto exitVerifyBMPPlanarImage;
        }
    }
#endif

    // Passed all verification checks.  Mark it as verified.
    verifiedImage = true;

exitVerifyBMPPlanarImage:
    // Returns the status of the verification process.
    return verifiedImage;
}

/*!
 * @brief Retrieves a read-only frame data pointer to attached image data.
 * data minus any header if one exists.
 *
 * @param attachedFrameSize     (Reference)Set to length of frame data or zero if none exist.
 * @param planarDataAttached    (Reference)Set to the planar data of attached frame.
 * @param frameOrientation      How is the frame oriented.
 *
 * @returns read-only pointer to attached frame data or nullptr if none exist in addition
 * the references (attachedFrameSize,planarDataAttached) are output settings as well.
 */
const uint8_t *FiBMPPlanar::GetAttachedFrameData(uint32_t& attachedFrameSize,
                                                 FAS_VIDEO::FRAME_PLANAR_DATA& planarDataAttached,
                                                 FAS_VIDEO::FRAME_ORIENTATION frameOrientation)
{
    const uint8_t *frameDataPtr = nullptr;              // Set the read-only pointer to null.
    bool frameCreatedOK         = false;                // Returned from the frame creator to catch error condition.
    bool frameReadyToUse        = false;                // Must get through all tests to be usable.

    uint32_t frameDataOffset = 0;                       // If image data has BMP header this will be offset to frame data.

    uint32_t frameDataLenMinusHdr = 0;                  // How many bytes are in the actual frame data.
    const uint8_t *frameDataPtrMinusHdr = nullptr;      // Will point to the actual frame data skipping the header.
    uint32_t frameHeight = 0;                           // Will hold the positive frame height value.

    uint32_t fourCCImage   = FAS_VIDEO::FOURCC_NONE;    // Type of frame data.
    uint32_t rowAlign      = GetRowAlign();

    // Set the caller's size to zero.
    attachedFrameSize = 0;

    // Initialize the passed in planar record to empty state.
    planarDataAttached = {0, {0}, {nullptr}};

    // Can't process if there is no bitmap image attached.
    if (status != FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        fi_log(LOG_ERR, "%s: No bitmap planar image registered as attached.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Make sure the header fourcc has a valid value in it.
    if (!IsFourCCValid(headerFourCC, __PRETTY_FUNCTION__))
        goto exitGetAttachedFrameData;

    frameHeight = (uint32_t) header.dib.height;

    // TODO_JLC - Orientation should be set to TOP_DOWN for planar data.
    // Consider this an error if it is not set up this way.
    if (frameOrientation == FAS_VIDEO::FRAME_ORIENTATION::BOTTOM_UP)
    {
        fi_log(LOG_ERR, "%s: Planar data orientation must be TOP_DOWN!", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Look at the attached image data package to find the bitmap frame data.
    frameDataPtr = FiImage::GetAttachedImageData();
    attachedFrameSize = FiImage::GetAttachedImageSize();

    // Check to see we were able to get attached frame information.
    if (frameDataPtr && (attachedFrameSize > 0))
    {
        if (isBmpY8((const char *)frameDataPtr))
        {
            // The data says the image is formated as Y8. The image data immediately follows the header.
            fourCCImage = FOURCC_Y8;
            frameDataOffset = header.data_offset;
        }
        else if (isBmpNV((const char *)frameDataPtr))
        {
            // The data says the image is formated as NV12. The image data immediately follows the header.
            fourCCImage = FOURCC_NV12;
            frameDataOffset = header.data_offset;
        }
        else
        {
            // We don't have a header on the image. This is the case when we are saving locally to a device
            // connected to the camera controller. Use the fourCC information in the pre-built header. Since
            // there is no header on the image data, there is no offset.
            fourCCImage = headerFourCC;
            frameDataOffset = 0;
        }

        // Prepare access to frame data only.
        frameDataLenMinusHdr = attachedFrameSize - frameDataOffset;
        frameDataPtrMinusHdr = frameDataPtr + frameDataOffset;
    }
    else
    {
        fi_log(LOG_ERR, "%s: The attached BMP Planar image is null or empty.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Using FiFrame factory create the frame using existing data that is already
    // set in the frame orientation that is requested.
    frameData = FiFrame::CreateFrame(frameCreatedOK,
                                     fourCCImage,
                                     (uint32_t)header.dib.width, frameHeight,
                                     rowAlign,
                                     frameOrientation,
                                     frameDataPtrMinusHdr, frameDataLenMinusHdr);

    // If unable to create the normalized frame then cannot continue.
    if (!frameCreatedOK)
    {
        const std::string fourCCReadable = FiVideo::GetFourCCReadableString(fourCCImage);
        fi_log(LOG_ERR, "%s: FiFrame::CreateFrame() failed.  Could external frame provided be corrupted?  Info: fourCC[%s] orient[%d] frameDataOffset[%d] attachedFrameSize[%d] w[%d] h[%d] frameDataLenMinusHdr[%d]",
               __PRETTY_FUNCTION__, fourCCReadable.c_str(), (uint32_t)frameOrientation, frameDataOffset, attachedFrameSize, header.dib.width, frameHeight, frameDataLenMinusHdr);
        goto exitGetAttachedFrameData;
    }

    // If we don't have any frame data don't attempt set it up.
    if (!frameData)
    {
        const std::string fourCCReadable = FiVideo::GetFourCCReadableString(fourCCImage);
        fi_log(LOG_ERR, "%s: FiFrame::CreateFrame() returned empty frameData.  Info: fourCC[%s] orient[%d] frameDataOffset[%d] attachedFrameSize[%d] w[%d] h[%d] frameDataLenMinusHdr[%d]",
               __PRETTY_FUNCTION__, fourCCReadable.c_str(), (uint32_t)frameOrientation, frameDataOffset, attachedFrameSize, header.dib.width, frameHeight, frameDataLenMinusHdr);
        goto exitGetAttachedFrameData;
    }

    // Simply use the attached image data's raw bitmap frame as it did not need to be flipped.
    frameDataPtr = frameDataPtrMinusHdr;
    attachedFrameSize = frameDataLenMinusHdr;

    // Using the created frame also initialize plane data.
    if (!frameData->GetPlanarData(framePlaneData))
    {
        fi_log(LOG_ERR, "%s: Unable to locate Planar data for BMP Planar image frame.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Fill in the planar data for the caller based on the attached frame data ptr.
    if (!GetAttachedFramePlanarData(planarDataAttached))
    {
        fi_log(LOG_ERR, "%s: Unable to verify Planar data for BMP Planar image frame.", __PRETTY_FUNCTION__);
        goto exitGetAttachedFrameData;
    }

    // Make it here and all checks and preparations for the frame were successful.
    frameReadyToUse = true;

exitGetAttachedFrameData:

    // If we did not reach ready-to-use status we will return a null pointer.
    if (!frameReadyToUse)
    {
        frameDataPtr = nullptr;
    }

    // The return read-only pointer to the bitmap data and the size was set in reference.
    return frameDataPtr;
}

/*!
 * @brief Retrieves planar information for frame data of attached image.
 *
 * The data returned provides read-only access to plane data if it exists.
 *
 * @param planarDataToSet  The planar information requested.
 *
 * @returns true if planar data is populated with at least one plane or false if not.
 */
bool FiBMPPlanar::GetAttachedFramePlanarData(FAS_VIDEO::FRAME_PLANAR_DATA& planarDataToSet)
{
    bool onePlaneExists = (framePlaneData.planeCount > 0) ? true : false;

    // Initialize the passed in record to empty state.
    planarDataToSet = {0, {0}, {nullptr}};

    uint32_t planeIndex;
    for (planeIndex = 0; planeIndex < FAS_VIDEO::MAX_PLANES; planeIndex++)
    {
        if (planeIndex < framePlaneData.planeCount)
        {
            planarDataToSet.rowsizePerPlane[planeIndex] = framePlaneData.rowsizePerPlane[planeIndex];
            planarDataToSet.planarData[planeIndex] = framePlaneData.planarData[planeIndex];
            //#COMMENT_JLC
            //fi_log(LOG_INFO, "%s: planeIndex[%d] rowsizePerPlane[%d] planarData[%ll].", "FiBMPPlanar",
            //       planeIndex, planarDataToSet.rowsizePerPlane[planeIndex], planarDataToSet.planarData[planeIndex]);
        }
    }
    planarDataToSet.planeCount = framePlaneData.planeCount;

#if 0
    if (planarDataToSet.planeCount > 0)
    {
        uint32_t deltaPlane = (uint32_t)(planarDataToSet.planarData[1] - planarDataToSet.planarData[0]);
        //#COMMENT_JLC
        //fi_log(LOG_INFO, "%s: deltaPlane[%d].", "FiBMPPlanar", deltaPlane);
    }
#endif

    // Return indication of at least one plane existing in the frame data.
    return onePlaneExists;
}

/*!
 * @brief Copy the BMP header information from the front of an image data array.
 *
 * @returns A true value if BMP header was copied, false if unable to copy.
 */
bool FiBMPPlanar::CopyBMPHeaderData(const uint8_t *memDataPtr, uint32_t imageDataSize)
{
    bool imageDataCopied = false;                   // Has the data been copied to internal structures?
    uint32_t bmpMagicNbrSize = BMP_MAGIC_SIZE;      // How big is BMP Magic string.
    uint32_t bmpHeaderSize = sizeof(BMP_HEADER);    // How big is BMP Header structure.

    const uint8_t *curImageDataPtr = memDataPtr;

    // Note - this isn't really used right now because we should never receive
    // headerless images but it's here just in case in the future it is needed.

    // The memDataPtr must not be a nullptr.
    if (memDataPtr == nullptr)
        goto exitCopyBMPHeaderData;

    // The image data must have a BMP_HEADER + BMP_MAGIC size.
    if (imageDataSize < (bmpMagicNbrSize + bmpHeaderSize))
        goto exitCopyBMPHeaderData;

    // The image data must have a BMP_MAGIC indicator.
    if (!isBmpNV((const char *)memDataPtr) && !isBmpY8((const char *)memDataPtr))
        goto exitCopyBMPHeaderData;

    // Adjust our interim data by the size of the BMP_MAGIC number.
    curImageDataPtr += bmpMagicNbrSize;

    // There is enough data to get the bitmap header so copy the data.
    memcpy(&header, curImageDataPtr, bmpHeaderSize);

    // The header is from image data.
    createdHeader = false;

    // Adjust our interim data by the size of the BMP_HEADER.
    //curImageDataPtr += bmpHeaderSize;

    // Set the hdr_size to the size of the bmp header data.
    hdr_size = (int) bmpHeaderSize;

    // We can assume the data was copied correctly.
    imageDataCopied = true;

exitCopyBMPHeaderData:
    return imageDataCopied;
}

bool FiBMPPlanar::isBmpY8(const char *imageDataPtr)
{
    bool rc = false;

    // If we have "created" an external header, then there won't be an "embedded"
    // header in the data stream. If we did not create an "external" header, then
    // we need to look.
    if (!createdHeader)
    {
        // Does the image data have the Y8 signature?
        if (strncmp(imageDataPtr, FI_BMP::BMP_MAGIC_Y8, BMP_MAGIC_SIZE) == 0)
        {
            // Looks like there is an embedded Y8 header.
            rc = true;
        }
    }

    return rc;
}

bool FiBMPPlanar::isBmpNV(const char *imageDataPtr)
{
    bool rc = false;

    // If we have "created" an external header, then there won't be an "embedded"
    // header in the data stream. If we did not create an "external" header, then
    // we need to look.
    if (!createdHeader)
    {
        // Does the image data have the Y8 signature?
        if (strncmp(imageDataPtr, FI_BMP::BMP_MAGIC_NV, BMP_MAGIC_SIZE) == 0)
        {
            // Looks like there is an embedded NV12 header.
            rc = true;
        }
    }

    return rc;
}

/*!
 * @brief Is the FourCC being used in this object valid for this object.
 *
 * @param fourCCToCheck  The FourCC to check.
 *
 * @returns true if fourCCToCheck is known to this object false if not.
 */
bool FiBMPPlanar::IsFourCCValid(uint32_t fourCCToCheck, const char *whoIsCalling)
{
    bool fourCCIsValid = false;

    if (fourCCToCheck == FOURCC_Y8)
        fourCCIsValid = true;
    if (fourCCToCheck == FOURCC_NV12)
        fourCCIsValid = true;

    // Log the error if it is not valid.
    if (!fourCCIsValid)
        fi_log(LOG_ERR, "%s: Invalid fourcc (0x%08x)", whoIsCalling, fourCCToCheck);

    return fourCCIsValid;
}

/*!
 * @brief Does BMP Planar Header exist.
 *
 * @returns A true value if header exists, false if not.
 */
bool FiBMPPlanar::headerExists()
{
    // If the hdr_size is zero then there is no header information.
    // It can be exactly equal to sizeof(BMP_HEADER) if there is no color table.
    // It will be the sizeof(BMP_HEADER) + sizeof(color_table) if it has a color table.
    // Right now we are only testing for the size to be >= sizeof(BMP_HEADER).
    return (hdr_size >= (int)sizeof(BMP_HEADER));
}


