/*!
 * @file fi_frame.cpp
 * @brief The Fastec Frame Class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "fi_frame.h"
#include "fas_video.h"
#include "log_if.h"

#include <cstring>

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------
const bool LOG_ALL_FI_FRAME_INFO = false;

} // End of FAS_VIDEO

using namespace FAS_VIDEO;

//--------------------------------------------------------------------
// FiFrame Base Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief FiFrame Construction.
 */
FiFrame::FiFrame() :
    frameStatus(FAS_VIDEO::FRAME_STATUS::FRAME_NEW),
    fourccType(FAS_VIDEO::FOURCC_NONE),
    frameWidth(0),
    frameHeight(0),
    rowAlign(1),
    nPlanes(0),
    frameData(nullptr),
    frameOrientation(FAS_VIDEO::FRAME_ORIENTATION::TOP_DOWN)
{
}

/*!
 * @brief FiFrame Destruction.
 */
FiFrame::~FiFrame()
{
}

/*!
 * @brief Factory for FiFrame object
 *
 * @param frameFourCC   One of the FAS_VIDEO::FOURCC* codes for the type of frame to create.
 * @param frameWidth    The number of pixel columns across a frame row.
 * @param frameHeight   The number of rows down a frame.
 * @param rowAlign      The row alignment, set to 1 if byte aligned,
 *     2 if 16 bit aligned, or
 *     4 if 32 bit aligned, or
 *     8 if 64 bit aligned.
 * @param frameOrientation      Is frame oriented top-down or bottom-up.
 * @param externalFrameDataPtr  If not a nullptr a temporary creation using external pointer to avoid frame copy.
 * @param externalFrameDataLen  If externalFrameDataPtr is not a nullptr this should have a full framesize value.
 *
 * @returns A std::unique_ptr<FAS_VIDEO::FiFrame> containing the FiFrame object or nullptr if not created.
 */
std::unique_ptr<FAS_VIDEO::FiFrame> FiFrame::CreateFrame(bool& frameCreatedOk,
                                                         uint32_t frameFourCC,
                                                         uint32_t frameWidth, uint32_t frameHeight,
                                                         uint32_t rowAlign,
                                                         FAS_VIDEO::FRAME_ORIENTATION frameOrientation,
                                                         const uint8_t *extFrameDataPtrToUse,
                                                         uint32_t extFrameDataLenToUse)
{
    std::unique_ptr<FAS_VIDEO::FiFrame> fiFramePtr = nullptr;

    uint32_t frameSize  = 0;

    // This will only turn true if all steps are accomplished without early termination.
    frameCreatedOk      = false;

    // Generate the FiFrame object.
    fiFramePtr = std::unique_ptr<FAS_VIDEO::FiFrame>(new FiFrame());

    // If the frame pointer was not allocated return nullptr.
    if (!fiFramePtr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to allocate a new FiFrame.", __PRETTY_FUNCTION__);
        goto exitCreateFrame;
    }

    // Make sure the row alignment is correctly specified.
    if (fiFramePtr->IsRowByteAlignValid(rowAlign))
    {
        fiFramePtr->rowAlign = rowAlign;
    }
    else
    {
        fiFramePtr->rowAlign = 1;
        fiFramePtr->frameStatus =  FAS_VIDEO::FRAME_STATUS::INVALID_ROW_ALIGN;
        fi_log(LOG_DEBUG, "%s: Invalid row alignment[%d].", __PRETTY_FUNCTION__, rowAlign);
        goto exitCreateFrame;
    }

    // Initialize the FiFrame object.
    fiFramePtr->frameWidth = frameWidth;
    fiFramePtr->frameHeight = frameHeight;
    fiFramePtr->nPlanes = 0;
    fiFramePtr->fourccType = frameFourCC;
    fiFramePtr->frameOrientation = frameOrientation;

    // Perform validation on the FourCC code and set any specific data based on format.
    switch (frameFourCC)
    {
    case FAS_VIDEO::FOURCC_ARGB:
        // ARGB: packed 8-bit RGBA.
        // Four bytes per pixel: alpha, red, green, blue.
        fiFramePtr->frameBytesPerPixel = 4;
        break;
    case FAS_VIDEO::FOURCC_RGB:
        // RGB: Basic Windows bitmap format, 24-bit/pixel.
        fiFramePtr->frameBytesPerPixel = 3;
        fiFramePtr->nPlanes = 1;
        break;
    case FOURCC_BYR1:
        //BYR1: Raw Bayer data with 8-bits per sample.
        fiFramePtr->frameBytesPerPixel = 1;
        break;
    case FAS_VIDEO::FOURCC_BYR2:
        //BYR2: Raw Bayer data with 12-bits per sample stored
        //in 16-bit words.
        fiFramePtr->frameBytesPerPixel = 2;
        break;
    case FOURCC_NV12:
        // NV12: two-plane 8-bit YUV 4:2:0.
        // The first plane contains Y,
        // the second plane contains U and V in pairs of bytes 1/4 size of Y plane.
        // NOTE_JLC come back to this when processing raw frame data.
        // The Y plane is width * height
        // the U plane is (width/2) * (height/2)
        // the V plane is (width/2) * (height/2)
        fiFramePtr->frameBytesPerPixel = 1;
        fiFramePtr->nPlanes = 2;
        break;
    case FAS_VIDEO::FOURCC_Y8:
        // Y8: 8-bit greyscale.
        // Only a single sample, 8 bit Y plane for monochrome images.
        fiFramePtr->frameBytesPerPixel = 1;
        fiFramePtr->nPlanes = 1;
        break;
    case FAS_VIDEO::FOURCC_Y16:
        // Y16: 16-bit greyscale.
        // Only a single sample, 16 bit Y plane for monochrome images.
        fiFramePtr->frameBytesPerPixel = 2;
        break;
    default:
        fiFramePtr->fourccType = FAS_VIDEO::FOURCC_NONE;
        fiFramePtr->frameBytesPerPixel = 1;
        fiFramePtr->frameStatus = FAS_VIDEO::FRAME_STATUS::INVALID_FOURCC;
        break;
    }

    // If the status isn't valid can't continue.
    if (fiFramePtr->frameStatus != FAS_VIDEO::FRAME_STATUS::FRAME_NEW)
    {
        fi_log(LOG_DEBUG, "%s: Unable to create FiFrame status[%d].", __PRETTY_FUNCTION__, (int)fiFramePtr->frameStatus);
        goto exitCreateFrame;
    }

    // Mark the frame as being ready to facilitate framesize calculations.
    fiFramePtr->frameStatus = FAS_VIDEO::FRAME_STATUS::FRAME_READY;

    // How big does the buffer need to be to hold a frame of data?
    frameSize = fiFramePtr->GetFrameBufferSize();

    // For debugging only.  Should always be false otherwise.
    if (LOG_ALL_FI_FRAME_INFO)
    {
        const std::string fourCCStr = FiVideo::GetFourCCReadableString(frameFourCC);
        fi_log(LOG_DEBUG, "%s: Calculated frameSize[%u] extFrameDataLenToUse[%u]. Fourcc[%s] w[%d] h[%d] align[%d] rowSize[%d]",
               __PRETTY_FUNCTION__, frameSize, extFrameDataLenToUse,
               fourCCStr.c_str(), frameWidth, frameHeight, rowAlign, fiFramePtr->GetRowBufferSize());
    }

    // Depending on how this is being created determines the frame data
    // pointer installed.
    if (extFrameDataPtrToUse)
    {
        // Does the length match what is expected?
        if (frameSize != extFrameDataLenToUse)
        {
            const std::string fourCCStr = FiVideo::GetFourCCReadableString(frameFourCC);
            fi_log(LOG_ERR, "%s: Received invalid external frame data.  Calculated frameSize[%u] != extFrameDataLenToUse[%u]. Fourcc[%s] w[%d] h[%d] rowAlign[%d]",
                   __PRETTY_FUNCTION__, frameSize, extFrameDataLenToUse,
                   fourCCStr.c_str(), frameWidth, frameHeight, rowAlign);

            // If we ever get an external frame that doesn't match our frame size calculation
            // we will consider the video file in error.
            goto exitCreateFrame;
        }

        // Assume this is valid and assign it to the object.
        fiFramePtr->externalFrameDataLen = extFrameDataLenToUse;
        fiFramePtr->externalFrameDataPtr = extFrameDataPtrToUse;
    }
    else
    {
        // Create a buffer large enough to hold the frame data and
        // initialize all the bytes to zero.
        // NOTE:  For some reason frameSize was being cast to uint64_t however
        // frameSize should only be a uint32_t - maximum size is 4,294,967,295
        // I don't know why this was made uint64_t so I leaving this note in
        // place to mark this change back to uint32_t made on June 11, 2021.
        fiFramePtr->frameData = std::unique_ptr<FAS_VIDEO::FRAME_DATA>
                                (new FAS_VIDEO::FRAME_DATA(static_cast<uint32_t>(frameSize), 0));

        // Was frame data allocated?
        if (!fiFramePtr->frameData)
        {
            fiFramePtr->frameStatus = FAS_VIDEO::FRAME_STATUS::FRAME_ALLOC_ERR;
            fi_log(LOG_DEBUG, "%s: Unable to allocate frame data.", __PRETTY_FUNCTION__);
            goto exitCreateFrame;
        }
    }

    // If make it to this point in the scope of this function the frame we assume
    // was created ok.
    if (fiFramePtr)
    {
        frameCreatedOk = true;
    }

exitCreateFrame:
    return fiFramePtr;
}

/*!
 * @brief Access to width of frame.
 *
 * @note This does not include row alignment setting just strict frame width.
 *       Use GetRowBufferSize() to get a full row-size in terms of buffer space.
 *
 * @returns Width of the frame as set by construction.
 */
uint32_t FiFrame::GetWidth()
{
    return frameWidth;
}

/*!
 * @brief Access to height of frame.
 *
 * @returns Height of the frame as set by construction.
 */
uint32_t FiFrame::GetHeight()
{
    return frameHeight;
}

/*!
 * @brief Access to row alignment setting for frame.
 *
 * @returns Row Alignment of the frame as set by construction.
 */
uint32_t FiFrame::GetRowAlign()
{
    return rowAlign;
}

/*!
 * @brief Access to fourCC setting for frame.
 *
 * @returns Fourcc type of the frame as set by construction.
 */
uint32_t FiFrame::GetFourCC()
{
    return fourccType;
}

/*!
 * @brief Get size of a full frame of data in memory.
 *
 *  @returns The number of bytes needed to store a full frame in memory.
 */
uint32_t FiFrame::GetFrameBufferSize()
{
    uint32_t frameSize = 0;

    // Is the frame in a valid ready state?
    if (frameStatus == FAS_VIDEO::FRAME_STATUS::FRAME_READY)
    {
        frameSize = GetRowBufferSize() * frameHeight;
        //#MP4IMAGEINPUT - FiFrame::GetFrameBufferSize - Add other plane format support here!
        if (fourccType == FOURCC_NV12)
        {
            // The UV-plane is 1/2 the size of the Y-plane.
            uint32_t secondPlaneSize = frameSize / 2;

            // The UV-plane is on a HEXWORD (64-byte) aligned boundary.
            frameSize = (((frameSize + 63) >> 6) << 6) + secondPlaneSize;
        }
    }

    return frameSize;
}

/*!
 * @brief Get buffersize of a full row of frame of data in memory.
 *
 *  @return The number of bytes needed to store a full row in memory.
 */
uint32_t FiFrame::GetRowBufferSize()
{
    uint32_t rowSize = 0;

    // Is the frame in a valid ready state?
    if (frameStatus == FAS_VIDEO::FRAME_STATUS::FRAME_READY)
    {
        rowSize = frameWidth * frameBytesPerPixel;

        // If rowAlign is set make sure rowSize is byte aligned on specified boundary.
        if (rowAlign > 1)
        {
            uint32_t rowAlignMinusOne = rowAlign - 1;

            if (IsRowByteAlignValid(rowAlign))
            {
                if ((rowSize & rowAlignMinusOne) != 0)
                {
                    rowSize = ((rowSize + rowAlignMinusOne) / rowAlign) * rowAlign;
                }
            }
        }
    }

    return rowSize;
}

/*!
 * @brief Provides access to frame data on read-only basis.
 *
 * @returns A read-only pointer to begining of frame data or
 *          nullptr if no frame data exists.
 */
const uint8_t *FiFrame::GetFrameDataPtr()
{
    const uint8_t *frameDataPtr = nullptr;

    if (frameStatus == FAS_VIDEO::FRAME_STATUS::FRAME_READY)
    {
        if (frameData)
        {
            frameDataPtr = frameData->data();
        }
        else if (externalFrameDataPtr)
        {
            frameDataPtr = externalFrameDataPtr;
        }
    }

    return frameDataPtr;
}

/*!
 * @brief Provides access to frame data size.
 *
 * @returns The number of bytes in the frame data or
 *          zero if no frame data exists.
 */
uint32_t FiFrame::GetFrameDataSize()
{
    uint32_t frameDataSize = 0;

    if (frameStatus == FAS_VIDEO::FRAME_STATUS::FRAME_READY)
    {
        if (frameData)
        {
            frameDataSize = (uint32_t)frameData->size();
        }
        else if (externalFrameDataPtr)
        {
            frameDataSize = externalFrameDataLen;
        }
    }

    return frameDataSize;
}

/*!
 * @brief Provides access to planar information from frame.
 *
 * @returns A true state if planar data contains at least one plane, false if not.
 */
bool FiFrame::GetPlanarData(FRAME_PLANAR_DATA& planarDataToSet)
{
    uint32_t planeIndex = 0;
    bool onePlaneExists = (nPlanes > 0) ? true : false;

    const uint8_t *planeDataPtr = GetFrameDataPtr();
    uint32_t rowSize = GetRowBufferSize();

    planarDataToSet.planeCount = nPlanes;

    for (planeIndex = 0; planeIndex < MAX_PLANES; planeIndex++)
    {
        // This could be improved upon but let's get something in place.
        if (planeIndex == 0)
        {
            // Right now the currently supported planar formats all
            // have a full frame size for the first frame.  This should
            // fall out correctly because of rowSize is correct for FOURCC type.
            planarDataToSet.rowsizePerPlane[planeIndex] = rowSize;
            planarDataToSet.planarData[planeIndex] = planeDataPtr;

            // Point to the next plane. The UV-plane is on a HEXWORD (64-byte)
            // aligned offset.
            planeDataPtr += ((((rowSize * frameHeight) + 63) >> 6) << 6);
        }
        else if (planeIndex == 1)
        {
            if (fourccType == FOURCC_NV12)
            {
                // Right now only NV21 is supporting a secondary
                // plane that we can find in our frame data.
                // The FFMpeg library expects the rowsize to be equal
                // to the full frame rowsize here even though the actual
                // data is only 1/2 of the framesize.  When the rowsize is
                // set to the true rowsize of the UV data it causes the
                // FFMpeg library to assert failure.  The UV data in a
                // NV21 planar image is only 25% for each U and V or 50%
                // of the full framesize because it samples in 4x4 sections.
                // Don't do this:
                //planarDataToSet.rowsizePerPlane[planeIndex] = (rowSize >> 1);
                planarDataToSet.rowsizePerPlane[planeIndex] = rowSize;

                // The planeDataPtr should be pointing to the start of the
                // UV data after processing plane 0.
                planarDataToSet.planarData[planeIndex] = planeDataPtr;
            }
            else
            {
                // Populate this plane as empty because right now we
                // only handle multiple planes for NV21 images.
                planarDataToSet.rowsizePerPlane[planeIndex] = 0;
                planarDataToSet.planarData[planeIndex] = nullptr;
            }
        }
        else
        {
            // Populate this plane as empty.
            planarDataToSet.rowsizePerPlane[planeIndex] = 0;
            planarDataToSet.planarData[planeIndex] = nullptr;
        }
    }
    return onePlaneExists;
}

/*!
 * @brief Retrieve the setting for the frame orientation.
 *
 * @returns Setting provided when frame data was created in factory.
 */
FRAME_ORIENTATION FiFrame::GetFrameOrientation()
{
    return frameOrientation;
}

/*!
 * @brief Query to determine if this FiFrame object can receive data.
 *
 * NOTE:  Only frameData created FiFrame objects can receive data.
 * FiFrames that are storing read-only external frame data cannot.
 *
 * @returns True if the object can receive data, false otherwise.
 */
bool FiFrame::FrameReadyToReceiveData()
{
    uint32_t frameBufSize = 0;

    // Do some simple tests on the frame object and return false if doesn't pass.
    if (frameStatus != FAS_VIDEO::FRAME_STATUS::FRAME_READY)
        return false;

    if (!frameData)
        return false;

    // How many uint8_t bytes does a full frame of data need to store itself.
    frameBufSize = GetFrameBufferSize();

    // Verify that the size of the frameData is large enough for the frame.
    return (frameData->size() >= frameBufSize);
}

/*!
 * @brief Takes external frame data and copies it to the internal frameData vector.
 * It will flip the data from bottom to top if requested.
 *
 * @param srcFrameData      The pointer to first top row of source frame data.
 * @param srcFrameSize      The number of byte in the source frame data.
 * @param flipIt            The flip indicator, true if frame data is "upside-down".
 *
 * @returns True if the data was copied successfully, false otherwise.
 */
bool FiFrame::CopyFrameData(const uint8_t *srcFrameData, uint32_t srcFrameSize, bool flipIt)
{
    bool dataWasCopied = false;
    try
    {
        // Make sure this frame can receive row data.
        if (!FrameReadyToReceiveData())
        {
            fi_log(LOG_DEBUG, "%s: Frame is unable to receive data.", __PRETTY_FUNCTION__);
            goto exitCopyFrameData;
        }

        // Verify that the size of the source frame fits in this frame.
        // If they are not the right size something is not set up properly.
        uint32_t dstFrameSize = GetFrameBufferSize();
        if (dstFrameSize < srcFrameSize)
        {
            fi_log(LOG_DEBUG, "%s: SrcFrameSize[%d] != DstFrameSize[%d].", __PRETTY_FUNCTION__, srcFrameSize, dstFrameSize);
            goto exitCopyFrameData;
        }

        // This will tell the loop how to increment to next source row.
        // NOTE: Any row alignment is applied here too.
        uint32_t rowSize = GetRowBufferSize();

        // Create a source row and destination row index.
        uint32_t srcRow = 0;
        uint32_t dstRow = 0;

        // If flipping the data change destination row.
        if (flipIt)
        {
            dstRow = (frameHeight - 1) * rowSize;
        }

        // How many rows need to be flipped?
        uint32_t rowCount = GetHeight();
        uint8_t *writePos = frameData->data();

        // Transfer the frame data to our local frameData vector.
        while (rowCount > 0)
        {
            // Use memcpy to move the frame's row data.
            memcpy(&writePos[dstRow], &srcFrameData[srcRow], rowSize);
            srcRow += rowSize;

            // If flipping frame data subtract the rowsize.
            if (flipIt)
                dstRow -= rowSize;
            else
                dstRow = srcRow;

            // Reduce the row count.
            rowCount -= 1;
        }
        dataWasCopied = true;
    }
    // Look for a specific resource allocation error to set state of the reader object.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception caught [%s].", __PRETTY_FUNCTION__, e.what());
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught.", __PRETTY_FUNCTION__);
    }

exitCopyFrameData:
    return dataWasCopied;
}

/*!
 * @brief Verifies if the row alignment is a valid one.
 *
 * @param rowAlignToCheck  The row byte alignment to check.
 *
 * @returns True if byte alignment is valid, false if not.
 */
bool FiFrame::IsRowByteAlignValid(uint32_t rowAlignToCheck)
{
    if ((rowAlignToCheck == 1) ||
        (rowAlignToCheck == 2) ||
        (rowAlignToCheck == 4) ||
        (rowAlignToCheck == 8))
    {
        return true;
    }

    return false;   // Only aligning on 1, 2, 4, or 8 byte boundaries.
}



