/*!
 * @file fi_image_pkg.cpp
 * @brief The Fastec Image Package Class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <memory>

#include "fi_image.h"
#include "fas_video.h"
#include "log_if.h"

#include "fi_image_pkg.h"

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} // End of FAS_VIDEO

using namespace FAS_VIDEO;

//--------------------------------------------------------------------
// FiImage Base Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief FiImage Construction.
 */
FiImagePkg::FiImagePkg() :
    externalImagePtr(nullptr),
    externalImagePtrLen(0),
    internalImageData(nullptr)
{
}

FiImagePkg::~FiImagePkg()
{
}

/*!
 * @brief Creates the image data package designed to package up image data coming
 * from different sources.
 *
 * In this factory method the data is being managed by an external source.
 *
 * This external image data is typically created by the CameraServer.
 *
 * @param memDataPtr  The externally managed read-only pointer to uint8_t image data.
 * @param memDataLen  The number of bytes of image data.
 *
 * @returns A std::unique_ptr<FAS_VIDEO::FiImagePkg> containing
 * the image data package object or a nullptr if unable to create one.
 */
std::unique_ptr<FAS_VIDEO::FiImagePkg> FiImagePkg::Create(const uint8_t *memDataPtr, uint32_t memDataLen)
{
    std::unique_ptr <FAS_VIDEO::FiImagePkg> theImageDataPkg = std::unique_ptr <FAS_VIDEO::FiImagePkg>(new FiImagePkg());
    if (theImageDataPkg)
    {
        if (memDataPtr && (memDataLen > 0))
        {
            theImageDataPkg->externalImagePtr = memDataPtr;
            theImageDataPkg->externalImagePtrLen = memDataLen;
            theImageDataPkg->internalImageData = nullptr;
        }
    }

    // return ownership the newly minted image data package.
    return theImageDataPkg;
}

/*!
 * @brief Creates the image data package designed to package up image data coming
 * from different sources.
 *
 * In this factory method the data is being managed by an internal source.
 *
 * This data is typically created internally in the fas_video library.
 *
 * @param imageDataToOwn   The internally managed image data vector.
 *
 * @returns A std::unique_ptr<FAS_VIDEO::FiImagePkg> containing
 * the image data package object or a nullptr if unable to create one.
 */
std::unique_ptr<FAS_VIDEO::FiImagePkg> FiImagePkg::Create(std::unique_ptr<FAS_VIDEO::IMAGE_DATA> imageDataToOwn)
{
    std::unique_ptr <FAS_VIDEO::FiImagePkg> theImageDataPkg = std::unique_ptr <FAS_VIDEO::FiImagePkg>(new FiImagePkg());
    if (theImageDataPkg)
    {
        if (imageDataToOwn)
        {
            theImageDataPkg->externalImagePtr = nullptr;
            theImageDataPkg->externalImagePtrLen = 0;
            theImageDataPkg->internalImageData = std::move(imageDataToOwn);
        }
    }

    // return ownership the newly minted image data package.
    return theImageDataPkg;
}

/*!
 * @brief Returns size of image data in the package.
 *
 * @returns Returns zero if no valid image data in package or valid byte count.
 */
uint32_t FiImagePkg::GetImageSize()
{
    uint32_t retSize = 0;   // Default return is zero bytes.

    // Is it internal?
    if (internalImageData)
    {
        retSize = (uint32_t) internalImageData->size();
    }
    // Is it external?
    else if (externalImagePtr && (externalImagePtrLen > 0))
    {
        retSize = externalImagePtrLen;
    }

    return retSize;
}

/*!
 * @brief Returns read-only pointer to image data.
 *
 * @returns Returns nullptr if no valid image data in package or valid image data pointer.
 */
const uint8_t *FiImagePkg::GetMemDataPtr()
{
    const uint8_t *retImageDataPtr = nullptr;

    // Is it internal?
    if (internalImageData)
    {
        retImageDataPtr = (const uint8_t *) internalImageData->data();
    }
    // Is it external?
    else if (externalImagePtr && (externalImagePtrLen > 0))
    {
        retImageDataPtr = externalImagePtr;
    }

    return retImageDataPtr;
}
