/*!
 * @file fi_jpeg.cpp
 * @brief Definitions and declarations for Fastec JPEG images.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <string.h>

#if defined( QT_CORE_LIB )

// Not for Mac or Linux
#if !( defined(__linux__) || ( defined(__MACH__) && defined(__APPLE__) ))
// For Windows these are defined in winsock.h but we don't want to
// pull in all that baggage in general.
#define htons(a) ( (((a) & 0xff) << 8 ) | (((a) & 0xff00) >> 8) )
#define ntohs(a) htons( (a) )
#endif

#else

#include <arpa/inet.h>      // Needed for ntohs(), et. al.

#endif /* !QT_CORE_LIB */

#include "fi_image.h"
#include "fi_jpeg.h"
#include "fas_video.h"
#include "log_if.h"

namespace FI_JPEG
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

// The maximum bits per pixel of a JPEG image (4:2:0) subsampled but not compressed
//const uint32_t FI_MAX_JPEG_BPP  = 2;

// Used to validate that we have a JPEG frame
const uint8_t JPEG_SOI[] = {0xff, 0xd8 };

// Used to look for Start of frame tag where we will extract the width and height
const uint8_t JPEG_SOF0[] = {0xff, 0xc0 };

// These offsets are used to extract key information from a Start Of Frame Zer0 tag
// stored in the JPEG_SOF0 (SOF0) array.
//
// These are within the SOF0 information tag which starts with the FF C0 tag identifier
//    followed by two bytes that indicate the size of the SOF Tag
//    followed by one byte that indicates the data precision (Number of bits/component pixel)
//    followed by two bytes that are the image height
//    followed by two bytes that are the image width
//    followed by one byte that tells us how many component layers are in the image.
//const uint8_t JPEG_SOF0_TAGSIZE = 2;
const uint8_t JPEG_SOF0_DATA_PRECISION_OFFSET = 4;
const uint8_t JPEG_SOF0_HEIGHT_OFFSET = 5;
const uint8_t JPEG_SOF0_WIDTH_OFFSET = 7;
const uint8_t JPEG_SOF0_COMPONENTS_OFFSET = 9;

// If the value in JPEG_SOF0_COMPONENT_OFFSET is 1 then there is only
// one component layer (and one quantization table) and we assume the
// image data 1 * DATA PRECISION or 8 bits per pixel.
const uint8_t JPEG_SOF0_COMPONENT_Y = 1;

// If the value in the JPEG_SOF0_COMPONENT_OFFSET is 3 then there is only
// one component layer (and 2 quantization tables) and we assume the
// image data is 3 * DATA PRECISION or 24 bits per pixel.
const uint8_t JPEG_SOF0_COMPONENTS_YCbCr = 3;

// Create a couple of default values for initializing data precision and components.
const uint32_t JPEG_SOF0_DATA_PRECISION_DEFAULT = 8;    // 8 bits per pixel per color component(s).
const uint32_t JPEG_SOF0_COMPONENT_DEFAULT = 3;         // 3 components based on past historical logic for 3 component layers.

} // End of FI_JPEG

// Make things easier for ourselves.
using namespace FI_JPEG;
using namespace FAS_VIDEO;

/*!
 * @brief Standard class constructor.
 */
FiJPEG::FiJPEG() : FiImage(),
    attachedImageWidth(0),
    attachedImageHeight(0),
    attachedComponentSize(JPEG_SOF0_COMPONENT_DEFAULT),
    attachedPrecision(JPEG_SOF0_DATA_PRECISION_DEFAULT)
{
}

/*!
 * @brief Create a JPEG header based on the specified parameters.
 *
 * @param fourcc    One of the supported FAS_VIDEO @ref fv_fourcc "FourCC" codes.
 * @param meta      Pointer to the image's metadata.
 *
 * @returns The number of bytes the header uses, or a negative number
 *          indicating an error.
 */
int FiJPEG::CreateHeader(uint32_t fourcc, const pm_metadata *meta)
{
    (void)fourcc;
    (void)meta;

    return 0;
}

/*!
 * @brief Get a unique pointer to an array of bytes containing
 *    the JPEG header.
 *
 * The returned data is ready to be prepended onto an image to
 * create a JPEG file.
 *
 * @returns A std::unique_ptr<char> containing the BMP header.
 */
std::unique_ptr<FAS_VIDEO::HEADER_DATA> FiJPEG::GetHeader()
{
    // Following changes made to FiVideo::CreateImageHeader(...)
    // Go ahead and return an "empty" header rather than a nullptr
    // header.  It differentiates between errors and emptiness.
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> emptyHeader;
    emptyHeader = std::unique_ptr<FAS_VIDEO::HEADER_DATA>(new FAS_VIDEO::HEADER_DATA());
    return emptyHeader;
}

/*!
 * @brief Query the image format supported by this object.
 *
 * @returns The IMAGE_FORMAT value.
 */
FAS_VIDEO::IMAGE_FORMAT FiJPEG::GetImageFormat()
{
    // We are JPEGs.
    return FAS_VIDEO::IMAGE_FORMAT::FI_JPEG_IMAGE;
}

/*!
 * @brief Query the image width of created header or attached image.
 *
 * @returns The image width or 0 if there is no known value.
 */
uint32_t FiJPEG::GetImageWidth()
{
    uint32_t imageWidth = 0;

    if (status == FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        imageWidth =  attachedImageWidth;
    }
    else if (HasKnownImageInfo())
    {
        imageWidth = FiImage::GetImageWidth();
    }
    return imageWidth;
}

/*!
 * @brief Query the image height of created header or attached image.
 *
 * @returns The image width or 0 if there is no known value.
 */
uint32_t FiJPEG::GetImageHeight()
{
    uint32_t imageHeight = 0;

    if (status == FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        imageHeight = attachedImageHeight;
    }
    else if (HasKnownImageInfo())
    {
        imageHeight = FiImage::GetImageHeight();
    }
    return imageHeight;
}

/*!
 * @brief Query the bits per pixel setting.
 *
 * @returns The best value to use for compressed JPEG images.
 */
uint32_t FiJPEG::GetBitsPerPixel()
{
    // A return of 0 is considered an unknown bits per pixel setting as
    // we will have never had an attached image and no image is attached.
    uint32_t bitsPerPixel = 0;

    // If we have a valid attached image we should know it's bits per pixel
    // based on the component size multiplied by the data precision (which is
    // the number of bits per pixel in each component layer).  The attachedPrecision
    // should be 8 bits per pixel setting.  The attachedComponentSize will be
    // 1 or 3 depending on if image is monochrome or has color planes respectively.
    if (status == FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        bitsPerPixel = attachedComponentSize * attachedPrecision;
    }
    // If we are unattached has the base class stored at least one valid image
    // and therefore are the properties usable from a previously attached image?
    else if (HasKnownImageInfo())
    {
        bitsPerPixel = FiImage::GetBitsPerPixel();
    }

    return bitsPerPixel;
}


/*!
 * @brief Returns the row alignment for JPEG image data.
 *
 * @returns The row alignment for JPEG frames.
 */
uint32_t FiJPEG::GetRowAlign()
{
    uint32_t rowAlign = 1;
    return rowAlign;
}

/*!
 * @brief Estimates frame size of compressed JPEG frame of data.
 *
 * Use this for calculations used for video formats that need
 * access to this information.
 *
 * @returns Best estimate of a compressed framesize.
 */
uint32_t FiJPEG::GetFullFramesize()
{
    uint32_t frameSize = 0;
    if (status == FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        //NOTE_JLC - the FI_MAX_JPEG_BPP (value of 2) constant
        // was used in old functionality to calculate a "best-guess"
        // of a full-frame containing compressed JPEG image data.
        // We shouldrevisit this to see if we need to lower the
        // attachedComponentSize down if > 1.
        // This is typically called to create a placeholder location
        // in an AVI index layout for missed frames.

        // Assume image size worst case with no compression
        frameSize = attachedImageWidth * attachedImageHeight * attachedComponentSize;
    }
    else if (HasKnownImageInfo())
    {
        // Call the base class to get "known" information.
        uint32_t theWidth = FiImage::GetImageWidth();
        uint32_t theHeight = FiImage::GetImageHeight();
        //NOTE_JLC - the same comment above applies here too.
        //uint32_t theBytesPerPixel = FI_JPEG::FI_MAX_JPEG_BPP;
        uint32_t theBytesPerPixel = (FiImage::GetBitsPerPixel() / 8);

        // Using a "worst case compressed frame size guestimate.
        frameSize = theWidth * theHeight * theBytesPerPixel;
    }

    return frameSize;
}

/*!
 * @brief Test the data in a JPEG frame to verify it and pull out dimensions.
 *
 * @param imageDataToVerify     The frame data to verify JPEG format and dimensions.
 * @param imageDataSize         How many uint8_t bytes are in this JPEG frame.
 *
 * @returns true if JPEG frame was verified and attach dimension data interpreted.
 */
bool FiJPEG::VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize)
{
    bool validJPEG = false;

    uint16_t width = 0;         // The width of the image data as interpreted by the SOF0 tag.
    uint16_t height = 0;        // The height of the image data as interpreted by the SOF0 tag.

    // The data precision (bits/sample) of image data as interpreted by SOF0 tag.
    uint8_t dataPrecision = JPEG_SOF0_DATA_PRECISION_DEFAULT;
    // The number of components in image data as interpreted by SOF0 tag.
    uint8_t componentSize = JPEG_SOF0_COMPONENT_DEFAULT;

    uint32_t i = 0;             // General purpose loop index
    uint16_t jpeg_marker_len;   // Used to search for jpeg markers

    const uint8_t *firstBytePtr = nullptr;  // Create a position pointer in the image blob data.

    // Set attached image information to default values.
    attachedImageWidth = 0;
    attachedImageHeight = 0;
    attachedComponentSize = JPEG_SOF0_COMPONENT_DEFAULT;
    attachedPrecision = JPEG_SOF0_DATA_PRECISION_DEFAULT;

    // Some Notes:
    // This is the JPEG image object and the attachment process for FiImage has
    // invoked this pure virtual function to validate the image data as well as
    // to set up internal properties in the JPEG image class in the process.
    // Therefore this method will interpret the imagedata and extract the
    // image width/height/bit-depth from the image data itself.

    if (imageDataToVerify != nullptr)
    {
        // Perform some validation on the image data blob and extract some image information.

        // Get pointer to the beginning of the image blob.
        firstBytePtr = imageDataToVerify;

        // NOTE_JLC - This is legacy code to note that Windows media players don't like
        // MJPEG files with a Y only (grayscale) JPEG images.
        // Comment from old code below:
        // Windows media player does not like MJPEG files with Y only (gray scale) JPEGS]
        // On line 1144 of the fi_avi_videoc.cpp the old implementation.
        //*raw_format = FI_PIX_FMT_RGB;

        // First check for SOI (Start of Image) marker at beginning.
        if (memcmp(firstBytePtr, JPEG_SOI, sizeof(JPEG_SOI)) != 0)
        {
            // The JPEG_SOI number was not verified.  Unable to validate the data being attached.
            fi_log(LOG_ERR, "%s: JPEG_SOI number not found.", __PRETTY_FUNCTION__);
            status = IMAGE_STATUS::BAD_IMAGE_FORMAT;
            goto exitVerifyJPEGImageData;
        }

        // Scan until we find the SOF (start of frame)
        for (i = 2 ; i < imageDataSize; i += jpeg_marker_len + sizeof(uint16_t))
        {
            if (memcmp(firstBytePtr + i, JPEG_SOF0, sizeof(JPEG_SOF0)) == 0)
                break;

            memcpy(&jpeg_marker_len, firstBytePtr + i + sizeof(uint16_t), sizeof(uint16_t));
            jpeg_marker_len = (uint16_t)ntohs(jpeg_marker_len);
        }

        // Did we find an SOF?
        if (i >= imageDataSize)
        {
            // The JPEG_SOI number was not verified.  Unable to validate the data being attached.
            fi_log(LOG_ERR, "%s: Invalid JPEG Image Data, SOF0 not found.", __PRETTY_FUNCTION__);
            status = IMAGE_STATUS::BAD_IMAGE_FORMAT;
            goto exitVerifyJPEGImageData;
        }

        // Extract the width and height from the SOF (Start Of Frame Zer0) data
        memcpy(&width, firstBytePtr + i + JPEG_SOF0_WIDTH_OFFSET, sizeof(uint16_t));
        memcpy(&height, firstBytePtr + i + JPEG_SOF0_HEIGHT_OFFSET, sizeof(uint16_t));
        memcpy(&dataPrecision, firstBytePtr + i + JPEG_SOF0_DATA_PRECISION_OFFSET, sizeof(uint8_t));
        memcpy(&componentSize, firstBytePtr + i + JPEG_SOF0_COMPONENTS_OFFSET, sizeof(uint8_t));

#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "%s: SOF0 Read:  w=%u h=%u dp=%u component(s)=%u.",
               __PRETTY_FUNCTION__, (uint32_t)htons(width), (uint32_t)htons(height),
               dataPrecision, componentSize);
#endif

        // Does the data precision and the component size make sense?
        if ((componentSize != JPEG_SOF0_COMPONENT_Y) && (componentSize != JPEG_SOF0_COMPONENTS_YCbCr))
        {
            // The JPEG_SOI component setting was not verified.  Unable to validate the data being attached.
            fi_log(LOG_DEBUG, "%s: SOF0 Components Information[%u] expecting 1 or 3.", __PRETTY_FUNCTION__, componentSize);
            status = IMAGE_STATUS::BAD_IMAGE_FORMAT;
            goto exitVerifyJPEGImageData;
        }

        // We will be expecting a data precision value of 8 bits per pixel per component layer (color plane).
        // An RGB image is 24 bits per pixel (8 bits data precision * 3 component color planes).
        // A Monochrome image is simply 8 bits per pixel (8 bits of data precision * 1 component plane).
        // Save the image data properties we were able to glean from the SOF0.
        if (dataPrecision != JPEG_SOF0_DATA_PRECISION_DEFAULT)
        {
            // The JPEG_SOI data precision is not expected.  We will mark the image in a bad format.
            fi_log(LOG_DEBUG, "%s: SOF0 Components Information[%u] expecting 1 or 3.", __PRETTY_FUNCTION__, componentSize);
            status = IMAGE_STATUS::BAD_IMAGE_FORMAT;
            goto exitVerifyJPEGImageData;
        }

        // If make it here in the process flow then save all of the image information in object.
        attachedImageWidth = (uint32_t)htons(width);
        attachedImageHeight = (uint32_t)htons(height);
        attachedComponentSize = componentSize;
        attachedPrecision = dataPrecision;

        // The JPEG frame has been verified and the attached data properties stored.
        validJPEG = true;
    }

exitVerifyJPEGImageData:

    return validJPEG;
}

