/*!
 * @file fas_stream_mp4.cpp
 * @brief The h.264 MP4 stream object of the Fastec Video library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <mutex>
#include <iostream>
#include <fstream>

#include "fas_video.h"
#include "fi_stream_mp4.h"
#include "log_if.h"

namespace FAS_MP4
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! If you want to turn on verbose logging for this object set this to 1
//static const bool VERBOSE_SAVE_LOG = true;
static const bool VERBOSE_SAVE_LOG = false;

//! The array of presets stored in the object has a default of
//! FAS_VIDEO::H264_SPEEDS::H264_MEDIUM which is currently set
//! to element of index 1 in the list of 0-3 indices built in the
//! constructor.
const uint32_t PRESET_DEFAULT_INDEX = 1;

//! The CRF (Constant Rate Factor) - the frame quality in essence
//! used by the H.264 as a default will be 22 (70% quality in our
//! range settings)
const uint32_t CRF_DEFAULT  = 22;

//! The CRF (Constant Rate Factor) in 8 bit encoding scheme has a
//! range of 0 to 51 where 0 is lossless (highest quality) and 51
//! is setting for lowest quality.
//!
//! This implementation has a CRF_WORST and CRF_BEST range a little
//! less than the highest and lowest quality settings available.
const uint32_t CRF_WORST_QUALITY    = 43;
const uint32_t CRF_BEST_QUALITY    = 13;

// TO CALCULATE THE CRF Using a Percentage:
//
// CRF_WORST - CRF_BEST = TOTAL_RANGE.   Our setting 43 - 13 = 30 as totalRange.
// TOTAL_RANGE * PCT / 100 = RANGE_VALUE.  At 70% we have 30 * 70 / 100 = 21 rangeValue.
// CRF_WORST - RANGE_VALUE = CRF SETTING.  43 - 21 = 22 crfQualitySetting...
//
// 43 - (30 * 100pct / 100) = 13 CRF (which is CRF_BEST_QUALITY)
// 43 - (30 *  90pct / 100) = 16 CRF
// 43 - (30 *  80pct / 100) = 21 CRF
// 43 - (30 *  70pct / 100) = 22 CRF
// 43 - (30 *  60pct / 100) = 25 CRF
// 43 - (30 *  50pct / 100) = 28 CRF
// 43 - (30 *  20pct / 100) = 37 CRF
// 43 - (30 *  10pct / 100) = 40 CRF
// Anything less than 10% should end up being 43 which is CRF_WORST_QUALITY.

} // End of FAS_MP4

using namespace FI_FFMPEG;
using namespace FAS_VIDEO;
using namespace FAS_MP4;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

extern std::mutex theLostFrameMutex;
std::mutex theLostFrameMutex;

FiVideoMP4Stream::FiVideoMP4Stream(uint32_t h264QualityPct,
                                   FAS_VIDEO::H264_SPEEDS h264Speed,
                                   uint32_t keyFrameIntervalToUse) : FiVideoStream(),
    ffmpegVideo(nullptr),
    receivedFirstFrame(false),
    mp4FramesLost(0),
    mp4TotalFramesLost(0)
{
#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: Constructing FiVideoMP4Stream", __PRETTY_FUNCTION__);
#endif
    // Calculate the crfValue for a frame quality that is 90%.
    uint32_t crfFor90PctFrameQuality = CalculateCRFOption(90);

    // Fill in the H.264 preset possible settings that are supported at this time.
    // NOTE: that the string constants match exactly what the ffmpeg h.264 encoder
    // library expects for the "preset" option.  This correlates to our h264Speed.
    presetsList[0] = {FAS_VIDEO::H264_SPEEDS::H264_SLOWER, "slower"};
    presetsList[1] = {FAS_VIDEO::H264_SPEEDS::H264_MEDIUM, "medium"};
    presetsList[2] = {FAS_VIDEO::H264_SPEEDS::H264_FAST, "fast"};
    presetsList[3] = {FAS_VIDEO::H264_SPEEDS::H264_VERY_FAST, "veryfast"};

    // Calculate the user's frame quality in our h.264 crfValue format.
    crfValue = FiVideoMP4Stream::CalculateCRFOption(h264QualityPct);

    // Set the user's compression speed as the h.264 "preset" option.
    presetValue = FiVideoMP4Stream::ValidatePresetOption(h264Speed);

    // Set the keyframe interval setting.
    keyFrameInterval = FiVideoMP4Stream::SetKeyFrameInterval(keyFrameIntervalToUse);

    // Go through the compression speed selected and make some
    // choices depending on quality of video selection.
    switch (h264Speed)
    {
    case H264_SLOWER:
        // The user wants better compression, force swOverride here.
        swOverrideFlag = true;
        break;
    case H264_MEDIUM:
    case H264_FAST:
        swOverrideFlag = false;
        // Note:  The crfValue will be based on the Frame Quality the user has
        // selected represented by a percentage of 1 to 100%.
        // The lower "crfValue" indicates better quality and 100% happens to
        // correlate with a "crfValue" of 13.  See CalculateCRFOption function
        // for the transformation from percentage value to h.264 "crfValue".
        // We will override a HW Encoder if the user has selected a frame
        // quality of 90 percent or higher currently for these two compression
        // speed settings.  The compression speed is the "preset" h.264
        // encoder option.
        if (crfValue <= crfFor90PctFrameQuality)
        {
            swOverrideFlag = true;  // Quality of video overrides HW encoder.
        }
        break;
    case H264_VERY_FAST:
        // In this case let HW override SW.
        swOverrideFlag = false;
        break;
    default:
        swOverrideFlag = false;
        break;
    }
}

/*!
 * @brief Video stream destruction.
 */
FiVideoMP4Stream::~FiVideoMP4Stream()
{
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_INFO, "%s: Destructing FiVideoMP4Stream.", __PRETTY_FUNCTION__);
    }
    LogStreamStatus();

    //NOTE_JLC - Do NOT attempt to close the stream in the destructor.
    // Give stream hierarchy chance to close down if it set as closed.
    //if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED)
    //    CloseStream();
}

/*!
 * @brief Prepare MP4 video stream object for writing.
 *
 * @param videoNameToCreate What to name the created video file.
 * @param frameRateToUse    The number of frames to render per second of video.
 * @param frameWidthToUse   The number of pixel columns across a row.
 * @param frameHeightToUse  The number of rows in a video frame.
 * @param frameCountToUse   The number of frames for this video.
 *
 * @returns The status of the video streamer.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::PrepareForWriting(const std::string& videoNameToCreate,
                                                             uint32_t frameRateToUse,
                                                             uint32_t frameWidthToUse,
                                                             uint32_t frameHeightToUse,
                                                             uint32_t frameCountToUse)
{
    // The stream should be in a status of construction in order to prepare for writing.
    if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_BUILT)
    {
        // This method is expecting a newly constructed video stream object.
        // Deny the request and leave the stream status set to its state.
        return FAS_VIDEO::STREAM_STATUS::REQUEST_DENIED;
    }

    // Set the number of frames expected.
    frameCount = frameCountToUse;

    // Is there a range of frames to stream?
    if (frameCount == 0)
    {
        streamStatus = FAS_VIDEO::STREAM_STATUS::BAD_RANGE;
        fi_log(LOG_DEBUG, "%s: frameCount set to zero.", __PRETTY_FUNCTION__);
    }

    // Make sure there has been no error already.
    // If no error create the ofstream object to receive data.
    if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_BUILT)
    {
        // NOTE:  Sometimes the frame dimensions are unknown at the time
        // of the stream writer preparation.  It will be up to the first
        // add frame to look at the setting of headerProcessed to see
        // if any "housekeeping" is needed and waiting on the first image
        // so that the additional information is provided from the image
        // itself.

        // Save the frame dimension information.
        frameWidth = frameWidthToUse;
        frameHeight = frameHeightToUse;

        // NOTE_JLC is there a minimum frame rate allowed?  Should we flag something under say 10?  TBD.
        frameRate = frameRateToUse;

        // If made it here and the status is still STREAM_BUILT the stream is prepared without error.
        if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_BUILT)
        {
            streamType = FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER;
            streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_READY;
            videoNameOfStream = videoNameToCreate;
        }
    }

    // The stream object is initially set but until the first frame is received
    // The object has no idea what type of image frames will be sent.  Therefore
    // it cannot prepare the proper FFmpeg objects for transcoding or encoding.
    receivedFirstFrame = false;

    // Log the stream status in a debug scenario.
    LogStreamStatus();

    // Return the status of this stream object as being prepared.
    return streamStatus;
}

/*!
 * @brief Set up the video stream object for reading.
 *
 * @param videoNameToOpen         Name of the video file to open for reading.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::SetH264Options(uint32_t frameQualityPct,
                                                          FAS_VIDEO::H264_SPEEDS compressionSpeed,
                                                          uint32_t keyFrameIntervalToUse)
{
    FAS_VIDEO::STREAM_STATUS rcStreamStatus = streamStatus;   // Set it to the stream status initially.

    // If the stream isn't set up as a writer this makes no sense.
    if (streamType != STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        rcStreamStatus = FAS_VIDEO::STREAM_STATUS::REQUEST_DENIED;
        fi_log(LOG_DEBUG, "%s: must be configured as a stream writer.", __PRETTY_FUNCTION__);
        goto exitSetH264Options;
    }

    // If the stream isn't in a ready to receive state cannot set the options.
    if (!IsStreamReady())
    {
        rcStreamStatus = streamStatus;  // Set the return code to the actual stream status.
        fi_log(LOG_DEBUG, "%s: must be configured as a stream writer.", __PRETTY_FUNCTION__);
        goto exitSetH264Options;
    }

    // Calculate the Constant Rate Factor based on percentage set by user.
    crfValue = FiVideoMP4Stream::CalculateCRFOption(frameQualityPct);

    // Translate the compression speed enumeration into our index.
    presetValue = FiVideoMP4Stream::ValidatePresetOption(compressionSpeed);

    // Set the keyframe interval setting.
    keyFrameInterval = FiVideoMP4Stream::SetKeyFrameInterval(keyFrameIntervalToUse);

exitSetH264Options:
    return rcStreamStatus;
}

/*!
 * @brief Set up the video stream object for reading.
 *
 * @param videoNameToOpen         Name of the video file to open for reading.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::PrepareForReading(const std::string& videoNameToOpen)
{
    (void)videoNameToOpen;

    // This stream does not read MP4 video at this time.
    return FAS_VIDEO::STREAM_STATUS::REQUEST_DENIED;
}

/*!
 * @brief Writes an image frame to the video stream.
 *
 * @param imageToAdd    The FiImage containing an attached image data frame.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::AddVideoFrame(std::shared_ptr<FAS_VIDEO::FiImage> imageToAdd)
{
    bool addImageAttachedCheck = false;     // We are adding so don't check for attached image at top.
    bool addedFrame = false;                // Will only be true if make it all way through without error.

    uint32_t imageFrameWidth = 0;
    uint32_t imageFrameHeight = 0;

    // Pull the count of frames lost since we have a valid one and reset
    // the count because we are duplicating here now.
    uint32_t framesLostCount = ClearMP4FramesLost();

    try
    {
        // Want to build up duration information for the time taken up for
        // processing new video frames in this stream.
        if (addVideoFrameStats == nullptr)
            addVideoFrameStats = std::shared_ptr<FiDurations>(new FiDurations());
        else
            addVideoFrameStats->MarkStart();

        // If the processingVideoFrame is true this is an invalid condition that should not occur!
        if (processingVideoFrame)
        {
            fi_log(LOG_ERR, "%s: !!! DANGER !!! RE-ENTRY occurred in MP4 stream writer!", __PRETTY_FUNCTION__);

            // Go ahead and leave the streamer in its current stream status state
            // and return the re-entry error condition right away.
            return FAS_VIDEO::STREAM_STATUS::STREAM_REENTRY_ERROR;
        }

        // Mark the condition of processing a video frame.
        processingVideoFrame = true;

        // Verify that data can be written to the stream.
        if (!OkToWrite(__PRETTY_FUNCTION__, addImageAttachedCheck))
            goto exitAddVideoFrame;

        // Attempt to attach the image data as a video frame for the writer.
        if (!attachVideoFrameToWrite(imageToAdd))
        {
            goto exitAddVideoFrame;
        }

        // If code executes to here the stream object has "attached" the shared pointer
        // imageToAdd by copying the shared pointer to this stream's internal shared pointer
        // called videoFrameToWrite.  It is same pointer just shared internally for stream
        // operations until error or the next frame is attached.

        // What type of images are supported by this stream writer?
        // #MPIMAGEINPUT
        if ((videoFrameToWrite->GetImageFormat() != FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE) &&
            (videoFrameToWrite->GetImageFormat() != FAS_VIDEO::IMAGE_FORMAT::FI_JPEG_IMAGE) &&
            (videoFrameToWrite->GetImageFormat() != FAS_VIDEO::IMAGE_FORMAT::FI_BMP_PLANAR_IMAGE))
        {
            fi_log(LOG_DEBUG, "%s: video stream does not support FAS_VIDEO::IMAGE_FORMAT[%d].",
                   __PRETTY_FUNCTION__, (int)videoFrameToWrite->GetImageFormat());
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
            goto exitAddVideoFrame;
        }

        // Is this the first image frame received for the stream writer?
        if (!receivedFirstFrame)
        {
            // Change the state to note that the first frame was received.
            // It may or may not be processed properly but it did make it
            // to this video stream.
            receivedFirstFrame = true;

            // Make sure the stream's frame information matches the data
            // in the image.  The stream creation sometimes doesn't know
            // the real information for the frames coming to it until the
            // first one is added.
            //
            // The opportunity to determine real information is now
            // available to reuse for additional AddVideoFrame calls.
            //
            // The FFmpeg stream support will now be initialized.
            //
            // A) The first frame has been received, attached, and verified.
            // B) The frame information should be valid and reliable.
            //
            // The fields being set here using this first attached image include:
            //
            // FiStreamer::frameWidth
            // FiStreamer::frameHeight
            // FiStreamer::fullFrameSize
            //
            imageFrameWidth = videoFrameToWrite->GetImageWidth();
            imageFrameHeight = videoFrameToWrite->GetImageHeight();
            fullFrameSize = videoFrameToWrite->GetFullFramesize();

            // Make sure that the image extents are valid.
            if ((imageFrameWidth == 0) || (imageFrameHeight == 0))
            {
                fi_log(LOG_ERR, "%s: Attached image size error width[%lu] height[%lu].",
                       __PRETTY_FUNCTION__, imageFrameWidth, imageFrameHeight);
                streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
                goto exitAddVideoFrame;
            }

            // Check the image frame width.
            if (frameWidth != imageFrameWidth)
            {
                // What if the frameWidth stored isn't zero?
                // NOTE_JLC - should this be considered an error?
                if (frameWidth != 0)
                {
                    fi_log(LOG_DEBUG, "%s: !receivedFirstFrame and frameWidth[%lu] not zero, imageFrameWidth[%lu]",
                           __PRETTY_FUNCTION__, frameWidth, imageFrameWidth);
                }
                frameWidth = imageFrameWidth;
            }
            if (frameHeight != imageFrameHeight)
            {
                // What if the frameHeight stored isn't zero?
                // NOTE_JLC - should this be considered an error?
                if (frameHeight != 0)
                {
                    fi_log(LOG_DEBUG, "%s: !receivedFirstFrame and frameHeight[%lu] not zero, imageFrameHeight[%lu]",
                           __PRETTY_FUNCTION__, frameHeight, imageFrameHeight);
                }
                frameHeight = imageFrameHeight;
            }

            // If we lost frames in the front the ffmpeg container must duplicate this one
            // in order to get an accurate frame count and keep track of duplications.
            uint32_t framesLostInFront = ClearMP4FramesLost();

            // Prepare the FFmpeg object depending on the type of image data being processed.
            ffmpegVideo = FFmpegContainer::CreateMP4Container(videoNameOfStream, videoFrameToWrite,
                                                              frameRate, frameCount,
                                                              crfValue,
                                                              presetsList[presetValue].speedPreset.c_str(),
                                                              keyFrameInterval,
                                                              swOverrideFlag,
                                                              framesLostInFront);

            // If the FFmpeg video container was not created or in error report the condition.
            if (!ffmpegVideo)
            {
                // This is fatal and is considered a codec setup error condition.
                streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_SETUP_ERROR;
                // Was unable to create the container and write out the first frame.
                fi_log(LOG_DEBUG, "%s: Unable to create MP4 streaming container.",
                       __PRETTY_FUNCTION__);
                goto exitAddVideoFrame;
            }

            // Ticket# 257
            // There may be lost frames prior to receiving a good frame.  In this case
            // when we create our container object we need to reuse the good frame duplicating
            // any lost frame(s) in the stream.   Remember that MP4 video cannot go back and reinsert
            // missing frames in the video so our only action is to duplicate either the last
            // known good frame or the first known good frame.  In this case we have a known
            // good frame and we must duplicate it at the beginning of the video file.
            if ((framesLostCount > 0) && (FFmpegContainerOK()))
            {
                // The FFmpeg container stream is ok go ahead and encode frame.
                if (!ffmpegVideo->EncodeFrame(videoFrameToWrite, framesLostCount))
                {
                    streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;

                    // Provide additional information about the video stream.
                    ffmpegVideo->GetStatusInfo(additionalStatusInfo);

                    // Unable to encode the image frame to the stream.
                    fi_log(LOG_ERR, "%s: Unable to duplicate %d lost frame(s) in front of MP4 video.",
                           __PRETTY_FUNCTION__, framesLostCount);
                    goto exitAddVideoFrame;
                }
            }
        }
        else
        {
            if (FFmpegContainerOK())
            {
                // The FFmpeg container stream is ok go ahead and encode frame.
                if (!ffmpegVideo->EncodeFrame(videoFrameToWrite, framesLostCount))
                {
                    streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;

                    // Provide additional information about the video stream.
                    ffmpegVideo->GetStatusInfo(additionalStatusInfo);

                    // Unable to encode the image frame to the stream.
                    fi_log(LOG_DEBUG, "%s: Unable to encode image frame to MP4 stream.",
                           __PRETTY_FUNCTION__);
                    goto exitAddVideoFrame;
                }
            }
            else
                goto exitAddVideoFrame;
        }

        // Make here and we can safely assume the frame has been added successfully to the stream.
        addedFrame = true;

    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while adding video frame.", __PRETTY_FUNCTION__, e.what());
        goto exitAddVideoFrame;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: All Catch Exception (...) caught while adding video frame.", __PRETTY_FUNCTION__);
        goto exitAddVideoFrame;
    }

exitAddVideoFrame:
    // Cleanup appropriately based on whether or not the frame was added.
    if (!addedFrame)
    {
        //videoFrameToWrite.reset();    // Do we reset a shared pointer?  TODO_JLC Double check this!!
    }
    // Reset the re-entry boolean so that we don't false flag a re-entry condition.
    processingVideoFrame = false;

    // Mark the interval.
    if (addVideoFrameStats)
        addVideoFrameStats->MarkInterval();

    return streamStatus;
}

/*!
 * @brief Handles the case of a lost frame for the MP4 video stream.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::AddPlaceholderFrame()
{
    bool addImageAttachedCheck      = false;     // Don't check for attached image at top.
    uint32_t framesAddedSoFar       = 0;
    uint32_t lostFramesSoFar;
    try
    {
        // Gather duration interval inforamtion for adding placeholder frames.
        if (addPlaceholderFrameStats == nullptr)
            addPlaceholderFrameStats = std::shared_ptr<FiDurations>(new FiDurations());
        else
            addPlaceholderFrameStats->MarkStart();


        // If the processingVideoFrame is true this is an invalid condition that should not occur!
        if (processingVideoFrame)
        {
            fi_log(LOG_ERR, "%s: !!! DANGER !!! RE-ENTRY occurred in MP4 stream writer!", __PRETTY_FUNCTION__);

            // Go ahead and leave the streamer in its current stream status state
            // and return the re-entry error condition right away.
            return FAS_VIDEO::STREAM_STATUS::STREAM_REENTRY_ERROR;
        }

        // Mark the condition of processing a video frame.
        processingVideoFrame = true;

        // Verify that data can be written to the stream.
        if (!OkToWrite(__PRETTY_FUNCTION__, addImageAttachedCheck))
            goto exitAddPlaceholderFrame;

        // If we get a request to add a placeholder frame that means that
        // the frame was lost - we can't go back and "edit" frames so the
        // next frame in the system will be set to go twice in its place.
        // NOTE:  Make sure this is done here because there could be
        // missing frames in front of the video and the container is not
        // created until we have at least the first valid one.
        lostFramesSoFar = IncrementMP4FramesLost();

        if (FFmpegContainerOK())
        {
            uint32_t lostTotalFramesSoFar   = GetMP4TotalFramesLost();
            uint32_t framesAddedPlusLost;

            // Get the number of real frames processed through the encoder.
            framesAddedSoFar = ffmpegVideo->GetFramesAdded();

            // Determine how many frames we have processed up to this point.
            framesAddedPlusLost = lostFramesSoFar + framesAddedSoFar;

            // If we are within our range of frameCount output a message.
            if (framesAddedPlusLost < frameCount)
            {
                fi_log(LOG_WARNING, "%s: <up to this point> frameCount[%d] Added [%d] frame(s). MP4 lost [%d] frame(s) this gap.  MP4 lost [%d] frame(s) in video.",
                       __PRETTY_FUNCTION__, frameCount, framesAddedSoFar, lostFramesSoFar, lostTotalFramesSoFar);
            }

            // Should we expect more possible missing frames or have we reached
            // the end of expected frames for this stream?
            if (framesAddedPlusLost == frameCount)
            {
                // GAP at the end that can't be processed.
                fi_log(LOG_WARNING, "%s: Last [%d] frame(s) in video lost and [%d] total frames were lost in video.",
                       __PRETTY_FUNCTION__, lostFramesSoFar, lostTotalFramesSoFar);

                // What we want to do here is fake out the frames processed
                // to frameCount which should flush the encoder so that we
                // can produce at least a playable video even with missing
                // end frames.  The Log data will note the missing frame(s).
                while ((framesAddedSoFar <= frameCount) && (lostFramesSoFar > 0))
                {
                    // Each pass adds to framesAddedSoFar value.
                    ffmpegVideo->PadWithEmptyFramesAtEnd();
                    framesAddedSoFar = ffmpegVideo->GetFramesAdded();
                    lostFramesSoFar--;  // Keep a control variable to make sure no endless loop occurs.
                    fi_log(LOG_DEBUG, "%s: After PadWithEmptyFramesAtEnd() framesAddedSoFar [%d] frame(s) and frameCount[%d]", __PRETTY_FUNCTION__, framesAddedSoFar, frameCount);
                }

                // #MISSING_FRAME - you can uncomment out this line to force an error
                // condition for testing purposes.
                //streamStatus = STREAM_STATUS::STREAM_INCOMPLETE;
                goto exitAddPlaceholderFrame;
            }
        }
        else
        {
            // If we have never received any valid frames yet we have not fully allocated
            // our ffmpeg container object.  In this case we need to keep track of the condition.
            if (!receivedFirstFrame)
            {
                fi_log(LOG_DEBUG, "%s: Lost %d frames in front of video so far.", __PRETTY_FUNCTION__, lostFramesSoFar);
            }
            else
            {
                fi_log(LOG_ERR, "%s: Attempting to add a lost frame but ffmpeg container is not OK!", __PRETTY_FUNCTION__);
            }
            goto exitAddPlaceholderFrame;
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while adding placeholder video frame.", __PRETTY_FUNCTION__, e.what());
        goto exitAddPlaceholderFrame;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: All Catch Exception (...) caught while adding placeholder video frame.", __PRETTY_FUNCTION__);
        goto exitAddPlaceholderFrame;
    }

exitAddPlaceholderFrame:

    // Reset the re-entry boolean so that we don't false flag a re-entry condition.
    processingVideoFrame = false;

    // Mark the interval.
    if (addPlaceholderFrameStats)
        addPlaceholderFrameStats->MarkInterval();

    return streamStatus;
}

/*!
 * @brief This is called when the system wants to "edit" the frame that
 * was lost in the stream.  But because we are not able to facilitate this
 * in an MP4 video we just pass it through.  Will output a log message that
 * indicates this state and the number of frames this stream is recording as
 * lost.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::EditVideoFrame(uint32_t frame_num, std::shared_ptr<FAS_VIDEO::FiImage> imageDataFrame)
{
    (void)frame_num;
    (void)imageDataFrame;

    // There is no way to "edit" an MP4 video so this frame will just be ignored
    // as the lost frame received in the stream was duplicated.
    fi_log(LOG_DEBUG, "%s: Passing through MP4 stream, frame_num[%d] of total frames lost[%d].", __PRETTY_FUNCTION__, frame_num, GetMP4TotalFramesLost());

    // Just return the stream status of this video streamer.
    return streamStatus;
}

/*!
 * @brief Queries complete state of this stream.
 *
 * @returns The status of the stream writer.  If it is complete
 * without error the status should be FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::IsStreamWriterComplete()
{
    // Make sure this is a stream writer otherwise this makes no sense.
    if (streamType != FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        return FAS_VIDEO::STREAM_STATUS::STREAM_TYPE_ERROR;
    }

    // For the MP4 stream when frame count and encoded count are equal the video
    // should be complete.  In the MP4 encoder it will base the allFramesEncoded
    // on the state of the ffmpeg encoder and whether or not it wrote the AV
    // trailer.  This is very different from the encoded frames being equal to
    // the frame count because if the trailer wasn't written by the frame writer
    // thread then the video isn't truly complete yet.  Because the frame writer
    // is a thread we don't know until it sets the condition and terminates the
    // thread.  This state is reflected in the setting of allFramesEncoded and
    // can be relied upon to truly indicate an MP4 video is complete.

    // Note:  I am going to catch any encoded count > than frame count.  I don't
    // believe this could happen but with the ffmpeg encoder you are not 100%
    // guaranteed that it will exactly equal the frame count.  I believe the
    // ffmpeg container object ensures the equality but for the purposes of this
    // function I would hate to miss a case.
    if (GetFrameCount() <= GetFramesEncodedCount())
    {
        // If the stream status says all is good with the stream writer
        // then mark the streamStatus as being complete if all frames were
        // encoded tells us that it is truly complete.
        if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
        {
            bool ffmpegEncoderHadError = false;

            // For MP4 stream writers when querying for the encoded count
            // it will know whether or not the AV trailer was written which
            // then and only then is the MP4 video truly complete.
            if (ffmpegVideo->IsFrameWriterComplete(ffmpegEncoderHadError))
            {
                // The ffmpeg frame writer completed the video.  There could
                // still have been some sort of error condition here.  For now
                // just log the unexpected setting.
                if (ffmpegEncoderHadError)
                {
                    fi_log(LOG_ERR, "%s: [UNEXPECTED] ffmpegVideo reports Frame Writer Thread encountered an ffmpegError condition but the AV Trailer was still written out completing the video.", __PRETTY_FUNCTION__);
                }
                // Transition the stream status to a completed stream for the stream writer.
                streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED;
            }
        }
    }

    // Return the status set.
    return streamStatus;
}

/*!
 * @brief Loads a video frame from a video file.
 * For MP4 there is no stream reader supported.
 *
 * @returns A null unique pointer because this operation is not supported.
 */
std::unique_ptr<FAS_VIDEO::FiImage> FiVideoMP4Stream::LoadVideoFrame(uint32_t frameNum)
{
    (void)frameNum;

    // This stream does not support loading individual frames at this time.
    std::unique_ptr<FAS_VIDEO::FiImage> noImageAvailable;
    noImageAvailable.reset();
    return noImageAvailable;
}

/*!
 * @brief Closes the video stream and releases any allocated resources.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoMP4Stream::CloseStream()
{
    FAS_VIDEO::STREAM_STATUS streamStatusBeforeClose = streamStatus;

    std::string durationInfoStr;
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_DEBUG, "%s: Closing the MP4 stream.", __PRETTY_FUNCTION__);
    }
    try
    {
        if (VERBOSE_SAVE_LOG)
        {
            FiDurations::LogDurationSeparator();
            // If collected interval durations log them.
            if (addVideoFrameStats)
            {
                durationInfoStr = "[FiVideoMP4Stream::AddVideoFrame() Total Time from Entry to Exit for ALL Frames]";
                addVideoFrameStats->LogDurationInfo(durationInfoStr.c_str());
            }
            if (addPlaceholderFrameStats)
            {
                durationInfoStr = "[MP4 AddPlaceholderFrame]";
                addPlaceholderFrameStats->LogDurationInfo(durationInfoStr.c_str());
            }
            FiDurations::LogDurationSeparator();
        }

        // If this is a stream writer flush encoder if needed.
        if (streamType == FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
        {
            uint32_t framesToEncodeStill = false;
            bool inError = false;

            if (ffmpegVideo != nullptr)
            {
                // Are there still frames being encoded?
                if (!ffmpegVideo->IsVideoComplete(framesToEncodeStill, inError))
                {
                    std::string ffmpegVideoStreamStatus;
                    // This is a problem with the user of this stream.
                    streamStatusBeforeClose = streamStatus = STREAM_STATUS::STREAM_INCOMPLETE;

                    // Pull the stream status into user friendly string.
                    LogStreamStatus(ffmpegVideoStreamStatus, false);

                    fi_log(LOG_ERR, "%s: framesToEncode[%d](should be zero) ffmpegFrameWriterInError[%s] %s.",
                           __PRETTY_FUNCTION__, framesToEncodeStill, inError ? "Y" : "N", ffmpegVideoStreamStatus.c_str());
                    // You want to allow this to run through the close process.
                    //goto exitMP4CloseStream;
                }

                // Clean up the ffmpeg objects.
                ffmpegVideo->CloseDown();

                // And clean up the container.
                ffmpegVideo.reset();
                ffmpegVideo = nullptr;
            }
            else
            {
                // Can't be complete if there is no container object.
                streamStatusBeforeClose = STREAM_STATUS::STREAM_INCOMPLETE;
                fi_log(LOG_ERR, "%s: The ffmpegContainer object was null!", __PRETTY_FUNCTION__);
            }
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while closing video stream.", __PRETTY_FUNCTION__, e.what());
        goto exitMP4CloseStream;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: All Catch Exception (...) caught while closing video stream.", __PRETTY_FUNCTION__);
        goto exitMP4CloseStream;
    }

exitMP4CloseStream:

    // If any inkling that this stream did not complete correctly rename the
    // resultant video file because it is not considered properly completed.
    if (streamStatusBeforeClose != FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED)
    {
        std::string videoNameWithErrorTag = videoNameOfStream + FiVideoStream::GetIncompleteVideoFileString();
        std::string statusLog;
        LogStreamStatus(statusLog, false);

        fi_log(LOG_ERR, "%s: MP4 streamStatus is not marked complete.  Writing %s to document incomplete status.", __PRETTY_FUNCTION__, videoNameWithErrorTag.c_str());
        // using ofstream constructors.
        std::ofstream outfile(videoNameWithErrorTag.c_str());
        outfile << "MP4 video creation is incomplete and was not fully processed." << std::endl;
        outfile << "Please review log data for possible error conditions." << std::endl;
        outfile << "Status:" << statusLog << std::endl;
        outfile.close();
        return streamStatusBeforeClose;
    }
    else
    {
        // Call the base class' CloseStream to set the stream status closed.
        FiVideoStream::CloseStream();
    }

    // Return the status of this stream writer.
    return streamStatus;
}

/*!
 * @brief Return the last frame number added count.
 *
 */
uint32_t FiVideoMP4Stream::GetLastFrameAdded()
{
    lastFrameAdded = 0;
    try
    {
        if (streamType == FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
        {
            if (ffmpegVideo)
            {
                lastFrameAdded = ffmpegVideo->GetFramesAdded();
            }
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s.", __PRETTY_FUNCTION__, e.what());
        goto exitGetLastFrameAdded;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: Exception (...).", __PRETTY_FUNCTION__);
        goto exitGetLastFrameAdded;
    }

exitGetLastFrameAdded:

    return lastFrameAdded;    // Return the last frame added to the stream.
}

/*!
 * @brief Return the count of encoded frames.
 * Polled by user of the video stream writer to access progress.
 *
 */
uint32_t FiVideoMP4Stream::GetFramesEncodedCount()
{
    uint32_t encodedFrameCount = 0;
    bool isPrepared = false;
    bool isComplete = false;
    bool hasError = false;
    try
    {
        // Must be a stream writer for this interface.
        if (streamType != FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
        {
            goto exitMP4GetEncodedFrameCount;
        }
        // Must have a ffmpeg container to query encoded frame count.
        if (!ffmpegVideo)
        {
            // This could be a nullptr in the preparation phase.
            goto exitMP4GetEncodedFrameCount;
        }

        // Query the encoded frame count.
        encodedFrameCount = ffmpegVideo->GetEncodedFrameCount(isPrepared, isComplete, hasError);

        // If this poll request indicates the codec error mark the stream with
        // a codec error condition.  We will not however overwrite any existing
        // stream error conditions that already are set for stream status.
        if (hasError)
        {
            // The FFmpeg codec is in an error condition mark the stream status.
            // But only do so if there isn't already an error condition marked.
            if ((streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY) ||
                (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED))
            {
                streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
            }
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s Flushing MP4 Frames.", __PRETTY_FUNCTION__, e.what());
        goto exitMP4GetEncodedFrameCount;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: Exception (...) Flushing MP4 Frames.", __PRETTY_FUNCTION__);
        goto exitMP4GetEncodedFrameCount;
    }

exitMP4GetEncodedFrameCount:

    return encodedFrameCount;    // Return the encoded frame count.
}

/*!
 * @brief Calculation to translate a h.264 percentage value into the
 * H.264 Constant Rate Factor option.
 *
 * @returns The CRF option value.
 */
uint32_t FiVideoMP4Stream::CalculateCRFOption(uint32_t h264QualityPct)
{
    uint32_t totalRangeValue = CRF_WORST_QUALITY - CRF_BEST_QUALITY;
    uint32_t crfCalculated = CRF_DEFAULT;

    // Keep it in range.
    if (h264QualityPct > 100)
        h264QualityPct = 100;

    // If zero - the else case will set it to the worst quality.

    // Convert the frame percentage value and store as preset value for codec.

    // CRF_WORST - CRF_BEST = TOTAL_RANGE.   Our setting 43 - 13 = 30 as totalRange.
    // TOTAL_RANGE * PCT / 100 = RANGE_VALUE.  At 70% we have 30 * 70 / 100 = 21 rangeValue.
    // CRF_WORST - RANGE_VALUE = CRF SETTING.  43 - 21 = 22 crfQualitySetting...

    // Just use integer math here without any rounding.
    if (h264QualityPct >= 10)
    {
        uint32_t rangeValue = totalRangeValue * h264QualityPct / 100;
        crfCalculated = CRF_WORST_QUALITY - rangeValue;
    }
    else
    {
        // Set to the worst quality if < 10%.
        crfCalculated = CRF_WORST_QUALITY;
    }

    fi_log(LOG_DEBUG, "%s: h264QualityPct[%d] crfCalculated[%d]", __PRETTY_FUNCTION__, h264QualityPct, crfCalculated);
    return crfCalculated;
}

/*!
 * @brief Validate and return proper Preset setting for h.264 preset.
 *
 * @returns Index of the preset option matching the user selected speed.
 */
uint32_t FiVideoMP4Stream::ValidatePresetOption(FAS_VIDEO::H264_SPEEDS h264Speed)
{
    uint32_t presetValueValidated = PRESET_DEFAULT_INDEX;  // This is medium setting.

    switch (h264Speed)
    {
    case FAS_VIDEO::H264_SPEEDS::H264_SLOWER:
        presetValueValidated = 0;
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_MEDIUM:
        presetValueValidated = 1;
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_FAST:
        presetValueValidated = 2;
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_VERY_FAST:
        presetValueValidated = 3;
        break;
    default:
        presetValueValidated = 1;   // Medium will be our default if it is invalid.
        break;
    }

    fi_log(LOG_DEBUG, "%s: h264Speed[%d] presetValueValidated[%d]", __PRETTY_FUNCTION__, (int)h264Speed, presetValueValidated);

    return presetValueValidated;
}

/*!
 * @brief Validate and return proper Preset setting for h.264 preset.
 *
 * @returns Index of the preset option matching the user selected speed.
 */
uint32_t FiVideoMP4Stream::SetKeyFrameInterval(uint32_t h264KeyFrameInterval)
{
    uint32_t h264KeyFrame;

    // Set the keyframe interval setting.
    if (h264KeyFrameInterval == 0)
        h264KeyFrame = 1;       // It can't be zero.
    else if (h264KeyFrameInterval > FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL)  // Keep constrained within default.
        h264KeyFrame = FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL;
    else
        h264KeyFrame = h264KeyFrameInterval;
    return h264KeyFrame;
}

/*!
 * @brief Determines if the FFmpeg Container is valid and available to use.
 *
 * @returns A boolean true if it is ok or false if not.  Logs any error condition.
 */
bool FiVideoMP4Stream::FFmpegContainerOK()
{
    bool FFmpegStreamIsOk = false;
    if (ffmpegVideo)
    {
        if (!ffmpegVideo->isFFmpegStreamOK())
        {
            fi_log(LOG_DEBUG, "%s: FFmpeg Container is not in streamable state.", __PRETTY_FUNCTION__);
            goto exitFFmpegContainerOK;
        }
        FFmpegStreamIsOk = true;
    }
exitFFmpegContainerOK:
    return FFmpegStreamIsOk;
}

/*!
 * @brief Common test to determine if the MP4 stream is able to write.
 *
 * @param whoIsAsking       The function that called this one for logging.
 * @param addImageAttachCheck   Does this test check to see if image is attached.
 *
 * @returns A boolean true if it is ok to write or false if not.  Logs
 * any error condition.
 */
bool FiVideoMP4Stream::OkToWrite(const char *whoIsAsking, bool addImageAttachCheck)
{
    bool okToWrite = false;

    try
    {
        // Is this AVI video stream set up for writing?
        if (streamType != FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
        {
            fi_log(LOG_DEBUG, "%s: MP4 stream not set for writing.", whoIsAsking);
            LogStreamStatus();
            goto exitOKToWrite;
        }
        // Is streamStatus in ready state?
        if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY)
        {
            fi_log(LOG_DEBUG, "%s: streamStatus is not set to STREAM_READY.", whoIsAsking);
            LogStreamStatus();
            goto exitOKToWrite;
        }

        // Does caller want to check for an attached image?
        if (addImageAttachCheck)
        {
            // Make sure there is a unique pointer for the image object.
            if (!videoFrameToWrite)
            {
                fi_log(LOG_DEBUG, "%s: missing FiImage shared pointer for stream writer.", __PRETTY_FUNCTION__);
                streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
                LogStreamStatus();
                goto exitOKToWrite;
            }
            // Is there an attached image and if not the the caller wants to know.
            if (!videoFrameToWrite->IsImageAttached())
            {
                fi_log(LOG_DEBUG, "%s: missing FiImage attachment for MP4 stream writer.", __PRETTY_FUNCTION__);
                streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
                LogStreamStatus();
                goto exitOKToWrite;
            }
        }

    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while validating video streamer.", __PRETTY_FUNCTION__, e.what());
        goto exitOKToWrite;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: All Catch Exception (...) caught while validating video streamer.", __PRETTY_FUNCTION__);
        goto exitOKToWrite;
    }

    // If made it here the tests were passed.
    okToWrite = true;

exitOKToWrite:
    return (okToWrite);
}

/*!
 * @brief Clears framesLost setting protecting access.
 * @returns The value in the field before clearing it.
 */
uint32_t FiVideoMP4Stream::ClearMP4FramesLost()
{
    std::lock_guard<std::mutex> guard(theLostFrameMutex);

    // Persist the setting.
    uint32_t mp4FramesLostB4Clear = mp4FramesLost;

    // Clear it.
    mp4FramesLost = 0;

    // Return the setting for a double duty function.
    return mp4FramesLostB4Clear;
}

/*!
 * @brief Retrieves the framesLost setting protecting access.
 *
 * @returns Value in our framesLost data member.
 */
uint32_t FiVideoMP4Stream::GetMP4FramesLost()
{
    std::lock_guard<std::mutex> guard(theLostFrameMutex);
    return mp4FramesLost;
}

/*!
 * @brief Retrieves the framesLost setting protecting access.
 *
 * @returns Value in our framesLost data member.
 */
uint32_t FiVideoMP4Stream::GetMP4TotalFramesLost()
{
    std::lock_guard<std::mutex> guard(theLostFrameMutex);
    return mp4TotalFramesLost;
}

/*!
 * @brief Increments the framesLost setting by one protecting access.
 * @returns The value in the field after incrementing it.
 */
uint32_t FiVideoMP4Stream::IncrementMP4FramesLost()
{
    std::lock_guard<std::mutex> guard(theLostFrameMutex);
    mp4FramesLost++;
    mp4TotalFramesLost++;
    return mp4FramesLost;
}
