/*!
 * @file fs_worker.cpp
 * @brief The implementation file for the Fastec Filesystem worker thread.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#define __STDC_FORMAT_MACROS
#include <unistd.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <openssl/evp.h>   // for an easy checksum

#include <iterator>
#include <map>
#include <memory>
#include <string>
#include <vector>

// Windows or Linux?
#if _WIN32
//
// Windows
#include "fas_trace.h"
#include "gige_msg.h"
#include "fas_log.h"

// These are the best alternatives I can find for Windows compatibility.
#define O_LARGEFILE     0
#define fsync(x)        fflush(NULL)

#else
//
// Linux and MacOS
#include <fastec/gige_msg.h>
#include <fas_trace.h>
#include <fas_log.h>
#endif

#ifdef MAC_OS
// MacOS compatibility
#define O_LARGEFILE     0
#define off64_t         off_t
#define lseek64         lseek
#endif

#include "fs_worker.h"

namespace FAS_FILESYSTEM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

enum FS_WORKER_TRACE
{
    FS_MESSAGE          = 0x00000010,
    FS_MESSAGE_NOISY    = 0x00000020,
    FS_WORKER_FUNC      = 0x00000040,
    FS_WORKER_FUNC_NOISY = 0x00000080
};


//! How many bytes in this many sectors.
#define SECTORS_TO_BYTES( b )   ((b) * SECTOR_LEN)

//! How many pages it takes to hold these bytes.
#define BYTES_TO_PAGES( b )     (((b) + SSD_PAGE_LEN - 1) / SSD_PAGE_LEN)

//! How many pages it takes to hold these sectors.
#define SECTORS_TO_PAGES( s )   (((s) + SECTORS_PER_PAGE - 1) / SECTORS_PER_PAGE)

//! 4096 alignment in memory needed for DMA buffers.
const int DMA_PAGE_ALIGN        = 4096;

//! LBA of the drive header.
const uint64_t DRIVE_HEADER_LBA = 0;

//! LBA of the first drive partition header.
const uint64_t FIRST_DPART_LBA  = 1;

//! FPGA appends 64 bytes (1 HEXWORDS) of Metadata onto the end of each frame.
const int FRAME_METADATA_HW_BYTES = 64;

//! Of the 64 bytes the FPGA appends, 32 bytes contain actual Metadata.
const int FRAME_METADATA_SIZE_BYTES = 32;

//! It is also convenient to have the size in 32-bit words.
const int FMETA_SIZE_WORDS = (FRAME_METADATA_SIZE_BYTES / 4);

//! The number of frames in a capture region needs to be a multiple of the number of
//! frame metadata records that fit in a page. This works out to 128.
const int FMETA_PER_PAGE = ((SECTORS_PER_PAGE * (SECTOR_LEN)) / FRAME_METADATA_SIZE_BYTES);

//! When reading back ranges of metadata from the SSD, we can use sector units.
const int FMETA_PER_SECTOR  = (SECTOR_LEN  / FRAME_METADATA_SIZE_BYTES);

//! The minimum number of frames required for a valid capture.
//! ALM: This was a totally arbitrary assignment. There is no "intent"
//!      behind this value. Feel free to adjust as needed.
const int MIN_FRAMES_REQUIRED = 3;

//! Message types for the worker thread.
enum FS_MESSAGE_TYPES
{
    FS_LAYOUT = 0x30000,            //!< Layout the drive, building the partition and capture header(s).
    FS_ANALYZE,                     //!< Analyze the drive to determine if it holds a Fastec Filesystem.
    FS_APPEND_CAPTURE,              //!< Append a capture to the end of a partition's space.
    FS_APPEND_PARTITION,            //!< Append a partition to the end of the drive's space.
    FS_CLEAR_CACHE,                 //!< Clear the worker thread cache.
    FS_DELETE_CAPTURE,              //!< Delete a capture from a partition's space.
    FS_DELETE_PARTITION,            //!< Delete a partition the drive's space.
    FS_GET_BOOKMARK,                //!< Retrieve the capture bookmark information.
    FS_GET_CORRECTION_TABLE,        //!< Retrieve the capture correction table information.
    FS_GET_GAIN_TABLE,              //!< Retrieve the capture gain table information.
    FS_GET_INFO,                    //!< Retrieve the partition information on a drive.
    FS_GET_PM_METADATA,             //!< Retrieve the capture metadata information.
    FS_RENAME,                      //!< Rename the drive.
    FS_RESIZE_CAPTURE,              //!< Resize a capture.
    FS_RESIZE_PARTITION,            //!< Resize a partition.
    FS_SET_BOOKMARK,                //!< Write the capture bookmark information.
    FS_SET_CORRECTION_INFO,         //!< Write the capture correction and gain table information.
    FS_SET_PM_METADATA,             //!< Write the capture metadata information.
    FS_TRY_LAYOUT,                  //!< Layout the drive, building the information without actually writing to the drive.
    FS_UPDATE_CAPTURE,              //!< Write updated capture information to the drive.
    FS_WRITE_PARTITION              //!< Write the partition information, typically created by FS_TRY_LAYOUT.
};

//--------------------------------------------------------------------
// Variable Declaration
//--------------------------------------------------------------------

} /* namespace FAS_FILESYSTEM */

using namespace FAS_FILESYSTEM;

// Our LR Filesystem worker thread.
std::unique_ptr<fsWorker> fsWorker::fs_proc;
std::mutex fsWorker::fs_mutex;

// For tracing, we need this local control defined outside the namespace.
static const uint32_t traceControl = 0 |
                                     FS_MESSAGE |
//                                     FS_MESSAGE_NOISY |
                                     FS_WORKER_FUNC |
//                                     FS_WORKER_FUNC_NOISY |
//                                     TRACE_0;        // Set our trace level to TRACE_0 to trace.
                                     TRACE_3;        // Set our trace level to TRACE_3 to not trace.

//--------------------------------------------------------------------
// Message Implementation
//--------------------------------------------------------------------

fsAnalyzeMessage::fsAnalyzeMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_ANALYZE;

    // Initialize our parameter.
    dev_path.clear();
    is_fastec = false;
}

fsAnalyzeMessage::~fsAnalyzeMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsAppendCaptureMessage::fsAppendCaptureMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_APPEND_CAPTURE;

    // Initialize our parameter.
    dev_path.clear();
    part_lba = 0;
    layout.clear();
}

fsAppendCaptureMessage::~fsAppendCaptureMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsAppendPartitionMessage::fsAppendPartitionMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_APPEND_PARTITION;

    // Initialize our parameter.
    dev_path.clear();
    layout.clear();
}

fsAppendPartitionMessage::~fsAppendPartitionMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsClearCacheMessage::fsClearCacheMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_CLEAR_CACHE;

    // Initialize our parameter.
    dev_path.clear();
}

fsClearCacheMessage::~fsClearCacheMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsDeleteCaptureMessage::fsDeleteCaptureMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_DELETE_CAPTURE;

    // Initialize our parameter.
    dev_path.clear();
    part_lba = 0;
    capture_lba = 0;
}

fsDeleteCaptureMessage::~fsDeleteCaptureMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsDeletePartitionMessage::fsDeletePartitionMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_DELETE_PARTITION;

    // Initialize our parameter.
    dev_path.clear();
    part_lba = 0;
}

fsDeletePartitionMessage::~fsDeletePartitionMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsGetBookmarkMessage::fsGetBookmarkMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_GET_BOOKMARK;

    // Initialize our parameter.
    dev_path.clear();
    capture_lba = 0;
    bk_buffer.clear();
}

fsGetBookmarkMessage::~fsGetBookmarkMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsGetCorrectionTableMessage::fsGetCorrectionTableMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_GET_CORRECTION_TABLE;

    // Initialize our parameter.
    dev_path.clear();
    capture_lba = 0;
    ct_buffer.clear();
}

fsGetCorrectionTableMessage::~fsGetCorrectionTableMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsGetGainTableMessage::fsGetGainTableMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_GET_GAIN_TABLE;

    // Initialize our parameter.
    dev_path.clear();
    capture_lba = 0;
    gt_buffer.clear();
}

fsGetGainTableMessage::~fsGetGainTableMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsGetInfoMessage::fsGetInfoMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_GET_INFO;

    // Initialize our parameter.
    dev_path.clear();
    name.clear();
    part_info.clear();
    assigned_space = 0;
}

fsGetInfoMessage::~fsGetInfoMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsGetPmMetadataMessage::fsGetPmMetadataMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_GET_PM_METADATA;

    // Initialize our parameter.
    dev_path.clear();
    capture_lba = 0;
    metadata.clear();
}

fsGetPmMetadataMessage::~fsGetPmMetadataMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsLayoutMessage::fsLayoutMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_LAYOUT;

    // Initialize our parameters structure.
    dev_path.clear();
    name.clear();
    layout.clear();
    num_sectors = 0;
}

fsLayoutMessage::~fsLayoutMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsRenameMessage::fsRenameMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_RENAME;

    // Initialize our parameters structure.
    dev_path.clear();
    name.clear();
}

fsRenameMessage::~fsRenameMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsResizeCaptureMessage::fsResizeCaptureMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_RESIZE_CAPTURE;

    // Initialize our parameter.
    dev_path.clear();
    cap_lba = 0;
    resize_amount = 0;
}

fsResizeCaptureMessage::~fsResizeCaptureMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsResizePartitionMessage::fsResizePartitionMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_RESIZE_PARTITION;

    // Initialize our parameter.
    dev_path.clear();
    part_lba = 0;
    resize_amount = 0;
}

fsResizePartitionMessage::~fsResizePartitionMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsSetBookmarkMessage::fsSetBookmarkMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_SET_BOOKMARK;

    // Initialize our parameter.
    dev_path.clear();
    capture_lba = 0;
    bk_buffer.clear();
}

fsSetBookmarkMessage::~fsSetBookmarkMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsSetCorrectionInfoMessage::fsSetCorrectionInfoMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_SET_CORRECTION_INFO;

    // Initialize our parameter.
    dev_path.clear();
    capture_lba = 0;
    ct_buffer.clear();
    gt_buffer.clear();
}

fsSetCorrectionInfoMessage::~fsSetCorrectionInfoMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsSetPmMetadataMessage::fsSetPmMetadataMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_SET_PM_METADATA;

    // Initialize our parameter.
    dev_path.clear();
    capture_lba = 0;
    metadata.clear();
}

fsSetPmMetadataMessage::~fsSetPmMetadataMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsTryLayoutMessage::fsTryLayoutMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_TRY_LAYOUT;

    // Initialize our parameters structure.
    dev_path.clear();
    layout.clear();
    part_info.clear();
    num_sectors = 0;
}

fsTryLayoutMessage::~fsTryLayoutMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsUpdateCaptureMessage::fsUpdateCaptureMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_UPDATE_CAPTURE;

    // Initialize our parameters structure.
    dev_path.clear();
    capture_info.lba = 0;
    capture_info.type = 0;
}

fsUpdateCaptureMessage::~fsUpdateCaptureMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}

fsWritePartitionMessage::fsWritePartitionMessage() :
    SyncMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);

    req = (FAS_THREAD::REQ_CODE)FS_WRITE_PARTITION;

    // Initialize our parameters structure.
    dev_path.clear();
    part_info.clear();
    num_sectors = 0;
}

fsWritePartitionMessage::~fsWritePartitionMessage()
{
    TP(FS_MESSAGE_NOISY, "%s", __PRETTY_FUNCTION__);
}


//--------------------------------------------------------------------
// Implementation: fsWorker
//--------------------------------------------------------------------

fsWorker::fsWorker()  :
    FAS_THREAD::MsgThread()
{
    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Can't check sizeof at compile time.
    if (sizeof(DRIVE_HD) > (unsigned)SECTOR_LEN)
    {
        fi_log(LOG_CRIT, "%s: DRIVE_HD too big (%" PRIu64 ")", __PRETTY_FUNCTION__, sizeof(DRIVE_HD));
    }

    // Initialize our members.
    num_sectors = 0;
    dev_path.clear();
    dev_fd = -1;
    sector_buf = std::make_unique<uint8_t[]>(SECTOR_LEN);
}

fsWorker::~fsWorker()
{
    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Tell the thread that we are stopping.
    FAS_THREAD::SyncMessage msg;
    msg.req = FAS_THREAD::HALT;

    // Pass it on.
    msg.doSync(this);

    // Try to join the thread.
    join();

    // Do we have an open file descriptor?
    if (dev_fd >= 0)
    {
        close(dev_fd);
        dev_fd = -1;
    }

    TP(FS_WORKER_FUNC, "%s: Done", __PRETTY_FUNCTION__);
}

fsWorker *fsWorker::factory()
{
    fsWorker *rc(nullptr);

    std::lock_guard<std::mutex> lock(fs_mutex);

    TP(FS_WORKER_FUNC_NOISY, "%s", __PRETTY_FUNCTION__);

    // We only allow one thread object.
    if (fs_proc.get() == nullptr)
    {
        // Create our thread.
        fs_proc = std::unique_ptr<fsWorker>(new fsWorker());
        if (fs_proc.get() != nullptr)
            fs_proc.get()->start();

        // Return the pointer to the singleton.
        rc = fs_proc.get();
    }

    // Check to see if we were halted.
    else if (fs_proc.get()->halted)
    {
        fi_log(LOG_DEBUG, "%s: Called after halted", __PRETTY_FUNCTION__);
    }
    else
    {
        // Everything looks OK.
        rc = fs_proc.get();
    }

    // Return our pointer.
    return rc;
}

/*!
 * @brief Save the path specified to the class field "fsWorker::dev_path" and
 *        the open the path and initialize the file descriptor "fsWorker::dev_fd".
 *
 * We open the path read-only because, for the most part, we only need read access.
 * When we want to write to the drive, we will open another file descriptor with
 * write access (see fsWorker::writeSectors).
 *
 * @param path  The fully qualified path to the device/file we want to use.
 * @return True if the path exists and we are able to successfully open it.
 */
bool fsWorker::setDrivePath(const std::string& path)
{
    TP(FS_WORKER_FUNC_NOISY, "%s: dev_path=[%s]", __PRETTY_FUNCTION__, path.c_str());

    // Is the new path the same as the old path?
    if (dev_path == path)
        return true;

    // Remember the path.
    dev_path = path;

    // Do we have a file open? It's OK, we just need to make
    // sure to close the old descriptor.
    if (dev_fd >= 0)
    {
        // Close the open descriptor.
        close(dev_fd);
        dev_fd = -1;
    }

    // Is the string empty?
    if (dev_path.empty())
        return true;

    // Open our device for reading.
    dev_fd = open(dev_path.c_str(), O_RDONLY | O_LARGEFILE);
    if (dev_fd < 0)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        fi_log(LOG_ERR, "%s: open(%s) for read failed", __PRETTY_FUNCTION__, dev_path.c_str());
    }

    // Return our result.
    return (dev_fd >= 0);
}

/*!
 * @brief Read the sector containing metadata for a specific frame.
 *
 * @param cap           Capture header already recovered from the drive.
 * @param sector_buf    Sector buffer for reading metadata, must be big enough for three sectors.
 * @param abs_frame     Absolute frame number of the frame whose metadata is to be read.
 *
 * @return Pointer to frame metadata for abs_frame within sector_buf or NULL if read failed.
 */
uint32_t *fsWorker::getFrameMetadata(const LR_CAPTURE_HDR *cap, uint32_t *sector_buf, uint32_t abs_frame)
{
    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Are we using a separate metadata area?
    if (cap->fmeta_lba != 0)
    {
        // Read the sector containing our metadata.
        if (readSectors((abs_frame / FMETA_PER_SECTOR) + cap->fmeta_lba, (uint8_t *)sector_buf, 1))
        {
            sector_buf += ((abs_frame % FMETA_PER_SECTOR) * FMETA_SIZE_WORDS);
        }
        else
        {
            fi_log(LOG_ERR, "%s: Read(%" PRIu64 ") failed - (%d:%s)", __PRETTY_FUNCTION__, (abs_frame / FMETA_PER_SECTOR) + cap->fmeta_lba,
                   errno, strerror(errno));
            sector_buf = nullptr;
        }
    }
    else
    {
        uint32_t pfmd_offset;
        off64_t frame_offset;

        // Get the offset in bytes.
        frame_offset = (((off64_t)cap->image_lba) * SECTOR_LEN) + (((off64_t)cap->frame_size) * abs_frame);

        // Get the amount to offset for the PFMD.
        pfmd_offset = FRAME_METADATA_HW_BYTES;

        // Adjust to the PFMD.
        frame_offset += (off64_t)(cap->frame_size - pfmd_offset);

        // Seek to the metadata and read it. Only read the actual data in the metadata block.
        // We don't care about the padding.
        if ((lseek64(dev_fd, frame_offset, SEEK_SET) != frame_offset) ||
            (read(dev_fd, sector_buf, FRAME_METADATA_SIZE_BYTES) != FRAME_METADATA_SIZE_BYTES))
        {
            fi_log(LOG_ERR, "%s: Read(%" PRIu64 ", %u) failed - (%d:%s)", __PRETTY_FUNCTION__, frame_offset,
                   FRAME_METADATA_SIZE_BYTES, errno, strerror(errno));
            sector_buf = nullptr;
        }
    }

    // Return the buffer pointer.
    return sector_buf;
}

/*!
 * @brief Resize the available partition on a drive.
 *
 * @param part_info     [IN] The partition information pointer we're interested in.
 * @param resize_amount [IN] The absolute number of sectors for the partition.
 * @return True if successful, false otherwise.
 */
bool fsWorker::resizeAvailPartition(PARTITION_INFO *part_info, uint64_t resize_amount)
{
    FI_SIG *sig;
    bool rc = false;
    PARTITION_UNION *part_ptr;

    TP(FS_WORKER_FUNC, "%s: resize_amount:%" PRIu64, __PRETTY_FUNCTION__, resize_amount);

    // Sanity checks.
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Invalid parameter, part_info is null!", __PRETTY_FUNCTION__);
        return false;
    }

    // Read and verify the partition header.
    if (!readSectors(part_info->part.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, part_info->part.lba);
        return false;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();
    if (!verifyPartitionHeader(sig))
    {
        fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, part_info->part.lba);
        return false;
    }

    // Overlay our partition header.
    part_ptr = (PARTITION_UNION *)(sector_buf.get() + sizeof(FI_SIG));

    // Make sure we don't exceed the drive limit.
    if ((part_ptr->avail_part.start_lba + resize_amount) > getNumSectors())
    {
        fi_log(LOG_DEBUG, "%s: Resizing would exceed the drive size!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Set the new size.
    part_ptr->avail_part.data_sectors = resize_amount;

    // Update our cache.
    memcpy(&part_info->part.hdr.avail_part, part_ptr, sizeof(AVAIL_PARTITION_HDR));

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the partition header to the drive.
    if (!writeSectors(part_info->part.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Everything looks OK.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Resize the LAST capture in a partition. Resizing can be to shrink the capture space,
 *        grow the capture space, or fit the capture space to the amount of data (number of
 *        frames) currently captured.
 *
 * @param cap_hdr       [IN/OUT] Header of the capture we want to modify.
 * @param resize_amount [IN] The amount to resize. 0: fit to current capture, otherwise the absoulte
 *                           number of sectors for the capture.
 * @param part_info     [IN] Pointer to the partition information for this capture.
 * @return True if successful, false otherwise.
 */
bool fsWorker::resizeLrCapture(LR_CAPTURE_HDR& cap_hdr, uint64_t resize_amount, const PARTITION_INFO *part_info)
{
    uint64_t part_end;
    bool rc = false;

    TP(FS_WORKER_FUNC, "%s: resize_amount:%" PRIu64, __PRETTY_FUNCTION__, resize_amount);

    // For now we need to enforce the rule that only the last capture in a partition can be mucked around with.
    if (cap_hdr.link.next_lba != 0)
    {
        fi_log(LOG_DEBUG, "%s: Only the last capture can be modified!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Calculate the end of the partition memory.
    part_end = part_info->part.hdr.lr_part.image_lba + part_info->part.hdr.lr_part.image_sectors;

    // How are we resizing the capture?
    if (resize_amount == 0)
    {
        // The user wants to fit the size of the capture to the number of frames currently captured.
        // Need to make sure that there ARE frames captured.
        if (cap_hdr.num_frames == 0)
        {
            // There are no frames captured. We can't do this operation.
            fi_log(LOG_DEBUG, "%s: There are no frames captured. Resize (fit) is not allowed!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update the sizes and counts to fit the captured data.
        if (cap_hdr.fmeta_sectors != 0)
            cap_hdr.fmeta_sectors = BYTES_TO_SECTORS(cap_hdr.num_frames * FRAME_METADATA_HW_BYTES);
        cap_hdr.image_sectors = BYTES_TO_SECTORS(cap_hdr.num_frames * cap_hdr.frame_size);
        cap_hdr.frame_count = cap_hdr.num_frames;
    }
    else
    {
        uint32_t frame_count = 0;
        uint64_t min_space_allowed;

        // This sanity checks only make sense if we actually have a capture.
        if (cap_hdr.frame_size > 0)
        {
            // Make sure that we meet the minimum requirements on space.
            min_space_allowed = BYTES_TO_SECTORS(cap_hdr.frame_size * MIN_FRAMES_REQUIRED);
            if (resize_amount < min_space_allowed)
            {
                fi_log(LOG_DEBUG, "%s: The resize_amount (%" PRId64 ") is too small!", __PRETTY_FUNCTION__, resize_amount);
                goto bail;
            }

            // Make sure we aren't too big.
            if ((cap_hdr.image_lba + resize_amount) > part_end)
            {
                // The resize amount won't work.
                fi_log(LOG_DEBUG, "%s: The resize amount (%" PRIu64 ") is too big.", __PRETTY_FUNCTION__, resize_amount);
                goto bail;
            }

            // How many frames are we talking about?
            frame_count = (uint32_t)(SECTORS_TO_BYTES(resize_amount) / cap_hdr.frame_size);

            // Do we have a separate PFMD section?
            if (cap_hdr.fmeta_sectors != 0)
            {
                uint64_t required_fmeta;
                uint64_t fmeta_end = part_info->part.hdr.lr_part.fmeta_lba + part_info->part.hdr.lr_part.fmeta_sectors;

                // How much PFMD space do we need for the new number of frames?
                required_fmeta = BYTES_TO_SECTORS(frame_count * FRAME_METADATA_HW_BYTES);

                // We know we'll be able to meet the minimum requirements, but do we have enough space
                // for the PFMD being requested?
                if ((cap_hdr.fmeta_lba + required_fmeta) > fmeta_end)
                {
                    fi_log(LOG_DEBUG, "%s: The resize amount (%" PRIu64 ") is too big for the frame metadata space.", __PRETTY_FUNCTION__, resize_amount);
                    goto bail;
                }

                // Set the frame metadata space.
                cap_hdr.fmeta_sectors = required_fmeta;
            }
        }

        // Set the number of sectors for the image data.
        cap_hdr.image_sectors = resize_amount;

        // Recalculate the number of frames that will fit in the new space.
        cap_hdr.frame_count = frame_count;
    }

    // Everything looks OK.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Resize the LAST partition on a drive. Resizing can be to shrink the space, grow
 *        the space, or fit the space to the currently captured data within the partition.
 *
 * @param part_info     [IN] The partition information pointer we're interested in.
 * @param resize_amount [IN] The amount to resize. 0: fit to current capture space, otherwise the absolute
 *                            number of sectors for the partition.
 * @return True if successful, false otherwise.
 */
bool fsWorker::resizeLrPartition(PARTITION_INFO *part_info, uint64_t resize_amount)
{
    FI_SIG *sig;
    bool rc = false;
    uint64_t cap_end;
    PARTITION_UNION *part_ptr;

    TP(FS_WORKER_FUNC, "%s: resize_amount:%" PRIu64, __PRETTY_FUNCTION__, resize_amount);

    // Sanity checks.
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Invalid parameter, part_info is null!", __PRETTY_FUNCTION__);
        return false;
    }

    // Read and verify the partition header.
    if (!readSectors(part_info->part.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, part_info->part.lba);
        return false;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();
    if (!verifyPartitionHeader(sig))
    {
        fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, part_info->part.lba);
        return false;
    }

    // Overlay our partition header.
    part_ptr = (PARTITION_UNION *)(sector_buf.get() + sizeof(FI_SIG));

    // Get the last capture of the partition.
    CAPTURE_DATA& last_cap = part_info->caps.back();

    // Sanity check.
    if (last_cap.hdr.link.next_lba != 0)
    {
        fi_log(LOG_ERR, "%s: The last capture in the partition has a next link!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get the end of the last capture of the partition.
    cap_end = last_cap.hdr.lr_cap.image_lba + last_cap.hdr.lr_cap.image_sectors;

    // How are we resizing the capture?
    if (resize_amount == 0)
    {
        // The user wants to fit the size of the partition to the number of frames currently captured.
        part_ptr->lr_part.image_sectors = cap_end;
        if (part_ptr->lr_part.fmeta_sectors != 0)
            part_ptr->lr_part.fmeta_sectors = last_cap.hdr.lr_cap.fmeta_lba + last_cap.hdr.lr_cap.fmeta_sectors;
    }
    else
    {
        // Make sure we don't clobber the last capture.
        if ((part_ptr->lr_part.image_lba + resize_amount) < cap_end)
        {
            fi_log(LOG_DEBUG, "%s: Resizing would cause the partition to be too small.", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Make sure we don't exceed the drive limit.
        if ((part_ptr->lr_part.image_lba + resize_amount) > getNumSectors())
        {
            fi_log(LOG_DEBUG, "%s: Resizing would exceed the drive size!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update the partition information.
        part_ptr->lr_part.image_sectors = resize_amount;
        if (part_ptr->lr_part.fmeta_sectors != 0)
        {
            // ALM_TODO: Because the captures can have different size frames, we don't know exactly
            //           how much PFMD area to reserve for the growing image area. We don't currently
            //           use a separate PFMD area, so I'm going to throw up my hands for now.
            fi_log(LOG_DEBUG, "%s: ==== UNDER CONSTRUCTION: resizing of the PFMD area is not currently supported. ====", __PRETTY_FUNCTION__);
        }
    }

    // Update our cache.
    memcpy(&part_info->part.hdr.lr_part, part_ptr, sizeof(LR_PARTITION_HDR));

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the partition header to the drive.
    if (!writeSectors(part_info->part.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Everything looks OK.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Resize the LAST capture in a partition. Resizing can be to shrink the capture space,
 *        grow the capture space, or fit the capture space to the amount of data (number of
 *        frames) currently captured.
 *
 * @param cap_hdr       [IN/OUT] Header of the capture we want to modify.
 * @param resize_amount [IN] The amount to resize. 0: fit to current capture, otherwise the absoulte
 *                           number of sectors for the capture.
 * @param part_info     [IN] Pointer to the partition information for this capture.
 * @return True if successful, false otherwise.
 */
bool fsWorker::resizeStorCapture(STOR_CAPTURE_HDR& cap_hdr, uint64_t resize_amount, const PARTITION_INFO *part_info)
{
    uint64_t part_end;
    bool rc = false;

    TP(FS_WORKER_FUNC, "%s: resize_amount:%" PRIu64, __PRETTY_FUNCTION__, resize_amount);

    // For now we need to enforce the rule that only the last capture in a partition can be mucked around with.
    if (cap_hdr.link.next_lba != 0)
    {
        fi_log(LOG_DEBUG, "%s: Only the last capture can be modified!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Calculate the end of the partition memory.
    part_end = part_info->part.hdr.stor_part.start_lba + part_info->part.hdr.stor_part.data_sectors;

    // How are we resizing the capture?
    if (resize_amount == 0)
    {
        // The user wants to fit the size of the capture to the number of sectors currently captured.
        // Need to make sure that there ARE sectors captured.
        if (cap_hdr.used_sectors == 0)
        {
            // There are no sectors captured. We can't do this operation.
            fi_log(LOG_DEBUG, "%s: There are no sectors captured. Resize (fit) is not allowed!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update the sizes and counts to fit the captured data.
        cap_hdr.data_sectors = cap_hdr.used_sectors;
    }
    else
    {
        // This sanity checks only make sense if we actually have a capture.
        if (cap_hdr.used_sectors > 0)
        {
            // Make sure that we meet the minimum requirements on space.
            if (resize_amount < STOR_CAPTURE_MIN_SECTORS)
            {
                fi_log(LOG_DEBUG, "%s: The resize_amount (%" PRId64 ") is too small!", __PRETTY_FUNCTION__, resize_amount);
                goto bail;
            }

            // Make sure we aren't too big.
            if ((cap_hdr.data_lba + resize_amount) > part_end)
            {
                // The resize amount won't work.
                fi_log(LOG_DEBUG, "%s: The resize amount (%" PRIu64 ") is too big.", __PRETTY_FUNCTION__, resize_amount);
                goto bail;
            }
        }

        // Set the number of sectors for the image data.
        cap_hdr.data_sectors = resize_amount;
    }

    // Everything looks OK.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Resize the LAST partition on a drive. Resizing can be to shrink the space, grow
 *        the space, or fit the space to the currently captured data within the partition.
 *
 * @param part_info     [IN] The partition information pointer we're interested in.
 * @param resize_amount [IN] The amount to resize. 0: fit to current capture space, otherwise the absolute
 *                            number of sectors for the partition.
 * @return True if successful, false otherwise.
 */
bool fsWorker::resizeStorPartition(PARTITION_INFO *part_info, uint64_t resize_amount)
{
    FI_SIG *sig;
    bool rc = false;
    uint64_t cap_end;
    PARTITION_UNION *part_ptr;

    TP(FS_WORKER_FUNC, "%s: resize_amount:%" PRIu64, __PRETTY_FUNCTION__, resize_amount);

    // Sanity checks.
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Invalid parameter, part_info is null!", __PRETTY_FUNCTION__);
        return false;
    }

    // Read and verify the partition header.
    if (!readSectors(part_info->part.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, part_info->part.lba);
        return false;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();
    if (!verifyPartitionHeader(sig))
    {
        fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, part_info->part.lba);
        return false;
    }

    // Overlay our partition header.
    part_ptr = (PARTITION_UNION *)(sector_buf.get() + sizeof(FI_SIG));

    // Get the last capture of the partition.
    CAPTURE_DATA& last_cap = part_info->caps.back();

    // Sanity check.
    if (last_cap.hdr.link.next_lba != 0)
    {
        fi_log(LOG_ERR, "%s: The last capture in the partition has a next link!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get the end of the last capture of the partition.
    cap_end = last_cap.hdr.stor_cap.data_lba + last_cap.hdr.stor_cap.data_sectors;

    // How are we resizing the capture?
    if (resize_amount == 0)
    {
        // The user wants to fit the size of the partition to the number of frames currently captured.
        part_ptr->stor_part.data_sectors = cap_end;
    }
    else
    {
        // Make sure we don't clobber the last capture.
        if ((part_ptr->stor_part.start_lba + resize_amount) < cap_end)
        {
            fi_log(LOG_DEBUG, "%s: Resizing would cause the partition to be too small.", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Make sure we don't exceed the drive limit.
        if ((part_ptr->stor_part.start_lba + resize_amount) > getNumSectors())
        {
            fi_log(LOG_DEBUG, "%s: Resizing would exceed the drive size!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update the partition information.
        part_ptr->stor_part.data_sectors = resize_amount;
    }

    // Update our cache.
    memcpy(&part_info->part.hdr.stor_part, part_ptr, sizeof(STOR_PARTITION_HDR));

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the partition header to the drive.
    if (!writeSectors(part_info->part.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Everything looks OK.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Update partition metadata using what we figured out from an aborted capture.
 *
 * This is for a capture that failed to complete normally but we detected that it wrapped.
 *
 * @param cap   Capture header recovered from the drive.
 * @fmeta_buf   Buffer containing metadata for the sector containing frame_num
 * @frame_num   Absolute frame number to use as end of capture.
 */
void fsWorker::simulateEndFrameWrap(LR_CAPTURE_HDR *cap, const uint32_t *fmeta_buf, uint32_t frame_num)
{
    TP(FS_WORKER_FUNC, "%s: frame_num=%u", __PRETTY_FUNCTION__, frame_num);

    // Fill in the fields that we normally update post-capture in the capture header.
    // Our trigger frame will be the last captured (as indicated by metadata) and
    // our first frame will be the one after that.
    // Near the last-first boundary, we might have metadata that doesn't match the actual images;
    // frames appearing at the start of the capture based on metadata might actually belong at
    // the end because a chunk of images are written to the drive before their corresponding
    // metadata overwrites the old data from frame_count frames ago.
    cap->num_frames = cap->frame_count;
    cap->first_frame_num = (frame_num + 1) % cap->frame_count;
    cap->trigger_frame_num = cap->frame_count - 1;
    cap->trigger_time_s = fmeta_buf[ FMETA_TIME_SECONDS ]; // SBS
    cap->trigger_time_ns = (fmeta_buf[ FMETA_TIME_USECS ] & FMETA_TIME_USEC_MASK) * 1000;   // SBUS
}


/*!
 * @brief Update partition metadata using what we figured out from an aborted capture.
 *
 * This is for a capture that failed to complete normally without wrapping.
 *
 * @param cap   Capture header recovered from the drive.
 * @fmeta_buf   Buffer containing metadata for the sector containing frame_num
 * @frame_num   Frame number to use as end of capture.
 */
void fsWorker::simulateEndFrameNoWrap(LR_CAPTURE_HDR *cap, const uint32_t *fmeta_buf, uint32_t frame_num)
{
    TP(FS_WORKER_FUNC, "%s: frame=%u", __PRETTY_FUNCTION__, frame_num);

    // Update the capture header.
    cap->num_frames = frame_num + 1;
    cap->first_frame_num = 0;

    // If it is not a ROC/BROC recording, set an end trigger.
    if ((cap->state != CAPTUREHD_APPENDING) && ((cap->flags & CAPHD_FLAG_APPENDMODE) == 0))
    {
        cap->trigger_frame_num = frame_num;
        cap->trigger_time_s = fmeta_buf[FMETA_TIME_SECONDS]; // SBS
        cap->trigger_time_ns = (fmeta_buf[FMETA_TIME_USECS] & FMETA_TIME_USEC_MASK) * 1000; // SBUS
    }

    // Make sure the trigger frame is zero for append mode.
    else if ((cap->state == CAPTUREHD_CAPTURING) && ((cap->flags & CAPHD_FLAG_APPENDMODE) != 0))
    {
        cap->trigger_frame_num = 0;
    }
}

/*!
 * @brief Analyze partial capture that did not wrap to find the end.
 *
 * Search through the capture metadata to locate the last record with a matching recording ID.
 * Because the recording did not wrap, the frame numbers/information is monotonically increasing.
 * This will allow us to be smart about how we find the end of the recording. We will be doing a
 * bsearch to find where the end of the capture is.
 *
 * @sa https://en.wikipedia.org/wiki/Binary_search_algorithm
 *
 * @param cap       Capture header already recovered from the drive.
 * @return true if it was possible to reconstruct the capture, false if not.
 */
bool fsWorker::findLrCaptureEnd(LR_CAPTURE_HDR *cap)
{
    uint32_t *meta_ptr;
    uint32_t frame_num;
    uint32_t low_frame;
    uint32_t high_frame;
    std::unique_ptr<uint32_t[]> pfmd_buf;

    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Setup our buffer.
    pfmd_buf = std::make_unique<uint32_t[]>(SECTOR_LEN / 4);

    // Get the PFMD for frame 0. Make sure to use the correct buffer!
    meta_ptr = getFrameMetadata(cap, pfmd_buf.get(), 0);
    if (meta_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to read PFMD for frame 0.", __PRETTY_FUNCTION__);
        return false;
    }

    // If the first metadata record doesn't match, we didn't save any metadata in this recording.
    if (meta_ptr[FMETA_RECORDING_ID] != cap->recording_id)
    {
        fi_log(LOG_WARNING, "%s: first metadata record mismatch (0x%08x!=0x%08x), no frames were saved",
               __PRETTY_FUNCTION__, meta_ptr[FMETA_RECORDING_ID], cap->recording_id);
        return false;
    }

    // Walk through the frames searching for the last valid frame based on PFMD.
    TP(FS_WORKER_FUNC, "%s: Looking through %u frames for the end.", __PRETTY_FUNCTION__, cap->frame_count);
    frame_num = 0;
    low_frame = 1;
    high_frame = cap->frame_count - 1;
    while (low_frame < high_frame)
    {
        frame_num = (uint32_t)((uint64_t)((uint64_t)low_frame + (uint64_t)high_frame) / 2);
        TP(FS_WORKER_FUNC_NOISY, "%s: Frame=%u [lo:%u, hi:%u]", __PRETTY_FUNCTION__, frame_num, low_frame, high_frame);

        // Get the PFMD for the frame.
        meta_ptr = getFrameMetadata(cap, pfmd_buf.get(), frame_num);
        if (meta_ptr == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to read PFMD for frame %u.", __PRETTY_FUNCTION__, frame_num);
            return false;
        }

        // Have we found the edge?
        if (meta_ptr[FMETA_RECORDING_ID] != cap->recording_id)
            high_frame = frame_num - 1;
        else
            low_frame = frame_num + 1;
    }

    // The last move COULD have put our frame number into a bad region. To resolve
    // this we might need to move back a few steps.
    TP(FS_WORKER_FUNC, "%s: End of BSEARCH - frame %d [cap_rec_id=0x%X, frm_id=0x%X]", __PRETTY_FUNCTION__, frame_num, cap->recording_id, meta_ptr[FMETA_RECORDING_ID]);
    while ((meta_ptr[FMETA_RECORDING_ID] != cap->recording_id) && (frame_num > (low_frame - 1)))
    {
        // Move back a frame.
        --frame_num;

        // Get the PFMD for the frame.
        meta_ptr = getFrameMetadata(cap, pfmd_buf.get(), frame_num);
    }

    // Update our meader data to indicate and end of the capture.
    TP(FS_WORKER_FUNC, "%s: Frame %d should be the end [cap_rec_id=0x%X, frm_id=0x%X]", __PRETTY_FUNCTION__, frame_num, cap->recording_id, meta_ptr[FMETA_RECORDING_ID]);
    simulateEndFrameNoWrap(cap, meta_ptr, frame_num);

    // All done.
    return true;
}

/*!
 * @brief Return the 64-bit time extracted from a frame metadata record.
 *
 * @param metadata_buf  Buffer containing FMETA_SIZE_WORDS words of frame metadata.
 *
 * @return (1M * SBS) + SBUS
 */
constexpr uint64_t fsWorker::fmetaTime(const uint32_t *metadata_buf)
{
    return ((uint64_t)metadata_buf[FMETA_TIME_SECONDS] * (1000L * 1000L)) + (uint64_t)metadata_buf[FMETA_TIME_USECS];
}

/*!
 * @brief Convert the message request code to a string.
 * @param req_code  The code to convert.
 * @return The corresponding string. "unknown" if not found.
 */
std::string fsWorker::reqToString(int req_code)
{
    std::string req_string;

    // Figure out the conversion.
    // Parse the message.
    switch (req_code)
    {
    case FS_ANALYZE:
        req_string.append("FS_ANALYZE");
        break;
    case FS_APPEND_CAPTURE:
        req_string.append("FS_APPEND_CAPTURE");
        break;
    case FS_APPEND_PARTITION:
        req_string.append("FS_APPEND_PARTITION");
        break;
    case FS_CLEAR_CACHE:
        req_string.append("FS_CLEAR_CACHE");
        break;
    case FS_DELETE_CAPTURE:
        req_string.append("FS_DELETE_CAPTURE");
        break;
    case FS_DELETE_PARTITION:
        req_string.append("FS_DELETE_PARTITION");
        break;
    case FS_GET_BOOKMARK:
        req_string.append("FS_GET_BOOKMARK");
        break;
    case FS_GET_CORRECTION_TABLE:
        req_string.append("FS_GET_CORRECTION_TABLE");
        break;
    case FS_GET_GAIN_TABLE:
        req_string.append("FS_GET_GAIN_TABLE");
        break;
    case FS_GET_INFO:
        req_string.append("FS_GET_INFO");
        break;
    case FS_GET_PM_METADATA:
        req_string.append("FS_GET_PM_METADATA");
        break;
    case FS_LAYOUT:
        req_string.append("FS_LAYOUT");
        break;
    case FS_RENAME:
        req_string.append("FS_RENAME");
        break;
    case FS_RESIZE_CAPTURE:
        req_string.append("FS_RESIZE_CAPTURE");
        break;
    case FS_RESIZE_PARTITION:
        req_string.append("FS_RESIZE_PARTITION");
        break;
    case FS_SET_BOOKMARK:
        req_string.append("FS_SET_BOOKMARK");
        break;
    case FS_SET_CORRECTION_INFO:
        req_string.append("FS_SET_CORRECTION_INFO");
        break;
    case FS_SET_PM_METADATA:
        req_string.append("FS_SET_PM_METADATA");
        break;
    case FS_TRY_LAYOUT:
        req_string.append("FS_TRY_LAYOUT");
        break;
    case FS_UPDATE_CAPTURE:
        req_string.append("FS_UPDATE_CAPTURE");
        break;
    case FS_WRITE_PARTITION:
        req_string.append("FS_WRITE_PARTITION");
        break;
    default:
        // Pass it on...
        req_string = MsgThread::reqToString(req_code);
        break;
    }

    return req_string;
}

const char *fsWorker::signatureToString(uint32_t signature)
{
    static char sig_string[5];

    // Build up our return string.
    sig_string[0] = (char)(signature & 0xFF);
    sig_string[1] = (char)((signature >> 8) & 0xFF);
    sig_string[2] = (char)((signature >> 16) & 0xFF);
    sig_string[3] = (char)((signature >> 24) & 0xFF);
    sig_string[4] = '\0';

    return sig_string;
}

int fsWorker::md5Length()
{
    return EVP_MD_size(EVP_md5());
}

uint8_t *fsWorker::md5Calculate(const unsigned char *buf, size_t buf_size, unsigned char *md)
{
    EVP_MD_CTX *mdctx;
    unsigned char *md5_digest;
    unsigned int md5_digest_len = md5Length();

    // MD5_Init
    mdctx = EVP_MD_CTX_new();
    EVP_DigestInit_ex(mdctx, EVP_md5(), NULL);

    // MD5_Update
    EVP_DigestUpdate(mdctx, buf, buf_size);

    // MD5_Final
    if (md == nullptr)
        md5_digest = (unsigned char *)OPENSSL_malloc(md5_digest_len);
    else
        md5_digest = md;
    EVP_DigestFinal_ex(mdctx, md5_digest, &md5_digest_len);
    EVP_MD_CTX_free(mdctx);

    return md5_digest;
}


/*!
 * @brief Analyze partial capture that wrapped to find the wrap point.
 *
 * Search through the capture metadata to locate a point where the time jumps backward.
 * We expect that this will only occur at a sector boundary but allow for the possibility that
 * something else happened.
 *
 * @param cap           Capture header already recovered from the drive.
 * @param metadat_buf   Sector buffer for reading metadata, must be big enough for three sectors.
 *
 * @return true if it was possible to reconstruct the capture, false if not.
 */
bool fsWorker::findLrCaptureWrap(LR_CAPTURE_HDR *cap)
{
    bool wrap_found = false;
    uint32_t *buf_ping_ptr;
    uint32_t *buf_pong_ptr;
    uint32_t *meta_ptr;
    uint32_t frame_num;
    uint64_t last_time;
    uint64_t frame_time;

    std::unique_ptr<uint32_t[]> pfmd_buf;       // Use the unique_ptr so we don't have to worry about freeing the memory.

    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Setup our buffer.
    pfmd_buf = std::make_unique<uint32_t[]>(SECTOR_LEN / 4);

    // Setup our ping/pong buffers.  Odd frame numbers use "ping", even use "pong".
    buf_ping_ptr = pfmd_buf.get();
    buf_pong_ptr = buf_ping_ptr + FMETA_SIZE_WORDS;

    // Get the PFMD for frame 0. Make sure to use the correct buffer!
    meta_ptr = getFrameMetadata(cap, buf_pong_ptr, 0);
    if (meta_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to read PFMD for frame 0.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // If the first metadata record doesn't match, something is seriously wrong.
    if (meta_ptr[FMETA_RECORDING_ID] != cap->recording_id)
    {
        fi_log(LOG_WARNING, "%s: first metadata record mismatch (%u!=%u), shouldn't happen here.",
               __FUNCTION__, meta_ptr[FMETA_RECORDING_ID], cap->recording_id);
        goto bail;
    }

    // Get the PFMD for the last frame. Make sure to use the correct buffer!
    //
    // Note that this is the last possible frame in the capture buffer, not necessarily
    // the last captured frame. We are still looking for that one.
    //
    // Also, we know that the frame at (frame_count - 1) has a valid recording ID, because
    // that is how we decided that the capture likely wrapped.
    meta_ptr = getFrameMetadata(cap, buf_ping_ptr, cap->frame_count - 1);
    if (meta_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: Unable to read PFMD for frame %u.", __PRETTY_FUNCTION__, cap->frame_count - 1);
        goto bail;
    }

    // Okay, things look somewhat sensible. Let's find the time discontinuity.
    last_time = fmetaTime(buf_pong_ptr);
    frame_time = fmetaTime(buf_ping_ptr);

    // Odds are against it, but it is possible that we wrapped right at the last frame.
    if (last_time <= frame_time)
    {
        // We COULD use the "no wrap" version, but the "wrap" version sets the trigger to the last
        // frame, which is more likely to have happened.
        simulateEndFrameWrap(cap, buf_ping_ptr, cap->frame_count - 1);
        wrap_found = true;
        goto bail;
    }

    // Walk through the frames searching for the last valid frame based on PFMD.
    // We are looking for a discontinuity in the time stamp or recording ID.
    for (frame_num = 1; frame_num < (cap->frame_count - 1); ++frame_num)
    {
        uint32_t *buf_ptr;

        // Odd frame numbers use "ping", even use "pong".
        if ((frame_num & 1) == 1)
            buf_ptr = buf_ping_ptr;
        else
            buf_ptr = buf_pong_ptr;

        // Get the PFMD for the frame.
        meta_ptr = getFrameMetadata(cap, buf_ptr, frame_num);
        if (meta_ptr == nullptr)
        {
            fi_log(LOG_ERR, "%s: Unable to read PFMD for frame %u.", __PRETTY_FUNCTION__, frame_num);
            goto bail;
        }

        // If we do not have a match, then we've just walked past the end
        // of the recording.
        if (meta_ptr[FMETA_RECORDING_ID] != cap->recording_id)
        {
            // Ok, this is the spot. Backup one.
            if (buf_ptr == buf_ping_ptr)
                buf_ptr = buf_pong_ptr;
            else
                buf_ptr = buf_ping_ptr;

            // Mark the capture done here.
            simulateEndFrameNoWrap(cap, buf_ptr, frame_num - 1);
            wrap_found = true;
            break;
        }

        // Get the time recorded for this frame.
        frame_time = fmetaTime(buf_ptr);

        // Did we find a time inversion?
        if (last_time > frame_time)
        {
            // Ok, this is the spot. Backup one.
            if (buf_ptr == buf_ping_ptr)
                buf_ptr = buf_pong_ptr;
            else
                buf_ptr = buf_ping_ptr;

            // Mark the capture done here.
            simulateEndFrameNoWrap(cap, buf_ptr, frame_num - 1);
            wrap_found = true;
            break;
        }

        // Save the frame time for next loop.
        last_time = frame_time;
    }

bail:
    // Odd...but we didn't find a wrap.
    if (!wrap_found)
    {
        TP(FS_WORKER_FUNC, "%s: No time wrap found?", __PRETTY_FUNCTION__);
    }
    return wrap_found;
}

/*!
 * @brief Look for the specified parition LBA in the drive header.
 *
 * @param drive_hdr     The drive header information.
 * @param part_lba      The LBA for the partition of interest. Use '0' to find the last entry.
 * @return The index into the drive header array, or MAX_DPARTS if not found.
 */
int fsWorker::findPartition(const DRIVE_HD& drive_hdr, uint64_t part_lba)
{
    int part_index;

    // Find the last partition entry.
    for (part_index = 0; part_index < MAX_DPARTS; ++part_index)
    {
        // Is this index a 0?
        if (drive_hdr.dpart_lba[part_index] == part_lba)
        {
            // Found it!
            break;
        }
    }

    // Are we looking for the last entry?
    if (part_lba == 0)
    {
        // We've gone one step too far. Back up our index and break out.
        --part_index;
    }

    // Make sure there was a partition!
    if (part_index < 0)
        part_index = 0;

    // Return what we found.
    return part_index;
}

/*!
 * @brief Analyze SSD content when we have a capture that did not complete gracefully.
 *
 * @param cap       Capture header already recovered from the drive.
 * @return True if it was possible to reconstruct a capture, false if not.
 */
bool fsWorker::analyzeLrPartialCapture(LR_CAPTURE_HDR *cap)
{
    bool rc = false;
    std::unique_ptr<uint32_t[]> pfmd_buf;       // Use the unique_ptr so we don't have to worry about freeing the memory.

    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Setup our buffer.
    pfmd_buf = std::make_unique<uint32_t[]>(SECTOR_LEN / 4);

    // Read the very last frame of the capture. "frame_count" is the number
    // of frames allocated in the space, not necessarily the number of captured
    // frames.
    if (getFrameMetadata(cap, pfmd_buf.get(), cap->frame_count - 1) == nullptr)
    {
        fi_log(LOG_ERR, "%s: Failed to read PFMD for frame %u", __PRETTY_FUNCTION__, cap->frame_count - 1);
        goto bail;
    }

    // Figure out whether we wrapped by looking at the last metadata record of the normal
    // metadata region.
    TP(FS_WORKER_FUNC, "%s: cap_recID=0x%0x, last_meta=0x%08x", __PRETTY_FUNCTION__, cap->recording_id, pfmd_buf[FMETA_RECORDING_ID]);
    if (pfmd_buf[FMETA_RECORDING_ID] == cap->recording_id)
    {
        // There is a small posibility that the recording ended EXACTLY on the last possible
        // frame, but let's be honest - we probably wrapped. Search the metadata for where
        // the frame time goes backward.
        rc = findLrCaptureWrap(cap);
    }
    else
    {
        // If we did not wrap, search the metadata for the first record that does not
        // have a matching recording ID. The one before that will become our end-trigger frame.
        rc = findLrCaptureEnd(cap);
    }

bail:
    return rc;
}

/*!
 * @brief Analyze the capture header, make sure it is a valid Long Record capture, and
 *        populate the capture information structure.
 *
 * @param hdr_ptr   [IN] Pointer to the raw header bytes read from the disk.
 * @param cap_info  [OUT] The capture informat cache we are filling.
 * @param next_lba  [OUT] The LBA for the next capture header. 0 if this is the last.
 * @return True if analysis was successful, false otherwise.
 */
bool fsWorker::analyzeLrCapture(const uint8_t *hdr_ptr, CAPTURE_DATA& cap_info, uint64_t& next_lba)
{
    LR_CAPTURE_HDR *cap_ptr;
    uint8_t cap_state;
    bool rc = false;
    bool bad_state = false;

    TP(FS_WORKER_FUNC, "%s: type='%s'/0x%08x@%" PRIu64, __PRETTY_FUNCTION__, signatureToString(cap_info.type), cap_info.type, cap_info.lba);

    // Sanity check.
    if (cap_info.type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Incorrect capture type [0x%0x]", __PRETTY_FUNCTION__, cap_info.type);
        goto bail;
    }

    // Copy the header information from the overlay.
    memcpy(&cap_info.hdr.lr_cap, (const LR_CAPTURE_HDR *)hdr_ptr, sizeof(LR_CAPTURE_HDR));

    TP(FS_WORKER_FUNC_NOISY, "%s: part=%" PRIu64 ", prev=%" PRIu64 ", next=%" PRIu64, __PRETTY_FUNCTION__,
       cap_info.hdr.lr_cap.link.dpart_lba, cap_info.hdr.lr_cap.link.previous_lba, cap_info.hdr.lr_cap.link.next_lba);
    TP(FS_WORKER_FUNC_NOISY, "%s: meta=%llu:%llu image=%llu:%llu state=%u flags=0x%x frames=%u/%u", __PRETTY_FUNCTION__,
       cap_info.hdr.lr_cap.fmeta_lba, cap_info.hdr.lr_cap.fmeta_sectors, cap_info.hdr.lr_cap.image_lba, cap_info.hdr.lr_cap.image_sectors,
       cap_info.hdr.lr_cap.state, cap_info.hdr.lr_cap.flags, cap_info.hdr.lr_cap.num_frames, cap_info.hdr.lr_cap.frame_count);

    // Return the next capture header's location.
    next_lba = cap_info.hdr.lr_cap.link.next_lba;

    // Get the capture state.
    cap_state = cap_info.hdr.lr_cap.state;

    // Get a pointer to our capture header.
    cap_ptr = &cap_info.hdr.lr_cap;

    // If the capture is not complete, but the append mode flag is set, we MUST still be in appending mode.
    if ((cap_state != CAPTUREHD_COMPLETE) && ((cap_ptr->flags & CAPHD_FLAG_APPENDMODE) != 0))
        cap_state = CAPTUREHD_APPENDING;

    // Do we have a complete LR capture?
    switch (cap_state)
    {
    case CAPTUREHD_ALLOCATED:
    case CAPTUREHD_READY:
        break;
    case CAPTUREHD_COMPLETE:
        // Sanity check.
        if (cap_ptr->num_frames > cap_ptr->frame_count)
        {
            // Try to correct the problem.
            fi_log(LOG_DEBUG, "%s: num_frames (%d) > frame_count (%d)", __PRETTY_FUNCTION__, cap_ptr->num_frames, cap_ptr->frame_count);
            cap_ptr->num_frames = cap_ptr->frame_count;
        }
        break;
    case CAPTUREHD_APPENDING:
        // Sort it out if the last ROC/BROC capture didn't terminate gracefully.
        // We know there cannot be a wrap so we only have to discover where it ended.
        rc = findLrCaptureEnd(cap_ptr);
        if (rc)
        {
            cap_ptr->flags |= CAPHD_FLAG_RECONSTRUCTED;
            cap_ptr->state = CAPTUREHD_COMPLETE;
        }
        else
        {
            fi_log(LOG_DEBUG, "%s: findLrCaptureEnd() failed.", __PRETTY_FUNCTION__);
            bad_state = true;
        }
        break;
    case CAPTUREHD_CAPTURING:
        // Sort it out if the last capture didn't terminate gracefully.
        rc = analyzeLrPartialCapture(cap_ptr);
        if (rc)
        {
            cap_ptr->flags |= CAPHD_FLAG_RECONSTRUCTED;
            cap_ptr->state = CAPTUREHD_COMPLETE;
        }
        else
        {
            fi_log(LOG_DEBUG, "%s: analyzeLrPartialCapture() failed.", __PRETTY_FUNCTION__);
            bad_state = true;
        }
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected capture state [%d].", __PRETTY_FUNCTION__, cap_state);
        bad_state = true;
        break;
    }

    // ==================================================================================
    // ALM_TODO:
    // In the old part_mgr_ltr.cpp/ltr_startup_analyze(), this is the point where
    // the GIGE_REG_LTR_MODE register would be updated.
    // ==================================================================================

    // As long as the state information looks OK, then we're good to go.
    if (!bad_state)
    {
        // Looks good.
        rc = true;
    }

bail:
    return rc;
}

/*!
 * @brief Analyze the capture header, make sure it is a valid storage capture, and
 *        populate the capture information structure.
 *
 * @param hdr_ptr   [IN] Pointer to the raw header bytes read from the disk.
 * @param cap_info  [OUT] The capture informat cache we are filling.
 * @param next_lba  [OUT] The LBA for the next capture header. 0 if this is the last.
 * @return True if analysis was successful, false otherwise.
 */
bool fsWorker::analyzeStorCapture(const uint8_t *hdr_ptr, CAPTURE_DATA& cap_info, uint64_t& next_lba)
{
    const STOR_CAPTURE_HDR *cap_ptr;
    uint8_t cap_state;
    bool rc = false;
    bool bad_state = false;

    TP(FS_WORKER_FUNC, "%s: type='%s'/0x%08x@%" PRIu64, __PRETTY_FUNCTION__, signatureToString(cap_info.type), cap_info.type, cap_info.lba);

    // Sanity check.
    if (cap_info.type != STOR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Incorrect capture type [0x%0x]", __PRETTY_FUNCTION__, cap_info.type);
        goto bail;
    }

    // Copy the header information from the overlay.
    memcpy(&cap_info.hdr.stor_cap, (const STOR_CAPTURE_HDR *)hdr_ptr, sizeof(STOR_CAPTURE_HDR));

    TP(FS_WORKER_FUNC_NOISY, "%s: name=[%s]", __PRETTY_FUNCTION__, cap_info.hdr.stor_cap.name);
    TP(FS_WORKER_FUNC_NOISY, "%s: part=%" PRIu64 ", prev=%" PRIu64 ", next=%" PRIu64, __PRETTY_FUNCTION__,
       cap_info.hdr.stor_cap.link.dpart_lba, cap_info.hdr.stor_cap.link.previous_lba, cap_info.hdr.stor_cap.link.next_lba);
    TP(FS_WORKER_FUNC_NOISY, "%s: data=%llu:%llu state=%u flags=0x%x used_sectors=%" PRIu64 ", trig_s=%d, trig_ns=%d", __PRETTY_FUNCTION__,
       cap_info.hdr.stor_cap.data_lba, cap_info.hdr.stor_cap.data_sectors, cap_info.hdr.stor_cap.state, cap_info.hdr.stor_cap.flags,
       cap_info.hdr.stor_cap.used_sectors, cap_info.hdr.stor_cap.trigger_time_s, cap_info.hdr.stor_cap.trigger_time_ns);

    // Return the next capture header's location.
    next_lba = cap_info.hdr.stor_cap.link.next_lba;

    // Get a pointer to our capture header.
    cap_ptr = &cap_info.hdr.stor_cap;

    // Get the capture state.
    cap_state = cap_info.hdr.stor_cap.state;

    // Do we have a complete capture?
    switch (cap_state)
    {
    case CAPTUREHD_COMPLETE:
        // We have a complete capture. Do a quick sanity check to make sure things
        // seem to be OK.
        if (cap_ptr->used_sectors > cap_ptr->data_sectors)
        {
            fi_log(LOG_DEBUG, "%s: used_sectors (%" PRIu64 ") > available (%" PRIu64 ")!", __PRETTY_FUNCTION__, cap_ptr->used_sectors, cap_ptr->data_sectors);

            // ALM_TODO:
            // Should probably do something here...
        }
        break;
    case CAPTUREHD_CAPTURING:
        fi_log(LOG_DEBUG, "%s: Attempt to save the capture failed.", __PRETTY_FUNCTION__);
        bad_state = true;
        break;
    case CAPTUREHD_ALLOCATED:
        // The space was allocated, but it is currently unused.
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected capture state [%d].", __PRETTY_FUNCTION__, cap_state);
        bad_state = true;
        break;
    }

    // As long as the state information looks OK, then we're good to go.
    if (!bad_state)
    {
        // Looks good.
        rc = true;
    }

bail:
    return rc;
}

/*!
 * @brief Analyze a Fastec formatted drive, building up the partition information structure(s).
 *
 * @param drive_hdr     [IN] Pointer to the drive header information.
 * @param drive_info    [OUT] The drive's partition information we want to build up.
 * @return True if the analysis was successfull, false otherwise.
 */
bool fsWorker::analyzeDrive(DRIVE_HD *hdr_ptr, PARTITION_INFO_VEC& drive_info)
{
    bool rc;
    int part_index;
    DRIVE_HD drive_hdr;

    // Copy the data into our structure. We need to do this, because some of the routines
    // we will be calling will reuse the sector_buf array.
    memcpy(&drive_hdr, hdr_ptr, sizeof(DRIVE_HD));

    TP(FS_WORKER_FUNC, "%s: drive size=%" PRIu64 ", first dpart@%" PRIu64, __PRETTY_FUNCTION__, drive_hdr.drive_size, drive_hdr.dpart_lba[0]);

    // Assume failure.
    rc = false;

    // Get the first partition's LBA.
    part_index = 0;
    if (drive_hdr.dpart_lba[part_index] == 0)
    {
        fi_log(LOG_DEBUG, "%s: There is no partition information.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Loop through the partitions.
    while (drive_hdr.dpart_lba[part_index] != 0)
    {
        FI_SIG *sig;
        PARTITION_INFO part_info;

        // Get the partition's LBA.
        part_info.part.lba = drive_hdr.dpart_lba[part_index];

        // Read and verify the partition header.
        if (!readSectors(part_info.part.lba, sector_buf.get(), 1))
        {
            fi_log(LOG_DEBUG, "%s: partition header %d@%" PRIu64 " read failed", __PRETTY_FUNCTION__, part_index, part_info.part.lba);
            goto bail;
        }

        // Overlay our signature header.
        sig = (FI_SIG *)sector_buf.get();
        if (!verifyPartitionHeader(sig))
        {
            fi_log(LOG_DEBUG, "%s: partition header %d@%" PRIu64 " verify failed", __PRETTY_FUNCTION__, part_index, part_info.part.lba);
            goto bail;
        }
        part_info.part.type = sig->signature;

        // Overlay our partition header.
        if (!analyzePartition(sector_buf.get() + sizeof(FI_SIG), part_info))
        {
            fi_log(LOG_DEBUG, "%s: partition header %d@%" PRIu64 " analyze failed", __PRETTY_FUNCTION__, part_index, part_info.part.lba);
            goto bail;
        }

        // Append this to our information vector.
        drive_info.push_back(part_info);

        // Move to the next partition.
        ++part_index;
    }

    // All is well.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Analyze a Fastec filesystem Available partition, making sure it is valid and building
 *        a cache of the information.
 *
 * @param hdr_ptr       [IN] Pointer to the raw header bytes read from the disk.
 * @param part_info     [OUT] The partition information cache we want to fill.
 * @return True if analysis was successful for a Fastec filesystem partition, false otherwise.
 */
bool fsWorker::analyzeAvailPartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info)
{
    // Sanity check.
    if (part_info.part.type != AVAIL_PART_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Incorrect partition type [0x%0x]", __PRETTY_FUNCTION__, part_info.part.type);
        return false;
    }

    // Copy the header information from the overlay.
    memcpy(&part_info.part.hdr.avail_part, (const AVAIL_PARTITION_HDR *)hdr_ptr, sizeof(AVAIL_PARTITION_HDR));
    TP(FS_WORKER_FUNC, "%s: data=%llu:%llu", __PRETTY_FUNCTION__, part_info.part.hdr.stor_part.start_lba, part_info.part.hdr.stor_part.data_sectors);

    // All good.
    return true;
}

/*!
 * @brief Analyze a Fastec filesystem Long Record partition, making sure it is valid and building
 *        a cache of the information.
 *
 * @param hdr_ptr       [IN] Pointer to the raw header bytes read from the disk.
 * @param part_info     [OUT] The partition information cache we want to fill.
 * @return True if analysis was successful for a Fastec filesystem partition, false otherwise.
 */
bool fsWorker::analyzeLrPartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info)
{
    bool rc = false;
    uint64_t lba;

    // Sanity check.
    if (part_info.part.type != LR_PART_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Incorrect partition type [0x%0x]", __PRETTY_FUNCTION__, part_info.part.type);
        return false;
    }

    // Copy the header information from the overlay.
    memcpy(&part_info.part.hdr.lr_part, (const LR_PARTITION_HDR *)hdr_ptr, sizeof(LR_PARTITION_HDR));

    // We are an LR recording partition.
    TP(FS_WORKER_FUNC, "%s: meta=%llu:%llu image=%llu:%llu caphd@%llu", __PRETTY_FUNCTION__,
       part_info.part.hdr.lr_part.fmeta_lba, part_info.part.hdr.lr_part.fmeta_sectors, part_info.part.hdr.lr_part.image_lba, part_info.part.hdr.lr_part.image_sectors,
       part_info.part.hdr.lr_part.capturehd_lba);

    // Get the first capture's LBA.
    if (part_info.part.hdr.lr_part.capturehd_lba == 0)
    {
        fi_log(LOG_DEBUG, "%s: There is no capture information.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Loop through the partition.
    lba = part_info.part.hdr.lr_part.capturehd_lba;
    while (lba != 0)
    {
        FI_SIG *sig;
        CAPTURE_DATA cap_info;

        // Read and verify the capture header.
        if (!readSectors(lba, sector_buf.get(), 1))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, lba);
            goto bail;
        }

        // Overlay our signature header and verify that the capture header structure is valid.
        sig = (FI_SIG *)sector_buf.get();
        if (!verifyCaptureHeader(sig, part_info.part.type))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, lba);
            goto bail;
        }
        cap_info.type = sig->signature;
        cap_info.lba = lba;

        // Analyze the capture information.
        switch (cap_info.type)
        {
        case LR_CAPTURE_SIG:
            // Analyze the LR capture header, making sure it is valid.
            if (!analyzeLrCapture(sector_buf.get() + sizeof(FI_SIG), cap_info, lba))
            {
                fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " analyze failed", __PRETTY_FUNCTION__, cap_info.lba);
                goto bail;
            }
            break;
        case STOR_CAPTURE_SIG:
            if (!analyzeStorCapture(sector_buf.get() + sizeof(FI_SIG), cap_info, lba))
            {
                fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " analyze failed", __PRETTY_FUNCTION__, cap_info.lba);
                goto bail;
            }
            break;
        default:
            break;
        }

        // Append this to our information vector.
        part_info.caps.push_back(cap_info);
    }

    // Looks good.
    rc = true;

bail:
    return rc;
}

/*!
 * @brief Analyze a Fastec filesystem Storage partition, making sure it is valid and building
 *        a cache of the information.
 *
 * @param hdr_ptr       [IN] Pointer to the raw header bytes read from the disk.
 * @param part_info     [OUT] The partition information cache we want to fill.
 * @return True if analysis was successful for a Fastec filesystem partition, false otherwise.
 */
bool fsWorker::analyzeStorPartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info)
{
    bool rc = false;
    uint64_t lba;

    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (part_info.part.type != STOR_PART_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Incorrect partition type [0x%0x]", __PRETTY_FUNCTION__, part_info.part.type);
        return false;
    }

    // Copy the header information from the overlay.
    memcpy(&part_info.part.hdr.stor_part, (const STOR_PARTITION_HDR *)hdr_ptr, sizeof(STOR_PARTITION_HDR));

    // We are a Storage partition.
    TP(FS_WORKER_FUNC, "%s: meta=%llu:%llu data=%llu:%llu caphd@%llu", __PRETTY_FUNCTION__,
       part_info.part.hdr.stor_part.meta_lba, part_info.part.hdr.stor_part.meta_sectors, part_info.part.hdr.stor_part.start_lba, part_info.part.hdr.stor_part.data_sectors,
       part_info.part.hdr.stor_part.capturehd_lba);

    // Get the first capture's LBA.
    if (part_info.part.hdr.stor_part.capturehd_lba == 0)
    {
        fi_log(LOG_DEBUG, "%s: There is no capture information.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Loop through the partition.
    lba = part_info.part.hdr.stor_part.capturehd_lba;
    while (lba != 0)
    {
        FI_SIG *sig;
        CAPTURE_DATA cap_info;

        // Read and verify the capture header.
        if (!readSectors(lba, sector_buf.get(), 1))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, lba);
            goto bail;
        }

        // Overlay our signature header and verify that the capture header structure is valid.
        sig = (FI_SIG *)sector_buf.get();
        if (!verifyCaptureHeader(sig, part_info.part.type))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, lba);
            goto bail;
        }
        cap_info.type = sig->signature;
        cap_info.lba = lba;

        // Analyze the storage capture header, making sure it is valid.
        if (!analyzeStorCapture(sector_buf.get() + sizeof(FI_SIG), cap_info, lba))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " analyze failed", __PRETTY_FUNCTION__, cap_info.lba);
            goto bail;
        }

        // Append this to our information vector.
        part_info.caps.push_back(cap_info);
    }

    // Looks good.
    rc = true;

bail:
    return rc;
}


/*!
 * @brief Analyze a Fastec filesystem partition, making sure it is valid and building
 *        a cache of the information.
 *
 * @param hdr_ptr       [IN] Pointer to the raw header bytes read from the disk.
 * @param part_info     [OUT] The partition information cache we want to fill.
 * @return True if analysis was successful for a Fastec filesystem partition, false otherwise.
 */
bool fsWorker::analyzePartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info)
{
    bool rc = false;

    TP(FS_WORKER_FUNC, "%s: type='%s'/0x%08x@%" PRIu64, __PRETTY_FUNCTION__, signatureToString(part_info.part.type), part_info.part.type, part_info.part.lba);

    // Continue processing type specific information.
    switch (part_info.part.type)
    {
    case AVAIL_PART_SIG:
        rc = analyzeAvailPartition(hdr_ptr, part_info);
        break;
    case LR_PART_SIG:
        rc = analyzeLrPartition(hdr_ptr, part_info);
        break;
    case STOR_PART_SIG:
        rc = analyzeStorPartition(hdr_ptr, part_info);
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unknown partition type [0x%0x]", __PRETTY_FUNCTION__, part_info.part.type);
        rc = false;
        break;
    }

    return rc;
}

/*!
 * @brief Append an LR capture to an existing LR partition.
 *
 * @param part_ptr  [IN] Pointer to the partition we're working in.
 * @param part_lba  [IN] The LBA for the partition.
 * @param layout    [OUT] The layout information we are creating.
 *
 * @return True on success, false otherwise.
 */
bool fsWorker::appendLrCapture(const LR_PARTITION_HDR *part_ptr, uint64_t part_lba, LAYOUT_CAPTURE_VEC& layout)
{
    bool rc = false;
    uint32_t max_ct_size;
    uint64_t cap_active_lba;
    uint64_t cap_lba;
    uint64_t cap_image_lba;
    uint64_t part_end;
    FI_SIG *sig;
    LR_CAPTURE_HDR *cap_ptr;
    LR_CAPTURE_HDR new_cap;
    LAYOUT_CAPTURE_VEC::iterator lay_it;

    TP(FS_WORKER_FUNC, "%s: part_lba=%" PRIu64, __PRETTY_FUNCTION__, part_lba);

    // Sanity checks.
    if (part_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: part_ptr is null?!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (layout.empty())
    {
        fi_log(LOG_DEBUG, "%s: empty layout!?", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (layout.size() != 1)
    {
        // ALM_TODO:
        fi_log(LOG_DEBUG, "%s: ==== Appending more than one capture is UNDER CONSTRUCTION ====", __PRETTY_FUNCTION__);
    }
    for (lay_it = layout.begin(); lay_it != layout.end(); ++lay_it)
    {
        // Make sure all the signatures are OK.
        if (lay_it->signature != LR_CAPTURE_SIG)
        {
            fi_log(LOG_DEBUG, "%s: Invalid signature '%s' [0x%08x]", __PRETTY_FUNCTION__, signatureToString(lay_it->signature), lay_it->signature);
            goto bail;
        }
    }

    // Get the first capture header's LBA.
    cap_lba = part_ptr->capturehd_lba;
    cap_ptr = nullptr;
    part_end = part_ptr->image_lba + part_ptr->image_sectors;

    // Walk the chain until we get to the last one.
    while (cap_lba != 0)
    {
        // Read and verify the capture header.
        if (!readSectors(cap_lba, sector_buf.get(), 1))
        {
            fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, cap_lba);
            goto bail;
        }

        // Overlay our signature header.
        sig = (FI_SIG *)sector_buf.get();
        if (!verifyCaptureHeader(sig, LR_PART_SIG))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, cap_lba);
            goto bail;
        }

        // Overlay our capture header.
        cap_ptr = (LR_CAPTURE_HDR *)(sector_buf.get() + sizeof(FI_SIG));

        // Get the next capture's LBA. Remember our LBA so we can update
        // the header information.
        cap_active_lba = cap_lba;
        cap_lba = cap_ptr->link.next_lba;
    }

    // Make sure we have a capture header.
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: cap_ptr is null. Problems finding the end of the chain.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Link to where our new header will go.
    cap_ptr->link.next_lba = cap_ptr->image_lba + cap_ptr->image_sectors;

    // Save our LBA for the new capture header.
    cap_lba = cap_ptr->link.next_lba;

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(cap_active_lba, sector_buf.get(), 1))
        goto bail;

    // Let's do this calculation once. The image LBA MUST be page aligned. The max_ct_size is the maximum
    // size of the correction table, be we need to calculate the total size of the correction/gain information.
    // This will be 1 sector for the header, max_ct_size for the correction table and another for the gain table.
    max_ct_size = ((2 * BYTES_TO_SECTORS(layout.at(0).max_ct_size) + 1)) * SECTOR_LEN;
    cap_image_lba = SECTORS_TO_PAGES(cap_lba + FiFilesystem::getReserved(LR_CAPTURE_SIG) + BYTES_TO_SECTORS(max_ct_size)) * SECTORS_PER_PAGE;

    // Are we supposed to use the remainder?
    if (layout.at(0).cap_size == 0)
    {
        layout.at(0).cap_size = part_end - cap_image_lba;
        fi_log(LOG_DEBUG, "%s: giving remainder to cap_size: %" PRIu64, __PRETTY_FUNCTION__, layout.at(0).cap_size);
    }

    // Make sure we still fit.
    if ((cap_image_lba + layout.at(0).cap_size) > part_end)
    {
        uint64_t new_size = part_end - cap_image_lba;
        fi_log(LOG_DEBUG, "%s: Adjusting capture size, %" PRIu64 " -> %" PRIu64, __PRETTY_FUNCTION__, layout.at(0).cap_size, new_size);
        layout.at(0).cap_size = new_size;
    }

    // Create our capture header.
    createLrCaptureHdr(&new_cap, part_lba, cap_active_lba, cap_lba, layout.at(0).cap_size, 0, 0, max_ct_size);

    // Reset our sector-aligned buffer memory.
    memset(sector_buf.get(), 0, SECTOR_LEN);

    // Copy the partition header into the memory.
    memcpy((sector_buf.get() + sizeof(FI_SIG)), &new_cap, sizeof(LR_CAPTURE_HDR));

    // Build up the signature for the partition.
    sig = (FI_SIG *)sector_buf.get();
    sig->signature = layout.at(0).signature;
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(cap_lba, sector_buf.get(), 1))
        goto bail;

    // Invalidate the cache.
    // ALM_TODO: Update the cache with the new information.
    cacheInvalidate(dev_path);

    // All is well.
    rc = true;

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Append storage capture to an existing storage partition.
 *
 * @param part_ptr  [IN] Pointer to the partition we're working in.
 * @param part_lba  [IN] The LBA for the partition.
 * @param layout    [OUT] The layout information we are creating.
 *
 * @return True on success, false otherwise.
 */
bool fsWorker::appendStorCapture(const STOR_PARTITION_HDR *part_ptr, uint64_t part_lba, LAYOUT_CAPTURE_VEC& layout)
{
    bool rc = false;
    uint64_t cap_active_lba;
    uint64_t cap_lba;
    uint64_t cap_image_lba;
    uint64_t part_end;
    FI_SIG *sig;
    STOR_CAPTURE_HDR *cap_ptr;
    STOR_CAPTURE_HDR new_cap;
    LAYOUT_CAPTURE_VEC::iterator lay_it;

    TP(FS_WORKER_FUNC, "%s: part_lba=%" PRIu64, __PRETTY_FUNCTION__, part_lba);

    // Sanity checks.
    if (part_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: part_ptr is null?!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (layout.empty())
    {
        fi_log(LOG_DEBUG, "%s: empty layout!?", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (layout.size() != 1)
    {
        // ALM_TODO:
        fi_log(LOG_DEBUG, "%s: ==== Appending more than one capture is UNDER CONSTRUCTION ====", __PRETTY_FUNCTION__);
    }
    for (lay_it = layout.begin(); lay_it != layout.end(); ++lay_it)
    {
        // Make sure all the signatures are OK.
        if (lay_it->signature != STOR_CAPTURE_SIG)
        {
            fi_log(LOG_DEBUG, "%s: Invalid signature '%s' [0x%08x]", __PRETTY_FUNCTION__, signatureToString(lay_it->signature), lay_it->signature);
            goto bail;
        }
    }

    // Get the first capture header's LBA.
    cap_lba = part_ptr->capturehd_lba;
    cap_ptr = nullptr;
    part_end = part_ptr->start_lba + part_ptr->data_sectors;

    // Walk the chain until we get to the last one.
    while (cap_lba != 0)
    {
        // Read and verify the capture header.
        if (!readSectors(cap_lba, sector_buf.get(), 1))
        {
            fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, cap_lba);
            goto bail;
        }

        // Overlay our signature header.
        sig = (FI_SIG *)sector_buf.get();
        if (!verifyCaptureHeader(sig, STOR_PART_SIG))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, cap_lba);
            goto bail;
        }

        // Overlay our capture header.
        cap_ptr = (STOR_CAPTURE_HDR *)(sector_buf.get() + sizeof(FI_SIG));

        // Get the next capture's LBA. Remember our LBA so we can update
        // the header information.
        cap_active_lba = cap_lba;
        cap_lba = cap_ptr->link.next_lba;
    }

    // Make sure we have a capture header.
    if (cap_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: cap_ptr is null. Problems finding the end of the chain.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Link to where our new header will go.
    cap_ptr->link.next_lba = cap_ptr->data_lba + cap_ptr->data_sectors;

    // Save our LBA for the new capture header.
    cap_lba = cap_ptr->link.next_lba;

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(cap_active_lba, sector_buf.get(), 1))
        goto bail;

    // Let's do this calculation once. The data LBA MUST be page aligned.
    cap_image_lba = SECTORS_TO_PAGES(cap_lba + FiFilesystem::getReserved(STOR_CAPTURE_SIG)) * SECTORS_PER_PAGE;

    // Are we supposed to use the remainder?
    if (layout.at(0).cap_size == 0)
    {
        layout.at(0).cap_size = part_end - cap_image_lba;
        fi_log(LOG_DEBUG, "%s: giving remainder to cap_size: %" PRIu64, __PRETTY_FUNCTION__, layout.at(0).cap_size);
    }

    // Make sure we still fit.
    if ((cap_image_lba + layout.at(0).cap_size) > part_end)
    {
        uint64_t new_size = part_end - cap_image_lba;
        fi_log(LOG_DEBUG, "%s: Adjusting capture size, %" PRIu64 " -> %" PRIu64, __PRETTY_FUNCTION__, layout.at(0).cap_size, new_size);
        layout.at(0).cap_size = new_size;
    }

    // Create our capture header.
    createStorCaptureHdr(&new_cap, part_lba, cap_active_lba, cap_lba, layout.at(0).cap_size);

    // Reset our sector-aligned buffer memory.
    memset(sector_buf.get(), 0, SECTOR_LEN);

    // Copy the partition header into the memory.
    memcpy((sector_buf.get() + sizeof(FI_SIG)), &new_cap, sizeof(STOR_CAPTURE_HDR));

    // Build up the signature for the partition.
    sig = (FI_SIG *)sector_buf.get();
    sig->signature = layout.at(0).signature;
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(cap_lba, sector_buf.get(), 1))
        goto bail;

    // Invalidate the cache.
    // ALM_TODO: Update the cache with the new information.
    cacheInvalidate(dev_path);

    // All is well.
    rc = true;

bail:
    // Return our result.
    return rc;
}


/*!
 * @brief Do the work of building up a layout. The information will be placed in the layout_info
 *        parameter.
 *
 * @param dev_path      [IN] The device we're interested in.
 * @param num_sectors   [IN] The total number of sectors we've allocated to work in.
 * @param part_info     [IN] The partition format information.
 * @param drv           [OUT] The header for the drive information.
 * @param layout_info   [OUT] The layout information we're building for the partition.
 * @return True if the layout was succesfull, false otherwise.
 */
bool fsWorker::buildLayout(const std::string& dev_path, uint64_t num_sectors, const LAYOUT_PARTITION_VEC& part_info, DRIVE_HD& drv, PARTITION_INFO_VEC& layout_info)
{
    bool rc;
    std::size_t layout_count;
    int64_t remaining_sectors;
    uint64_t part_lba;

    // Assume failure.
    rc = false;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Remember some of the information we're working with.
    setDrivePath(dev_path);
    setNumSectors(num_sectors);

    // Initialize the drive header.
    memset((void *)&drv, 0, sizeof(drv));
    drv.alloc_unit = SECTORS_PER_PAGE;
    drv.drive_size = num_sectors;

    // Initialize where the first drive partition will be located.
    part_lba = FIRST_DPART_LBA;

    // The number of available sectors is the total amount of sectors
    // AFTER the partition header. The partition header will take a
    // single sector.
    remaining_sectors = drv.drive_size - (part_lba + 1);

    // Loop through each entry in the layout vector.
    for (layout_count = 0; layout_count < part_info.size(); ++layout_count)
    {
        uint64_t part_size;
        PARTITION_INFO part_entry;
        LAYOUT_PARTITION part_layout = part_info.at(layout_count);

        // Sanity check. Make sure we have more sectors.
        if (remaining_sectors <= 0)
        {
            fi_log(LOG_DEBUG, "%s: We've run out of sectors!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // How many sectors are we allocating for this partition? This allocation amount
        // includes the partition header, all capture headers, and all data space.
        if (part_layout.part_size == 0)
            part_size = remaining_sectors;
        else
            part_size = part_layout.part_size;

        // Make sure this makes sense.
        if (part_size > (uint64_t)remaining_sectors)
        {
            fi_log(LOG_DEBUG, "%s: Adjusting part_size, %" PRIu64 " -> %" PRIu64, __PRETTY_FUNCTION__, part_size, remaining_sectors);
            part_size = remaining_sectors;
        }

        // Build the appropriate layout.
        switch (part_layout.signature)
        {
        case AVAIL_PART_SIG:
            rc = buildAvailLayout(&part_layout, part_lba, part_size, part_entry);
            break;
        case LR_PART_SIG:
            rc = buildLrLayout(&part_layout, part_lba, part_size, part_entry);
            break;
        case STOR_PART_SIG:
            rc = buildStorLayout(&part_layout, part_lba, part_size, part_entry);
            break;
        default:
            fi_log(LOG_DEBUG, "%s: Unexpected signature [0x%x]", __PRETTY_FUNCTION__, part_layout.signature);
            rc = false;
            break;
        }
        if (!rc)
        {
            fi_log(LOG_DEBUG, "%s: There was an error with the build routine for signature 0x%x", __PRETTY_FUNCTION__, part_layout.signature);
            goto bail;
        }

        // Add this to our cache information.
        layout_info.push_back(part_entry);

        // Remember where this partition goes.
        drv.dpart_lba[layout_count] = part_lba;

        // Move to the next area.
        part_lba += part_size;

        // Update our remaining number of sectors.
        remaining_sectors -= part_size;
    }

    // Everything looks good.
    rc = true;

bail:
    // Close the drive handle.
    setDrivePath(std::string());

    // Return our result.
    return rc;
}

/*!
 * @brief Process the messages for the thread.
 */
void fsWorker::messageHandle(FAS_THREAD::Message *message)
{
    // Sanity check.
    if (message == nullptr)
    {
        fi_log(LOG_ERR, "%s: Got a null message!", __PRETTY_FUNCTION__);
        return;
    }

    TP(FS_MESSAGE, "%s: %d[%s] message.", __PRETTY_FUNCTION__, (int)message->req, reqToString((int)message->req).c_str());

    try
    {
        // Parse the message.
        switch ((int)message->req)
        {
        case FS_ANALYZE:
            handleAnalyzeRequest(static_cast<fsAnalyzeMessage *>(message));
            break;
        case FS_APPEND_CAPTURE:
            handleAppendCaptureRequest(static_cast<fsAppendCaptureMessage *>(message));
            break;
        case FS_APPEND_PARTITION:
            handleAppendPartitionRequest(static_cast<fsAppendPartitionMessage *>(message));
            break;
        case FS_CLEAR_CACHE:
            handleClearCacheRequest(static_cast<fsClearCacheMessage *>(message));
            break;
        case FS_DELETE_CAPTURE:
            handleDeleteCaptureRequest(static_cast<fsDeleteCaptureMessage *>(message));
            break;
        case FS_DELETE_PARTITION:
            handleDeletePartitionRequest(static_cast<fsDeletePartitionMessage *>(message));
            break;
        case FS_GET_BOOKMARK:
            handleGetBookmarkRequest(static_cast<fsGetBookmarkMessage *>(message));
            break;
        case FS_GET_CORRECTION_TABLE:
            handleGetCorrectionTableRequest(static_cast<fsGetCorrectionTableMessage *>(message));
            break;
        case FS_GET_GAIN_TABLE:
            handleGetGainTableRequest(static_cast<fsGetGainTableMessage *>(message));
            break;
        case FS_GET_INFO:
            handleGetInfoRequest(static_cast<fsGetInfoMessage *>(message));
            break;
        case FS_GET_PM_METADATA:
            handleGetPmMetadataRequest(static_cast<fsGetPmMetadataMessage *>(message));
            break;
        case FS_LAYOUT:
            handleLayoutRequest(static_cast<fsLayoutMessage *>(message));
            break;
        case FS_RENAME:
            handleRenameRequest(static_cast<fsRenameMessage *>(message));
            break;
        case FS_RESIZE_CAPTURE:
            handleResizeCaptureRequest(static_cast<fsResizeCaptureMessage *>(message));
            break;
        case FS_RESIZE_PARTITION:
            handleResizePartitionRequest(static_cast<fsResizePartitionMessage *>(message));
            break;
        case FS_SET_BOOKMARK:
            handleSetBookmarkRequest(static_cast<fsSetBookmarkMessage *>(message));
            break;
        case FS_SET_CORRECTION_INFO:
            handleSetCorrectionInfoRequest(static_cast<fsSetCorrectionInfoMessage *>(message));
            break;
        case FS_SET_PM_METADATA:
            handleSetPmMetadataRequest(static_cast<fsSetPmMetadataMessage *>(message));
            break;
        case FS_TRY_LAYOUT:
            handleTryLayout(static_cast<fsTryLayoutMessage *>(message));
            break;
        case FS_UPDATE_CAPTURE:
            handleUpdateCapture(static_cast<fsUpdateCaptureMessage *>(message));
            break;
        case FS_WRITE_PARTITION:
            handleWritePartition(static_cast<fsWritePartitionMessage *>(message));
            break;

        case FAS_THREAD::DUMP:
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::HALT:
            halted = true;
            message->resp = FAS_THREAD::OK;
            break;
        case FAS_THREAD::NOP:
            message->resp = FAS_THREAD::OK;
            break;
        default:
            message->resp = FAS_THREAD::FAULT;
            break;
        }

        // Forward on the messsage.
        message->forward();
    }
    catch (FAS_THREAD::Exception& e)
    {
        fi_log(LOG_DEBUG, "Exception in jpegThread:\n%s", e.what().c_str());
        message->resp = FAS_THREAD::FAULT;
        message->forward();
    }
}

void fsWorker::handleAnalyzeRequest(fsAnalyzeMessage *msg)
{
    DRIVE_HD *drive_hdr;
    FS_CACHE_ENTRY entry;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        return;
    }

    // Remember the drive path we're working with. We don't sanity check
    // the path here because the instant we try to use an invalid path we
    // will get a failure.
    if (!setDrivePath(msg->dev_path))
    {
        fi_log(LOG_DEBUG, "%s: Unable to set the path to [%s]", __PRETTY_FUNCTION__, msg->dev_path.c_str());
        return;
    }

    // Read the drive header.
    if (!readSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the drive header.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Verify that we have a Fastec Filesystem drive.
    if (!verifyDriveHeader((FI_SIG *)sector_buf.get()))
    {
        // No Fastec Filesystem here.
        TP(FS_MESSAGE_NOISY, "%s: drive [%s] is not a Fastec Filesystem.", __PRETTY_FUNCTION__, dev_path.c_str());
        msg->is_fastec = false;
        msg->resp = FAS_THREAD::OK;
        goto bail;
    }

    // Get our pointer to the drive header. Save the number of sectors allocated
    // to the filesystem in our cache entry.
    drive_hdr = (DRIVE_HD *)(sector_buf.get() + sizeof(FI_SIG));
    entry.drive_size = drive_hdr->drive_size;

    // Ok, so we definitely have a Fastec filesystem based on the drive header.
    msg->is_fastec = true;

    // Analyze the drive and see if we have a VALID filesystem.
    //
    // ALM_TODO: If there are ANY valid partitions in the cache, we should maybe
    //           save that and let the user know that SOME of the information is
    //           OK, even if there is corrupt data.
    if (analyzeDrive(drive_hdr, entry.part_info))
    {
        std::map<std::string, FS_CACHE_ENTRY>::iterator it;

        // It is a valid Fastec filesystem, so add information to cache.
        // Make sure we don't already have an entry in the cache for this device.
        // If we do, remove it so we don't create a duplicate.
        it = cache.find(dev_path);
        if (it != cache.end())
        {
            // Remove the existing entry.
            TP(FS_MESSAGE_NOISY, "%s: Found an existing entry. Removing.", __PRETTY_FUNCTION__);
            cache.erase(it);
        }

        // Save the information to the cache.
        TP(FS_MESSAGE_NOISY, "%s: Inserting part_info for [%s]", __PRETTY_FUNCTION__, dev_path.c_str());
        cache.insert(std::make_pair(dev_path, entry));

        // All good.
        msg->resp = FAS_THREAD::OK;
    }

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleAppendCaptureRequest(fsAppendCaptureMessage *msg)
{
    int part_index;
    bool enough_space = false;
    PARTITION_INFO *part_info;
    CAP_UNION_VEC::iterator cap_it;

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    TP(FS_MESSAGE, "%s: [%s] lba:%" PRIu64, __PRETTY_FUNCTION__, msg->dev_path.c_str(), msg->part_lba);

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        return;
    }

    // Remember some of the information we're working with.
    setDrivePath(msg->dev_path);

    // Get the partition information from our cache.
    part_info = cacheGetPartition(msg->dev_path, msg->part_lba, part_index);
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to locate partition %" PRIu64, __PRETTY_FUNCTION__, msg->part_lba);
        goto bail;
    }

    // We don't "append" captures to an available partition.
    if (part_info->part.type == AVAIL_PART_SIG)
    {
        fi_log(LOG_DEBUG, "%s: Available Partitions do not support append capture operations.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Find the last capture.
    for (cap_it = part_info->caps.begin(); cap_it != part_info->caps.end(); ++cap_it)
    {
        bool found_it = false;

        switch (part_info->part.type)
        {
        case LR_PART_SIG:
            // Is this the last one?
            if (cap_it->hdr.lr_cap.link.next_lba == 0)
            {
                uint64_t cap_end = cap_it->hdr.lr_cap.image_lba + cap_it->hdr.lr_cap.image_sectors;
                uint64_t part_end = part_info->part.hdr.lr_part.image_lba + part_info->part.hdr.lr_part.image_sectors;

                // Found it.
                found_it = true;

                // We need to figure out if there enough space left for what the
                // user is asking for. If the cap_size requested is 0, then the
                // user is asking for whatever's left.
                if (msg->layout.at(0).cap_size == 0)
                {
                    // There is a minimum amount of space required for a LR capture.
                    uint64_t min_space;
                    min_space = BYTES_TO_SECTORS(cap_it->hdr.lr_cap.frame_size * MIN_FRAMES_REQUIRED);
                    min_space += FiFilesystem::getReserved(LR_CAPTURE_SIG);

                    // User wants to fit in the space.
                    if ((part_end - cap_end) > min_space)
                        enough_space = true;
                    else
                        fi_log(LOG_DEBUG, "%s: min_space:%" PRIu64 ", avail:%" PRIu64, __PRETTY_FUNCTION__, min_space, (part_end - cap_end));
                }
                else
                {
                    // User wants to fit in the space.
                    if ((part_end - cap_end) > msg->layout.at(0).cap_size)
                        enough_space = true;
                    else
                        fi_log(LOG_DEBUG, "%s: Not enough space, avail:%" PRIu64, __PRETTY_FUNCTION__, (part_end - cap_end));
                }
            }
            break;
        case STOR_CAPTURE_SIG:
            // Is this the last one?
            if (cap_it->hdr.stor_cap.link.next_lba == 0)
            {
                uint64_t cap_end = cap_it->hdr.stor_cap.data_lba + cap_it->hdr.stor_cap.data_sectors;
                uint64_t part_end = part_info->part.hdr.stor_part.start_lba + part_info->part.hdr.stor_part.data_sectors;

                // Found it.
                found_it = true;

                // We need to figure out if there enough space left for what the
                // user is asking for. If the cap_size requested is 0, then the
                // user is asking for whatever's left.
                if (msg->layout.at(0).cap_size == 0)
                {
                    // Make sure there's at least the minimum amount.
                    if ((part_end - cap_end) > STOR_CAPTURE_MIN_SECTORS)
                        enough_space = true;
                    else
                        fi_log(LOG_DEBUG, "%s: min_space:%" PRIu64 ", avail:%" PRIu64, __PRETTY_FUNCTION__, STOR_CAPTURE_MIN_SECTORS, (part_end - cap_end));
                }
                else
                {
                    // User wants to fit in the space.
                    if ((part_end - cap_end) > msg->layout.at(0).cap_size)
                        enough_space = true;
                    else
                        fi_log(LOG_DEBUG, "%s: Not enough space, avail:%" PRIu64, __PRETTY_FUNCTION__, (part_end - cap_end));
                }
            }
            break;
        default:
            fi_log(LOG_DEBUG, "%s: Unsupported partition type [0x%x]", __PRETTY_FUNCTION__, part_info->part.type);
            goto bail;
        }

        // Did we find the one we were looking for?
        if (found_it)
            break;
    }
    if (!enough_space)
    {
        fi_log(LOG_DEBUG, "%s: Insufficient space for the append operation.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Continue processing type specific information.
    switch (part_info->part.type)
    {
    case LR_PART_SIG:
        if (appendLrCapture(&part_info->part.hdr.lr_part, msg->part_lba, msg->layout))
            msg->resp = FAS_THREAD::OK;
        break;
    case STOR_PART_SIG:
        if (appendStorCapture(&part_info->part.hdr.stor_part, msg->part_lba, msg->layout))
            msg->resp = FAS_THREAD::OK;
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unknown partition type [0x%0x]", __PRETTY_FUNCTION__, part_info->part.type);
        break;
    }

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleAppendPartitionRequest(fsAppendPartitionMessage *msg)
{
    FAS_THREAD::RESP_CODE rc;
    int last_part_index;
    int part_index;
    int64_t remaining_sectors;
    uint64_t part_lba;
    FI_SIG *sig;
    DRIVE_HD drive_hdr;
    LAYOUT_PARTITION_VEC::iterator lay_it;
    PARTITION_INFO *part_info;
    PARTITION_INFO_VEC new_part_vec;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we will fail.
    rc = FAS_THREAD::FAULT;

    // Continue sanity checks.
    if (msg->layout.size() != 1)
    {
        // ALM_TODO:
        fi_log(LOG_DEBUG, "%s: ==== Appending more than one partition is UNDER CONSTRUCTION ====", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Remember some of the information we're working with.
    setDrivePath(msg->dev_path);

    // Read and verify the drive header.
    if (!readSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: drive header read failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();
    if (!verifyDriveHeader(sig))
    {
        fi_log(LOG_DEBUG, "%s: drive header verify failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Copy over the drive header information so we can reuse the buffer.
    memcpy(&drive_hdr, (sector_buf.get() + sizeof(FI_SIG)), sizeof(DRIVE_HD));

    // Remember how many sectors we have to work with.
    setNumSectors(drive_hdr.drive_size);

    // Find the last partition entry.
    last_part_index = findPartition(drive_hdr, 0);

    // Make sure we have room in the drive for another partition.
    if (last_part_index == MAX_DPARTS)
    {
        fi_log(LOG_WARNING, "%s: No room left for any more partitions", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Get the partition information from our cache.
    part_info = cacheGetPartition(msg->dev_path, drive_hdr.dpart_lba[last_part_index], part_index);
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to locate partition %" PRIu64, __PRETTY_FUNCTION__, drive_hdr.dpart_lba[last_part_index]);
        goto bail;
    }
    if (last_part_index != part_index)
    {
        fi_log(LOG_DEBUG, "%s: last_part_index (%d) != part_index (%d)?", __PRETTY_FUNCTION__, last_part_index, part_index);
    }

    // The new partition will go after the last partition.
    part_lba = part_info->part.hdr.lr_part.image_lba + part_info->part.hdr.lr_part.image_sectors;

    // How much space is left on the drive?
    remaining_sectors = getNumSectors() - part_lba;

    // Move the partition index up to where the new partitions will start.
    part_index = last_part_index + 1;

    // Walk through the layout and build up our partition(s).
    for (lay_it = msg->layout.begin(); lay_it != msg->layout.end(); ++lay_it)
    {
        bool result;
        uint64_t part_size;
        LAYOUT_PARTITION layout_info = *lay_it;
        PARTITION_INFO part_entry;

        // Sanity check. Make sure we have more sectors.
        if (remaining_sectors <= 0)
        {
            fi_log(LOG_DEBUG, "%s: We've run out of sectors!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // How many sectors are we allocating for this partition? This allocation amount
        // includes the partition header, all capture headers, and all data space.
        if (layout_info.part_size == 0)
            part_size = remaining_sectors;
        else
            part_size = layout_info.part_size;

        // Make sure this makes sense.
        if (part_size > (uint64_t)remaining_sectors)
        {
            fi_log(LOG_DEBUG, "%s: Adjusting part_size, %" PRIu64 " -> %" PRIu64, __PRETTY_FUNCTION__, part_size, remaining_sectors);
            part_size = remaining_sectors;
        }

        // Build the appropriate layout.
        switch (layout_info.signature)
        {
        case LR_PART_SIG:
            result = buildLrLayout(&layout_info, part_lba, part_size, part_entry);
            break;
        case AVAIL_PART_SIG:
            result = buildAvailLayout(&layout_info, part_lba, part_size, part_entry);
            break;
        case STOR_PART_SIG:
            result = buildStorLayout(&layout_info, part_lba, part_size, part_entry);
            break;
        default:
            fi_log(LOG_DEBUG, "%s: Unexpected signature [0x%x]", __PRETTY_FUNCTION__, layout_info.signature);
            result = false;
            break;
        }
        if (!result)
        {
            fi_log(LOG_DEBUG, "%s: There was an error with the build routine for signature 0x%x", __PRETTY_FUNCTION__, layout_info.signature);
            goto bail;
        }

        // Push the partition entry onto the vector.
        new_part_vec.push_back(part_entry);

        // Remember where this partition goes.
        drive_hdr.dpart_lba[part_index] = part_lba;

        // Move to the next area.
        part_lba += part_size;
        ++part_index;

        // Update our remaining number of sectors.
        remaining_sectors -= part_size;
    }

    // Write the new layout to the drive.
    if (!writeLayout(&drive_hdr, new_part_vec))
    {
        fi_log(LOG_DEBUG, "%s: writeLayout() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All done.
    rc = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());

    // Set our response code.
    msg->resp = rc;
}

void fsWorker::handleClearCacheRequest(fsClearCacheMessage *msg)
{
    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Are we clearing the entire map, or just one entry?
    if (msg->dev_path.empty())
    {
        // Erase the entire cache.
        cache.clear();
    }
    else
    {
        // Invalidate the cache.
        cacheInvalidate(msg->dev_path);
    }

    // All done.
    msg->resp = FAS_THREAD::OK;
}

void fsWorker::handleDeleteCaptureRequest(fsDeleteCaptureMessage *msg)
{
    int part_index;
    FI_SIG *sig;
    CAPTURE_UNION *cap_ptr;
    PARTITION_INFO *part_info;

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    TP(FS_MESSAGE, "%s: [%s] part_lba:%" PRIu64 ", cap_lba:%" PRIu64, __PRETTY_FUNCTION__, msg->dev_path.c_str(), msg->part_lba, msg->capture_lba);

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        return;
    }

    // Remember some of the information we're working with.
    setDrivePath(msg->dev_path);

    // Get the partition information from our cache.
    part_info = cacheGetPartition(msg->dev_path, msg->part_lba, part_index);
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to locate partition %" PRIu64, __PRETTY_FUNCTION__, msg->part_lba);

        // Close the drive handle.
        setDrivePath(std::string());
        return;
    }

    // Make sure the capture we're trying to delete is the last one.
    CAPTURE_DATA& cap_info = part_info->caps.back();
    if (cap_info.lba != msg->capture_lba)
    {
        fi_log(LOG_DEBUG, "%s: The cache is out of sync!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (cap_info.hdr.link.next_lba != 0)
    {
        fi_log(LOG_DEBUG, "%s: We can only delete the LAST capture!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Remove the capture header from the cache.
    part_info->caps.pop_back();

    // Get the information for the new last capture.
    cap_info = part_info->caps.back();

    // Terminate the chain.
    cap_info.hdr.link.next_lba = 0;

    // Read and verify the capture header.
    if (!readSectors(cap_info.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, cap_info.lba);
        goto bail;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();
    if (!verifyCaptureHeader(sig, part_info->part.type))
    {
        fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, cap_info.lba);
        goto bail;
    }

    // Overlay our capture header.
    cap_ptr = (CAPTURE_UNION *)(sector_buf.get() + sizeof(FI_SIG));

    // Terminate the link
    cap_ptr->link.next_lba = 0;

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the updated capture header to the drive.
    if (!writeSectors(cap_info.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleDeletePartitionRequest(fsDeletePartitionMessage *msg)
{
    FAS_THREAD::RESP_CODE rc;
    int part_index;
    FI_SIG *sig;
    DRIVE_HD *drive_hdr;
    std::map<std::string, FS_CACHE_ENTRY>::iterator it;

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    TP(FS_MESSAGE, "%s: [%s] part_lba:%" PRIu64, __PRETTY_FUNCTION__, msg->dev_path.c_str(), msg->part_lba);

    // Assume we will fail.
    rc = FAS_THREAD::FAULT;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Remember some of the information we're working with.
    setDrivePath(msg->dev_path);

    // Read and verify the drive header.
    if (!readSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: drive header read failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();
    if (!verifyDriveHeader(sig))
    {
        fi_log(LOG_DEBUG, "%s: drive header verify failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Copy over the drive header information so we can reuse the buffer.
    drive_hdr = (DRIVE_HD *)(sector_buf.get() + sizeof(FI_SIG));

    // Remember how many sectors we have to work with.
    setNumSectors(drive_hdr->drive_size);

    // Find the specified partition entry.
    part_index = findPartition(*drive_hdr, msg->part_lba);

    // Make sure we have room in the drive for another partition.
    if (part_index == MAX_DPARTS)
    {
        fi_log(LOG_WARNING, "%s: Couldn't find the partition [%" PRIu64, __PRETTY_FUNCTION__, msg->part_lba);
        goto bail;
    }

    // At least for now, the partition be removed MUST be the last partition.
    if (drive_hdr->dpart_lba[part_index + 1] != 0)
    {
        fi_log(LOG_DEBUG, "%s: Only deleting the LAST partition is supported!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Clear the entry.
    drive_hdr->dpart_lba[part_index] = 0;

    // Update the checksum.
    setBlockChecksum(sig);

    // Write the drive header.
    if (!writeSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSectors(DRIVE_HD) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Clear this partition from the cache.
    it = cache.find(dev_path);
    if (it != cache.end())
    {
        // Sanity check.
        if (msg->part_lba != it->second.part_info.back().part.lba)
        {
            fi_log(LOG_WARNING, "%s: cache is out of sync!", __PRETTY_FUNCTION__);
        }

        // Remove the partition entry.
        it->second.part_info.pop_back();
    }

    // All good.
    rc = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());

    // Set our response code.
    msg->resp = rc;
}

void fsWorker::handleGetBookmarkRequest(fsGetBookmarkMessage *msg)
{
    uint32_t buf_size;
    uint64_t start_sector;
    CAPTURE_DATA *cap_data;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
        return;
    }
    setDrivePath(msg->dev_path);

    // Get our cached capture data.
    cap_data = cacheFindCapture(msg->dev_path, msg->capture_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: capture (%" PRIu64 ") not found.", __PRETTY_FUNCTION__, msg->capture_lba);
        goto bail;
    }

    // Verify that the capture header is of a type we would expect.
    if (cap_data->type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: capture header type ('%s' [0x%08x]) is not valid!", __PRETTY_FUNCTION__, signatureToString(cap_data->type), cap_data->type);
        goto bail;
    }

    // Allocate enough buffer to read the bookmarks.
    start_sector = BYTES_TO_SECTORS((msg->capture_lba * SECTOR_LEN) + cap_data->hdr.lr_cap.bk_offset);
    buf_size = BYTES_TO_SECTORS(cap_data->hdr.lr_cap.bk_bytes) * SECTOR_LEN;

    // Sanity check. Make sure there IS a bookmark.
    if (buf_size == 0)
    {
        // This isn't wrong, but it is strange. I assume we're trying to read a bookmark table
        // because we think there IS a bookmark table.
        fi_log(LOG_DEBUG, "%s: bookmark size is 0.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Read the bookmark information from the drive.
    msg->bk_buffer.resize(buf_size, 0);
    if (!readSectors(start_sector, msg->bk_buffer.data(), BYTES_TO_SECTORS(buf_size)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the bookmark information.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Shrink down the vector to just the size required (not SECTOR aligned).
    msg->bk_buffer.resize(cap_data->hdr.lr_cap.bk_bytes);

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleGetCorrectionTableRequest(fsGetCorrectionTableMessage *msg)
{
    bool header_found(false);
    uint32_t buf_size;
    uint64_t start_sector;
    CT_GT_HDR *hdr_ptr;
    CAPTURE_DATA *cap_data;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
        return;
    }
    setDrivePath(msg->dev_path);

    // Get our cached capture data.
    cap_data = cacheFindCapture(msg->dev_path, msg->capture_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: capture (%" PRIu64 ") not found.", __PRETTY_FUNCTION__, msg->capture_lba);
        goto bail;
    }

    // Verify that the capture header is of a type we would expect.
    if (cap_data->type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: capture header type ('%s' [0x%08x]) is not valid!", __PRETTY_FUNCTION__, signatureToString(cap_data->type), cap_data->type);
        goto bail;
    }

    // Find the start sector for the correction table (or the correction table header). Convert
    // the capture LBA to bytes so we can add in the ct_offset value. Then convert everything
    // back to a sector (LBA) value.
    start_sector = BYTES_TO_SECTORS((msg->capture_lba * SECTOR_LEN) + cap_data->hdr.lr_cap.ct_offset);

    // Calculate the size to read the correction information, assuming an old style CT. Make
    // sure the length is on a sector boundary. We'll re-adjust as necessary later.
    buf_size = BYTES_TO_SECTORS(cap_data->hdr.lr_cap.ct_bytes) * SECTOR_LEN;

    // Sanity check. Make sure there IS a correction table.
    if (buf_size == 0)
    {
        // This isn't wrong, but it is strange. I assume we're trying to read a correction table
        // because we think there IS a correction table.
        fi_log(LOG_DEBUG, "%s: correction table size is 0.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::OK;
        goto bail;
    }

    // Read a single sector to see if it contains a CT header, if not then this is pre-CT
    // header times, and we're only looking at a correction table.
    if (!readSectors(start_sector, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the correction table.", __PRETTY_FUNCTION__);
        goto bail;
    }
    hdr_ptr = (CT_GT_HDR *)sector_buf.get();

    // Do we a header, or is it an old style CT?
    if (hdr_ptr->magic == CT_GT_HDR_ID)
    {
        // Typically, the CT will follow immediately after the header, so the ct_offset SHOULD be
        // one sector. Of course, this might change in the future.
        if (hdr_ptr->ct_offset == SECTOR_LEN)
        {
            uint32_t ct_gt_len;

            // Calculate the lengths and make sure we have a match for what we think it should be.
            // The tables both reside on a sector boundary, the GT immediately following the CT which
            // immediately follows the header.
            ct_gt_len = SECTORS_TO_BYTES(1 + BYTES_TO_SECTORS(hdr_ptr->ct_bytes) + BYTES_TO_SECTORS(hdr_ptr->gt_bytes));
            if (ct_gt_len == buf_size)
            {
                // We can now be pretty confident that we have a header followed by data, and we
                // are not looking at an old style CT.
                header_found = true;
            }
        }
    }

    // If we have a header, adjust information so we read the actual correction table.
    if (header_found)
    {
        // The correction table is offset from the header.
        start_sector += BYTES_TO_SECTORS(hdr_ptr->ct_offset);

        // Utilize the length in the header.
        buf_size = BYTES_TO_SECTORS(hdr_ptr->ct_bytes) * SECTOR_LEN;
    }

    // Read the correction table from the drive.
    msg->ct_buffer.resize(buf_size, 0);
    if (!readSectors(start_sector, msg->ct_buffer.data(), BYTES_TO_SECTORS(buf_size)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the correction table.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Shrink down the vector to just the size required (not SECTOR aligned).
    if (header_found)
        msg->ct_buffer.resize(hdr_ptr->ct_bytes);
    else
        msg->ct_buffer.resize(cap_data->hdr.lr_cap.ct_bytes);

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleGetGainTableRequest(fsGetGainTableMessage *msg)
{
    bool header_found(false);
    uint32_t buf_size;
    uint64_t start_sector;
    CT_GT_HDR *hdr_ptr;
    CAPTURE_DATA *cap_data;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
        return;
    }
    setDrivePath(msg->dev_path);

    // Get our cached capture data.
    cap_data = cacheFindCapture(msg->dev_path, msg->capture_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: capture (%" PRIu64 ") not found.", __PRETTY_FUNCTION__, msg->capture_lba);
        goto bail;
    }

    // Verify that the capture header is of a type we would expect.
    if (cap_data->type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: capture header type ('%s' [0x%08x]) is not valid!", __PRETTY_FUNCTION__, signatureToString(cap_data->type), cap_data->type);
        goto bail;
    }

    // Find the start sector for the correction/gain table header. Convert the capture LBA to bytes so
    // we can add in the ct_offset value. Then convert everything back to a sector (LBA) value.
    start_sector = BYTES_TO_SECTORS((msg->capture_lba * SECTOR_LEN) + cap_data->hdr.lr_cap.ct_offset);

    // Calculate the total size of the correction and gain table information. Make
    // sure the length is on a sector boundary. We'll re-adjust as necessary later.
    buf_size = BYTES_TO_SECTORS(cap_data->hdr.lr_cap.ct_bytes) * SECTOR_LEN;

    // Sanity check. Make sure there IS a correction/gain table header.
    if (buf_size == 0)
    {
        // This isn't wrong, but it is strange. I assume we're trying to read a gain table
        // because we think there IS a gain table.
        fi_log(LOG_DEBUG, "%s: correction/gain table size is 0.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::OK;
        goto bail;
    }

    // Read a single sector to see if it contains a CT header, if not then this is pre-CT
    // header times, and we're only looking at a correction table.
    if (!readSectors(start_sector, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the correction/gain header.", __PRETTY_FUNCTION__);
        goto bail;
    }
    hdr_ptr = (CT_GT_HDR *)sector_buf.get();

    // Do we a header, or is it an old style CT?
    if (hdr_ptr->magic == CT_GT_HDR_ID)
    {
        // Typically, the CT will follow immediately after the header, so the ct_offset SHOULD be
        // one sector. Of course, this might change in the future.
        if (hdr_ptr->ct_offset == SECTOR_LEN)
        {
            uint32_t ct_gt_len;

            // Calculate the lengths and make sure we have a match for what we think it should be.
            // The tables both reside on a sector boundary, the GT immediately following the CT which
            // immediately follows the header.
            ct_gt_len = SECTORS_TO_BYTES(1 + BYTES_TO_SECTORS(hdr_ptr->ct_bytes) + BYTES_TO_SECTORS(hdr_ptr->gt_bytes));
            if (ct_gt_len == buf_size)
            {
                // We can now be pretty confident that we have a header followed by data, and we
                // are not looking at an old style CT.
                header_found = true;
            }
        }
    }

    // If we have don't have a header, then we don't have a gain table.
    if (!header_found)
    {
        fi_log(LOG_INFO, "%s: No CT_GT_INFO header found, so no gain table!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // The gain table is offset from the header.
    start_sector += BYTES_TO_SECTORS(hdr_ptr->gt_offset);

    // Utilize the length in the header.
    buf_size = BYTES_TO_SECTORS(hdr_ptr->gt_bytes) * SECTOR_LEN;

    // Read the gain table from the drive.
    msg->gt_buffer.resize(buf_size, 0);
    if (!readSectors(start_sector, msg->gt_buffer.data(), BYTES_TO_SECTORS(buf_size)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the gain table.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Shrink down the vector to just the size required (not SECTOR aligned).
    msg->gt_buffer.resize(hdr_ptr->gt_bytes);

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleGetInfoRequest(fsGetInfoMessage *msg)
{
    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
    }
    else
    {
        DRIVE_HD *drv;

        // Set the drive path.
        setDrivePath(msg->dev_path);

        // Read the drive header.
        if (!readSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
        {
            fi_log(LOG_DEBUG, "%s: Unable to read the drive header.", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Verify that we have a Fastec Filesystem drive.
        if (!verifyDriveHeader((FI_SIG *)sector_buf.get()))
        {
            fi_log(LOG_DEBUG, "%s: drive header verify failed", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Get our pointer to the drive header. Save the number of sectors allocated
        // to the filesystem in our cache entry.
        drv = (DRIVE_HD *)(sector_buf.get() + sizeof(FI_SIG));
        msg->assigned_space = drv->drive_size;
        msg->name.assign((const char *)drv->name);

        // Try to locate the information in the cache.
        std::map<std::string, FS_CACHE_ENTRY>::iterator it;
        it = cache.find(dev_path);
        if (it != cache.end())
        {
            // Found it. Copy the partition information over.
            msg->part_info = it->second.part_info;
            msg->resp = FAS_THREAD::OK;
        }
        else
        {
            // Couldn't find the device.
            TP(FS_MESSAGE_NOISY, "%s: device [%s] not found", __PRETTY_FUNCTION__, msg->dev_path.c_str());
        }
    }

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleGetPmMetadataRequest(fsGetPmMetadataMessage *msg)
{
    uint32_t buf_size;
    uint64_t start_sector;
    CAPTURE_DATA *cap_data;
    pm_metadata *metadata_ptr;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
        return;
    }
    setDrivePath(msg->dev_path);

    // Get our cached capture data.
    cap_data = cacheFindCapture(msg->dev_path, msg->capture_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: capture (%" PRIu64 ") not found.", __PRETTY_FUNCTION__, msg->capture_lba);
        goto bail;
    }

    // Verify that the capture header is of a type we would expect.
    if (cap_data->type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: capture header type ('%s' [0x%08x]) is not valid for pm_metadata!", __PRETTY_FUNCTION__, signatureToString(cap_data->type), cap_data->type);
        goto bail;
    }

    // Allocate enough buffer to read the PM.
    start_sector = BYTES_TO_SECTORS((msg->capture_lba * SECTOR_LEN) + cap_data->hdr.lr_cap.pm_offset);
    buf_size = BYTES_TO_SECTORS(cap_data->hdr.lr_cap.pm_bytes) * SECTOR_LEN;

    // Sanity check. Make sure there IS a pm_metadata structure.
    if (buf_size == 0)
    {
        fi_log(LOG_DEBUG, "%s: pm_bytes is 0.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Read the metadata from the drive.
    msg->metadata.resize(buf_size, 0);
    if (!readSectors(start_sector, msg->metadata.data(), BYTES_TO_SECTORS(buf_size)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the pm_metadata.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Shrink down the vector to just the size required (not SECTOR aligned).
    msg->metadata.resize(cap_data->hdr.lr_cap.pm_bytes);

    // Overlay the buffer with our pm_metadata structure.
    metadata_ptr = (pm_metadata *)msg->metadata.data();

    // Now fill in the fields that are were updated post-capture in the capture header
    // but not re-written in the PM.
    metadata_ptr->frame_count = cap_data->hdr.lr_cap.num_frames;    // Number captured, not allocated.
    metadata_ptr->first_frame = cap_data->hdr.lr_cap.first_frame_num;
    metadata_ptr->trig_time_s = cap_data->hdr.lr_cap.trigger_time_s;
    metadata_ptr->trig_time_ns = cap_data->hdr.lr_cap.trigger_time_ns;

    // Everywhere except pm_metadata we maintain the trigger frame as relative to
    // the first frame number. But in pm_metadata it is an absolute frame number
    // so we need to convert from relative to absolute.
    if ((cap_data->hdr.lr_cap.trigger_frame_num + cap_data->hdr.lr_cap.first_frame_num) < cap_data->hdr.lr_cap.frame_count)
        metadata_ptr->trig_frame = cap_data->hdr.lr_cap.trigger_frame_num + cap_data->hdr.lr_cap.first_frame_num;
    else
        metadata_ptr->trig_frame = (cap_data->hdr.lr_cap.trigger_frame_num + cap_data->hdr.lr_cap.first_frame_num) - cap_data->hdr.lr_cap.frame_count;

    fi_log(LOG_DEBUG, "%s: CAP(%u, %u, %u) PM(%u, %u, %u)", __PRETTY_FUNCTION__,
           cap_data->hdr.lr_cap.frame_count, cap_data->hdr.lr_cap.first_frame_num, cap_data->hdr.lr_cap.trigger_frame_num,
           metadata_ptr->frame_count, metadata_ptr->first_frame, metadata_ptr->trig_frame);

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

/*!
 * @brief Process the layout request message.
 *
 * The layout request message is used to establish the initial layout for the SSD.
 *
 * @param msg   Pointer to the layout message. The layout message contains all
 *              the parameters required to perform the operation.
 */
void fsWorker::handleLayoutRequest(fsLayoutMessage *msg)
{
    DRIVE_HD drv;
    FS_CACHE_ENTRY entry;
    std::map<std::string, FS_CACHE_ENTRY>::iterator it;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Build up our layout. This has the side effect of setting up the drive path.
    if (!buildLayout(msg->dev_path, msg->num_sectors, msg->layout, drv, entry.part_info))
    {
        fi_log(LOG_DEBUG, "%s: buildLayout() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we set up our path before writing.
    setDrivePath(msg->dev_path);

    // Save the drive name.
    std::strncpy((char *)drv.name, msg->name.c_str(), MAX_DRV_NAME_LEN);

    // Write everrything to disk.
    if (!writeLayout(&drv, entry.part_info))
    {
        fi_log(LOG_DEBUG, "%s: writeLayout() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Add our information to the cache.
    entry.drive_size = drv.drive_size;

    // It is a valid Fastec filesystem, so add information to cache.
    // Make sure we don't already have an entry in the cache for this device.
    // If we do, remove it so we don't create a duplicate.
    it = cache.find(msg->dev_path);
    if (it != cache.end())
    {
        // Remove the existing entry.
        TP(FS_MESSAGE_NOISY, "%s: Found an existing entry. Removing.", __PRETTY_FUNCTION__);
        cache.erase(it);
    }

    // Save the information to the cache.
    TP(FS_MESSAGE_NOISY, "%s: Inserting part_info for [%s]", __PRETTY_FUNCTION__, msg->dev_path.c_str());
    cache.insert(std::make_pair(msg->dev_path, entry));

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

/*!
 * @brief Process the layout request message.
 *
 * The layout request message is used to establish the initial layout for the SSD.
 *
 * @param msg   Pointer to the layout message. The layout message contains all
 *              the parameters required to perform the operation.
 */
void fsWorker::handleRenameRequest(fsRenameMessage *msg)
{
    DRIVE_HD *drive_hdr;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Make sure we're going to the right drive.
    setDrivePath(msg->dev_path);

    // Read the drive header.
    if (!readSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the drive header.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Verify that we have a Fastec Filesystem drive.
    if (!verifyDriveHeader((FI_SIG *)sector_buf.get()))
    {
        // No Fastec Filesystem here.
        fi_log(LOG_DEBUG, "%s: drive [%s] is not a Fastec Filesystem.", __PRETTY_FUNCTION__, dev_path.c_str());
        goto bail;
    }

    // Get our pointer to the drive header.
    drive_hdr = (DRIVE_HD *)(sector_buf.get() + sizeof(FI_SIG));

    // Clear out the old drive name and save the new.
    std::memset(drive_hdr->name, 0, MAX_DRV_NAME_LEN + 1);
    std::strncpy((char *)drive_hdr->name, msg->name.c_str(), MAX_DRV_NAME_LEN);

    // Update our checksum.
    setBlockChecksum((FI_SIG *)sector_buf.get());

    // Write everrything to disk.
    if (!writeSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSectors(DRIVE_HD) failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleResizeCaptureRequest(fsResizeCaptureMessage *msg)
{
    int part_index;
    FI_SIG *sig;
    CAPTURE_UNION *cap_ptr;
    CAPTURE_DATA *cap_data;
    PARTITION_INFO *part_info;

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    TP(FS_MESSAGE, "%s: [%s] lba:%" PRIu64 ", resize:%" PRId64, __PRETTY_FUNCTION__, msg->dev_path.c_str(), msg->cap_lba, msg->resize_amount);

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        return;
    }

    // Remember some of the information we're working with.
    setDrivePath(msg->dev_path);

    // See if we can find the entries in the cache related to this capture. We will
    // need the partition containing this capture and it would be useful to have
    // the cache entry for this capture as well.
    cap_data = cacheFindCapture(msg->dev_path, msg->cap_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Unable to find cache entry for cap_lba:%" PRIu64, __PRETTY_FUNCTION__, msg->cap_lba);
        goto bail;
    }
    part_info = cacheGetPartition(msg->dev_path, cap_data->hdr.link.dpart_lba, part_index);
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Cache entry for '%s', part_lba:%" PRIu64 " not found", __PRETTY_FUNCTION__, msg->dev_path.c_str(), cap_data->hdr.link.dpart_lba);
        goto bail;
    }

    // Read and verify the capture header.
    if (!readSectors(msg->cap_lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, msg->cap_lba);
        goto bail;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();
    if (!verifyCaptureHeader(sig, part_info->part.type))
    {
        fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, msg->cap_lba);
        goto bail;
    }

    // Overlay our capture header.
    cap_ptr = (CAPTURE_UNION *)(sector_buf.get() + sizeof(FI_SIG));

    // What sort of a capture are we trying to resize?
    switch (sig->signature)
    {
    case LR_CAPTURE_SIG:
        // Resize the capture header.
        if (!resizeLrCapture(cap_ptr->lr_cap, msg->resize_amount, part_info))
        {
            fi_log(LOG_DEBUG, "%s: resizeLrCapture() failed!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update our cache values.
        memcpy(&cap_data->hdr.lr_cap, &cap_ptr->lr_cap, sizeof(LR_CAPTURE_HDR));
        break;
    case STOR_CAPTURE_SIG:
        // Resize the capture header.
        if (!resizeStorCapture(cap_ptr->stor_cap, msg->resize_amount, part_info))
        {
            fi_log(LOG_DEBUG, "%s: resizeStorCapture() failed!", __PRETTY_FUNCTION__);
            goto bail;
        }

        // Update our cache values.
        memcpy(&cap_data->hdr.stor_cap, &cap_ptr->stor_cap, sizeof(STOR_CAPTURE_HDR));
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected signature [0x%x]", __PRETTY_FUNCTION__, sig->signature);
        goto bail;
    }

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the updated capture header to the drive.
    if (!writeSectors(msg->cap_lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleResizePartitionRequest(fsResizePartitionMessage *msg)
{
    int part_index;
    PARTITION_INFO *part_info;

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    TP(FS_MESSAGE, "%s: [%s] lba:%" PRIu64 ", resize:%" PRIu64, __PRETTY_FUNCTION__, msg->dev_path.c_str(), msg->part_lba, msg->resize_amount);

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have our buffer. It should have been allocated
    // during the creation of the object.
    if (sector_buf.get() == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: sector_buf is not allocated!", __PRETTY_FUNCTION__);
        return;
    }

    // Remember some of the information we're working with.
    setDrivePath(msg->dev_path);

    // See if we can find the entries in the cache for this partition.
    part_info = cacheGetPartition(msg->dev_path, msg->part_lba, part_index);
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Cache entry for '%s', part_lba:%" PRIu64 " not found", __PRETTY_FUNCTION__, msg->dev_path.c_str(), msg->part_lba);
        goto bail;
    }

    // For now we need to enforce the rule that only the last partition in the drive can be resized.
    if (part_index != (int)(cache[msg->dev_path].part_info.size() - 1))
    {
        fi_log(LOG_DEBUG, "%s: Only the last partition in the drive can be modified!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Set the number of sectors allocated from this drive.
    setNumSectors(cache[msg->dev_path].drive_size);

    // What sort of a partition are we trying to resize?
    switch (part_info->part.type)
    {
    case AVAIL_PART_SIG:
        // Update the Long Record partition header.
        if (!resizeAvailPartition(part_info, msg->resize_amount))
        {
            fi_log(LOG_DEBUG, "%s: resizeAvailPartition() failed!", __PRETTY_FUNCTION__);
            goto bail;
        }
        break;
    case LR_PART_SIG:
        // Update the Long Record partition header.
        if (!resizeLrPartition(part_info, msg->resize_amount))
        {
            fi_log(LOG_DEBUG, "%s: resizeLrPartition() failed!", __PRETTY_FUNCTION__);
            goto bail;
        }
        break;
    case STOR_PART_SIG:
        // Update the Long Record partition header.
        if (!resizeStorPartition(part_info, msg->resize_amount))
        {
            fi_log(LOG_DEBUG, "%s: resizeStorPartition() failed!", __PRETTY_FUNCTION__);
            goto bail;
        }
        goto bail;
    default:
        fi_log(LOG_DEBUG, "%s: Unexpected signature [0x%x]", __PRETTY_FUNCTION__, part_info->part.type);
        goto bail;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleSetBookmarkRequest(fsSetBookmarkMessage *msg)
{
    uint32_t buf_size;
    uint64_t start_sector;
    FI_SIG *sig;
    CAPTURE_DATA *cap_data;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
        return;
    }
    setDrivePath(msg->dev_path);

    // Get our cached capture data.
    cap_data = cacheFindCapture(msg->dev_path, msg->capture_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: capture (%" PRIu64 ") not found.", __PRETTY_FUNCTION__, msg->capture_lba);
        goto bail;
    }

    // Verify that the capture header is of a type we would expect.
    if (cap_data->type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: capture header type ('%s' [0x%08x]) is not valid!", __PRETTY_FUNCTION__, signatureToString(cap_data->type), cap_data->type);
        goto bail;
    }

    // Allocate enough buffer to contain all the data.
    buf_size = BYTES_TO_SECTORS(msg->bk_buffer.size()) * SECTOR_LEN;

    // If the incoming bookmark table is empty, and the existing bookmark table is empty,
    // then we are done.
    if ((buf_size == 0) && (cap_data->hdr.lr_cap.bk_bytes == 0))
    {
        TP(FS_WORKER_FUNC_NOISY, "%s: Both incoming and existing are empty.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::OK;
        goto bail;
    }

    // Make sure the incoming bookmarks will fit.
    if (buf_size > CAPTURE_BOOKMARK_SIZE)
    {
        fi_log(LOG_DEBUG, "%s: Bookmark table is too big!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Figure out where the bookmark data goes.
    start_sector = cap_data->lba + BYTES_TO_SECTORS(cap_data->hdr.lr_cap.bk_offset);

    // Read the bookmark information from the drive.
    msg->bk_buffer.resize(BYTES_TO_SECTORS(CAPTURE_BOOKMARK_SIZE) * SECTOR_LEN, 0);
    if (!writeSectors(start_sector, msg->bk_buffer.data(), BYTES_TO_SECTORS(CAPTURE_BOOKMARK_SIZE)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the bookmark information.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Since we have changed the size of the bookmark information, we need
    // to update the capture header. First we will update our cache entry.
    cap_data->hdr.lr_cap.bk_bytes = msg->bk_buffer.size();

    // Reset our sector-aligned buffer memory.
    memset(sector_buf.get(), 0, SECTOR_LEN);

    // Copy the capture header from the cache into the memory.
    memcpy((sector_buf.get() + sizeof(FI_SIG)), &cap_data->hdr.lr_cap, sizeof(LR_CAPTURE_HDR));

    // Build up the signature for the capture header.
    sig = (FI_SIG *)sector_buf.get();
    sig->signature = cap_data->type;
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(cap_data->lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleSetCorrectionInfoRequest(fsSetCorrectionInfoMessage *msg)
{
    uint32_t old_size;
    uint32_t buf_size;
    FI_SIG *sig;
    CAPTURE_DATA *cap_data;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
        return;
    }
    setDrivePath(msg->dev_path);

    // Get our cached capture data.
    cap_data = cacheFindCapture(msg->dev_path, msg->capture_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: capture (%" PRIu64 ") not found.", __PRETTY_FUNCTION__, msg->capture_lba);
        goto bail;
    }

    // Verify that the capture header is of a type we would expect.
    if (cap_data->type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: capture header type ('%s' [0x%08x]) is not valid!", __PRETTY_FUNCTION__, signatureToString(cap_data->type), cap_data->type);
        goto bail;
    }

    // Write out the correction and gain table information. The returned valued
    // is the total number of bytes used by the information.
    buf_size = writeCorrectionInfo(cap_data, msg);

    // If the incoming correction table is empty, and the existing correction table is empty,
    // then we are done.
    if ((buf_size == 0) && (cap_data->hdr.lr_cap.ct_bytes == 0))
    {
        TP(FS_WORKER_FUNC_NOISY, "%s: Both incoming and existing are empty.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::OK;
        goto bail;
    }

    // Since we have changed the size of the correction/gain table information, we need
    // to update the capture header. First we will update our cache entry and change
    // the size of the correction/gain table information.
    old_size = cap_data->hdr.lr_cap.ct_bytes;
    cap_data->hdr.lr_cap.ct_bytes = buf_size;

    // The correction table uses "image" data, so we need to adjust those fields as well.
    cap_data->hdr.lr_cap.image_lba = SECTORS_TO_PAGES(cap_data->lba + FiFilesystem::getReserved(LR_CAPTURE_SIG) + BYTES_TO_SECTORS(cap_data->hdr.lr_cap.ct_bytes)) * SECTORS_PER_PAGE;

    // If the correction table size grew, then the amount of image data shrank by that amount.
    // Otherwise, the correction table size shrank, so the image data space increased.
    if (old_size < cap_data->hdr.lr_cap.ct_bytes)
        cap_data->hdr.lr_cap.image_sectors -= BYTES_TO_SECTORS(cap_data->hdr.lr_cap.ct_bytes - old_size);
    else
        cap_data->hdr.lr_cap.image_sectors += BYTES_TO_SECTORS(old_size - cap_data->hdr.lr_cap.ct_bytes);

    // Reset our sector-aligned buffer memory.
    memset(sector_buf.get(), 0, SECTOR_LEN);

    // Copy the capture header from the cache into the memory.
    memcpy((sector_buf.get() + sizeof(FI_SIG)), &cap_data->hdr.lr_cap, sizeof(LR_CAPTURE_HDR));

    // Build up the signature for the capture header.
    sig = (FI_SIG *)sector_buf.get();
    sig->signature = cap_data->type;
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(cap_data->lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleSetPmMetadataRequest(fsSetPmMetadataMessage *msg)
{
    uint32_t buf_size;
    uint64_t start_sector;
    FI_SIG *sig;
    CAPTURE_DATA *cap_data;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we'll fail.
    msg->resp = FAS_THREAD::FAULT;

    // Make sure we have a device path.
    if (msg->dev_path.empty())
    {
        // We need the device information!
        fi_log(LOG_DEBUG, "%s: dev_path is empty!", __PRETTY_FUNCTION__);
        return;
    }
    setDrivePath(msg->dev_path);

    // Get our cached capture data.
    cap_data = cacheFindCapture(msg->dev_path, msg->capture_lba);
    if (cap_data == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: capture (%" PRIu64 ") not found.", __PRETTY_FUNCTION__, msg->capture_lba);
        goto bail;
    }

    // Verify that the capture header is of a type we would expect.
    if (cap_data->type != LR_CAPTURE_SIG)
    {
        fi_log(LOG_DEBUG, "%s: capture header type ('%s' [0x%08x]) is not valid!", __PRETTY_FUNCTION__, signatureToString(cap_data->type), cap_data->type);
        goto bail;
    }

    // Allocate enough buffer to contain all the data.
    buf_size = BYTES_TO_SECTORS(msg->metadata.size()) * SECTOR_LEN;

    // If the incoming metadata table is empty, and the existing metadata table is empty,
    // then we are done.
    if ((buf_size == 0) && (cap_data->hdr.lr_cap.pm_bytes == 0))
    {
        TP(FS_WORKER_FUNC_NOISY, "%s: Both incoming and existing are empty.", __PRETTY_FUNCTION__);
        msg->resp = FAS_THREAD::OK;
        goto bail;
    }

    // Make sure the incoming metadata will fit.
    if (buf_size > CAPTURE_METADATA_SIZE)
    {
        fi_log(LOG_DEBUG, "%s: Metadata table is too big!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Figure out where the bookmark data goes.
    start_sector = cap_data->lba + BYTES_TO_SECTORS(cap_data->hdr.lr_cap.pm_offset);

    // Read the bookmark information from the drive.
    msg->metadata.resize(BYTES_TO_SECTORS(CAPTURE_METADATA_SIZE) * SECTOR_LEN, 0);
    if (!writeSectors(start_sector, msg->metadata.data(), BYTES_TO_SECTORS(CAPTURE_METADATA_SIZE)))
    {
        fi_log(LOG_DEBUG, "%s: Unable to read the bookmark information.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Since we have changed the size of the metadata information, we need
    // to update the capture header. First we will update our cache entry.
    cap_data->hdr.lr_cap.pm_bytes = msg->metadata.size();

    // Reset our sector-aligned buffer memory.
    memset(sector_buf.get(), 0, SECTOR_LEN);

    // Copy the capture header from the cache into the memory.
    memcpy((sector_buf.get() + sizeof(FI_SIG)), &cap_data->hdr.lr_cap, sizeof(LR_CAPTURE_HDR));

    // Build up the signature for the capture header.
    sig = (FI_SIG *)sector_buf.get();
    sig->signature = cap_data->type;
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(cap_data->lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

/*!
 * @brief Process the layout request message without actually writing anything
 *        to the drive.
 *
 * The "try layout" request message is used to prototype the initial layout for
 * the drive without actually writing anything to the drive.
 *
 * @param msg   Pointer to the layout message. The layout message contains all
 *              the parameters required to perform the operation.
 */
void fsWorker::handleTryLayout(fsTryLayoutMessage *msg)
{
    DRIVE_HD drv;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Build up our layout. This has the side effect of setting up the drive path.
    if (!buildLayout(msg->dev_path, msg->num_sectors, msg->layout, drv, msg->part_info))
    {
        fi_log(LOG_DEBUG, "%s: buildLayout() failed!", __PRETTY_FUNCTION__);
    }
    else
    {
        // All done.
        msg->resp = FAS_THREAD::OK;
    }

    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleUpdateCapture(fsUpdateCaptureMessage *msg)
{
    FI_SIG *sig;
    CAPTURE_DATA *cache_entry;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Remember some of the information we're working with.
    setDrivePath(msg->dev_path);

    // Read and verify the capture header.
    if (!readSectors(msg->capture_info.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: partition header @%" PRIu64 " read failed", __PRETTY_FUNCTION__, msg->capture_info.lba);
        goto bail;
    }

    // Overlay our signature header.
    sig = (FI_SIG *)sector_buf.get();

    // What capture are we trying to update.
    switch (msg->capture_info.type)
    {
    case LR_CAPTURE_SIG:
        // Verify our capture header.
        if (!verifyCaptureHeader(sig, LR_PART_SIG))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, msg->capture_info.lba);
            goto bail;
        }
        else
        {
            LR_CAPTURE_HDR *lr_cap_ptr;

            // Overlay our capture header.
            lr_cap_ptr = (LR_CAPTURE_HDR *)(sector_buf.get() + sizeof(FI_SIG));

            // Update the contents.
            memcpy(lr_cap_ptr, &msg->capture_info.hdr.lr_cap, sizeof(LR_CAPTURE_HDR));
        }
        break;
    case STOR_CAPTURE_SIG:
        // Verify our capture header.
        if (!verifyCaptureHeader(sig, STOR_PART_SIG))
        {
            fi_log(LOG_DEBUG, "%s: capture header @%" PRIu64 " verify failed", __PRETTY_FUNCTION__, msg->capture_info.lba);
            goto bail;
        }
        else
        {
            STOR_CAPTURE_HDR *stor_cap_ptr;

            // Overlay our capture header.
            stor_cap_ptr = (STOR_CAPTURE_HDR *)(sector_buf.get() + sizeof(FI_SIG));

            // Update the contents.
            memcpy(stor_cap_ptr, &msg->capture_info.hdr.stor_cap, sizeof(STOR_CAPTURE_HDR));
        }
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Capture type 0x%08x [%s] not supported", __PRETTY_FUNCTION__, msg->capture_info.lba, signatureToString(msg->capture_info.lba));
        goto bail;
    }

    // Update our checksum.
    setBlockChecksum(sig);

    // Write the capture header to the drive.
    if (!writeSectors(msg->capture_info.lba, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: writeSector() failed", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Do we know about this in our cache?
    cache_entry = cacheFindCapture(msg->dev_path, msg->capture_info.lba);
    if (cache_entry != nullptr)
    {
        // Update the cache.
        *cache_entry = msg->capture_info;
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

bail:
    // Close the drive handle.
    setDrivePath(std::string());
}

void fsWorker::handleWritePartition(fsWritePartitionMessage *msg)
{
    DRIVE_HD drv;
    FS_CACHE_ENTRY entry;

    TP(FS_MESSAGE, "%s", __PRETTY_FUNCTION__);

    // Sanity check.
    if (msg == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Null msg pointer!", __PRETTY_FUNCTION__);
        return;
    }

    // Assume we will fail.
    msg->resp = FAS_THREAD::FAULT;

    // Initialize our settings.
    setDrivePath(msg->dev_path);
    setNumSectors(msg->num_sectors);

    // Add our information to the cache.
    entry.drive_size = msg->num_sectors;

    // It is a valid Fastec filesystem, so add information to cache.
    // Make sure we don't already have an entry in the cache for this device.
    // If we do, remove it so we don't create a duplicate.
    std::map<std::string, FS_CACHE_ENTRY>::iterator it;
    it = cache.find(msg->dev_path);
    if (it != cache.end())
    {
        // Remove the existing entry.
        TP(FS_MESSAGE_NOISY, "%s: Found an existing entry. Removing.", __PRETTY_FUNCTION__);
        cache.erase(it);
    }

    // Is there any information?
    if (msg->part_info.size() > 0)
    {
        // Write everrything to disk.
        if (!writeLayout(&drv, entry.part_info))
        {
            fi_log(LOG_DEBUG, "%s: writeLayout() failed!", __PRETTY_FUNCTION__);

            // Close the drive handle.
            setDrivePath(std::string());
            return;
        }

        // Save the information to the cache.
        TP(FS_MESSAGE_NOISY, "%s: Inserting part_info for [%s]", __PRETTY_FUNCTION__, msg->dev_path.c_str());
        cache.insert(std::make_pair(msg->dev_path, entry));
    }

    // All done.
    msg->resp = FAS_THREAD::OK;

    // Close the drive handle.
    setDrivePath(std::string());
}

/*!
 * @brief This builds all the headers associated with an available partition. The partition headers are all written to the drive.
 *
 * @param part_info     [IN] The partition layout information.
 * @param part_lba      [IN] Where the partition header is located.
 * @param part_sectors  [IN] The number of sectors allocated to this partition.
 * @param cache_info    [OUT] The cache information we're building for the partition.
 * @return True if the layout was succesfull, false otherwise.
 */
bool fsWorker::buildAvailLayout(const LAYOUT_PARTITION *part_info, uint64_t part_lba, uint64_t part_sectors, PARTITION_INFO& cache_info)
{
    bool rc = false;

    TP(FS_WORKER_FUNC, "%s: part_lba=0x%" PRIx64 ", part_sectors=%" PRIu64, __PRETTY_FUNCTION__, part_lba, part_sectors);

    // Sanity check.
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: null part_info pointer.", __PRETTY_FUNCTION__);
        goto bail;
    }
    if (part_info->signature != AVAIL_PART_SIG)
    {
        fi_log(LOG_DEBUG, "%s: part_info is not for an AVAIL partition.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Create our basic partition header.
    createAvailPartHdr(&cache_info.part.hdr.avail_part, part_lba, part_sectors);
    cache_info.part.type = AVAIL_PART_SIG;
    cache_info.part.lba = part_lba;

    // Everything is looking good.
    rc = true;

bail:
    // Return the result.
    return rc;
}

/*!
 * @brief This builds all the headers associated with an LR partition, including the capture
 *        headers. The partition and capture headers are all written to the drive.
 *
 * @param part_info     [IN] The partition layout information.
 * @param part_lba      [IN] Where the partition header is located.
 * @param part_sectors  [IN] The number of sectors allocated to this partition.
 * @param cache_info    [OUT] The cache information we're building for the partition.
 * @return True if the layout was succesfull, false otherwise.
 */
bool fsWorker::buildLrLayout(const LAYOUT_PARTITION *part_info, uint64_t part_lba, uint64_t part_sectors, PARTITION_INFO& cache_info)
{
    bool rc = false;
    std::size_t cap_count;
    uint64_t cap_lba;
    uint64_t fmeta_lba;
    uint64_t fmeta_sectors;
    uint64_t prev_cap_lba;
    uint64_t part_end;

    TP(FS_WORKER_FUNC, "%s: part_lba=0x%" PRIx64 ", part_sectors=%" PRIu64, __PRETTY_FUNCTION__, part_lba, part_sectors);

    // Create our basic partition header.
    createLrPartHdr(&cache_info.part.hdr.lr_part, part_lba, part_sectors, part_info->min_frame_size);
    cache_info.part.type = LR_PART_SIG;
    cache_info.part.lba = part_lba;

    // Loop through the information for creating capture headers for this partition.
    prev_cap_lba = 0;
    fmeta_lba = cache_info.part.hdr.lr_part.fmeta_lba;
    fmeta_sectors = 0;
    part_end = cache_info.part.hdr.lr_part.image_lba + cache_info.part.hdr.lr_part.image_sectors;
    cap_lba = cache_info.part.hdr.lr_part.image_lba;

    // Loop through all the capture headers we want to create.
    for (cap_count = 0; cap_count < part_info->cap_info.size(); cap_count++)
    {
        CAPTURE_DATA cap_data;
        uint64_t cap_image_lba;
        uint32_t max_ct_size;
        LAYOUT_CAPTURE cap_info = part_info->cap_info.at(cap_count);

        // Initialize our capture data structure.
        cap_data.type = LR_CAPTURE_SIG;
        cap_data.lba = cap_lba;

        // Let's do this calculation once. The image LBA MUST be page aligned. The max_ct_size is the maximum
        // size of the correction table, be we need to calculate the total size of the correction/gain information.
        // This will be 1 sector for the header, max_ct_size for the correction table and another for the gain table.
        max_ct_size = ((2 * BYTES_TO_SECTORS(cap_info.max_ct_size) + 1)) * SECTOR_LEN;
        cap_image_lba = SECTORS_TO_PAGES(cap_lba + FiFilesystem::getReserved(LR_CAPTURE_SIG) + BYTES_TO_SECTORS(max_ct_size)) * SECTORS_PER_PAGE;

        // Sanity check. Is there enough room? There is a minimum number of frames
        // we need to be able to record, however we don't know what the frame size
        // for the capture is. The best we can do is make sure we have room for the
        // structures we now about right now.
        if (cap_image_lba >= part_end)
        {
            fi_log(LOG_DEBUG, "%s: The capture header [LBA 0x%" PRIx64 "] exceeds the partition!", __PRETTY_FUNCTION__, cap_lba);
            goto bail;
        }

        // Are we supposed to use the remainder?
        if (cap_info.cap_size == 0)
        {
            cap_info.cap_size = part_end - cap_image_lba;
            fi_log(LOG_DEBUG, "%s: giving remainder to cap_size: %" PRIu64, __PRETTY_FUNCTION__, cap_info.cap_size);
        }

        // Make sure we still fit.
        if ((cap_image_lba + cap_info.cap_size) > part_end)
        {
            uint64_t new_size = part_end - cap_image_lba;
            fi_log(LOG_DEBUG, "%s: Adjusting capture %d size, %" PRIu64 " -> %" PRIu64, __PRETTY_FUNCTION__, cap_count, cap_info.cap_size, new_size);
            cap_info.cap_size = new_size;
        }

        // Create our capture header.
        createLrCaptureHdr(&cap_data.hdr.lr_cap, part_lba, prev_cap_lba, cap_lba, cap_info.cap_size, fmeta_lba, fmeta_sectors, max_ct_size);

        // Is this the first capture header? If it is then we need to link the
        // partition header to this capture header. Otherwise we want to link the
        // previous capture header.
        if (cap_count == 0)
            cache_info.part.hdr.lr_part.capturehd_lba = cap_lba;
        else
            cache_info.caps.back().hdr.link.next_lba = cap_lba;

        // Update where the next capture header is located.
        prev_cap_lba = cap_lba;
        cap_lba = cap_data.hdr.lr_cap.image_lba + cap_data.hdr.lr_cap.image_sectors;

        // Push our new capture header onto our list.
        cache_info.caps.push_back(cap_data);
    }

    // Everything is looking good.
    rc = true;

bail:
    // Return the result.
    return rc;
}

/*!
 * @brief This builds all the headers associated with a storage partition, including the capture
 *        headers. The partition and capture headers are all written to the drive.
 *
 * @param part_info     [IN] The partition layout information.
 * @param part_lba      [IN] Where the partition header is located.
 * @param part_sectors  [IN] The number of sectors allocated to this partition.
 * @param cache_info    [OUT] The cache information we're building for the partition.
 * @return True if the layout was succesfull, false otherwise.
 */
bool fsWorker::buildStorLayout(const LAYOUT_PARTITION *part_info, uint64_t part_lba, uint64_t part_sectors, PARTITION_INFO& cache_info)
{
    bool rc = false;
    std::size_t cap_count;
    uint64_t cap_lba;
    uint64_t prev_cap_lba;
    uint64_t part_end;

    TP(FS_WORKER_FUNC, "%s: part_lba=0x%" PRIx64 ", part_sectors=%" PRIu64, __PRETTY_FUNCTION__, part_lba, part_sectors);

    // Create our basic partition header.
    createStorPartHdr(&cache_info.part.hdr.stor_part, part_lba, part_sectors);
    cache_info.part.type = STOR_PART_SIG;
    cache_info.part.lba = part_lba;

    // Loop through the information for creating capture headers for this partition.
    prev_cap_lba = 0;
    part_end = cache_info.part.hdr.stor_part.start_lba + cache_info.part.hdr.stor_part.data_sectors;
    cap_lba = cache_info.part.hdr.stor_part.start_lba;

    // Loop through all the capture headers we want to create.
    for (cap_count = 0; cap_count < part_info->cap_info.size(); cap_count++)
    {
        CAPTURE_DATA cap_data;
        uint64_t cap_data_lba;
        LAYOUT_CAPTURE cap_info = part_info->cap_info.at(cap_count);

        // Initialize our capture data structure.
        cap_data.type = STOR_CAPTURE_SIG;
        cap_data.lba = cap_lba;

        // Let's do this calculation once. The image LBA MUST be page aligned. The max_ct_size is the maximum
        // size of the correction table, be we need to calculate the total size of the correction/gain information.
        // This will be 1 sector for the header, max_ct_size for the correction table and another for the gain table.
        cap_data_lba = SECTORS_TO_PAGES(cap_lba + FiFilesystem::getReserved(STOR_CAPTURE_SIG)) * SECTORS_PER_PAGE;

        // Sanity check. Is there enough room? There is a minimum number of frames
        // we need to be able to record, however we don't know what the frame size
        // for the capture is. The best we can do is make sure we have room for the
        // structures we now about right now.
        if (cap_data_lba >= part_end)
        {
            fi_log(LOG_DEBUG, "%s: The capture header [LBA 0x%" PRIx64 "] exceeds the partition!", __PRETTY_FUNCTION__, cap_lba);
            goto bail;
        }

        // Are we supposed to use the remainder?
        if (cap_info.cap_size == 0)
        {
            cap_info.cap_size = part_end - cap_data_lba;
            fi_log(LOG_DEBUG, "%s: giving remainder to cap_size: %" PRIu64, __PRETTY_FUNCTION__, cap_info.cap_size);
        }

        // Make sure we still fit.
        if ((cap_data_lba + cap_info.cap_size) > part_end)
        {
            uint64_t new_size = part_end - cap_data_lba;
            fi_log(LOG_DEBUG, "%s: Adjusting capture %d size, %" PRIu64 " -> %" PRIu64, __PRETTY_FUNCTION__, cap_count, cap_info.cap_size, new_size);
            cap_info.cap_size = new_size;
        }

        // Create our capture header.
        createStorCaptureHdr(&cap_data.hdr.stor_cap, part_lba, prev_cap_lba, cap_lba, cap_info.cap_size);

        // Is this the first capture header? If it is then we need to link the
        // partition header to this capture header. Otherwise we want to link the
        // previous capture header.
        if (cap_count == 0)
            cache_info.part.hdr.stor_part.capturehd_lba = cap_lba;
        else
            cache_info.caps.back().hdr.link.next_lba = cap_lba;

        // Update where the next capture header is located.
        prev_cap_lba = cap_lba;
        cap_lba = cap_data.hdr.stor_cap.data_lba + cap_data.hdr.stor_cap.data_sectors;

        // Push our new capture header onto our list.
        cache_info.caps.push_back(cap_data);
    }

    // Everything is looking good.
    rc = true;

bail:
    // Return the result.
    return rc;
}


/*!
 * @brief Write the disk layout to the drive.
 *
 * @param drv           [IN] The filesystem drive header structure.
 * @param part_info     [IN] The layout to write to the drive.
 * @return True if we're successful, false otherwise.
 */
bool fsWorker::writeLayout(const DRIVE_HD *drv_ptr, const PARTITION_INFO_VEC& part_info)
{
    std::size_t part_count;
    bool rc;
    FI_SIG *sig;

    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Assume failure.
    rc = false;

    // Are we supposed to write the drive header?
    if (drv_ptr != nullptr)
    {
        // Reset our sector-aligned buffer memory.
        memset(sector_buf.get(), 0, SECTOR_LEN);

        // Copy the drive header into the buffer.
        memcpy((void *)(sector_buf.get() + sizeof(FI_SIG)), (void *)drv_ptr, sizeof(DRIVE_HD));

        // Create the signature for our drive header.
        sig = (FI_SIG *)sector_buf.get();
        sig->signature = DRIVE_SIG;
        sig->size = sizeof(DRIVE_HD);
        setBlockChecksum(sig);

        // Write the drive header.
        if (!writeSectors(DRIVE_HEADER_LBA, sector_buf.get(), 1))
        {
            fi_log(LOG_DEBUG, "%s: writeSectors(DRIVE_HD) failed!", __PRETTY_FUNCTION__);
            goto bail;
        }
    }

    // Loop through each of the partitions and write out the data.
    for (part_count = 0; part_count < part_info.size(); ++part_count)
    {
        std::size_t cap_count;
        const PARTITION_INFO *pi_ptr = &part_info[part_count];

        // Now write out all the capture headers we just built.
        for (cap_count = 0; cap_count < pi_ptr->caps.size(); cap_count++)
        {
            uint64_t cap_lba;

            // Reset our sector-aligned buffer memory.
            memset(sector_buf.get(), 0, SECTOR_LEN);

            // Copy the capture header into the memory.
            switch (pi_ptr->caps[cap_count].type)
            {
            case LR_CAPTURE_SIG:
                memcpy((sector_buf.get() + sizeof(FI_SIG)), &pi_ptr->caps[cap_count].hdr.lr_cap, sizeof(LR_CAPTURE_HDR));
                break;
            case STOR_CAPTURE_SIG:
                memcpy((sector_buf.get() + sizeof(FI_SIG)), &pi_ptr->caps[cap_count].hdr.stor_cap, sizeof(STOR_CAPTURE_HDR));
                break;
            default:
                fi_log(LOG_DEBUG, "%s: Unexpected capture header type: 0x%08x [%s]", __PRETTY_FUNCTION__, pi_ptr->caps[cap_count].type, signatureToString(pi_ptr->caps[cap_count].type));
                goto bail;
            }

            // Build up the signature for the capture header.
            sig = (FI_SIG *)sector_buf.get();
            sig->signature = pi_ptr->caps[cap_count].type;
            setBlockChecksum(sig);

            // Figure out where this header is located on the drive.
            cap_lba = pi_ptr->caps[cap_count].lba;

            // Write the capture header to the drive. Note that we're only writing the
            // one sector, even though the capture STRUCTURE might cover more. This is
            // fine, because the elements which exceed the one sector aren't initialized
            // at this point. Everything that references those "out-of-sector" elements
            // are properly initialized.
            if (!writeSectors(cap_lba, sector_buf.get(), 1))
                goto bail;
        }

        // Now we can write out the partition header.
        // Reset our sector-aligned buffer memory.
        memset(sector_buf.get(), 0, SECTOR_LEN);

        // Copy the partition header into the memory.
        switch (pi_ptr->part.type)
        {
        case AVAIL_PART_SIG:
            memcpy((sector_buf.get() + sizeof(FI_SIG)), &pi_ptr->part.hdr.avail_part, sizeof(AVAIL_PARTITION_HDR));
            break;
        case LR_PART_SIG:
            memcpy((sector_buf.get() + sizeof(FI_SIG)), &pi_ptr->part.hdr.lr_part, sizeof(LR_PARTITION_HDR));
            break;
        case STOR_PART_SIG:
            memcpy((sector_buf.get() + sizeof(FI_SIG)), &pi_ptr->part.hdr.stor_part, sizeof(STOR_PARTITION_HDR));
            break;
        default:
            fi_log(LOG_DEBUG, "%s: Unexpected partition header type: 0x%08x [%s]", __PRETTY_FUNCTION__, pi_ptr->part.type, signatureToString(pi_ptr->part.type));
            goto bail;
        }

        // Build up the signature for the partition.
        sig = (FI_SIG *)sector_buf.get();
        sig->signature = pi_ptr->part.type;
        setBlockChecksum(sig);

        // Write the partition header to the drive.
        if (!writeSectors(pi_ptr->part.lba, sector_buf.get(), 1))
            goto bail;
    }

    // All went well.
    rc = true;

bail:
    return rc;

}

/*!
 * @brief Hunt through all partitions until we find the specified capture.
 *
 * @param dev_path  [IN] The device we're interested in.
 * @param cap_lba   [IN] The capture's LBA, uniquely identifies the capture within the partition.
 * @return Pointer to the found capture, or nullptr if not found.
 */
CAPTURE_DATA *fsWorker::cacheFindCapture(const std::string& dev_path, uint64_t cap_lba)
{
    int cap_index;
    CAPTURE_DATA *cap_ptr;
    PARTITION_INFO_VEC part_vec;
    PARTITION_INFO_VEC::iterator part_it;

    TP(FS_WORKER_FUNC, "%s: dev_path:%s, cache entries:%d cap_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), cache.count(dev_path), cap_lba);

    // Assume we fail.
    cap_ptr = nullptr;

    // Make sure we have information about this drive.
    if (cache.count(dev_path) == 0)
        goto bail;

    // Loop through each of the partitions.
    for (part_it = cache[dev_path].part_info.begin(); part_it != cache[dev_path].part_info.end(); ++part_it)
    {
        CAP_UNION_VEC *cap_vec_ptr;

        // Find the capture.
        cap_vec_ptr = &part_it->caps;
        TP(FS_WORKER_FUNC, "%s: cap_vec_ptr->size:%d", __PRETTY_FUNCTION__, (int)cap_vec_ptr->size());
        for (cap_index = 0; cap_index < (int)cap_vec_ptr->size(); ++cap_index)
        {
            // Is this the one?
            TP(FS_WORKER_FUNC, "%s: [%d] lba=%" PRIu64, __PRETTY_FUNCTION__, cap_index, cap_vec_ptr->at(cap_index).lba);
            if (cap_vec_ptr->at(cap_index).lba == cap_lba)
            {
                cap_ptr = &cap_vec_ptr->at(cap_index);
                goto bail;
            }
        }
    }

bail:
    return cap_ptr;
}

/*!
 * @brief Retrieve the capture information for the drive (dev_path) and partition (part_lba), uniquely
 *        identified by the LBA (cap_lba).
 *
 * @param dev_path  [IN] The device we're interested in.
 * @param part_lba  [IN] The partition's LBA, uniquely identifies the partition within the drive.
 * @param cap_lba   [IN] The capture's LBA, uniquely identifies the capture within the partition.
 * @return Pointer to the found capture, or nullptr if not found.
 */
CAPTURE_DATA *fsWorker::cacheGetCapture(const std::string& dev_path, uint64_t part_lba, uint64_t cap_lba)
{
    int cap_index;
    CAPTURE_DATA *cap_ptr;
    CAP_UNION_VEC *cap_vec_ptr;
    PARTITION_INFO_VEC part_vec;
    PARTITION_INFO_VEC::iterator part_it;

    TP(FS_WORKER_FUNC, "%s: dev_path:%s, part_lba:%" PRIu64 ", cap_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), part_lba, cap_lba);

    // Assume we fail.
    cap_ptr = nullptr;

    // Make sure we have information about this drive.
    if (cache.count(dev_path) == 0)
        goto bail;

    // Find the partition.
    for (part_it = cache[dev_path].part_info.begin(); part_it != cache[dev_path].part_info.end(); ++part_it)
    {
        // Is this the one?
        if (part_it->part.lba == part_lba)
            break;
    }
    if (part_it == cache[dev_path].part_info.end())
        goto bail;

    // Find the capture.
    cap_vec_ptr = &part_it->caps;
    for (cap_index = 0; cap_index < (int)cap_vec_ptr->size(); ++cap_index)
    {
        // Is this the one?
        if (cap_vec_ptr->at(cap_index).lba == cap_lba)
        {
            cap_ptr = &cap_vec_ptr->at(cap_index);
            break;
        }
    }

bail:
    return cap_ptr;
}

/*!
 * @brief Retrieve the capture information for the partition (part_info), uniquely
 *        identified by the LBA (cap_lba).
 *
 * @param part_info [IN] Pointer to the FAS_FILESYSTEM::PARTITION_INFO we expect will contain the capture header.
 * @param cap_lba   [IN] The capture's LBA, uniquely identifies the capture within the partition.
 * @return Pointer to the found capture, or nullptr if not found.
 */
CAPTURE_DATA *fsWorker::cacheGetCapture(PARTITION_INFO *part_info, uint64_t cap_lba)
{
    int cap_index;
    CAPTURE_DATA *cap_ptr;
    CAP_UNION_VEC *cap_vec_ptr;

    // Sanity check.
    if (part_info == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Invalid argument, part_info is null!", __PRETTY_FUNCTION__);
        return nullptr;
    }

    TP(FS_WORKER_FUNC, "%s: part_lba:%" PRIu64 ", cap_lba:%" PRIu64, __PRETTY_FUNCTION__, part_info->part.lba, cap_lba);

    // Assume we fail.
    cap_ptr = nullptr;

    // Find the capture.
    cap_vec_ptr = &part_info->caps;
    for (cap_index = 0; cap_index < (int)cap_vec_ptr->size(); ++cap_index)
    {
        // Is this the one?
        if (cap_vec_ptr->at(cap_index).lba == cap_lba)
        {
            cap_ptr = &cap_vec_ptr->at(cap_index);
            break;
        }
    }

    // Return the capture pointer we found.
    return cap_ptr;
}

/*!
 * @brief Retrieve the partition information for the partition of the device (dev_path) identified
 *        by the LBA value (part_lba).
 *
 * @param dev_path      [IN] The device we're interested in.
 * @param part_lba      [IN] The partition's LBA, uniquely identifies the partition within the drive.
 * @param part_index    [OUT] The index of the partition in the FAS_FILESYSTEM::PARTITION_INFO_VEC. -1 if not found.
 * @return Pointer to the found partition, or nullptr if not found.
 */
PARTITION_INFO *fsWorker::cacheGetPartition(const std::string& dev_path, uint64_t part_lba, int& part_index)
{
    PARTITION_INFO *part_ptr;
    PARTITION_INFO_VEC *part_vec_ptr;

    TP(FS_WORKER_FUNC, "%s: dev_path:%s, part_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), part_lba);

    // Assume we fail.
    part_ptr = nullptr;
    part_index = -1;

    // Make sure we have information about this drive.
    if (cache.count(dev_path) == 0)
        goto bail;

    // Find the partition.
    part_vec_ptr = &cache[dev_path].part_info;
    for (part_index = 0; part_index < (int)part_vec_ptr->size(); ++part_index)
    {
        // Is this the one?
        if (part_vec_ptr->at(part_index).part.lba == part_lba)
        {
            part_ptr = &part_vec_ptr->at(part_index);
            break;
        }
    }

bail:
    return part_ptr;
}

/*!
 * @brief Clear a cache entry.
 * @param dev_path  [IN] The cache entry we want to clear.
 */
void fsWorker::cacheInvalidate(const std::string& dev_path)
{
    TP(FS_WORKER_FUNC, "%s: dev_path:%s", __PRETTY_FUNCTION__, dev_path.c_str());

    // Iterate through the cache to find the matching entry.
    std::map<std::string, FS_CACHE_ENTRY>::iterator it;
    it = cache.find(dev_path);
    if (it != cache.end())
    {
        // Remove the existing entry.
        cache.erase(it);
    }
}

/*!
 * @brief Create an available partition header.
 *
 * @param dpart             [IN][OUT] Pointer to the partition structure we want to fill.
 * @param start_lba         [IN] Starting LBA. The partition header goes here.
 * @param num_sectors       [IN] Size of the partition in sectors, including the header.
 */
void fsWorker::createAvailPartHdr(AVAIL_PARTITION_HDR *dpart, uint64_t start_lba, uint64_t num_sectors)
{
    TP(FS_WORKER_FUNC, "%s: start_lba=0x%" PRIx64 ", num_sectors=%" PRIu64, __PRETTY_FUNCTION__, start_lba, num_sectors);

    // Clear the memory.
    memset(dpart, 0, sizeof(AVAIL_PARTITION_HDR));

    // The number of sectors we need for the partition.
    dpart->reserved_sectors = FiFilesystem::getReserved(AVAIL_PART_SIG);

    // Available memory starts immediately following the partition header.
    dpart->start_lba = start_lba + dpart->reserved_sectors;

    // The entire usable space is available.
    dpart->data_sectors = num_sectors - dpart->reserved_sectors;

    TP(FS_WORKER_FUNC, "%s: AVAIL_PART(0x%" PRIx64 ":%" PRIu64 ")", __PRETTY_FUNCTION__, start_lba, num_sectors);
}

/*!
 * @brief Create a Long Recording capture header.
 *
 * @param cap           [IN][OUT] Capture header we're going to initialize.
 * @param part_lba      [IN] LBA where enclosing partition header is stored.
 * @param prev_cap_lba  [IN] LBA where previous capture header is stored.
 * @param cap_lba       [IN] LBA where this capture header is going to be placed.
 * @param image_sectors [IN] The number of sectors allocated for image memory.
 * @param fmeta_lba     [IN] LBA where the per-frame metadata is to be located.
 * @param fmeta_sectors [IN] The number of sectors allocated for PFMD.
 * @param max_ct_size   [IN] The maximum number of bytes to allocate for the correction table.
 */
void fsWorker::createLrCaptureHdr(
    LR_CAPTURE_HDR *cap,
    uint64_t part_lba,
    uint64_t prev_cap_lba,
    uint64_t cap_lba,
    uint64_t image_sectors,
    uint64_t fmeta_lba,
    uint64_t fmeta_sectors,
    uint32_t max_ct_size
)
{
    TP(FS_WORKER_FUNC, "%s: cap_lba=0x%" PRIx64 ", image_sectors=%" PRIu64 ", part_lba=0x%" PRIx64 ", prev_cap_lba=0x%" PRIx64 ", fmeta_lba=0x%" PRIx64 ", fmeta_sectors=%" PRIu64 ", max_ct_size=%u",
       __PRETTY_FUNCTION__, cap_lba, image_sectors, part_lba, prev_cap_lba, fmeta_lba, fmeta_sectors, max_ct_size);

    // Clear the memory.
    memset(cap, 0, sizeof(LR_CAPTURE_HDR));

    // Initialize our links.
    cap->link.dpart_lba = part_lba;
    cap->link.previous_lba = prev_cap_lba;
    cap->link.next_lba = 0;

    // Initialize what we know. We will allocate reserved space for the bookmark
    // table and the capture metadata. The lengths will remain at zero until we
    // actually populate the memory with data. Make sure we're sector aligned.
    cap->bk_offset = BYTES_TO_SECTORS(sizeof(LR_CAPTURE_HDR) + sizeof(FI_SIG)) * SECTOR_LEN;
    cap->bk_bytes = 0;
    cap->pm_offset = BYTES_TO_SECTORS(cap->bk_offset + CAPTURE_BOOKMARK_SIZE) * SECTOR_LEN;
    cap->pm_bytes = 0;
    cap->ct_offset = BYTES_TO_SECTORS(cap->pm_offset + CAPTURE_METADATA_SIZE) * SECTOR_LEN;
    cap->ct_bytes = 0;

    // The number of sectors used for the capture header is the one sector for
    // the actual header, the sectors for the capture metadata, and the sectors
    // for the bookmark table. The correction table information MAY be added later.
    cap->reserved_sectors = FiFilesystem::getReserved(LR_CAPTURE_SIG) + BYTES_TO_SECTORS(max_ct_size);

    // The start of the image data has to happen after the header. Everything else
    // is pretty straight forward. Make sure the image_lba is page aligned.
    cap->image_lba        = SECTORS_TO_PAGES(cap_lba + cap->reserved_sectors) * SECTORS_PER_PAGE;
    cap->image_sectors    = image_sectors;
    cap->fmeta_lba        = fmeta_lba;
    cap->fmeta_sectors    = fmeta_sectors;

    // No capture-specific fields are set, only the image, frame metadata, and capture reserved
    // locations.
    cap->state = CAPTUREHD_ALLOCATED;

    TP(FS_WORKER_FUNC, "%s: LR_CAP(0x%" PRIx64 "), image(0x%" PRIx64 ":%" PRIu64 "), state=%u flags=0x%x",
       __PRETTY_FUNCTION__, cap_lba, cap->image_lba, cap->image_sectors, cap->state, cap->flags);
}

/*!
 * @brief Create a Long Record partition header.
 *
 * @param dpart             [IN][OUT] Pointer to the partition structure we want to fill.
 * @param start_lba         [IN] Starting LBA. The partition header goes here.
 * @param num_sectors       [IN] Size of the partition in sectors, including the header.
 * @param min_frame_size    [IN] Frame size in bytes to use in allocating the frame metadata region.
 */
void fsWorker::createLrPartHdr(LR_PARTITION_HDR *dpart, uint64_t start_lba, uint64_t num_sectors, uint32_t min_frame_size)
{
    TP(FS_WORKER_FUNC, "%s: start_lba=0x%" PRIx64 ", num_sectors=%" PRIu64 ", min_frame_size=%u (0x%x)",
       __PRETTY_FUNCTION__, start_lba, num_sectors, min_frame_size, min_frame_size);

    // Clear the memory.
    memset(dpart, 0, sizeof(LR_PARTITION_HDR));

    // The number of sectors we need for the partition.
    dpart->reserved_sectors = FiFilesystem::getReserved(LR_PART_SIG);

    // Available memory starts immediately following the partition header.
    dpart->image_lba = start_lba + dpart->reserved_sectors;

    // For implementations which do not separate out the PFMD, the entire usable space
    // is available for the image data. This includes capture headers too.
    dpart->image_sectors = num_sectors - dpart->reserved_sectors;

    // To maintain backwards compatibility, clear these fields. Currently only the series3 cameras
    // separate out the per-frame metadata from the image data.
    dpart->fmeta_lba = 0;
    dpart->fmeta_sectors = 0;

    // The minimum frame size we're allowed. This value goes into the calculation for
    // the number of per-frame metadata sectors we need. Currently only the series3 cameras
    // make use of this field.
    dpart->min_frame_size = min_frame_size;

    TP(FS_WORKER_FUNC, "%s: LR_PART(0x%" PRIx64 ":%" PRIu64 "), image(0x%" PRIx64 ":%" PRIu64 "), fmeta(0x%" PRIx64 ":%" PRIu64 ")",
       __PRETTY_FUNCTION__, start_lba, num_sectors, dpart->image_lba, dpart->image_sectors, dpart->fmeta_lba, dpart->fmeta_sectors);
}

/*!
 * @brief Create a storage capture header.
 *
 * @param cap           [IN][OUT] Capture header we're going to initialize.
 * @param part_lba      [IN] LBA where enclosing partition header is stored.
 * @param prev_cap_lba  [IN] LBA where previous capture header is stored.
 * @param cap_lba       [IN] LBA where this capture header is going to be placed.
 * @param data_sectors  [IN] The number of sectors allocated for the data.
 */
void fsWorker::createStorCaptureHdr(STOR_CAPTURE_HDR *cap, uint64_t part_lba, uint64_t prev_cap_lba, uint64_t cap_lba, uint64_t data_sectors)
{
    TP(FS_WORKER_FUNC, "%s: cap_lba=0x%" PRIx64 ", data_sectors=%" PRIu64 ", part_lba=0x%" PRIx64 ", prev_cap_lba=0x%" PRIx64,
       __PRETTY_FUNCTION__, cap_lba, data_sectors, part_lba, prev_cap_lba);

    // Clear the memory.
    memset(cap, 0, sizeof(LR_CAPTURE_HDR));

    // Initialize our links.
    cap->link.dpart_lba = part_lba;
    cap->link.previous_lba = prev_cap_lba;
    cap->link.next_lba = 0;
    cap->reserved_sectors = FiFilesystem::getReserved(STOR_CAPTURE_SIG);

    // The start of the image data has to happen after the header. Everything else
    // is pretty straight forward. Make sure the image_lba is page aligned.
    cap->data_lba        = SECTORS_TO_PAGES(cap_lba + cap->reserved_sectors) * SECTORS_PER_PAGE;
    cap->data_sectors    = data_sectors;

    // No capture-specific fields are set, only the image, frame metadata, and capture reserved
    // locations.
    cap->state = CAPTUREHD_ALLOCATED;

    TP(FS_WORKER_FUNC, "%s: STOR_CAP(0x%" PRIx64 "), data(0x%" PRIx64 ":%" PRIu64 "), state=%u flags=0x%x",
       __PRETTY_FUNCTION__, cap_lba, cap->data_lba, cap->data_sectors, cap->state, cap->flags);
}

/*!
 * @brief Create a storage partition header.
 *
 * @param dpart             [IN][OUT] Pointer to the partition structure we want to fill.
 * @param start_lba         [IN] Starting LBA. The partition header goes here.
 * @param num_sectors       [IN] Size of the partition in sectors, including the header.
 */
void fsWorker::createStorPartHdr(STOR_PARTITION_HDR *dpart, uint64_t start_lba, uint64_t num_sectors)
{
    TP(FS_WORKER_FUNC, "%s: start_lba=0x%" PRIx64 ", num_sectors=%" PRIu64, __PRETTY_FUNCTION__, start_lba, num_sectors);

    // Clear the memory.
    memset(dpart, 0, sizeof(STOR_PARTITION_HDR));

    // The number of sectors we need for the partition.
    dpart->reserved_sectors = FiFilesystem::getReserved(STOR_PART_SIG);

    // Available memory starts immediately following the partition header.
    dpart->start_lba = start_lba + dpart->reserved_sectors;

    // The entire usable space is available for the image data. This includes capture headers too.
    dpart->data_sectors = num_sectors - dpart->reserved_sectors;

    // For future expandibility, currently there's no metadata associated with the storage partition.
    dpart->meta_lba = 0;
    dpart->meta_sectors = 0;

    TP(FS_WORKER_FUNC, "%s: STOR_PART(0x%" PRIx64 ":%" PRIu64 ")", __PRETTY_FUNCTION__, start_lba, num_sectors);
}

/*!
 * @brief Read sector(s) from the path specified in the class field "fsWorker::dev_path".
 *
 * @param start_sector  LBA of first sector to read.
 * @param buf           Page aligned buffer to receive data.
 * @param num_sectors   Number of sectors to read.
 *
 * @return True for success, false for failure.
 */
bool fsWorker::readSectors(uint64_t start_sector, uint8_t *buf, uint32_t num_sectors)
{
    bool    rc = false;
    off64_t off;
    ssize_t size;

    TP(FS_WORKER_FUNC_NOISY, "%s: dev_fd=%d, start_sector=%" PRIu64 ", buf=0x%" PRIx64 ", num_sectors=%u", __PRETTY_FUNCTION__, dev_fd, start_sector, ((uint64_t)buf), num_sectors);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: There is no dev_path specified!", __PRETTY_FUNCTION__);
        goto bail;
    }
    if ((buf == nullptr) || (num_sectors == 0))
    {
        fi_log(LOG_DEBUG, "%s: Bad parameters [buf=0x%" PRIx64 ", num_sectors=%u]", __PRETTY_FUNCTION__, ((uint64_t)buf), num_sectors);
        goto bail;
    }

    // The device should already be opened.
    if (dev_fd == -1)
    {
        fi_log(LOG_DEBUG, "%s: The device is not open!", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Covert to byte values.
    off = start_sector * SECTOR_LEN;
    size = num_sectors * SECTOR_LEN;

    // Seek to and read the data.
    if ((lseek64(dev_fd, off, SEEK_SET) != off) || (read(dev_fd, buf, size) != size))
    {
        fi_log(LOG_ERR, "%s: Read(%" PRIu64 ", %u) failed - (%d) %s", __PRETTY_FUNCTION__, start_sector, num_sectors, errno, strerror(errno));
    }
    else
    {
        rc = true;
    }

bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Set the checksum field in a block based on the size that was already set.
 *
 * @param sig   Points to the signature structure at the start of the block
 *      which is followed immedately by the data. The size field of the
 *      signature block must already be set.
 */
void fsWorker::setBlockChecksum(FI_SIG *sig)
{
    unsigned char digest[md5Length()];

    TP(FS_WORKER_FUNC_NOISY, "%s", __PRETTY_FUNCTION__);

    md5Calculate((uint8_t *)(sig + 1), sig->size, digest);
    memcpy((void *)&sig->checksum, digest, sizeof(sig->checksum));
}

/*!
 * @brief Verify the signature and checksum field in a capture header.
 *
 * @param sig       Points to the signature structure at the start of the block
 *                  which is followed immedately by the data, of size specified
 *                  in the signature block.
 * @param part_sig  The signature/type of the enclosing partition. Make sure
 *                  the capture header is compatible.
 *
 * @return True if it's good, false if not.
 */
bool fsWorker::verifyCaptureHeader(FI_SIG *sig, uint32_t part_sig)
{
    unsigned char digest[md5Length()];

    TP(FS_WORKER_FUNC_NOISY, "%s", __PRETTY_FUNCTION__);

    // Check the signature. Make sure it's a valid capture signature.
    switch (sig->signature)
    {
    case LR_CAPTURE_SIG:
    case STOR_CAPTURE_SIG:
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unknown capture type '%s' [0x%08x]", __PRETTY_FUNCTION__, signatureToString(sig->signature), sig->signature);
        return false;
    }

    // Make sure the capture signature is valid for the partition signature.
    switch (part_sig)
    {
    case AVAIL_PART_SIG:
        fi_log(LOG_DEBUG, "%s: Available partitions can not have captures.", __PRETTY_FUNCTION__);
        return false;

    case LR_PART_SIG:
        if (sig->signature != LR_CAPTURE_SIG)
        {
            fi_log(LOG_DEBUG, "%s: Incompatible signature types [part='%s'/0x%08x, cap='%s'/0x%08x]", __PRETTY_FUNCTION__, signatureToString(part_sig), part_sig, signatureToString(sig->signature), sig->signature);
            return false;
        }
        break;
    case STOR_PART_SIG:
        if (sig->signature != STOR_CAPTURE_SIG)
        {
            fi_log(LOG_DEBUG, "%s: Incompatible signature types [part='%s'/0x%08x, cap='%s'/0x%08x]", __PRETTY_FUNCTION__, signatureToString(part_sig), part_sig, signatureToString(sig->signature), sig->signature);
            return false;
        }
        break;

    default:
        fi_log(LOG_DEBUG, "%s: Unknown partition signature '%s' [0x%08x]", __PRETTY_FUNCTION__, signatureToString(part_sig), part_sig);
        return false;
    }

    // Compute the digest of the data following the signature block.
    md5Calculate((uint8_t *)(sig + 1), sig->size, digest);
    if (memcmp((void *)&sig->checksum, digest, sizeof(sig->checksum)) != 0)
        return false;

    // Signature validates.
    return true;
}

/*!
 * @brief Verify the signature and checksum field in a drive header.
 *
 * @param sig   Points to the signature structure at the start of the block
 *              which is followed immedately by the data, of size specified
 *              in the signature block.
 *
 * @return True if it's good, false if not.
 */
bool fsWorker::verifyDriveHeader(FI_SIG *sig)
{
    unsigned char digest[md5Length()];

    TP(FS_WORKER_FUNC_NOISY, "%s", __PRETTY_FUNCTION__);

    // Check the signature.
    if (sig->signature != DRIVE_SIG)
        return false;

    // Compute the digest of the data following the signature block.
    md5Calculate((uint8_t *)(sig + 1), sig->size, digest);
    if (memcmp((void *)&sig->checksum, digest, sizeof(sig->checksum)) != 0)
        return false;

    // Signature validates.
    return true;
}

/*!
 * @brief Verify the signature and checksum field in a partition header.
 *
 * @param sig   Points to the signature structure at the start of the block
 *              which is followed immedately by the data, of size specified
 *              in the signature block.
 *
 * @return True if it's good, false if not.
 */
bool fsWorker::verifyPartitionHeader(FI_SIG *sig)
{
    unsigned char digest[md5Length()];

    TP(FS_WORKER_FUNC_NOISY, "%s", __PRETTY_FUNCTION__);

    // Check the signature. Make sure it's a valid partition signature.
    switch (sig->signature)
    {
    case AVAIL_PART_SIG:
    case LR_PART_SIG:
    case STOR_PART_SIG:
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unknown partition type '%s' [0x%08x]", __PRETTY_FUNCTION__, signatureToString(sig->signature), sig->signature);
        return false;
    }

    // Compute the digest of the data following the signature block.
    md5Calculate((uint8_t *)(sig + 1), sig->size, digest);
    if (memcmp((void *)&sig->checksum, digest, sizeof(sig->checksum)) != 0)
        return false;

    // Signature validates.
    return true;
}

/*!
 * @brief Write sectors to the SSD
 *
 * @param start_sector  LBA of first sector to write.
 * @param buf           Buffer containing data.
 * @param num_sectors   Number of sectors to write.
 *
 * @return True for success, false for failure
 */
bool fsWorker::writeSectors(uint64_t start_sector, const uint8_t *buf, uint32_t num_sectors)
{
    bool    rc = false;
    int     fd;
    off64_t off;
    ssize_t size;

    TP(FS_WORKER_FUNC_NOISY, "%s: start_sector=%" PRIu64 ", buf=0x%" PRIx64 ", num_sectors=%u", __PRETTY_FUNCTION__, start_sector, ((uint64_t)buf), num_sectors);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: There is no dev_path specified!", __PRETTY_FUNCTION__);
        return false;
    }
    if ((buf == nullptr) || (num_sectors == 0))
    {
        fi_log(LOG_DEBUG, "%s: Bad parameters [buf=0x%" PRIx64 ", num_sectors=%u]", __PRETTY_FUNCTION__, (uint64_t)buf, num_sectors);
        return false;
    }

    // Open the device.
    fd = open(dev_path.c_str(), O_WRONLY | O_LARGEFILE);
    if (fd < 0)
    {
        fi_log(LOG_ERR, "%s: open(%s) for write failed - (%d) %s", __PRETTY_FUNCTION__,
               dev_path.c_str(), errno, strerror(errno));
        return false;
    }

    // Convert to byte values.
    off = start_sector * SECTOR_LEN;
    size = num_sectors * SECTOR_LEN;

    // Seek to and write the data.
    if ((lseek64(fd, off, SEEK_SET) != off) || (write(fd, buf, size) != size) || (fsync(fd) < 0))
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        fi_log(LOG_ERR, "%s: Write(%" PRIu64 ", %u) to [%s] failed", __PRETTY_FUNCTION__, start_sector, num_sectors, dev_path.c_str());
    }
    else
    {
        rc = true;
    }

    // Close the device.
    close(fd);

    // Return our result.
    return rc;
}

/*!
 * @brief Write out the correction and gain table information to the drive.
 *
 * @param cap_data  [IN] Information about the current capture.
 * @param msg       [IN] The new tables we are going to save.
 * @return The total number of bytes required for the CT/GT information.
 */
uint32_t fsWorker::writeCorrectionInfo(const CAPTURE_DATA *cap_data, fsSetCorrectionInfoMessage *msg)
{
    uint32_t total_size(0);
    uint32_t buf_size;
    uint64_t start_sector;
    CT_GT_HDR *hdr_info;

    TP(FS_WORKER_FUNC, "%s", __PRETTY_FUNCTION__);

    // Are we "deleting" the correction table information, there's really
    // nothing to do here.
    if ((msg->ct_buffer.size() == 0) && (msg->gt_buffer.size() == 0))
    {
        TP(FS_WORKER_FUNC, "%s: deleting the correction/gain table information.", __PRETTY_FUNCTION__);
        return 0;
    }

    // Reset our sector-aligned buffer memory.
    memset(sector_buf.get(), 0, SECTOR_LEN);

    // Overlay our header onto the sector buffer.
    hdr_info = (CT_GT_HDR *)sector_buf.get();

    // The layout of the correction and gain table information is the header
    // immediately followed by the correction table. The gain table then follows
    // the correction table. Everything is aligned on a sector boundary.
    hdr_info->magic = CT_GT_HDR_ID;
    hdr_info->ct_bytes = msg->ct_buffer.size();
    if (hdr_info->ct_bytes == 0)
        hdr_info->ct_offset = 0;
    else
        hdr_info->ct_offset = SECTOR_LEN;
    hdr_info->gt_bytes = msg->gt_buffer.size();
    if (hdr_info->gt_bytes == 0)
        hdr_info->gt_offset = 0;
    else
        hdr_info->gt_offset = SECTORS_TO_BYTES(1 + BYTES_TO_SECTORS(hdr_info->ct_bytes));

    // Figure out where the correction table goes.
    start_sector = cap_data->lba + BYTES_TO_SECTORS(cap_data->hdr.lr_cap.ct_offset);
    TP(FS_WORKER_FUNC, "%s: cap_lba=0x%" PRIx64 ", ct_offset=0x%x, start_sector=0x%" PRIx64, __PRETTY_FUNCTION__, cap_data->lba, cap_data->hdr.lr_cap.ct_offset, start_sector);

    // Write the correction/gain table header to the drive.
    if (!writeSectors(start_sector, sector_buf.get(), 1))
    {
        fi_log(LOG_DEBUG, "%s: Unable to write the correction/gain table header.", __PRETTY_FUNCTION__);
        goto bail;
    }

    // Move our write location past the header.
    start_sector += 1;

    // Only writing if there is a correction table.
    if (msg->ct_buffer.size() > 0)
    {
        // Round the correction table size up to a sector boundary.
        buf_size = BYTES_TO_SECTORS(msg->ct_buffer.size()) * SECTOR_LEN;

        // Make sure the buffer is updated to the sector aligned value.
        msg->ct_buffer.resize(buf_size, 0);

        // Write the correction table data to the drive.
        if (!writeSectors(start_sector, msg->ct_buffer.data(), BYTES_TO_SECTORS(buf_size)))
        {
            fi_log(LOG_DEBUG, "%s: Unable to write the correction table information.", __PRETTY_FUNCTION__);
            goto bail;
        }
    }

    // Move our write location past the correction table.
    start_sector += BYTES_TO_SECTORS(hdr_info->ct_bytes);

    // Only writing if there is a gain table.
    if (msg->gt_buffer.size() > 0)
    {
        // Round the gain table size up to a sector boundary.
        buf_size = BYTES_TO_SECTORS(msg->gt_buffer.size()) * SECTOR_LEN;

        // Make sure the buffer is updated to the sector aligned value.
        msg->gt_buffer.resize(buf_size, 0);

        // Write the gain table data to the drive.
        if (!writeSectors(start_sector, msg->gt_buffer.data(), BYTES_TO_SECTORS(buf_size)))
        {
            fi_log(LOG_DEBUG, "%s: Unable to write the gain table information.", __PRETTY_FUNCTION__);
            goto bail;
        }
    }

    // Calculate the total number of bytes taken up by the correction and gain table information.
    total_size = SECTORS_TO_BYTES(1 + BYTES_TO_SECTORS(hdr_info->ct_bytes) + BYTES_TO_SECTORS(hdr_info->gt_bytes));

bail:
    // Return the total number of bytes used by the correction and gain table information.
    return total_size;
}
